/*global cc*/
/*global require*/
/*global Game*/
/*global BrainShiftSettings*/
/*global BrainShiftModel*/
/*global BrainShiftView*/
/*global LumosityUtils*/

require( "src/Game/Base/Game.js" );
require( "src/Game/BrainShiftSettings.js" );
require( "src/Game/BrainShiftModel.js" );
require( "src/Game/BrainShiftView.js" );
require( "src/Utility/LumosityUtils.js" );

var BrainShift = Game.extend(
{
    /** @lends BrainShift.prototype */

    /**
     *  BrainShift is the main class for the game, Brain Shift!
     *
     *  @constructs
     */
    ctor:function()
    {
        this._super();

        this.timeRemaining = this.settings.gameLengthInSeconds;
    },

    onBegin: function()
    {
        this._super();

        this.view.enableButtons( false );
    },

    onStartGameplay:function()
    {
        this._super();

        this.createNewStimuli();

        this.view.playAnimation( "problem" );

        this.view.setHintVisibility( true );

        this._setHintsShowing( true );

        this.model.hintsShowing = true;

        this.view.enableButtons( true );
    },

    onEndGameplay:function()
    {
        this._super();

        this._setHintsShowing( false );

        this.model.hintsShowing = false;

        this.view.enableButtons( false );
    },

    onRestart: function()
    {
        this._super();

        this.model.score = 0;

        this.view.hud.setScore( 0 );

        this._setHintsShowing( false );

        this.model.hintsShowing = true;

        this.view.setHintVisibility( false );

        this.view.enableButtons( false );
    },

    checkAnswer:function( yesPressed )
    {
        if ( this.isBlockingInput )
        {
            return;
        }

        this.model.currentTrial.response = yesPressed;
        // Using number comparison
        if ( this.model.isUsingTopStimulus )
        {
            // Correct
            if ( this.settings.doEvensContain( this.model.topNumber ) === yesPressed )
            {
                this.factorResults( true );
            }
            // Incorrect
            else
            {
                this.factorResults( false );
            }
        }
        // Using letter comparison
        else
        {
            // Correct
            if ( this.settings.doVowelsContain( this.model.bottomLetter ) === yesPressed )
            {
                this.factorResults( true );
            }
            // Incorrect
            else
            {
                this.factorResults( false );
            }
        }
    },

    factorResults:function( success )
    {

        // Increase how many responses we have seen this game
        this.model.totalResponseCount = this.model.totalResponseCount + 1;

        // Block user input now
        this.model.isBlockingInput = true;

        // Initialize Metadata Trial Object
        var trial = {};

        trial.progress = this._components[ GameComponentScoreMultiplier.prototype.getType() ]._meterValue - 1;

        trial.multiplier = this._components[ GameComponentScoreMultiplier.prototype.getType() ]._multiplier;

        // Correct answer, give some points
        if ( success )
        {

            trial.points = 50 * trial.multiplier;

            // Update the score
            this.adjustScoreForCorrectAnswer();

            // Increase consecutive total
            this.model.consecutiveCorrectCount = this.model.consecutiveCorrectCount + 1;
            if ( this.model.isUsingTopStimulus )
            {
                this.model.consecutiveCorrectNumberCount = this.model.consecutiveCorrectNumberCount + 1;
            }
            else
            {
                this.model.consecutiveCorrectLetterCount = this.model.consecutiveCorrectLetterCount + 1;
            }

            this.model.totalCorrectResponseCount = this.model.totalCorrectResponseCount + 1;

        }
        // Incorrect answer, give penalty
        else
        {

            trial.points = 0;

            //update the score
            this.adjustScoreForIncorrectAnswer();

            //update correct count
            this.model.consecutiveCorrectCount = 0;
        }

        var scoreMultiplierComponent = this._components[ GameComponentScoreMultiplier.prototype.getType() ];

        // Reaction time
        var now = new Date().getTime();
        var reactionTime = now - this.model.lastAnswerTime;
        this.model.averageReactionTime = this.model.averageReactionTime
                                        * ( 1.0 - 1.0 / this.model.totalResponseCount )
                                        + reactionTime * ( 1.0 / this.model.totalResponseCount );
        this.model.lastAnswerTime = now;


        // METADATA
        trial.correct = (success) ? "T" : "F";

        // trial.points = this.model.currentTrial.points;
        trial.response = LumosityUtils.booleanToChar( this.model.currentTrial.response );

        trial.response_type = this.model.currentTrial.response_type;

        trial.stimulus = this.model.currentTrial.stimulus;

        trial.switch = LumosityUtils.booleanToChar( this.model.currentTrial.switch );

        trial.aided = LumosityUtils.booleanToChar( this.model.hintsShowing );

        this.endTrial( trial, success, this.model.currentTrial.switch, false, true, false, true );

        //update the values displayed and in the instance
        this.view.setScore( this.model.score );

        //give a new card
        this.createNewStimuli();


        // ui: feedback, instructions
        if ( success )
        {
            this.model.addTrialResult( true );
            // instructions
            //hide the help text if we've been AWESOME enough
            if ( this.model.consecutiveCorrectCount >= this.settings.removeInstructionsConsecutiveCountRequired &&
                this.model.consecutiveCorrectLetterCount >= this.settings.removeInstructionsLetterCountRequired &&
                this.model.consecutiveCorrectNumberCount >= this.settings.removeInstructionsNumberCountRequired )
            {
                this._setHintsShowing( false );
                this.model.hintsShowing = false;
            }

            // Play feedback animation

            if ( this.model.isNewMultiplier )
            {
               this.view.showFeedback( true , scoreMultiplierComponent._multiplier, true );
            }
            else
            {
                this.view.showFeedback( true, 1, true );
            }

        }
        else
        {
            if (this.model.shouldShowHints())
            {
                this._setHintsShowing( true );
                this.model.hintsShowing = true;
            }

            this.model.addTrialResult( false );

            // play feedback animation
            this.view.showFeedback( false , scoreMultiplierComponent._multiplier, true );

        }
    },



    createNewStimuli:function()
    {
        //Initialize Metadata
        this.getMetadataComponent().beginMetadataTrial();

        //generate our value to go in the panel
        var number = "";
        var letter = "";
        var value = "";
        // even
        if ( Math.random() < this.settings.numberIsEvenChance )
        {
            number = this.settings.evenCharacterArray[ Math.floor( Math.random() * this.settings.evenCharacterArray.length ) ];
        }
        // odd
        else
        {
            number = this.settings.oddCharacterArray[ Math.floor( Math.random() * this.settings.oddCharacterArray.length ) ];
        }

        // vowel
        if ( Math.random() < this.settings.letterIsVowelChance )
        {
            letter = this.settings.consonantCharacterArray[ Math.floor( Math.random() * this.settings.consonantCharacterArray.length ) ];
        }
        // consonant
        else
        {
            letter = this.settings.vowelCharacterArray[ Math.floor( Math.random() * this.settings.vowelCharacterArray.length ) ];
        }
        // number before letter
        if ( Math.random() < this.settings.numberBeforeLetterChance )
        {

            value = number + letter;
            this.model.numberThenLetter = true;
        }
        // letter before number
        else
        {

            value = letter + number;
            this.model.numberThenLetter = false;
        }


        //place the value in the panel
        // use top panel - the first stimulus is always the top, the second is always the bottom
        if ( this.model.totalResponseCount === 0 || ( Math.random() < this.settings.numberStimulusChance && this.model.totalResponseCount !== 1 ) )
        {
            this.model.setTopStimulus( value, number, letter, this.model.numberThenLetter );

            this.model.setMetadataSwitch( "E" );

            this.model.currentTrial.response_type = "E";
        }
        // use bottom panel
        else
        {
            this.model.setBottomStimulus( value, number, letter, this.model.numberThenLetter );

            this.model.setMetadataSwitch( "V" );

            this.model.currentTrial.response_type = "V";
        }
    },

    adjustScoreForCorrectAnswer: function()
    {
        var scoreMultiplierComponent = this._components[ GameComponentScoreMultiplier.prototype.getType() ];

        // incrememnt the multiplier
        this.model.correctAnswerStreak = this.model.correctAnswerStreak + 1;

        //clamp the multiplier index - TODO, the model should probably do this itself
        if ( this.model.scoreMultiplierIndex >= this.settings.getMultiplierCount() )
        {
            this.model.scoreMultiplierIndex = this.settings.getMultiplierCount() - 1;
        }

        this.model.isNewMultiplier = false;

        // Check if the streak is complete
        if ( this.model.correctAnswerStreak === this.settings.getMultiplierStreak( this.model.scoreMultiplierIndex ) )
        {
            this.model.correctAnswerStreak = 0;
            if ( scoreMultiplierComponent._multiplier < 10 )
            {
               this.model.scoreMultiplierIndex = this.model.scoreMultiplierIndex + 1;
               this.model.isNewMultiplier = true;
            }
        }

        scoreMultiplierComponent.increaseMultiplierProgress();
    },

    adjustScoreForIncorrectAnswer: function()
    {
        if ( this.model.correctAnswerStreak > 0 )
        {
            this.model.correctAnswerStreak = 0;
        }
        else
        {
            this.model.scoreMultiplierIndex = this.model.scoreMultiplierIndex - 1;

            //clamp the multiplier index
            if ( this.model.scoreMultiplierIndex < 0 )
            {
                this.model.scoreMultiplierIndex = 0;
            }
        }

        var scoreMultiplierComponent = this._components[ GameComponentScoreMultiplier.prototype.getType() ];

        scoreMultiplierComponent.decreaseMultiplierProgress();
    },

    // This only adjusts the view, it doesn't calculate an actual new stimulus
    cue_showNextStimulus:function()
    {
        // The two parameters to be passed into the updateStimuli function
        var numberParameter;
        var letterParameter;

        if( this.model.isUsingTopStimulus )
        {
            numberParameter = this.model.topNumber;
            letterParameter = this.model.topLetter;
        }
        else
        {
            numberParameter = this.model.bottomNumber;
            letterParameter = this.model.bottomLetter;
        }

        this.view.updateStimuli( numberParameter, letterParameter, this.model.isUsingTopStimulus, this.model.numberThenLetter );
        this.model.isBlockingInput = false;
    },


    feedback_isLargest:function( feedbackWidget )
    {
        this.view.playAnimation( "problem" );
    },

    cue_showCountdown: function()
    {
        this.view.showCountdown();
    },

    _setHintsShowing: function( showing, animated )
    {
        if ( showing )
        {
            this.view.showHints( true );
        }
        else
        {
            this.view.hideHints( true );
        }
    },

    _createSettings:function()
    {
        return new BrainShiftSettings( "settings.json" );
    },

    _createModel:function()
    {
        return new BrainShiftModel();
    },

    _createView:function( screenSize )
    {
        return new BrainShiftView( screenSize, this.settings, this.model, this, true );
    },

    _buttonPressed:function( buttonName )
    {
        this._super( buttonName );

        if ( this.model.isPlaying() )
        {
            if ( buttonName === "yes" )
            {
                this.checkAnswer( true );
            }
            else if ( buttonName === "no" )
            {
                this.checkAnswer( false );
            }
        }
    }
});