/*global cc*/
/*global require*/
/*global ll*/
/*global lu*/
/*global Game*/
/*global DisillusionSettings*/
/*global DisillusionModel*/
/*global DisillusionView*/
/*global DisillusionRoundModel*/
/*global DisillusionBoard*/
/*global DisillusionBoardPiece*/
/*global DisillusionPieceOrientation*/
/*global DisillusionPieceType*/
/*global InteractionType*/
/*global SwipeGestureDirection*/


require( "src/Utility/LumosityUtils.js" );
require( "src/Utility/LumosityLogger.js" );
require( "src/Game/Base/Game.js" );
require( "src/Game/DisillusionBlock.js" );
require( "src/Game/DisillusionSettings.js" );
require( "src/Game/DisillusionModel.js" );
require( "src/Game/DisillusionView.js" );
require( "src/Game/DisillusionRoundModel.js" );
require( "src/Game/DisillusionBoard.js" );
require( "src/Game/DisillusionBoardPiece.js" );
var lu = LumosityUtils;


var Disillusion = Game.extend(
{
    /** @lends Disillusion.prototype */
    onFinishEnabled: true,
    
    /**
     *  Disillusion is the main class for the game, Disillusion!
     *
     *  @constructs
     */
    ctor: function()
    {
        // lumosity.game.params.short_game = true;  // XXX Is there a way to set short game only in my LumosityMobile?
        // lumosity.game.params.debug_mode = true;  // XXX Is there a way to set debug only in my LumosityMobile?
        this._super();
        if ( lumosity.device.id === "html5" ) 
        {
            if (!lu.isDebugModeOn())
            {
                ll.CurrentLoggingLevel = LoggingLevel.Normal;
            }
            try 
            {
                lumosity.game.instance = this;
            }
            catch (err) 
            {
            }
        }
        if (lu.isDebugModeOn())
        {
            var labelText = "Win level " + this.model.maxLevel.toString();
            this.view.addDebugMenuButton( { labelText: labelText, 
                callback: this._winLevelMax, 
                target: this, closeMenuOnCallback: true } );
        }
        this.block = new DisillusionBlock();
    },

    /**
     * MOBGAMES-1898 GAME-1877 Adapted from Pet Detective, if at level select screen and pressing system back button, then exit game prematurely.  Consistent with Pet Detective and Train of Thought, completed, although inconsistent with GameView.  Test case:  2014-11-10 Android.  Disillusion.  Tap system back button.  Angel Cruz expects to exit game and return to pregame screen in the app.
     */
    _backButtonTapped: function()
    {
        if (this.model.persistentData.isLevelSelectMode)
        {
            this.onFinish(false);
            return;
        }
        this._super();
    },
    
    _winLevelMax: function()
    {
        this.model._winLevelMax();
        this.endCurrentRound();
    },

    _createSettings: function()
    {
        var settings = new DisillusionSettings( "settings.json" );
        return settings;
    },

    _createModel:function()
    {
        this.model = new DisillusionModel(this.settings);
        if ( lu.isDebugModeOn() && !lu.isShortGamesModeOn() ) {
            this.model.initSessionLevel( this.model.maxLevel - 8 );
        }
        return this.model;
    },

    _createView:function( screenSize )
    {
        return new DisillusionView( screenSize, this.settings, this.model, this );
    },

    _createComponents:function()
    {
        this._super();
    },

    configureMetaDataKeys: function()
    {
        this._super();
    },

    /**
     * XXX Save trials to private metadata variables.
     * GAME-2503 Stomp shared metadata faulty num_correct with this game's num_correct.  Test case:  2015-01-05 Short game.  No answer.  [~aaron] expects num_correct 0.  Got 1.
     * GAME-2016 Test case:  2015-01-23 Restart.  Expect num correct 0.  Got num correct of previous session.
     * In the metadata component "onEndGameplay":  One improvement would be for each key to ignore writing a value if they already contains a defined value.  Test case:  2014-12-13 Trouble Brewing has a clean and completely portable Metadata class that manages all the game-specific values, including "num_correct", "num_trials".  This generic metadata stomps those values with a complicated, rigid expectation to use the beginMetadataTrial and endMetadataTrial functions.  This component's onEndGameplay function is called after "finalizeMetadata" function, so there is no forward-compatible place to overwrite the metadata's bloated "onEndGameplay" values.
     * Unfortunately, time still gets stomped.  Instead, time_remaining is pre-bonus time remaining.
     */
    finalizeMetadata: function()
    {
        this._super();
        var component = this.getMetadataComponent();
        component._trialCorrectTimes = [];
        component._trialCorrectTimes.length = this.model.num_correct;
    },

    /**
     * Clear tappable pieces before constructing model.
     * Do not clear tappable pieces at all.  Test case:  2014-10-22 How to play mode.  Answer 2.  Skip.  Play.  Answer 1.  Pause.  How to play.  Expect to see spots to tap.  Got none.
     * Test case:  2014-10-21 Play.  Pause.  How to play.  Expect spot to tap.  Got none.
     */
    onBegin: function()
    {
        this._super();

        // ll.verbose( "Disillusion", "onBegin" );
        this.timeBetweenRounds = 0;
        this.view.destroyTappablePieces( true );
        this.model.onBegin();
        this.view.onBegin();
        this.onFinishEnabled = true;
    },

    /**
     * GAME-1405 Stop all animations in view before base class plays end tutorial animations.  Test case:  2014-10-22 Complete first trial and first board.  You cleared board animation begins. Immediately click Skip tutorial.  Expect popup.  Got nothing happens.
     * Stop tutor.  Race precaution:  Program hides the orientation hint and prompt if they would still appear when skipping.  There is an animation to make them disappear.
     * GAME-1405 End game now.
     */
    onCompleteTutorial: function( skipped )
    {
        this._endGameNow();
        this.model.tutor.shared.stop();
        this.view.onCompleteTutorial(skipped);
        this._super(skipped);
    },

    /**
     * Do not trigger ending when time bonus rolls down.
     */
    onComponentTimerEnd:function()
    {
        if ( this.model.bonusScorePoints <= 0 ) 
        {
            this._super();
        }
    },

    /**
     * Guard until not in level select.
     */
    onStartGameplay:function()
    {
        if (this.model.persistentData.isLevelSelectMode) 
        {
            return;
        }
        this._super();
        // ll.verbose( "Disillusion", "onStartGameplay" );
        this._components.GameComponentTimer.reset( 
            this.model.onStartGameplay() );
        if (!this.model.persistentData.isPracticeMode 
        && this.model.levelHintMax() <= this.model.currentLevel) 
        {
            this.view.showPrompt(false);
        }
        this.startNextRound();
        this.block.reset();
        if (this.model.persistentData.isPracticeMode) {
            this.view.playAnimation("practiceTrialBegin");
        }
    },

    /**
     * If increasing next session level, disable default finish to show illusions.
     * GAME-1403 Practice: Complete tutorial instead of end game.
     * MOBGAMES-1792 If end gameplay already called, ignore second call.  2014-10-30 Practice: End game.  QA Temp1 expects to see end tutorial message once.
     */
    onEndGameplay: function()
    {
        if (this._endGameNow())
        {
            // ll.verbose( "Disillusion", "onEndGameplay" );
            if (this.model.persistentData.isPracticeMode) 
            {
                this.onCompleteTutorial( false );
            }
            else 
            {
                this.view.cue_hidePauseButton();
                this.onFinishEnabled = this.model.wouldFinish();
                this._super();
                if (!this.onFinishEnabled)
                {
                    this.view.illusionLoadInBackground(1.5);
                }
            }
        }
    },

    /**
     * @return  If this call ends the game.
     * GAME-2506 Record milliseconds from time consumed, which is paused between trials, rather than from metadata running time.  Test case:  2015-01-23 Chart normalized, stratified score percentiles of users who never played a prequel (Disillusion 1 or Disillusion 2).  Daniel Sternberg expects median score of mobile users no less than about 5% lower than web users.  Got more than 20% lower for users who may or may not have played a prequel.
     */
    _endGameNow: function()
    {
        var ending = this.model.isGameRunning;
        if (ending)
        {
            var consumed = this._components.GameComponentTimer.getTimeConsumed();
            this.model.onEndGameplay(consumed);
            this.view.onEndGameplay();
        }
        return ending;
    },

    onFinish: function( wasQuit )
    {
        if (this.onFinishEnabled) 
        {
            this._super( wasQuit );
        }
    },

    animationFinished_endTutorial: function()
    {
        this.onEndTutorial();
    },

    animationFinished_gameEnd: function()
    {
        if (!this.onFinishEnabled) {
            this.view.illusionBegin();
        }
    },

    /**
     * GAME-1400 If last tile was a block and it is removed, then end the current round.
     * GAME-1403 After block clears, can tap there.
     * GAME-1405 If game complete, and not started and finished transferring time, then pause timer.
     * Test case:  2014-10-15 Answer in last second.  +10 bonus.  Expect only one popup.  Got 2.  Expect game not end in illusion view.  Got ends.  (+2014-10-20)
     * MOBGAME-1822 Tutor updates with accumulated time between rounds.  Test case:  End tutor.  Expect text transitions off.
     * GAME-1405 Between rounds accumulate time but do not reset accumulated time in next round.  Test case:  2014-11-08 Second tutor board.  Expect immediate prompt.  Got delayed for about 4 seconds.
     * GAME-2506 Pause timer between trials.  I logged 200ms of animation between trials and another 200ms of animation between rounds.  For 46 trials and 4 rounds this would cost 10 seconds, which could lower a score around 20%.  I also adjusted metadata to record game timer rather than the shared metadata component timer, which does not consider pauses.  Test case:  2015-01-23 Chart score percentiles.  [~daniel] expects median score mobile approximates web or better.  Got mobile was more than 20% lower.
     */
    onUpdate: function( dt )
    {
        var timer = this._components.GameComponentTimer;
        timer.setIsComponentPaused(
            transferring || !this.isCurrentRound() || this.model.isBlockingInput );
        this._super( dt );
        var consumed = timer.getTimeConsumed();
        if (this.model.isGameRunning)
        {
            this.model.timeConsumed = consumed;
            if (!this.model.currentRound)
            {
                this.timeBetweenRounds += dt;
            }
        }
        this.model.tutor.shared.update(consumed + this.timeBetweenRounds);

        var transferring = this.model.transferringTime();
        if ( this.model.transferTimeToPoints )
        {
            if ( transferring )
            {
                this.model.transferTime(dt);
            }
            else
            {
                this.model.transferTimeToPoints = false;
                this.view.playAnimation( "bonusPointsEnd" );
                timer._timeRemaining = 1;
                timer._timeDelta = 0;
            }
            this.view.onUpdateTransferTime();
        }
        else {
            if (this.block.unblock(consumed)) {
                if ( this.model.isBoardEmpty() )
                {
                    this.endCurrentRound();
                }
                else {
                    this.setTappablePieces();
                }
            }
        }
    },

    /**
     * Prepares next round ahead of time, unless it already exists.
     */
    prepareNextRound: function()
    {
        if ( undefined == this.model.nextRound )
        {
            var params = {};
            if (this.model.persistentData.isPracticeMode) {
            }
            else {
                // ll.verbose( "Disillusion", "prepareNextRound" );
                if ( lu.isShortGamesModeOn() )
                {
                    params.fillHeight = 1;
                    params.fillWidth = 2;
                    params.roundMax = 1;
                    params.fillCount = 2;
                }
            }
            this.model.prepareNextRound(params);
        }
    },

    /**
     * After new board is finished sliding into the view:
     * If not already constructed, generate next board.
     * Set up and mark current round as started.
     * Color the grey piece silhouettes.
     * Start the next trial.
     * GAME-1405 If round already started, do not start again.
     */
    startNextRound:function()
    {
        if (this.model && this.model.currentRound)
        {
            return;
        }
        this.prepareNextRound();
        if ( !this.view.isConstructed() )
        {
            this.view.constructBoard( this.model.nextRound.board );
        }
        var consumed = this._components.GameComponentTimer.getTimeConsumed();
        // ll.log( "Disillusion: startNextRound: time consumed " + consumed );
        this.model.startNextRound(consumed);

        var puzzlePieces = this.model.currentRound.board.getNonEmptyBoardPieces();
        this.view.updatePieces( puzzlePieces );

        this.startNextTrial();
    },

    /**
     * First round fades in current.  Other rounds fade in next.
     */
    cue_boardFadeIn: function()
    {
        // ll.verbose( "Disillusion", "boardFadeIn" );

        var fadeRound;
        if ( this.model.currentRound && this.model.currentRound.isFinished !== true )
        {
            fadeRound = this.model.currentRound;
        }
        else if ( this.model.nextRound )
        {
            fadeRound = this.model.nextRound;
        }
        if (fadeRound) {
            var piecesToFadeIn = fadeRound.board.getNonEmptyBoardPieces();
            this.view.fadePiecesIn( piecesToFadeIn );
        }
    },

    /**
     * GAME-1405 Guard start next trial, if not in a round.  Test case:  2014-10-23 iPad mini:  Last moment.  Answer at very last moment.  Out of time.  Next tile being generated.  Next tile overlaps.
     */
    isCurrentRound: function()
    {
        return this.model && this.model.isGameRunning && !this.model.isGameOver 
            && this.model.currentRound && !this.model.currentRound.isFinished 
            && this.view;
    },

    /**
     * Guard against double-call.
     * Clear tappable pieces first before ending, so that the current round is still defined.
     * Test case:  How to Play.  Tap.  Spots remain.
     * GAME-1602 At end of round, if any pieces remaining, fade them out.  In Boilder, I don't notice pieces fading out.  2014-10-23 [~wouter] expects to fade out remaining pieces.
     */
    endCurrentRound: function()
    {
        if ( this.isCurrentRound() )
        {
            this.view.showTappablePieces();
            var piecesToFadeOut = this.model.currentRound.board.getNonEmptyBoardPieces();
            this.view.fadePiecesOut( piecesToFadeOut, true, true );
            var consumed = this._components.GameComponentTimer.getTimeConsumed();
            // ll.log( "Disillusion.endCurrentRound: time consumed " + consumed );
            this.model.onEndCurrentRound(consumed);
            this.view.updateBonusPoints();
            if ( this.model.transferringTime() ) {
                this.view.playAnimation( "bonusPoints" );
            }
            else {
                this.animationFinished_bonusPointsEnd();
            }
        }
    },

    animationFinished_bonusPoints: function()
    {
        this.model.timeRemained = this.model.timeRemaining;
        this.model.transferTimeToPoints = true;
    },

    /**
     * End game or transit to next round.
     * GAME-1403 If continuing practice, animate checkmark.
     */
    animationFinished_bonusPointsEnd: function()
    {
        if ( this.model.wouldEndGame() )
        {
            this.onEndGameplay();
        }
        else
        {
            if (this.model.persistentData.isPracticeMode) {
                this.view.getWidget( GameWidgetFeedback.prototype.getType() ).play(true);
            }
            else 
            {
                this.view.playAnimation( "nextRound", this.startNextRound );
            }
        }
    },

    animationFinished_tutor_clear: function()
    {
        this.view.playAnimation( "nextRound", this.startNextRound );
    },

    /**
     * Starts new trial (new piece)
     * Show next piece
     * Generate and show tappable pieces
     * Stop input blocking
     */
    startNextTrial: function()
    {
        if ( !this.isCurrentRound() )
        {
            return;
        }

        if ( this.model.nextPiece === undefined )
        {
            this.model.prepareNextTrial();
        }

        var vertical = this.model.nextPiece.isVertical();
        this.model.tutor.vertical(vertical);
        this.setTappablePieces();
        this.view.setupNextPiece(this.model.nextPiece);
        var consumed = this._components.GameComponentTimer.getTimeConsumed();
        // ll.log( "Disillusion.startNextTrial: time consumed " + consumed );
        this.model.isBlockingInput = false;
        this.beginTrial();
    },

    setTappablePieces: function()
    {
        this.model.currentRound.board.generateTappablePieces( this.model.nextPiece );
        this.view.showTappablePieces();
    },

    /**
     *  Checks if user placed new piece in a correct place
     *
     *  @param newPiecePosition position where new piece was placed
     */
    checkAnswer: function( newPiecePosition )
    {
        // ll.verbose( "Disillusion", "checkAnswer" );

        var newPiece = this.model.currentRound.board.getPiece( newPiecePosition );
        var piecesToCheck = this.model.currentRound.board.getNeighborBoardPieces( newPiecePosition );
        var pieceToCheck;
        var piecesToShake = []; // case of failure
        var piecesToFadeToEmpty = []; // case of success
        var i;
        var vertical = newPiece.orientation === DisillusionPieceOrientation.Vertical;
        var attribute = vertical ? "pieceColor" : "pieceType";
        var correct = true;
        for ( i = 0; i < piecesToCheck.length; i++ )
        {
            pieceToCheck = this.model.currentRound.board.getPiece( piecesToCheck[ i ] );
            if ( pieceToCheck.isPuzzlePiece() )
            {
                if ( pieceToCheck[attribute] !== newPiece[attribute] )
                {
                    piecesToShake.push( piecesToCheck[i] );
                    correct = false;
                }
                else
                {
                    piecesToFadeToEmpty.push( piecesToCheck[i] );
                }
            }
        }
        this.model.lastPiecePlacedPosition = newPiecePosition;
        var consumed = this._components.GameComponentTimer.getTimeConsumed();
        this.model.lastTrialTimeConsumed = consumed;
        this.model.checkAnswer( correct, piecesToFadeToEmpty.length, vertical );
        this.endTrial( this.model.trial, correct, false, false, false, false );
        var audioPath;
        if ( correct )
        {
            // Reset the piece that was just placed
            newPiece.setEmpty();

            // Add newPiece to the fade out array
            piecesToFadeToEmpty.unshift( newPiecePosition );

            // Update pieces
            for ( i = 0; i < piecesToFadeToEmpty.length; i++ )
            {
                var pieceToFade = this.model.currentRound.board.getPiece( piecesToFadeToEmpty[ i ] );
                if ( pieceToFade )
                {
                    pieceToFade.setEmpty();
                    pieceToFade.toBeRemoved = true;

                }
            }

            // Update view with score
            this.view.setScore( this.model.score );

            // Fade out pieces
            this.model.numberOfPiecesFading = piecesToFadeToEmpty.length;
            this.view.fadePiecesOut( piecesToFadeToEmpty, true, true );
            audioPath = this.randomPlaceTileSound();
        }
        else
        {
            var boardObject = this.model.currentRound.board;
            this.block.block(boardObject, newPiece, consumed);

            for ( i = 0; i < piecesToShake.length; i++ )
            {
                this.view.showWrongSelectionFeedback( piecesToShake[i] );
            }

            this.model.prepareNextTrial();
            audioPath = "sounds/answerWrong.mp3";
        }
        audio.playEffect( audioPath );
    },

    /**
     * @return {string} Audio path to random tile, 1 to 5.
     */
    randomPlaceTileSound: function()
    {
        var r = DisillusionBoard.getRandomInt(1, 5);
        return "sounds/placeTile" + r + ".mp3";
    },

    /**
     *  Callback for when piece on a board is done shaking.
     */
    onShakePieceComplete: function()
    {
        if ( this.model.lastPiecePlacedPosition !== undefined )
        {
            // Update the board
            var gridPosition = this.model.lastPiecePlacedPosition;
            this.view.updatePiece( gridPosition );

            // Clear last piece placed
            this.model.lastPiecePlacedPosition = undefined;

            // Start next trial
            this.startNextTrial();
        }
    },

    /**
     *  Callback for when piece on a board is done fading in.
     */
    pieceView_onFadeInComplete: function( pieceModel )
    {
        this.model.numberOfPiecesFading--;

        if ( this.model.numberOfPiecesFading === 0 )
        {
            if ( this.model.isGameRunning )
            {
                this.model.currentRound.isStarted = true;
                this.startNextTrial();
            }
        }
    },

    /**
     *  Callback for when piece on a board is done fading out.
     * if pieces left, next trial; otherwise if no blocks, end game.
     * GAME-1405 When last piece has been removed, except there is still a block, then remove previous tappable pieces, while still waiting for block to be removed before ending the round.  Test case:  2014-10-23 Level 12.  2 tiles remaining that are diagonally adjacent.  Wrong.  Quickly match 2 tiles in less than 3 seconds.  Expect spots disappear.  Got spots remain until block is removed.
     *
     * After fading out the last piece, before asking current round, see if current round exists.
     * Test case:  2014-12-17 John Tighe noticed one Android user on Crashlytics expects no race between ending round and fading pieces out.  
     * Disillusion.js line 653 TypeError: this.model.currentRound is undefined
     * https://www.crashlytics.com/lumos-labs-inc/android/apps/com.lumoslabs.lumosity/issues/5491761865f8dfea155c7527/sessions/5491b13203ba000179c57d365b937757
     * GAME-2506 If all pieces blocked, unblock input to allow timer to unpause.  Test case:  2015-01-27 Block.  Only blocked tile remains.  
     */
    pieceView_onFadeOutComplete: function( pieceModel )
    {
        this.model.numberOfPiecesFading--;
        if (!this.isCurrentRound()) {
            return;
        }
        if (this.model.numberOfPiecesFading === 0 
         || this.model.numberOfPiecesFading < 0)
        {
            var blockCount = this.block.pieceViewsBlocked.length;
            var fillCount = this.model.currentRound.board.getNonEmptyCount();
            if (blockCount <= 0 && fillCount <= 0) 
            {
                this.endCurrentRound();
            }
            else if (blockCount < fillCount)
            {
                this.startNextTrial();
            }
            else if (blockCount == fillCount)
            {
                this.view.showTappablePieces();
                this.model.isBlockingInput = false;
            }
        }
    },

    pieceView_onSlideInComplete: function( pieceModel )
    {
        this.model.isSlidingIn = false;
        this.checkAnswer( pieceModel.boardPosition );
    },

    /**
     * GAME-1602 Workaround animation export: If not in level select mode, yet the button is enabled by the animation frames, then disable level select button and next button, as if no button were pressed.  Ideally, the button would not be enabled because it was not animated into view and was set to invisible during How To Play.  Steps to reproduce false positive case:  2014-10-20 Select mode How To Play.  Click Disillusion.  At first screen of tutorial, click at bottom center of screen.  Expect nothing happens.  Got HUD and Pause button appear and time number increases to many digits.  Steps to reproduce false negative case:  How To Play mode.  Skip Tutorial.  Level Select. Click Next.  Expect level buttons exit and see HUD.  Got nothing happens.
     *
     * [~eli] Have you seen this before?  Test case:  2014-10-21 Boiler.  How To Play.  Second board.  Tap area.  Expect place tile.  Got "btn_station_10" TypeError:  JSON.stringify cannot serialize cyclic structures.
     * GAME-1602 No next button.  Instead, tapping a level button starts the game.  2014-10-22 Unlike the web versions, for any mobile game with level select, Sushmita expects no Next button.
     * I tried to disable button down state by overriding view onTouchBegan, _handleTouchBegan, and _handleTouchEnded.  Each was fragile and awkward.  Test case:  2014-10-23 Tap level button.  Spam.  Expect transition without changing state.
     */
    _buttonReleased:function( buttonName )
    {
        // ll.verbose("Disillusion._buttonReleased: " + buttonName);
        if (this.model.isBlockingInput) 
        {
            return;
        }
        if (this.model.persistentData.isLevelSelectMode) {
            var selectedMilestone = this.view.selectLevelName(buttonName);
            if (1 <= selectedMilestone) 
            {
                this.model.setMilestone(selectedMilestone);
                this.selectLevel();
            }
            return;
        }
        switch ( buttonName ) 
        {
            case "btnNext":
                if (this.model.persistentData.isLevelSelectMode) {
                    this.selectLevel();
                }
                else {
                    ll.log("Disillusion._buttonReleased: btnNext is not visible.");
                }
                break;
            default:
                this._super( buttonName );
        }
    },

    /**
     * GAME-1405 Do not disable level buttons now, because their visual appearance would change and distract.
     */
    selectLevel: function()
    {
        this.model.isBlockingInput = true;
        this.view.selectLevel();
        audio.playEffect( this.randomPlaceTileSound() );
    },

    /** 
     * Prepare for the first round.
     * Enable starting play after playing "gameplay" animation.
     * GAME-1405 After removing level buttons from screen, disable level buttons.  Test case:  2014-10-24 Boiler.  Play.  Pause.  Click Restart.  Expect to restart.  Got level button intercepts click.
     */
    animationFinished_transition: function()
    {
        this.prepareNextRound();
        this.view.disableLevelButtons();
        this.view.showPrompt(true);
        this.view.playAnimation( "gameplay", this.onStartGameplayFromLevelSelect );
    },

    /**
     * Test case:  2014-11-06 Fade in.  Rapidly tap level.  Expect spots appear once.  Got double spots.  Tap.  Expect to place.  Got crash.
     */
    onStartGameplayFromLevelSelect: function()
    {
        this.model.persistentData.isLevelSelectMode = false;
        this.onStartGameplay();
    },

    /**
     * This cue sets the level.
     */
    cue_setLevel: function()
    {
        this.view.setLevel();
    },

    /**
     * This cue appears triggers generation of the playable board.
     */
    cue_generateBoardView: function()
    {
        // ll.verbose( "Disillusion", "cue_generateBoardView", "Cue called: generateBoardView" );

        if ( !this.model.nextRound || !this.model.nextRound.board )
        {
            this.prepareNextRound();
        }
        if ( this.model.nextRound && this.model.nextRound.board )
        {
            this.view.constructBoard( this.model.nextRound.board );
        }
        /*
        else if ( this.model.currentRound && this.model.currentRound.board )
        {
            this.view.constructBoard( this.model.currentRound.board );
        }
        */
        else
        {
            ll.error( "Disillusion", "cue_generateBoardView", "Error! Could not generate new board view" );
        }
    },

    /**
     * To avoid race with base class finish callback, wait until next animation finished to enable finishing the game.  The user might quit the game while looking at an illusion.
     */
    animationFinished_illusionBegin: function()
    {
        this.onFinishEnabled = true;
    },

    hideIllusionView: function()
    {
        if ( this.view )
        {
            this.view.playAnimation( "illusionEnd", this.onFinish );
            this.view.illusionSwipeView.show(false);
        }
    },

    previousIllusion: function()
    {
        this.view.illusionSwipeView.previousPage();
    },

    nextIllusion: function()
    {
        this.view.illusionSwipeView.nextPage();
    },

    interaction_received: function( interaction )
    {
        this._super( interaction, true );

        if( this._ignoreSwipes === true || this.model.isGamePaused === true )
        { 
            return; 
        }
        if ( !this.isGameRunning )
        {
            if ( interaction.getType() === InteractionType.Swipe )
            {
                if ( interaction.getSwipeDirection() === SwipeGestureDirection.Right )
                {
                    this.view.illusionSwipeView.previousPage();
                }
                else if ( interaction.getSwipeDirection() === SwipeGestureDirection.Left )
                {
                    this.view.illusionSwipeView.nextPage();
                }
            }
        }
    }
});
