/*global cc*/
/*global require*/
/*global ll*/
/*global lu*/
/*global DisillusionBoardPiece*/
/*global DisillusionPieceOrientation*/
/*global DisillusionPieceType*/

require( "src/Game/DisillusionBoardPiece.js" );
require( "src/Utility/LumosityUtils.js" );

var DisillusionBoard = cc.Class.extend(
{
    /** @lends DisillusionBoard.prototype */

    _boardHeight: 0,
    _boardWidth: 0,
    _boardSetup: [],
    model: undefined,

    _colorCount: 3,
    _patternCount: 3,

    _patternSelectionArray: [],
    _colorSelectionArray: [],

    /**
     * Disillusion Board
     *
     * @constructs
     */
    ctor: function( model )
    {
        this.model = model;

        this._boardSetup = [];
        this._colorCount = 3;
        this._patternCount = 3;

        this._patternSelectionArray = [];
        this._colorSelectionArray = [];
    },

    /**
     *  Creates a new board layout with a given size
     *
     *  @param round Round to create new board for.
     */
    generateNewBoard: function()
    {
        var colVertical = Math.random() < 0.5;
        if (undefined !== this.model.seedVerticalIndex) {
            colVertical = this.model.seedVerticalIndex != 1;
        }
        this._boardHeight = this.model.fillHeight + 2;
        this._boardWidth = this.model.fillWidth + 2;
        var fillCount = this.model.fillCount;
        /*
        ll.verbose( "generateNewBoard: fillCount: " + fillCount 
                + ", fillWidth: " + this.model.fillWidth 
                + ", fillHeight: " + this.model.fillHeight )
         */

        var x;
        this._boardSetup = [];
        for ( x = 0; x < this._boardWidth; x++ )
        {
            this._boardSetup[x] = [];
        }

        var minPuzzleX = Math.floor( (this._boardWidth - this.model.fillWidth) / 2 );
        var maxPuzzleX = minPuzzleX + this.model.fillWidth - 1;
        var minPuzzleY = Math.floor( (this._boardHeight - this.model.fillHeight) / 2 );
        var maxPuzzleY = minPuzzleY + this.model.fillHeight - 1;

        // ll.verbose( "minPuzzleX: " + minPuzzleX + ", maxPuzzleX: " + maxPuzzleX + ", minPuzzleY: " + minPuzzleY + ", maxPuzzleY: " + maxPuzzleY );

        var dominoCount = 0;
        var y;
        var colorSymbols;
        if (null != this.model.diagram) {
            colorSymbols = this.model.diagram.split(",");
        }
        for ( x = 0; x < this._boardWidth; x++ )
        {
            var rowVertical = colVertical;
            for ( y = 0; y < this._boardHeight; y++ )
            {
                var orientation;
                if ( rowVertical )
                {
                    orientation = "v";
                }
                else
                {
                    orientation = "h";
                }

                var pieceType;
                var pieceColor;
                var inBounds = ( x >= minPuzzleX && x <= maxPuzzleX ) 
                            && ( y >= minPuzzleY && y <= maxPuzzleY );
                if ( dominoCount < fillCount && inBounds )
                {
                    if (null == colorSymbols) {
                        pieceType = this._getNextRandomPieceType();
                        pieceColor = this._getNextRandomPieceColor();
                    }
                    else {
                        var colorSymbol = colorSymbols[dominoCount];
                        var spec = DisillusionBoardPiece.prototype.setSpec(colorSymbol);
                        orientation = spec.orientation;
                        pieceColor = spec.pieceColor;
                        pieceType = spec.pieceType;
                    }
                    dominoCount++;
                }
                else
                {
                    pieceType = DisillusionPieceType.Empty;
                }

                this._boardSetup[x][y] = new DisillusionBoardPiece( pieceType, pieceColor, orientation, cc.p( x, y ) );

                rowVertical = !rowVertical;
            }

            colVertical = !colVertical;
        }
    },

    /**
     * GAME-1405 Match Disillusion 1 algorithm:  For each tile in a board, roll a random symbol.  
     *
     * @returns {*} index [1..3] or null
     * @private
     */
    _getNextRandomPieceType: function()
    {
        return DisillusionBoard.getRandomInt(1, this._patternCount);
    },

    /**
     * GAME-1405 Match Disillusion 1 algorithm:  For each tile in a board, roll a random color.  
     *
     * @returns {*} index [1..3] or null
     * @private
     */
    _getNextRandomPieceColor: function()
    {
        return DisillusionBoard.getRandomInt(1, this._colorCount);
    },

    /**
     *  Adds tappable pieces to the board
     *
     */
    generateTappablePieces: function( nextPiece )
    {
        // ll.verbose( "DisillusionBoard", "generateTappablePieces" );

        if ( nextPiece )
        {
            var x;
            var y;
            var numTappablePieces = 0;

            for ( y = 0; y < this._boardHeight; y++ )
            {
                for ( x = 0; x < this._boardWidth; x++ )
                {
                    if ( this._boardSetup[x][y].isEmpty() && this._boardSetup[x][y].orientation === nextPiece.orientation )
                    {
                        // Check for neighbors
                        var neighbors = [];
                        if ( x > 0 && this._boardSetup[x-1][y].isPuzzlePiece() )
                        {
                            neighbors.push( this._boardSetup[x-1][y] );
                        }
                        if ( x < this._boardWidth - 1 && this._boardSetup[x+1][y].isPuzzlePiece() )
                        {
                            neighbors.push( this._boardSetup[x+1][y] );
                        }
                        if ( y > 0 && this._boardSetup[x][y-1].isPuzzlePiece() )
                        {
                            neighbors.push( this._boardSetup[x][y-1] );
                        }
                        if ( y < this._boardHeight - 1 && this._boardSetup[x][y+1].isPuzzlePiece() )
                        {
                            neighbors.push( this._boardSetup[x][y+1] );
                        }

                        var i;
                        var correctAnswer = true;
                        for ( i = 0; i < neighbors.length; i++ )
                        {
                            if ( nextPiece.orientation === DisillusionPieceOrientation.Horizontal && nextPiece.pieceType !== neighbors[ i ].pieceType )
                            {
                                correctAnswer = false;
                                break;
                            }
                            else if ( nextPiece.orientation === DisillusionPieceOrientation.Vertical && nextPiece.pieceColor !== neighbors[ i ].pieceColor )
                            {
                                correctAnswer = false;
                                break;
                            }
                        }

                        // If there is a neighbor, make this tappable
                        if ( neighbors.length > 0 )
                        {
                            this._boardSetup[x][y].setTappable( correctAnswer );
                            numTappablePieces++;
                        }
                    }
                }
            }
            // ll.verbose( "DisillusionBoard", "generateTappablePieces", "Marked " + numTappablePieces + " pieces as tappable." );
        }
        else
        {
            ll.error( "DisillusionBoard", "generateTappablePieces", "NextPiece is null, no tappable pieces will be generated." );
        }
    },

    /**
     * Clears tappable pieces from the board
     *
     */
    clearTappablePieces: function()
    {
        var x;
        var y;

        for ( y = 0; y < this._boardHeight; y++ )
        {
            for ( x = 0; x < this._boardWidth; x++ )
            {
                if ( this._boardSetup[x][y].isTappable() )
                {
                    this._boardSetup[x][y].setEmpty();
                }
            }
        }
    },


    getTappablePiecesThatSatisfy: function( nextPiece )
    {
        var tappablePieces = [];

        if ( nextPiece )
        {
            var x;
            var y;

            for ( y = 0; y < this._boardHeight; y++ )
            {
                for ( x = 0; x < this._boardWidth; x++ )
                {
                    if ( this._boardSetup[x][y].isTappable() && this._boardSetup[x][y].correctAnswer )
                    {
                        tappablePieces.push( cc.p( x, y ) );
                    }
                }
            }
        }

        return tappablePieces;
    },


    getTappablePiecesThatNotSatisfy: function( nextPiece )
    {
        var tappablePieces = [];

        if ( nextPiece )
        {
            var x;
            var y;

            for ( y = 0; y < this._boardHeight; y++ )
            {
                for ( x = 0; x < this._boardWidth; x++ )
                {
                    if ( this._boardSetup[x][y].isTappable() && this._boardSetup[x][y].correctAnswer === false )
                    {
                        tappablePieces.push( cc.p( x, y ) );
                    }
                }
            }
        }

        return tappablePieces;
    },


    /**
     *  Calculates what possible locations could be tapped,
     *  this will be used to determine what next piece should be
     * MOBGAMES-1799 Do not filter out existing pieces.  Test case:  Nov 3 15:33:53 LumosityiPad[112] : Cocos2d: JS: DisillusionBoard.findPossibleMatches: Could not find valid variations that are not present on the board. Nov 3 15:33:53 LumosityiPad[112] : Cocos2d: JS: Disillusion.prepareNextTrial: Error! Error! Cannot determine next piece.
     * @param   previousNextPiece   Exclude tiles that would match this off-match color or symbol.
     * GAME-2506 As a precaution, exclude tiles that would duplicate the previous beforehand to have same distribution as Flash version.  Test case:  2015-01-23 Chart median scores.  Daniel Sternberg expects approximately equal score.
     *
     *  @return Array of possible next pieces.
     */
    findPossibleMatches:function(previousNextPiece)
    {
        var pieces = [];

        var x;
        var y;
        var i;

        for ( y = 0; y < this._boardHeight; y++ )
        {
            for ( x = 0; x < this._boardWidth; x++ )
            {
                var piece = this._boardSetup[x][y];
                if ( piece.isEmpty() || piece.isBlocked() )
                {
                    var neighborPositions = this.getNeighborBoardPieces( cc.p( x, y ) );
                    if ( neighborPositions.length > 0 )
                    {
                        var validNeighborCount = 0;
                        var types = [];
                        var colors = [];
                        for ( i = 0; i < neighborPositions.length; i++ )
                        {
                            var neighbor = this.getPiece( neighborPositions[i] );
                            if ( neighbor && neighbor.isPuzzlePiece() )
                            {
                                validNeighborCount++;
                                if ( types.indexOf( neighbor.pieceType ) === -1 )
                                {
                                    types.push( neighbor.pieceType );
                                }
                                if ( colors.indexOf( neighbor.pieceColor ) === -1 )
                                {
                                    colors.push( neighbor.pieceColor );
                                }
                            }
                        }
                        var orientation;
                        var pieceType;
                        var pieceColor;
                        var n;
                        if ( validNeighborCount > 0 )
                        {
                            if ( piece.isHorizontal() && types.length === 1 )
                            {
                                for ( i = 1; i <= this._colorCount; i++ )
                                {
                                    orientation = DisillusionPieceOrientation.Horizontal;
                                    pieceType = types[ 0 ];
                                    pieceColor = i;
                                    if (!previousNextPiece || previousNextPiece.pieceColor != pieceColor) {
                                        pieces.push( new DisillusionBoardPiece( pieceType, pieceColor, orientation ) );
                                    }
                                }
                            }
                            else if ( piece.isVertical() && colors.length === 1 )
                            {
                                for ( i = 1; i <= this._patternCount; i++ )
                                {
                                    orientation = DisillusionPieceOrientation.Vertical;
                                    pieceType = i;
                                    pieceColor = colors[ 0 ];

                                    if (!previousNextPiece || previousNextPiece.pieceType != pieceType) {
                                        pieces.push( new DisillusionBoardPiece( pieceType, pieceColor, orientation ) );
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        if ( pieces.length <= 0 )
        {
            ll.error( "DisillusionBoard", "findPossibleMatches", 
                "Could not find valid variations that are not present on the board." );
        }
        return pieces;
    },

    getPiece: function( boardPosition )
    {
        var retval = undefined;

        if ( boardPosition.x >= 0 && boardPosition.x < this._boardWidth && boardPosition.y >= 0 && boardPosition.y < this._boardHeight )
        {
            retval = this._boardSetup[ boardPosition.x ][ boardPosition.y ];
        }
        else
        {
            ll.error( "DisillusionBoard", "getPiece", "Count not find board piece at (" + boardPosition.x + ", " + boardPosition.y + ")" );
        }

        return retval;
    },


    /**
     * Returns all of the board pieces in an array
     *
     * @returns {Array}
     */
    getAllBoardPieces: function()
    {
        var retval = [];

        if ( this._boardSetup.length > 0 )
        {
            var x;
            var y;

            for ( y = 0; y < this._boardSetup[0].length; y++ )
            {
                for ( x = 0; x < this._boardSetup.length; x++ )
                {
                    retval.push( this._boardSetup[x][y] );
                }
            }
        }

        return retval;
    },


    /**
     *  Calculates number of non empty board tiles
     *
     *  @return number of non empty board tiles
     */
    getNonEmptyCount: function()
    {
        var nonEmptyCount = 0;

        if ( this._boardSetup.length > 0 )
        {
            var x;
            var y;

            for ( y = 0; y < this._boardSetup[0].length; y++ )
            {
                for ( x = 0; x < this._boardSetup.length; x++ )
                {
                    if ( !this._boardSetup[x][y].isEmpty() )
                    {
                        nonEmptyCount++;
                    }
                }
            }
        }

        // ll.verbose( "DisillusionBoard", "getNonEmptyCount", "Found non empty " + nonEmptyCount + " board pieces." );
        return nonEmptyCount;
    },

    /**
     *  Generates array of non empty board tiles
     *
     *  @return array of non empty board tiles
     */
    getNonEmptyBoardPieces: function()
    {
        var retval = [];

        if ( this._boardSetup.length > 0 )
        {
            var x, y;
            for ( y = 0; y < this._boardSetup[0].length; y++ )
            {
                for ( x = 0; x < this._boardSetup.length; x++ )
                {
                    if ( !this._boardSetup[x][y].isEmpty() )
                    {
                        retval.push( cc.p( x, y ) );
                    }
                }
            }
        }

        return retval;
    },

    /**
     *  Generates array of empty board tiles
     *
     *  @return array of empty board tiles
     */
    getEmptyBoardPieces: function()
    {
        var retval = [];

        if ( this._boardSetup.length > 0 )
        {
            var x, y;
            for ( y = 0; y < this._boardSetup[0].length; y++ )
            {
                for ( x = 0; x < this._boardSetup.length; x++ )
                {
                    if ( this._boardSetup[x][y].isEmpty() )
                    {
                        retval.push( cc.p( x, y ) );
                    }
                }
            }
        }

        return retval;
    },

    /**
     *  Generates array of neighbor board tiles
     *
     *  @return array of neighbor board tiles
     */
    getNeighborBoardPieces: function( boardPosition )
    {
        var retval = [];

        // Top Row
        if ( boardPosition.y > 0 )
        {
            retval.push( cc.p( boardPosition.x, boardPosition.y - 1 ) );
        }

        // Mid Row
        if ( boardPosition.x > 0 )
        {
            retval.push( cc.p( boardPosition.x - 1, boardPosition.y ) );
        }

        if ( boardPosition.x < this._boardSetup.length - 1 )
        {
            retval.push( cc.p( boardPosition.x + 1, boardPosition.y ) );
        }

        // Bottom Row
        if ( boardPosition.y < this._boardSetup[boardPosition.x].length - 1 )
        {
            retval.push( cc.p( boardPosition.x, boardPosition.y + 1 ) );
        }

        return retval;
    },

    getPieceNeighbors: function( pieceModel )
    {
        var neighbors = [];

        var pieceToAdd = undefined;

        if ( pieceModel )
        {
            // left
            if ( pieceModel.boardPosition.x > 0 )
            {
                pieceToAdd = this.getPiece( cc.p( pieceModel.boardPosition.x - 1, pieceModel.boardPosition.y ) );
                if ( pieceToAdd )
                {
                    neighbors.push( pieceToAdd );
                }
            }

            // right
            if ( pieceModel.boardPosition.x < this._boardWidth - 1 )
            {
                pieceToAdd = this.getPiece( cc.p( pieceModel.boardPosition.x + 1, pieceModel.boardPosition.y ) );
                if ( pieceToAdd )
                {
                    neighbors.push( pieceToAdd );
                }
            }

            // up
            if ( pieceModel.boardPosition.y > 0 )
            {
                pieceToAdd = this.getPiece( cc.p( pieceModel.boardPosition.x, pieceModel.boardPosition.y - 1 ) );
                if ( pieceToAdd )
                {
                    neighbors.push( pieceToAdd );
                }
            }

            // down
            if ( pieceModel.boardPosition.y < this._boardHeight - 1 )
            {
                pieceToAdd = this.getPiece( cc.p( pieceModel.boardPosition.x, pieceModel.boardPosition.y + 1 ) );
                if ( pieceToAdd )
                {
                    neighbors.push( pieceToAdd );
                }
            }
        }

        return neighbors;
    },



    getTopLeftTile: function()
    {
        "use strict";

        return this.getPiece( cc.p( 0, this._boardHeight - 1 ) );
    }
});

/**
 * GAME-2506 As a precaution, replace Lumosity utility random number generator (based on trigonometric transformation) with standard RNG, same as Flash version:  Math.random.  Test case:  2015-01-23 Chart median scores.  Daniel Sternberg expects approximately equal score.
 */
DisillusionBoard.getRandomInt = function(min, max)
{
    return Math.floor(Math.random() * (max - min + 1) + min);
}

/**
 * GAME-2506 As a precaution, replace Lumosity utility random number generator (based on trigonometric transformation) with standard RNG, same as Flash version:  Math.random.  Test case:  2015-01-23 Chart median scores.  Daniel Sternberg expects approximately equal score.
 */
DisillusionBoard.shuffle = function(anArray)
{
    for (var i = anArray.length - 1; 1 <= i; i--) {
        var r = Math.floor(Math.random() * i + 1);
        var swap = anArray[i];
        anArray[i] = anArray[r];
        anArray[r] = swap;
    }
}
