var require = require;

var cc = cc;

var DisillusionPieceType =
{
    Empty: 0,

    Type1: 1,
    Type2: 2,
    Type3: 3,
    Type4: 4,
    Type5: 5,

    Blocked: 98,
    Tappable: 99
};

var DisillusionPieceOrientation =
{
    Horizontal: "h",
    Vertical: "v"
};

var DisillusionBoardPiece = cc.Class.extend(
{
    /** @lends DisillusionBoardPiece.prototype */

    pieceType : DisillusionPieceType.Empty,
    pieceColor : 0,

    orientation : DisillusionPieceOrientation.Horizontal,

    boardPosition: cc.p( 0, 0 ),

    isRippled: false,
    toBeRemoved : false,

    correctAnswer: false,

    /**
     * Disillusion Board Piece
     *
     * @constructs
     */
    ctor: function( pieceType, pieceColor, pieceOrientation, boardPosition )
    {
        this.pieceType = pieceType;
        this.pieceColor = pieceColor;
        this.orientation = pieceOrientation;

        this.boardPosition = boardPosition;

        this.isRippled = false;
        this.toBeRemoved = false;

        this.correctAnswer = false;
    },

    setEmpty: function()
    {
        this.pieceType = DisillusionPieceType.Empty;
    },

    isEmpty: function()
    {
        return this.pieceType === DisillusionPieceType.Empty;
    },

    setTappable: function( correctAnswer )
    {
        this.pieceType = DisillusionPieceType.Tappable;
        this.correctAnswer = correctAnswer;
    },

    isTappable: function()
    {
        return this.pieceType === DisillusionPieceType.Tappable;
    },

    isPuzzlePiece: function()
    {
        return !this.isEmpty() && !this.isBlocked() && !this.isTappable();
    },

    /**
     * GAME-1400 Cannot tap or spawn a puzzle piece here.  
     * Set timer to auto remove after 5 seconds.
     */
    block: function(consumed)
    {
        var blockTime = 5.0;
        this.pieceType = DisillusionPieceType.Blocked;
        this.timeConsumedRemove = consumed + blockTime;
        // ll.log("DisillusionBoardPiece.block: " + consumed + " until " + this.timeConsumedRemove);
    },

    unblock: function(consumed)
    {
        var unblockNow = false;
        if (this.isBlocked()) {
            if (this.timeConsumedRemove < consumed) {
                // ll.log("DisillusionBoardPiece.unblock: " + consumed + " due " + this.timeConsumedRemove);
                this.pieceType = DisillusionPieceType.Empty;
                unblockNow = true;
            }
        }
        return unblockNow;
    },

    isBlocked: function()
    {
        return this.pieceType === DisillusionPieceType.Blocked;
    },

    isHorizontal: function()
    {
        return this.orientation === DisillusionPieceOrientation.Horizontal;
    },

    isVertical: function()
    {
        return this.orientation === DisillusionPieceOrientation.Vertical;
    },

    /**
     * Notation: color-orientation symbol
     * 
     * rgy colors
     * Tall: R
     * Wide: r
     * +@* symbols
     * , tile delimiter
     * 
     * Examples:
     * 
     * b+,R@,y*,R*,r@,B+
     * 
     * y*,Y+,b@,B*,R+,y+
     * 
     * @param   colorSymbol     Color-orientation symbol.  Examples "b+", "R@".
     * @param   target  {Object} Modified in place if provided.  
     * @return  target  {Object} Object containing these keys: {orientation, pieceColor, pieceType}.
     *
     * @static
     */
    setSpec: function(colorSymbol, target)
    {
        // ll.verbose("DisillusionBoardPiece.setSpec: " + colorSymbol);
        var colorSymbols = colorSymbol.split("");
        var color = colorSymbols[0].toLowerCase();
        var colorIndex = this.colors.indexOf(color);
        var shape = colorSymbols[1];
        var shapeIndex = this.shapes.indexOf(shape);
        var target = target || {};
        target.orientation = 0 <= this.verticals.indexOf(colorSymbols[0]) 
            ? DisillusionPieceOrientation.Vertical
            : DisillusionPieceOrientation.Horizontal;
        target.pieceColor = colorIndex + 1;
        target.pieceType = shapeIndex + 1;
        return target;
    },

    verticals: "RBY".split(""),
    colors: "rby".split(""),
    shapes: "@+*".split(""),

    /**
     * @return color Symbol specification.  If empty, return "".
     */
    toSpec: function()
    {
        if (this.isEmpty())
        {
            return "";
        }
        var shape = this.shapes[this.pieceType - 1];
        var orientationColors = this.isVertical() ? this.verticals : this.colors;
        var color = orientationColors[this.pieceColor - 1];
        return color + shape;
    },

    toString: function()
    {
        var retval = "";

        if ( this.isHorizontal() )
        {
            retval += "horizontal";
        }
        else if ( this.isVertical() )
        {
            retval += "vertical";
        }

        retval += " type: " + this.pieceType;
        retval += " color: " + this.pieceColor;

        if ( this.boardPosition )
        {
            retval += " position: ( " + this.boardPosition.x + ", " + this.boardPosition.y + " )";
        }

        if ( this.pieceType === DisillusionPieceType.Tappable )
        {
            retval += " correctAnswer " + this.correctAnswer;
        }

        return retval;
    }
});
