/*global require*/
/*global cc*/
/*global LumosityButton*/
/*global LumositySprite*/
/*global LumosityUtils*/
/*global DisillusionBoardPiece*/
/*global DisillusionPieceOrientation*/


var DisillusionBoardPieceView;
(function()
{
    "use strict";

    require( "src/View/LumosityButton.js" );
    require( "src/View/LumositySprite.js" );
    require( "src/Utility/LumosityUtils.js" );
    require( "src/Game/DisillusionBoardPiece.js" );

    DisillusionBoardPieceView = cc.Node.extend(
    {
        /** @lends DisillusionBoardPieceView.prototype */

        blockedFrames: undefined,
        pieceModel: null,

        _backgroundSprite: null,
        _symbolSprite: null,

        _settings: null,
        _delegate: null,

        designSize: cc.size ( 140, 140 ),

        PuzzleBackgroundImageName: "pieceShape.png",
        NonPuzzleBackgroundImageName: "pieceDimArea.png",

        /**
         * @static
         */
        addSpriteFrames: function()
        {
            cc.SpriteFrameCache.getInstance().addSpriteFrames( "piece.plist", "piece.png" );
            this.blockedFrames = {};
            this.blockedFrames[ DisillusionPieceOrientation.Vertical ] = "pieceBlockVertical.png";
            this.blockedFrames[ DisillusionPieceOrientation.Horizontal ] = "pieceBlockHorizontal.png";
        },

        /**
         * Tile.
         *
         * @constructs
         */
        ctor: function( pieceModel, boardPosition, designSize, settings, delegate )
        {
            this._super();

            this._settings = settings;
            this._delegate = delegate;

            this.nodeName = "boardNode" + boardPosition.x + "x" + boardPosition.y;

            // Model
            this.pieceModel = pieceModel;
            this.pieceModel.pieceView = this;

            this.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this.setContentSize( designSize );

            var spritePosition = cc.p( designSize.width / 2, designSize.height / 2 );

            // Background Sprite
            var backgroundImage;
            if ( !pieceModel.isPuzzlePiece() )
            {
                backgroundImage = "pieceDimArea.png";
            }
            else
            {
                backgroundImage = this._settings.getDarkBoardPieceImage( pieceModel.pieceType );
            }

            this._backgroundSprite = cc.Sprite.createWithSpriteFrameName ( backgroundImage );
            this._backgroundSprite.nodeName = "BackgroundSprite";
            this._backgroundSprite.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this._backgroundSprite.setPosition( spritePosition );
            this.addChild( this._backgroundSprite );

            // Sprite image
            var symbolImage;
            if ( !pieceModel.isPuzzlePiece() )
            {
                symbolImage = "pieceDimArea.png";
            }
            else
            {
                symbolImage = this._settings.getBoardPieceImage( pieceModel.pieceType );
            }

            this._symbolSprite = cc.Sprite.createWithSpriteFrameName ( symbolImage );
            this._symbolSprite.nodeName = "SymbolSprite";
            this._symbolSprite.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this._symbolSprite.setPosition( spritePosition );
            this.addChild( this._symbolSprite );

            // Sprite color
            var color = cc.WHITE;
            if ( pieceModel.isPuzzlePiece() )
            {
                color = this._settings.getBoardPieceColor( pieceModel.pieceColor );
            }
            this._backgroundSprite.setColor( color );

            // Starting opacity
            if ( !pieceModel.isPuzzlePiece() )
            {
                this._symbolSprite.setOpacity ( 0 );
            }
            else
            {
                this._backgroundSprite.setOpacity( 255 );
                this._symbolSprite.setOpacity( 0 );
            }

            // Orientation
            if ( pieceModel.orientation === DisillusionPieceOrientation.Vertical )
            {
                this.setRotation( 0 );
            }
            else
            {
                this.setRotation( 90 );
            }

            this.scaleForContentSize( this._backgroundSprite, this.designSize );
            this.scaleForContentSize( this._symbolSprite, this.designSize );
        },

        /**
         * GAME-1400 If blocked, show vertical or horizontal image as symbol.
         */
        updateFromModel: function()
        {
            var symbolImage;
            var backgroundImage;
            var backgroundColor;
            var opacity;

            if ( !this.pieceModel.isPuzzlePiece() )
            {
                if ( this.pieceModel.isBlocked() ) {
                    var orientation = this.pieceModel.orientation;
                    var images = this.blockedFrames;
                    symbolImage = images[orientation];
                    opacity = 255;
                }
                else {
                    symbolImage = "pieceDimArea.png";
                    opacity = 0;
                }
                backgroundImage = "pieceDimArea.png";
                backgroundColor = cc.WHITE;
            }
            else
            {
                symbolImage = this._settings.getBoardPieceImage( this.pieceModel.pieceType );
                backgroundImage = "pieceShape.png";
                backgroundColor = this._settings.getBoardPieceColor( this.pieceModel.pieceColor );
                opacity = 255;
            }

            if ( this._symbolSprite )
            {
                this._symbolSprite.setDisplayFrame( cc.SpriteFrameCache.getInstance().getSpriteFrame( symbolImage ) );
                this._symbolSprite.setOpacity( opacity );
            }

            if ( this._backgroundSprite )
            {
                this._backgroundSprite.setDisplayFrame( cc.SpriteFrameCache.getInstance().getSpriteFrame( backgroundImage ) );
                this._backgroundSprite.setColor( backgroundColor );
                this._backgroundSprite.setOpacity( opacity );
            }
        },

        setOpacity: function( opacity )
        {
            if ( this._symbolSprite )
            {
                this._symbolSprite.setOpacity( opacity );
            }

            if ( this._backgroundSprite )
            {
                this._backgroundSprite.setOpacity( opacity );
            }
        },

        fadeIn: function()
        {
            var fadeDuration = 0.05;

            // Symbol
            //var symbolFadeAnimation = cc.FadeTo.create( fadeDuration, 255 );
            //var finalSequence = cc.Sequence.create( symbolFadeAnimation, callbackAction );
            //this._symbolSprite.runAction( finalSequence );

            // Background
            var backgroundFadeAnimation = cc.FadeTo.create( fadeDuration, 255 );
            var callbackAction = cc.CallFunc.create( this._onFadeInComplete, this );
            var finalSequence = cc.Sequence.create( backgroundFadeAnimation, callbackAction );
            this._backgroundSprite.runAction( finalSequence );
        },

        _onFadeInComplete: function()
        {
            if ( this._delegate && this._delegate.pieceView_onFadeInComplete )
            {
                this._delegate.pieceView_onFadeInComplete( this.pieceModel );
            }
        },

        fadeOut: function( callDelegate, updateFromModel )
        {
            if ( callDelegate === undefined )
            {
                callDelegate = true;
            }

            if ( updateFromModel === undefined )
            {
                updateFromModel = false;
            }

            var fadeDuration = 0.05;

            // Symbol
            var symbolFadeAnimation = cc.FadeTo.create( fadeDuration, 0 );
            this._symbolSprite.runAction( symbolFadeAnimation );

            // Background
            var backgroundFadeAnimation = cc.FadeTo.create( fadeDuration, 0 );
            this._backgroundSprite.runAction( backgroundFadeAnimation );

            if ( callDelegate )
            {
                var waitAction = cc.DelayTime.create( fadeDuration );
                var callbackAction = cc.CallFunc.create( this._onFadeOutComplete, this );
                var finalSequence = cc.Sequence.create( waitAction, callbackAction );

                if ( updateFromModel )
                {
                    var updateFromModelAction = cc.CallFunc.create( this.updateFromModel, this );
                    finalSequence = cc.Sequence.create( finalSequence, updateFromModelAction );
                }

                this.runAction( finalSequence );
            }
        },

        _onFadeOutComplete: function()
        {
            if ( this._delegate && this._delegate.pieceView_onFadeOutComplete )
            {
                this._delegate.pieceView_onFadeOutComplete( this.pieceModel );
            }
        },

        /**
         * GAME-1581 Tap spot.  Immediately check answer; previously the program constructed a slide in animation.  Test case:  2014-09-30 Tap spot.  [~wouter] expects no tile slide in.
         */
        slideIn: function()
        {
            this.updateFromModel();

            var positionOffset = cc.p( 0, -10 );
            var position = this.getPosition();
            position.y -= positionOffset.y;
            this.setPosition( position );

            var moveAction = cc.MoveBy.create( 0.05, positionOffset );
            var moveActionEase = cc.EaseIn.create( moveAction, 2 );
            var callbackAction = cc.CallFunc.create( this._onSlideInComplete, this );
            var finalSequence = cc.Sequence.create( moveActionEase, callbackAction );
            this.runAction( finalSequence );
        },

        _onSlideInComplete: function()
        {
            if ( this._delegate && this._delegate.pieceView_onSlideInComplete )
            {
                this._delegate.pieceView_onSlideInComplete( this.pieceModel );
            }
        },

        scaleForContentSize: function( node, size )
        {
            var currentSize = node.getContentSize();
            var currentScale = cc.p( 1, 1 );

            if ( currentSize.x !== 0 && currentSize.y !== 0 )
            {
                node.setScaleX( currentScale.x * ( size.width / currentSize.width ) );
                node.setScaleY( currentScale.y * ( size.height / currentSize.height ) );
            }
        }
    });
}());
