/*global metadataAggregateKeys*/

var require = require;

require( "src/Utility/LumosityUtils.js" );
require( "src/Game/Base/GameModel.js" );
require( "src/Game/DisillusionTutor.js" );

var GameModel = GameModel;
var lu = LumosityUtils;

var DisillusionModel = GameModel.extend(
{
    /** @lends DisillusionModel.prototype */

    lastPiecePlacedPosition: {},

    nextPiece: {},
    previousNextPiece: null,

    bonus: 0,
    score: 0,
    bonusScorePoints: 0,

    fillCount: 0,
    fillHeight: 0,
    fillWidth: 0,
    fills: [],
    firstLevel: 12,
    /**
     * Maximum float that resolves to 4 significant digits is  2 ** (53 - 10), since 2 ** 10 = 1024.
     * 2 ** 43 seconds is over 200,000 years.
     * http://stackoverflow.com/questions/307179/what-is-javascripts-max-int-whats-the-highest-integer-value-a-number-can-go-t
     */
    MILLISECOND_MAX_VALUE: Math.pow(2, 53 - 10),

    timeRemaining: 60,
    timeRemained: 60,

    ranOutOfTime: false,

    numberOfPiecesFading: 0,

    currentLevel: 1,
    session_level: 1,
    user_level: 1,

    currentRound: undefined,
    nextRound: undefined,

    difficulty: 0,
    milestoneParams: [],
    tutorialParams: [],
    roundCount: 0,
    roundMax: 4,
    maxLevel: 96,

    scorePerTime: 100,
    scorePerTrial: 1000,
    secondsPerMinute: 60,
    seconds: 60,
    trialsCount: 0,

    lastTrialTimeConsumed: 0.0,

    currentIllusion: 0,
    unknown: "(?)",

    /**
     * GAME-1405 Latest results of "{}".  Test case:  2014-10-28 Boiler.  How to Play.  Quit.  Boiler expects JSON string to parse.  Got "undefined".
     * GAME-2504 If first time playing, load first level.  Test case:  2015-02-04 Locomotive.  Tap "Reset game data".  Tap Disillusion.  Skip Tutorial.  Look at level select.  Expect first button is highlighted.  Got last button is highlighted with text "NaN".
     * @constructs
     */
    ctor:function( settings )
    {
        this._super();

        this.firstLevel = settings.firstLevel;
        this.fills = settings.fills;
        this.milestoneParams = settings.difficultyProgressionSettings;
        this.tutorialParams = settings.tutorialParams;
        this.seconds = settings.getGameLengthInSeconds();
        this.secondsOriginal = settings.gameLengthInSeconds;

        this.nextPiece = undefined;
        this.num_correct = 0;
        this.difficulty = 0;
        this.roundCount = 0;
        this.fillCount = 0;
        this.fillHeight = 0;
        this.fillWidth = 0;
        this.latestResults = undefined === this.latestResults ? "{}" : this.latestResults;

        this.trialsCount = 0;

        this.lastTrialTimeConsumed = 0.0;

        this.currentIllusion = 0;
        this.initSessionLevel(this.loadFromLastResults(this.firstLevel));
        this.tutor = new DisillusionTutor();
    },

    /**
     * GAME-1403 At least first milestone.  2014-10-16 Finish session.  Do not level up.  Start next session.  Expect to fix first milestone available.  Got no milestone available.
     */
    initSessionLevel: function( newSessionLevel )
    {
        this.session_level = Math.max(this.firstLevel, this.session_level, 
            Math.min(newSessionLevel, this.maxLevel));
        this.currentLevel = this.session_level;
        this.user_level = this.session_level;
    },

    /**
     * Increment trials counts.  Add to score for each piece removed.
     * GAME-1403 As in web version, if wrong and at low level or in first round, then display prompt again.  2014-10-23 Wrong.  Ashley expects reminder of rules.
     * @param   correct {[boolean]}
     * @param   piecesRemoved {[int]}
     * @param   vertical {[boolean]}
     * GAME-1404 If placed piece does not match, record one error.  At beginning of a session, or practice, reset to 0.  2014-10-24 Aaron expects to record num_errors.  
     * GAME-1404 #comment Match number correct adds one for each piece matched.  Example:  42 tiles placed which matched 48 tiles.
     */
    checkAnswer: function( correct, piecesRemoved, vertical )
    {
        this.trialsCount++;
        this.currentRound.num_tries++;
        if (correct) 
        {
            this.num_correct += piecesRemoved;
            this.currentRound.num_correct += piecesRemoved;
            this.score += piecesRemoved * this.scorePerTrial;
            this.tutor.correct(vertical);
            this.tutor.shared.prepare = false;
        }
        else 
        {
            //- this.metadata.num_errors++;
            if (this.persistentData.isPracticeMode) {
                this.tutor.wrong(vertical);
            }
            if (this.prepare()) {
                this.tutor.shared.prepare = true;
            }
        }
        this._recordTrialMetadata(correct);
    },

    /**
     * 2014-10-24 Aaron expects position row, column string.
     */
    _recordTrialMetadata: function(correct)
    {
        this.trial = {};
        this.trial.correct = correct ? "T" : "F";
        this.trial.position = this.lastPiecePlacedPosition.y.toString()
            + this.lastPiecePlacedPosition.x.toString();
        var piece = this.currentRound.board.getPiece( this.lastPiecePlacedPosition );
        this.trial.piece = piece.toSpec();
        this.trial.round = this.roundCount;
        this.trial.vertical = piece.isVertical() ? "T" : "F";
    },

    /**
     * WG-1275 More conservatively estimates next board tiles and next session level.  Increase ease from 1 to 3, which reduces confidence when few tiles are removed.  Hopefully this reduces rubberbanding of scores between sessions.  2014-08-04 scale group has more jagged score percentile per session.
     */
    confidence: function(num_correct)
    {
        var ease = // 1;
                       // 2;
                       3;
        return Math.max(1, num_correct - ease) / Math.max(1, num_correct);
    },

    wouldEndGame: function()
    {
        var wouldEnd;
        if (this.persistentData.isPracticeMode) {
            wouldEnd = this.tutorialParams.length <= this.tutor.roundIndex;
        }
        else {
            wouldEnd = this.ranOutOfTime || this.complete();
        }
        return wouldEnd;
    },

    /**
     * GAME-1405 Continue round when time is up, yet tapped piece is still sliding in.  2014-10-31 Test case:  Short game.  Wait until 1 second remaining.  Tap a tile just in time.  Animation might not complete.  Popup score 1000.  Game results.  Derek Bender could expect 1000.  Got 0.
     */
    isBoardEmpty: function()
    {
        return this.currentRound.board.getNonEmptyCount() <= 0 
            && !this.isSlidingIn;
    },

    /**
     * Creates a results JSON string to send back to the application layer.
     * Save next session level to gameStats.user_level.  
     * Analogous to Penguin Pursuit (endingLevelIndex) and Train of Thought (sessionLevel), yet their keys differ.
     * A standard key would be more clear.
     *
     * Following Memory Matrix, Disillusion also exports total tiles cleared into metadata.
     * Test case: Lumosity Mobile.  Play Disillusion.  Match 2 tiles.  End.  Results screen.  Expect to read "2 Tiles".
     *
     * @param  {Object} stats An object that contains keys describing particular stats.
     * @return {String}       String version of results object.
     *
     * GAME-2418 Post game, display minimum of tiles per minute or next session level; was number of tiles cleared.  Test case:  2014-12-29 Play several sessions.  Post game.  Remember Best Tiles and Past Best Tiles.  Replay game.  Observe current level.  John Tighe expects current maximum of Best Tiles and Past Best Tiles to be lower or equal to current level.  Got Past Best Tiles was higher.
     */
    exportResults: function( stats, bestStatKey )
    {
        if ( !stats )
        {
            stats = {};
        }
        stats[ this.StatsKeys.TilesStat ] = Math.min(this.user_level, this.metadata.performance_level);
                                            //- this.num_correct;
        return this._super( stats, this.StatsKeys.TilesStat );
    },

    configure: function(tilesPerMinute)
    {
        var params = this.fillCountPerRound(tilesPerMinute);
        params.level = tilesPerMinute;
        return params;
    },

    fillCountPerRound: function(tilesPerMinute)
    {
        var params = {};
        var fillCount = Math.ceil(tilesPerMinute / this.roundMax);
        params.fillCount = Math.min(fillCount, this.fillMax());
        return params;
    },

    /**
     * Expects width and height are transposed in settings, such that dimensions are tall.  Mobile game has portrait aspect ratio.  Web game has landscape aspect ratio.
     */
    fill: function(params, tileCount)
    {
        for (var f = 0; f < this.fills.length; f++) {
            var aFill = this.fills[f];
            var area = aFill.fillWidth * aFill.fillHeight;
            if (tileCount <= area) {
                params.fillWidth = aFill.fillWidth;
                params.fillHeight = aFill.fillHeight;
                break;
            }
        }
        if (this.fills.length <= f) {
            throw new Error("Expected to find a fill for " + tileCount + " tiles.");
        }
        return params;
    },

    fillMax: function()
    {
        var max = this.fills[this.fills.length - 1];
        var area = max.fillWidth * max.fillHeight;
        return area;
    },

    initTutor: function(prepare)
    {
        prepare = prepare === undefined ? this.prepare() : prepare;
        this.tutor.init();
        this.tutor.shared.active = this.persistentData.isPracticeMode;
        this.tutor.shared.intro = this.persistentData.isPracticeMode;
        this.tutor.shared.prepare = prepare;
        this.tutor.correctMin = this.persistentData.isPracticeMode ? 1 : 2;
    },

    levelHintMax: function()
    {
        return this.tilesPerMinuteMin(2);
    },

    /**
     * After overriding fill count, pick the closest width and height to the aspect ratio.
     */
    loadLevel: function(level, overrideParams)
    {
        var params = this.configure(level);
        if (null != overrideParams) {
            for (var param in overrideParams) {
                params[param] = overrideParams[param];
            }
        }
        this.fill(params, params.fillCount);
        this.populate(params);
    },

    complete: function()
    {
        return this.roundMax <= this.roundCount;
    },

    getUserLevel: function()
    {
        var nextSessionLevel = this.session_level;
        if (this.session_level <= this.currentLevel) {
            nextSessionLevel = this.scaleLevelUp(0, this.num_correct, this.lastTrialTimeConsumed);
            if (nextSessionLevel < this.session_level + 1)
            {
                if (this.complete()) {
                    nextSessionLevel = Math.min(this.session_level + 1, this.maxLevel);
                }
                else {
                    nextSessionLevel = this.session_level <= this.firstLevel ? 0 : 
                        this.session_level;
                }
            }
        }
        return nextSessionLevel;
    },

    _recordUserLevel: function(isComplete)
    {
        this.user_level = this.getUserLevel();
        this.metadata[ metadataAggregateKeys.userLevel ] = this.user_level;
    },

    /**
     * GAME-1405 If at max level, would not finish Disillusion session before seeing illusions.  Test case:  2014-11-06 Win level 96.   Next session.  Win level 96.  Expect to see all illusions.
     */
    wouldFinish: function()
    {
        return this.user_level <= this.session_level && this.user_level < this.maxLevel;
    },
   
    /**
     * Set level select by not being in practice mode so as not to override start play at end of gameBegin animation.
     * GAME-1403 May show tutor message.
     * GAME-2503 On begin, reset number correct to 0.  Test case:  2015-01-23 Chart number correct.  Expect 96 or less.  Got 1% of sessions had number correct 97 to 4287.  Score was still accurate.
     */
    onBegin: function()
    {
        //- this.metadata.num_errors = 0;
        this.metadata.num_trials = 0;
        this.metadata.time_consumed = undefined;
        this.num_correct = 0;
        this.persistentData.isLevelSelectMode = !this.persistentData.isPracticeMode;
        if (this.persistentData.isPracticeMode) {
            this.initTutor();
        }
        else {
            this.isBlockingInput = false;
        }
    },

    /**
     * GAME-1403 May outro tutor.  Practice: Recognize round cleared.  Enable tutor to update.  
     */
    onEndCurrentRound: function(timeConsumed)
    {
        var endMilliseconds = Math.round(1000 * timeConsumed);
        if (this.persistentData.isPracticeMode) {
            this.tutor.roundIndex++;
            this.tutor.clear = true;
            this.tutor.updating = true;
            this.tutor.shared.promptEnd();
            this.tutor.shared.outro = this.tutorialParams.length <= this.tutor.roundIndex;
        }
        else {
            this.tutor.shared.outro = this.levelHintMax() < this.currentLevel;
            this.tutor.shared.update();
        }
        this.currentRound.isFinished = true;
        this.roundCount++;
        var bonusPoints = 0;
        if (this.complete()) {
            this.scorePerTime = this.scaleScorePerSecond(this.num_correct, this.lastTrialTimeConsumed);
            bonusPoints = Math.round(this.timeRemaining) * this.scorePerTime;
            this.metadata.time_remaining = (this.seconds * 1000) - endMilliseconds;
            this.metadata.time_consumed = endMilliseconds;
        }

        // ll.verbose( "DisillusionModel", "onEndCurrentRound", "bonusPoints " + bonusPoints );
        this.bonus = bonusPoints;
        this._recordRoundMetadata(endMilliseconds);
        this.currentRound = undefined;
    },

    /**
     * 2014-10-24 At end of each round, Aaron expects diagram of tiles.  
     * GAME-1404 Record end milliseconds.  Record new length of rounds.  Test case:  2014-11-03 End with a round incomplete.  View metadata.  Aaron would expect time_remaining and time defined.  And he would expect round number 1.
     */
    _recordRoundMetadata: function(endMilliseconds)
    {
        var round = {};
        round.fillCount = this.fillCount;
        round.num_correct = this.currentRound.num_correct;
        round.num_tries = this.currentRound.num_tries;
        round.time_remaining = (this.seconds * 1000) - endMilliseconds;
        round.time = endMilliseconds - this.currentRound.startMilliseconds;
        round.height = this.fillHeight;
        round.round = this.rounds.length + 1;
        round.width = this.fillWidth;
        round.tile_layout = this.currentRound.tile_layout;
        this.rounds.push(round);
    },
    
    /**
     * 2014-10-24 Aaron could expect to record an unfinished round.
     * GAME-1405 Disable prepare of tutor.  Test case: 2014-10-30 Short game.  End.  Not sure how.  Sometimes shows message to prepare briefly.
     */
    onEndGameplay: function(timeConsumed)
    {
        var endMilliseconds = Math.round(1000 * timeConsumed);
        if (undefined === this.metadata.time_consumed) {
            this.metadata.time_consumed = endMilliseconds;
        }
        this.tutor.shared.prepare = false;
        this.tutor.shared.stop();
        this.scorePerTime = this.scaleScorePerSecond(this.num_correct, this.lastTrialTimeConsumed);
        this.score = this._score(this.num_correct, this.scorePerTrial, this.timeRemaining);
        this.metadata.scorePerSecond = this.scorePerTime;
        this.metadata.performance_level = this.scaleTilesPerMinute(
            this.num_correct, this.lastTrialTimeConsumed);
        this.metadata.selected_level = this.currentLevel;
        this.metadata.session_level = this.session_level;
        if (this.currentRound && this.rounds.length < this.roundMax)
        {
            this._recordRoundMetadata(endMilliseconds);
        }
        this.metadata.round_csv = lu.jsToCsv(this.rounds);
        this._recordUserLevel();
    },

    /**
     * block input
     * get board coordinates from the button
     * remove model of buttons
     * set up board piece to update
     */
    onPieceTapped: function( button )
    {
        var position;
        if ( this.isGameRunning && button )
        {
            if ( !this.isBlockingInput )
            {
                // ll.log("DisillusionModel.onPieceTapped: " + button.nodeName + " time consumed " + this.timeConsumed);
                this.isBlockingInput = true;
                var x = button.x;
                var y = button.y;
                var position = cc.p( x, y );
                this.currentRound.board.clearTappablePieces();
                var boardPieceToUpdate = this.currentRound.board.getPiece( position );
                if ( boardPieceToUpdate )
                {
                    boardPieceToUpdate.pieceType = this.nextPiece.pieceType;
                    boardPieceToUpdate.pieceColor = this.nextPiece.pieceColor;
                    boardPieceToUpdate.orientation = this.nextPiece.orientation;
                    boardPieceToUpdate.toBeRemoved = true;
                }
                this.previousNextPiece = this.nextPiece;
                this.nextPiece = undefined;
            }
        }
        return position;
    },

    /**
     * @return  Time remaining.  If practice mode, return max value.
     *
     * Change log:
     * GAME-1403 If practice mode, return max value of time remaining.  
     * Test cases:  
     * 2014-10-07 Practice:  Wait 2 minutes. Expect no tick or time out.  
     * 2014-10-07 Practice:  Set blocked tile.  Clear all tiles.  Wait 5 seconds.  
     * Expect blocked tile has removed itself and board has cleared.
     */
    onStartGameplay: function()
    {
        this.num_correct = 0;
        this.roundCount = 0;
        this.score = 0;
        this.rounds = [];
        var timeRemaining;
        if (this.persistentData.isPracticeMode) {
            timeRemaining = this.MILLISECOND_MAX_VALUE;
        }
        else {
            this.initTutor(false);
            if ( lu.isDebugModeOn() && !lu.isShortGamesModeOn() ) {
                timeRemaining = 720;
            }
            else {
                timeRemaining = this.seconds;
            }
        }
        return timeRemaining;
    },

    populate: function(params)
    {
        for (var param in params) {
            this[param] = params[param];
            // trace("    " + param + ": " + params[param]);
        }
        if (null != this.queueDiagram) {
            this.queue = this.queueDiagram.split(",");
        }
    },

    hint: function()
    {
        return this.currentLevel <= this.levelHintMax();
    },

    /**
     * GAME-1295 In first round, show message to prepare.  2014-08-01 In first round of any group at any level, [~aaron] expects to read message to prepare.
     */
    prepare: function()
    {
        return this.roundCount <= 1 || this.hint();
    },
    
    /**
     * GAME-1403 If in practice mode, then load a tutorial level.
     * If playing quickly, outro tutor.
     */
    prepareNextRound: function(params)
    {
        if (this.persistentData.isPracticeMode) {
            this.tutor.clear = false;
            this.tutor.updating = true;
            if (1 <= this.tutor.roundIndex) {
                this.tutor.colorCount = 1;
                this.tutor.symbolCount = 1;
            }
            params = this.tutorialParams[this.tutor.roundIndex];
        }
        else {
            var tilesPerMinute = this.currentLevel;
            var consumed = Math.floor(this.lastTrialTimeConsumed);
            if (1 <= this.roundCount) {
                tilesPerMinute = this.scaleLevelUp(this.currentLevel, 
                    this.num_correct, consumed);
                params.fillCount = this.roundFillCount(
                    this.num_correct, consumed,
                    this.roundCount, this.roundMax, tilesPerMinute);
                this.tutor.shared.outro = this.levelHintMax() < tilesPerMinute;
            }
        }
        this.loadLevel(tilesPerMinute, params);
        this.metadata.num_trials += this.fillCount;
        this.nextRound = new DisillusionRoundModel( this );
    },

    /**
     * Keep previous next piece for logic
     * Clear previous next piece
     * Figure out which board to use
     * Get possible next pieces to choose next piece from
     *
     * GAME-1403 Practice: 
     *      If queue was diagrammed, then dequeue a specific piece.
     *      If previous was wrong, then reuse same piece.
     */
    prepareNextTrial:function()
    {
        // ll.verbose( "Disillusion", "prepareNextTrial" );
        this.tutor.updating = true;
        this.nextPiece = undefined;
        var board;
        if ( this.currentRound && this.currentRound.board )
        {
            board = this.currentRound.board;
        }
        if ( board )
        {
            var previousNextPiece = this.previousNextPiece;
            if (this.persistentData.isPracticeMode) {
                if (this.tutor.colorWrong || this.tutor.symbolWrong) {
                    this.nextPiece = this.previousNextPiece;
                    return;
                }
            }
            if (null != this.queue && 1 <= this.queue.length)
            {
                this.nextPiece = new DisillusionBoardPiece();
                var colorSymbol = this.queue.shift();
                DisillusionBoardPiece.prototype.setSpec(colorSymbol, this.nextPiece);
                return;
            }
            var possibleNextPieces = board.findPossibleMatches(previousNextPiece);
            this.nextPiece = this._varyPreviousPiece(previousNextPiece, possibleNextPieces);
        }
        else
        {
            ll.error( "DisillusionModel", "prepareNextTrial", "Error! No valid board to use." );
        }
    },
    
    /**
     * MOBGAMES-1799 
     * Port essential algorithm of Disillusion 1:
     *      # Shuffle tiles that would match to an empty spot or blocked spot.
     *      # Next tile has 50% chance tall or wide.
     *      # If there was a previously placed tile, find different next tile.
     *      # If no such tile found, then draw a match.
     *      # Construct the next tile to match.
     * Test case:  Nov 3 15:33:53 LumosityiPad[112] : Cocos2d: JS: DisillusionBoard.findPossibleMatches: Could not find valid variations that are not present on the board. Nov 3 15:33:53 LumosityiPad[112] : Cocos2d: JS: Disillusion.prepareNextTrial: Error! Error! Cannot determine next piece.
     */
    _varyPreviousPiece: function(previousNextPiece, possibleNextPieces)
    {
        var varied;
        if ( possibleNextPieces.length <= 0 )
        {
            ll.error( "DisillusionModel", "_varyPreviousPiece", "Error! No piece would match." );
            return varied;
        }

        DisillusionBoard.shuffle(possibleNextPieces);
        if (null == previousNextPiece)
        {
            varied = possibleNextPieces[0];
        }
        else 
        {
            var isVertical = Math.random() < 0.5;
            for (var p = 0; p < possibleNextPieces.length; p++)
            {
                var possible = possibleNextPieces[p];
                if (isVertical == possible.isVertical())
                {
                    var differs = possible.pieceType != previousNextPiece.pieceType
                        || possible.pieceColor != previousNextPiece.pieceColor;
                    if (differs)
                    {
                        varied = possible;
                        break;
                    }
                }
            }
            if (null == varied)
            {
                varied = possibleNextPieces[0];
            }
        }
        return varied;
    },

    /**
     * GAME-1404 At start of round, initialize notation of tiles, number correct, and number of tries.  2014-10-24 Aaron expects to record tile layout, number of tries, number correct.
     */
    startNextRound: function(timeConsumed)
    {
        var startMilliseconds = Math.round(timeConsumed);
        this.currentRound = this.nextRound;
        this.nextRound = undefined;
        this.currentRound.num_correct = 0;
        this.currentRound.isStarted = true;
        this.currentRound.num_tries = 0;
        var pieces = this.currentRound.board.getAllBoardPieces();
        var specs = "";
        for (var p = 0; p < pieces.length; p++) 
        {
            specs += pieces[p].toSpec();
        }
        this.currentRound.tile_layout = specs;
        this.currentRound.startMilliseconds = startMilliseconds;
        if (undefined === this.startMilliseconds) 
        {
            this.startMilliseconds = startMilliseconds;
        }
    },

    /**
     * @return  {int}   Distributed number of tiles for next round given remaining rounds.
     *
     * Changelog:
     * WG-1203 Distribute confident tiles in remaining time to remaining boards.  4 boards total.  2014-07-23 [~aaron] may expect 4 boards.  Tyler expects least 70 tiles per minute.  Extrapolating, expert subscribers will probably expect up to 96 tiles per minute.  2014-07-01 Konstantin expects at most 24 tiles per board.
     * WG-1279 Pace number of tiles in next round, at least remaining proportion of selected number of tiles.  2014-09-04 [~tyler]  2014-09-04 Level 78.  Tyler expects 78 tiles.  Got 63:  Tiles in each round (20, 14, 14, 15).  Might expect tiles in each round to be (20, 19, 19, 20).  Being at level 78 might be spurious.  After this correction, Tyler's performance might betwee 70 and 75 tiles per minute.  
     */
    roundFillCount: function(num_correct, timeConsumed,
            round, roundMax, tilesPerMinuteTotal)
    {
        var total = this._scaleFillCount(num_correct, timeConsumed);
        var roundsRemaining = roundMax - round;
        var scaleRemaining = total / roundsRemaining;
        var goalPace = (tilesPerMinuteTotal - num_correct) / roundsRemaining;
        var perRound = Math.floor(Math.max(goalPace, scaleRemaining));
        var fillMin = 3;
        perRound = Math.max(fillMin, perRound);
        perRound = Math.min(perRound, this.fillMax());
        ll.verbose("DisillusionModel.roundFillCount: " + perRound + " roundsRemaining " + roundsRemaining 
            + " tilesPerMinuteTotal " + tilesPerMinuteTotal);
        return perRound;
    },

    /**
     * Round up to nearest 10.  So it is always at least 10.
     */
    roundUp: function(points)
    {
        var rounding = 10;
        var rounded = Math.round(points / rounding) * rounding;
        rounded = Math.max(rounded, rounding);
        // ll.verbose("DisillusionModel.roundUp: " + rounded + " points " + points);
        return rounded;
    },

    /**
     * @return  unbounded.
     */
    _scaleFillCount: function(num_correct, timeConsumed)
    {
        var rate = this.scaleTilesPerMinute(num_correct, timeConsumed);
        var fillCount = Math.floor(rate * (this.seconds - timeConsumed) / this.secondsPerMinute);
        ll.verbose("DisillusionModel._scaleFillCount: " + fillCount 
            + " timeConsumed " + timeConsumed + " rate " + rate);
        return fillCount;
    },

    scaleLevelUp: function(minimum, num_correct, timeConsumed)
    {
        var rate = this.scaleTilesPerMinute(num_correct, timeConsumed);
        return Math.max(minimum, Math.min(this.maxLevel, rate));
    },

    /**
     * GAME-1538 Measure fraction of second since last entry.  Was measuring whole seconds at end of game.  This potentially doubles session score percentile resolution.  2014-09-12 Currently 52% of a sample of session scores had duplicates with a time bonus.
     * @param   timeConsumed    Seconds with fractions.
     */
    _scaleTilesPerMinute: function(num_correct, timeConsumed)
    {
        var rate = num_correct * this.secondsPerMinute / Math.max(1.0, timeConsumed);
        rate *= this.confidence(num_correct);
        ll.verbose("DisillusionModel._scaleTilesPerMinute: " + rate.toFixed(3));
        return rate;
    },

    /**
     * @return  Cautious estimate of number of tiles could be cleared in a minute at current rate.
     */
    scaleTilesPerMinute: function(num_correct, timeConsumed)
    {
        var rate = this._scaleTilesPerMinute(num_correct, timeConsumed);
        var rounded = Math.round(rate);
        // ll.verbose("DisillusionModel.scaleTilesPerMinute: " + rounded);
        return rounded;
    },

    /**
     * @return  Confident tiles per second * score per trial.  Round to nearest 10.
     * 2014-07-28 Ben expects to fix not higher score with fewer tiles and equal performance in tiles per minute.
     */
    scaleScorePerSecond: function(num_correct, timeConsumed)
    {
        var perSec = this._scaleTilesPerMinute(num_correct, timeConsumed) 
            * this.scorePerTrial / this.secondsPerMinute;
        var power = 4;
        perSec *= Math.pow(this.confidence(num_correct), power);
        perSec = this.roundUp(perSec);
        perSec = Math.min(perSec, this.scorePerTrial);
        ll.verbose("DisillusionModel.scaleScorePerSecond: " + perSec 
            + " num_correct " + num_correct + " timeConsumed " + timeConsumed);
        return perSec;
    },

    /**
     * Ceiling of seconds.  Tyler expects time on clock matches time bonus.
     */
    _score: function(num_correct, accuracy, seconds)
    {
        var value = Math.floor(num_correct * accuracy) + seconds * this.scorePerTime 
            + this.bonusScorePoints;
        return value;
    },

    /**
     * GAME-1405 [~wouter]: This may be convenient to win the game quickly with a large board.  Set Debug mode.  Click Debug menu.  Click button "Win level 96".  Set level to max, round to max, tiles to max, time to a minute or less, and end the current round.  Test case:  2014-10-24 Debug menu.  Win level 96.  Expect to conveniently view all illusions.
     */
    _winLevelMax: function()
    {
        this.currentLevel = this.maxLevel;
        this.roundCount = this.roundMax;
        this.num_correct = this.maxLevel;
        this.lastTrialTimeConsumed = Math.min(60.0, this.lastTrialTimeConsumed);
    },

    setMilestone: function(milestone)
    {
        this.currentLevel = this.milestoneToLevel(milestone);
        // ll.verbose("DisillusionModel.setMilestone: " + this.currentLevel);
    },

    getMilestone: function()
    {
        return this.milestone(this.currentLevel);
    },

    /**
     * @return  Level corresponding to milestone or max achieved, whichever is greater.  
     */
    milestoneToLevel: function(milestone)
    {
        var level;
        if (milestone != this.milestoneUnlocked()) {
            level = this.tilesPerMinuteMin(milestone);
        }
        else {
            level = this.levelUnlocked();
        }
        return level;
    },

    levelUnlocked: function()
    {
        return Math.max(this.currentLevel, this.session_level, this.user_level);
    },

    milestoneUnlocked: function()
    {
        this.levelUnlocked();
        return this.milestone(this.levelUnlocked());
    },
    
    /**
     * @return  WG-1410 Usually display text of milestone to tiles.  Special case: If milestone is the highest, and it is locked, display "(?)".  2014-08-28 To clarify that highest milestone has not yet been unlocked, Aaron expects to obfuscate tile count at max.  Wouter expects a simple milestone display.  2014-08-14 To simplify, Abhi expects current tile count on current milestone.
     */
    milestoneToLevelText: function(milestone)
    {
        var text;
        if (milestone < this.milestoneMax() || this.milestoneMax() <= this.milestoneUnlocked()) {
            text = this.milestoneToLevel(milestone).toString();
        }
        else {
            text = this.unknown;
        }
        return text;
    },

    milestoneMax: function()
    {
        return this.milestoneParams.length - 1;
    },

    /**
     * Ignore short games or debug mode seconds.
     * Test case:  Select short games mode or debug mode.  Level select.  Expect number 12 on first.
     */
    tilesPerMinuteMin: function(milestoneIndex)
    {
        var params = this.milestoneParams[milestoneIndex];
        var milestoneTilesPerMinute = params.fillCount * this.secondsPerMinute / this.secondsOriginal;
        return milestoneTilesPerMinute;
    },

    /**
     * @param   tilesPerMinute  Confident of how many tiles can be cleared in 60 seconds.  
     * @return  Index into Disillusion 1 levels [0..11], to unlock a milestone.  
     */
    milestone: function(tilesPerMinute)
    {
        for (var index = 0; index < this.milestoneParams.length; index++) {
            if (tilesPerMinute < this.tilesPerMinuteMin(index)) {
                break;
            }
        }
        index = Math.max(0, index - 1);
        return index;
    },

    tapHeightMax: function()
    {
        var max = this.fills[this.fills.length - 1].fillHeight + 2;
        return max;
    },

    tapWidthMax: function()
    {
        var max = this.fills[this.fills.length - 1].fillWidth + 2;
        return max;
    },

    transferringTime: function()
    {
        return !this.persistentData.practiceMode && this.timeRemaining > 0 && 
            this.complete();
    },

    /**
     * Step of rolling down time and rolling up bonus, clamp at end.
     * Not cc.ActionTween, because that expects a cc.Node, and there are two parameters tweened.
     *
     * Change log:
     * GAME-1399 End. Rolldown remaining time and rollup score in steps of 10.  2014-08-05 Ben expects to rolldown time and rollup score in steps of 10.  Flash version has an end screen to read a time bonus.  Mobile version does not.
     */
    transferTime: function( dt )
    {
        var speed = this.timeRemained;
        this.timeRemaining = Math.max(0.0, this.timeRemaining - speed * dt);
        if (this.timeRemaining <= 0.0) 
        {
            this.bonusScorePoints = this.bonus;
            this.score = this._score(this.num_correct, this.scorePerTrial, this.timeRemaining);
        }
        else 
        {
            var fraction = 1.0 - this.timeRemaining / this.timeRemained;
            var bonusLerp = lu.lerp(fraction, 0, this.bonus);
            bonusLerp = this.roundUp(bonusLerp);
            this.bonusScorePoints = Math.min(this.bonus, bonusLerp);
            this.score = this._score(this.num_correct, this.scorePerTrial, 0.0);
        }
    }
});
