/*global require*/
/*global cc*/
/*global BirdFormationSettings*/
/*global BirdFormationModel*/
/*global BirdFormationView*/
/*global LumositySprite*/
/*global LayoutLoader*/
/*global LumosityUtils*/
/*global SwipeGestureDirection*/

require( "src/Game/Base/Game.js" );
require( "src/Game/BirdFormationSettings.js" );
require( "src/Game/BirdFormationModel.js" );
require( "src/Game/BirdFormationView.js" );
require( "src/View/LumositySprite.js" );
require( "src/Animation/LayoutLoader.js" );
require( "src/Utility/LumosityUtils.js" );

var BirdFormation;
( function()
{
    "use strict";

    BirdFormation = cc.Class.extend(
    {
        settings: null,
        model: null,

        /** @lends BirdFormation.prototype */
        
        /**
         *  BirdFormation is the main class for a bird formation!
         *
         *  @constructs
         *  @param {array} Settings that define the bird formation.
         */
        ctor: function( settings )
        {
            this._createSettings( settings );
            this._createModel();
            this._createView();
        },

        onEndGameplay: function()
        {
            var toOpacity = 0;
            var fadeToAction = null;
            var sequence = null;

            fadeToAction = cc.FadeTo.create( this.settings.endGameFadeOutDuration, toOpacity );
            sequence = cc.Sequence.create( fadeToAction );
            this.view.setBirdsAnimation( sequence );
        },

        onRestart: function()
        {
            this.view.removeAllBirds();
        },

        /**
        * Randomizes a formation.
        */
        randomizeFormation: function()
        {
            this.view.hideBoundingBox();

            // Get the formations center offset rect.
            var randomFormationIndex = this.settings.getRandomFormationIndex();
            var formation = this.settings.getFormation( randomFormationIndex );
            var formationCenterOffsetRect = this._getFormationCenterOffsetRect( formation );

            // Calculate and set the views content size.
            var imageName = this.settings.birdImage;
            var formationContentSize = this._calculateContentSize( formationCenterOffsetRect );
            this.view.setContentSize( formationContentSize );

            // Calculate and gather properties needed to contruct the view.
            var centerBirdRotation = this.settings.getRandomBirdImageRotation();
            var otherBirdsRotation = this.settings.getRandomBirdImageRotation();

            var originOffsetPoint = this._getOriginOffsetPoint( formationCenterOffsetRect );
            var offsetPos = this._calculateTextureOffsetPos( this.settings.birdContentSize, originOffsetPoint );

            // Create the formation
            this._createFormation( formation, imageName, centerBirdRotation, otherBirdsRotation, offsetPos );

            // Apply scale to the formationContentSize because the formationContainer has a factor scale.
            var formationContainer = this._getFormationContainer();
            var formationContainerScale = cc.p( formationContainer.getScaleX(), formationContainer.getScaleY() );
            var scaledFormationContentSize = cc.size( formationContentSize.width * formationContainerScale.x,
                                                      formationContentSize.height * formationContainerScale.y );

            // Caclulate the area in which the formation can be placed
            var birdContainer = this._getBirdContainer();
            var birdContainerContentSize = birdContainer.getContentSize();
            var residualArea = this._calculateResidualArea( birdContainerContentSize, scaledFormationContentSize );

            // Calculate and set the position to place the formation.
            var formationRandomPosition = this._calculateFormationRandomPosition( residualArea );

            // Convert position into formation container space.
            var formationConvertedRandomPosition = birdContainer.convertToWorldSpace( formationRandomPosition );
            formationConvertedRandomPosition = formationContainer.convertToNodeSpace( formationConvertedRandomPosition );
            this.view.setPosition( formationConvertedRandomPosition );

            // Set the model data
            this.model.centerBirdRotation = centerBirdRotation;
            this.model.otherBirdsRotation = otherBirdsRotation;
            this.model.stimulusLayout = randomFormationIndex;

            // Debug: Comment in to display the bounding box of the bird formation view.
            //this.view.drawBoundingBox( 2.0, cc.c4f( 1, 0, 0, 1 ) );
        },

        /**
        * Returns the center bird's facing direction as a swipe gesture direction.
        *
        * @returns {int} Swipe gesture direction.
        */
        getCenterBirdDirection: function()
        {
            var centerBirdDirection = this._getSwipeGestureDirection( this.model.centerBirdRotation );
            return centerBirdDirection;
        },

        /**
        * Returns the other birds' facing direction as a swipe gesture direction.
        *
        * @returns {int} Swipe gesture direction.
        */
        getOtherBirdsDirection: function()
        {
            var otherBirdsRotation = this._getSwipeGestureDirection( this.model.otherBirdsRotation );
            return otherBirdsRotation;
        },

        /**
        * Creates the bird formation given a set of properties.
        *
        * @param {object} Formation.
        * @param {string} Image Name.
        * @param {int} Degrees in which to rotate the center bird.
        * @param {int} Degrees in which to rotate all other birds.
        * @param {cc.p} Offset position to place all birds in the formation.
        *
        */
        _createFormation: function( formation, imageName, centerBirdRotation, otherBirdsRotation, offsetPos )
        {
            var i;
            var info = null;
            var bird = null;
            var birdPos = cc.PointZero();
            var birdRot = 0;
            var formationLength = 0;
            var birdsLength = 0;
            var reductionLength = 0;

            formationLength = formation.length;
            birdsLength = this.view.getBirdsLength();

            // Reduce the size of the formation. 
            reductionLength =  birdsLength - formationLength;
            if( reductionLength > 0 )
            {
                this._reduceFormation( reductionLength );
            }

            for( i = 0; i < formationLength; i++ )
            {
                info = formation[ i ];
                birdPos = cc.p( ( info.x * this.settings.birdContentSize.width ) + offsetPos.x,
                                ( info.y * this.settings.birdContentSize.height ) + offsetPos.y );

                if( info.center === true )
                {
                    this.model.centerBirdPos = birdPos;
                    birdRot = centerBirdRotation;
                }
                else
                {
                    birdRot = otherBirdsRotation;
                }

                // Recycle the bird if it already exists.
                bird = this._getBird( i );
                if( bird )
                {
                    bird.setPosition( birdPos );
                    bird.setRotation( birdRot );
                }
                else
                {
                    bird = this._createBird( imageName, birdPos, birdRot );
                    this.view.addBird( bird );
                }
            }
        },

        /**
        * Reduces the size of the formation based on length passed in. 
        *
        * @param {int} Reduction Length.
        *
        * @returns {cc.Sprite} Bird sprite.
        */
        _reduceFormation: function( reductionLength )
        {
            var i;
            for( i = 0; i < reductionLength; i++ )
            {
                this.view.removeBird( i );
            }
        },

        /**
        * Creates and adds a bird sprite to the view.
        *
        * @param {string} Image Name.
        * @param {cc.p} Position of the image in the view.
        * @param {int} Degrees in which to rotate the image.
        *
        * @returns {cc.Sprite} Bird sprite.
        */
        _createBird: function( imageName, position, rotation )
        {
            var bird = null;

            bird = new LumositySprite( imageName );
            bird.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            bird.setPosition( position );
            bird.setRotation( rotation );
            bird.setContentSize( this.settings.birdContentSize, true );
            bird.nodeName = "Bird";

            return bird;
        },

        /**
        * Returns a bird sprite at given index.
        *
        * @param {int} Index.
        *
        * @returns {cc.Sprite} Bird sprite.
        */
        _getBird: function( index )
        {
            var bird = null;

            bird = this.view.getBird( index );

            return bird;
        },

        /**
        * Returns a cc.Rect that contains offset values from
        * the center object of the formation.
        * (Used to calculate the formation's content size when
        * multiplied by the object's content size).
        *
        * @param {object} Formation.
        *
        * @returns {cc.Rect} Rect containing offset positions from the center object.
        */
        _getFormationCenterOffsetRect: function( formation )
        {
            var centerOffsetRect;
            var minX = 0;
            var minY = 0;
            var maxX = 0;
            var maxY = 0;

            formation.forEach( function ( info )
            {
                if ( minX > info.x )
                {
                    minX = info.x;
                }
                else if ( maxX < info.x )
                {
                    maxX = info.x;
                }

                if ( minY > info.y )
                {
                    minY = info.y;
                }
                else if ( maxY < info.y )
                {
                    maxY = info.y;
                }
            } );

            centerOffsetRect = cc.rect( minX, minY, maxX, maxY);
            return centerOffsetRect;
        },

        /**
        * Returns a cc.Point that contains offset values from
        * the center origin of the Rect.
        *
        * @param {cc.Rect} Rect containing offset positions from the center object.
        *
        * @returns {cc.p} Point containing offset positions from the origin of the Rect.
        */
        _getOriginOffsetPoint: function( centerOffsetRect )
        {
            var originOffsetPoint;
            originOffsetPoint = cc.p( Math.abs( centerOffsetRect.x ) + 0.5,
                                      Math.abs( centerOffsetRect.y ) + 0.5 );
            return originOffsetPoint;
        },

        /**
        * Returns a calculated content size.
        *
        * @param {string} Name of the image.
        * @param {cc.Rect} Rect containing offset positions from the center object.
        *
        * @returns {cc.size} Content size.
        */
        _calculateContentSize: function( centerOffsetRect )
        {
            var breadthFactor = this._calculateBreadthFactor( centerOffsetRect );
            var depthFactor = this._calculateDepthFactor( centerOffsetRect );
            var dimensionsFactor = cc.p( breadthFactor, depthFactor );

            var contentSize = cc.size( this.settings.birdContentSize.width * dimensionsFactor.x,
                                       this.settings.birdContentSize.height * dimensionsFactor.y );

            return contentSize;
        },

        /**
        * Returns a breadth factor value.
        * (Used to calculate a content size when
        * multiplied by the width of the content).
        *
        * @param {cc.Rect} Rect containing offset positions from the center.
        *
        * @returns {float} Breadth factor value.
        */
        _calculateBreadthFactor: function( centerOffsetRect )
        {
            var breadthFactor;

            breadthFactor = ( Math.abs( centerOffsetRect.x ) + 1 ) + centerOffsetRect.width;
            return breadthFactor;
        },

        /**
        * Returns a depth factor value.
        * (Used to calculate a content size when
        * multiplied by the height of the content).
        *
        * @param {cc.Rect} Rect containing offset positions from the center.
        *
        * @returns {float} Depth factor value.
        */
        _calculateDepthFactor: function( centerOffsetRect )
        {
            var depthFactor;

            depthFactor = ( Math.abs( centerOffsetRect.y ) + 1 ) + centerOffsetRect.height;
            return depthFactor;
        },

        /**
        * Calculates and returns a random formation position.
        *
        * @returns {cc.p} Random formation position.
        */
        _calculateFormationRandomPosition: function( residualArea )
        {
            var randomPosition = cc.PointZero();
            var radiusFromLastFormation = cc.PointZero();
            var contentSize = cc.SizeZero();
            var centerPos = cc.PointZero();
            var posDelta = cc.PointZero();
            var centerBirdPos = cc.PointZero();
            var newPos = cc.PointZero();
            
            // Generate a random position in the offset size space.
            randomPosition.x = LumosityUtils.getRandomInt( residualArea.x, residualArea.width );
            randomPosition.y = LumosityUtils.getRandomInt( residualArea.y, residualArea.height );
            
            // Get the center bird position relative to the randomPosition.
            contentSize = this.view.getContentSize();
            centerPos = cc.p( contentSize.width >> 1, contentSize.height >> 1 );
            posDelta = cc.pSub( this.model.centerBirdPos, centerPos );
            centerBirdPos = cc.pAdd( randomPosition, posDelta );

            // Check to see if your random position is greater that the minimum radius from the last position.
            radiusFromLastFormation.x = Math.abs( this.model.currentPos.x - centerBirdPos.x );
            radiusFromLastFormation.y = Math.abs( this.model.currentPos.y - centerBirdPos.y );

            if( radiusFromLastFormation.x < this.settings.minRadiusFromLastFormation )
            {
                newPos.x = this.model.currentPos.x + this.settings.minRadiusFromLastFormation;
                if( newPos.x > residualArea.width )
                {
                    newPos.x = Math.max( residualArea.x, this.model.currentPos.x - this.settings.minRadiusFromLastFormation );
                }
               
                randomPosition.x = newPos.x;
            }

            if( radiusFromLastFormation.y < this.settings.minRadiusFromLastFormation )
            {
                newPos.y = this.model.currentPos.y + this.settings.minRadiusFromLastFormation;
                if( newPos.y > residualArea.height )
                {
                    newPos.y = Math.max( residualArea.y, this.model.currentPos.y - this.settings.minRadiusFromLastFormation );
                }

                randomPosition.y = newPos.y; 
            }

            // Set the model's currentPos to the new centerBirdPos.
            centerBirdPos = cc.pAdd( randomPosition, posDelta );
            this.model.currentPos = centerBirdPos;

            // Debug valid window rect.
            this.model.validWinRect = residualArea;

            return randomPosition;
        },

        /**
        * Calculates and returns the area in which the formation can be placed.
        *
        * @param {cc.size} Allowed area for the formation.
        * @param {cc.size} Content size of the formation.
        *
        * @returns {cc.rect} Formation's residual area.
        */
        _calculateResidualArea: function( allowedArea, formationContentSize )
        {
            var halfFormationContentSize = cc.SizeZero();
            var residualArea = cc.RectZero();

            halfFormationContentSize = cc.size( formationContentSize.width >> 1,
                                                formationContentSize.height >> 1 );
            residualArea = cc.rect( halfFormationContentSize.width,
                                    halfFormationContentSize.height,
                                    allowedArea.width - halfFormationContentSize.width,
                                    allowedArea.height - halfFormationContentSize.height );
            return residualArea;
        },

        /**
        * Returns a cc.Point that contains the offset position from
        * an origin offset point.
        *
        * @param {cc.size} Content size of a texture.
        *
        * @returns {cc.p} Offset position.
        */
        _calculateTextureOffsetPos: function( textureContentSize, originOffsetPoint )
        {
            var offsetPos = cc.p( originOffsetPoint.x * textureContentSize.width,
                                  originOffsetPoint.y * textureContentSize.height );
            return offsetPos;
        },

        /**
        * Returns swipe gesture direction based on a rotation in degrees.
        *
        * @param {int} Degrees of rotation.
        *
        * @returns {int} Swipe gesture direction.
        */
        _getSwipeGestureDirection: function( rotation )
        {
            var normRotation = rotation % 360;
            var swipeGestureDirection;

            if( normRotation === 0 )
            {
                swipeGestureDirection = SwipeGestureDirection.Up;
            }
            else if( normRotation === 90 )
            {
                swipeGestureDirection = SwipeGestureDirection.Right;
            }
            else if( normRotation === 180 )
            {
                swipeGestureDirection = SwipeGestureDirection.Down;
            }
            else if( normRotation === 270 )
            {
                swipeGestureDirection = SwipeGestureDirection.Left;
            }

            return swipeGestureDirection;
        },

        /**
        * Returns a node defined as the formationContainer within the layout.fla.
        *
        * @returns {cc.Node} Formation Container node.
        */
        _getFormationContainer: function()
        {
            return this.view.getParent();
        },

        /**
        * Returns a node defined as the birdContainer within the layout.fla.
        *
        * @returns {cc.Node} Bird Container node.
        */
        _getBirdContainer: function()
        {
            return this.view.getParent().getParent();
        },

        _createSettings: function( settings )
        {
            this.settings = new BirdFormationSettings( settings );
        },

        _createModel: function()
        {
             this.model = new BirdFormationModel();
        },

        _createView: function()
        {
            this.view = new BirdFormationView( true );
            this.view.nodeName = "BirdFormation";
            this.view.setAnchorPoint( cc.p( 0.5, 0.5 ) );
        }
    } );
}() );
