/*global require*/
/*global cc*/
/*global LumosityAspectNode*/
/*global LumosityUtils*/

require( "src/Game/Base/Game.js" );
require( "src/View/LumosityAspectNode.js" );
require( "src/Utility/LumosityUtils.js" );

var BirdFormationView;
( function()
{
    "use strict";

    BirdFormationView = LumosityAspectNode.extend(
    {
        _birds: null,

        _boundingBox: null,

        /** @lends BirdFormationView.prototype */
        
        /**
         *  BirdFormationView is the main class for a bird formation view!
         *
         *  @constructs
         */
        ctor: function( scaleToMin )
        {
            this._super( scaleToMin );

            this._birds = [];
        },

        /**
        * Adds a bird sprite to the view.
        */
        addBird: function( bird )
        {
            this._birds.push( bird );
            this.addChild( bird );
        },

        /**
        * Removes a bird sprite from the view.
        */
        removeBird: function( index )
        {
            var bird = null;

            if( index >= this._birds.length ||
                index < 0 )
            {
                return;
            }

            bird = this._birds[ index ];
            this.removeChild( bird );
            this._birds.splice( index, 1 );
        },

        /**
        * Returns a bird sprite.
        *
        * @param {int} Index of bird.
        *
        * @returns {cc.Sprite} Bird sprite.
        */
        getBird: function( index )
        {
            var bird = null;

            if( index >= this._birds.length ||
                index < 0 )
            {
                return bird;
            }

            bird = this._birds[ index ];
            return bird;
        },

        /**
        * Removes all bird sprites from the view.
        */
        removeAllBirds: function()
        {
            var self = this;
            this._birds.forEach( function( bird )
            {
                self.removeChild( bird );
            } );

            this._birds.length = 0;
        },

        /**
        * Returns the length of the birds array.
        *
        * @returns {int} Birds array length.
        */
        getBirdsLength: function()
        {
            var length = 0;
            length = this._birds.length;

            return length;
        },

        /**
        * Animates the bird sprites with a given sequence.
        *
        * @param {cc.Sequence} Animation sequence.
        */
        setBirdsAnimation: function( sequence )
        {
            this._birds.forEach( function( bird )
            {
                bird.runAction( sequence.clone() );
            } );
        },

        /**
        * Debug: Renders a bounding box around the view.
        *
        * @param {float} Radius of the border.
        * @param {cc.c4f} Color of the border.
        *
        * @name Debug#showBoundingBox
        */
        drawBoundingBox: function( outlineRadius, outlineColor )
        {
            if( this._boundingBox !== null )
            {
                return;
            }

            var contentSize = this.getContentSize();
            
            var points = [ cc.p( 0, 0), 
                           cc.p( contentSize.width, 0 ),
                           cc.p( contentSize.width, contentSize.height ),
                           cc.p( 0, contentSize.height ) ];

            this._boundingBox = LumosityUtils.createDrawNodePoly( points, outlineRadius, outlineColor );
            this.addChild( this._boundingBox );
        },

        /**
        * Debug: Stops rendering a bounding box around the view.
        *
        * @name Debug#hideBoundingBox
        */
        hideBoundingBox: function()
        {
            if( this._boundingBox === null )
            {
                return;
            }

            this.removeChild( this._boundingBox );
            this._boundingBox = null;
        }
    } );
}() );
