/*global require*/
/*global cc*/
/*global CloudsSettings*/
/*global CloudsModel*/
/*global CloudsView*/
/*global LumositySprite*/
/*global LumosityUtils*/

require( "src/Game/Base/Game.js" );
require( "src/Game/CloudsSettings.js" );
require( "src/Game/CloudsModel.js" );
require( "src/Game/CloudsView.js" );
require( "src/View/LumositySprite.js" );
require( "src/Utility/LumosityUtils.js" );

var Clouds;
( function()
{
    "use strict";

    Clouds = cc.Class.extend(
    {
        settings: null,

        model: null,

        /** @lends Clouds.prototype */
        
        /**
         *  Clouds is the main class for a clouds!
         *
         *  @constructs
         *  @param {array} Settings that define the clouds.
         */
        ctor: function( settings )
        {
            this._createSettings( settings );
            this._createModel();
            this._createView();
        },

        onEndGameplay: function()
        {
            var toOpacity = 0;
            var fadeToAction = null;
            var sequence = null;

            fadeToAction = cc.FadeTo.create( this.settings.endGameFadeOutDuration, toOpacity );
            sequence = cc.Sequence.create( fadeToAction );
            this.view.setCloudsAnimation( sequence );
        },

        onRestart: function()
        {
            this.view.removeAllClouds();
            this.model.cloudObjects.length = 0;
        },

        /**
        * Randomizes clouds.
        */
        randomizeClouds: function()
        {
            var clouds = null;
            var positions = null;

            // Calculate the views parent inverse scale content size.
            this.model.parentInverseScaleContentSize = this._calculateParentInverseScaleContentSize();

            // Create random cloud sprites.
            clouds = this._createRandomClouds( this.settings.maxCount );

            // Calculate and set the clouds random positions.
            positions = this._calculateRandomPositions( clouds );
            this._setSpritesPosition( clouds, positions );
        },

        /**
        * Creates and returns random cloud sprites.
        *
        * @param {int} Count of clouds to randomize.
        *
        * @returns {array} Random cloud sprites.
        */
        _createRandomClouds: function( count )
        {
            var i;
            var imageInfo = null;
            var imageName = null;
            var contentSize = cc.SizeZero();
            var opacity = 100;
            var velocity = cc.PointZero();
            var cloud = null;
            var clouds = [];
            for( i = 0; i < count; i++ )
            {
                // Create a new cloud.
                imageInfo = this.settings.imagesInfo.getRandomElement();
                imageName = imageInfo.imageName;
                contentSize = imageInfo.contentSize;
                opacity = this.settings.opacities.getRandomElement();
    
                cloud = this._createCloud( imageName, contentSize, opacity );
                clouds.push( cloud );

                // Add cloud objects to the model data.
                velocity = this.settings.velocities.getRandomElement();
                this._addCloudObject( cloud, velocity );
            }

            return clouds;
        },

        /**
        * Creates and adds a cloud sprite to the view.
        *
        * @param {string} Image Name.
        * @param {cc.p} Position of the image in the view.
        *
        * @returns {cc.Sprite} Cloud sprite.
        */
        _createCloud: function( imageName, contentSize, opacity )
        {
            var cloud = null;

            cloud = new LumositySprite( imageName );
            cloud.setAnchorPoint( cc.PointZero() );
            cloud.setContentSize( contentSize, true );
            cloud.setOpacity( opacity );

            this.view.addCloud( cloud );
            return cloud;
        },

        /**
        * Adds a cloud object to the model's cloud objects array.
        *
        *  @param {object} Sprite object.
        *  @param {cc.p} Velocity.
        */
        _addCloudObject: function( sprite, velocity )
        {
            var cloudObject = 
            {
                sprite:sprite,
                velocity:velocity
            };

            this.model.cloudObjects.push( cloudObject );
        },

        /**
        * Calculates and returns random positions for the
        * sprite images.
        *
        * @param {array} Array of sprites.
        *
        * @returns {array} Random positions.
        */
        _calculateRandomPositions: function( sprites )
        {
            var i;
            var sprite = null;
            var spriteContentSize = cc.SizeZero();
            var offsetPosition = cc.PointZero();
            var paddingY = 0;
            var randomPosition = cc.PointZero();
            var randomPositions = [];

            paddingY = this._getImagePaddingY();
            
            // Set the screen width to the offset position x
            // ( You can set this to 0 to view clouds immediately. )
            offsetPosition.x = 0;

            for( i = 0; i < sprites.length; i++ )
            {
                sprite = sprites[ i ];
                spriteContentSize = sprite.getContentSize();

                // Calculate a random position
                offsetPosition.y += spriteContentSize.height;
                randomPosition = this._calculateRandomPosition( this.model.parentInverseScaleContentSize, offsetPosition );
                offsetPosition.y += paddingY;
                randomPositions.push( randomPosition );
            }

            // Debug valid window rect.
            this.model.validWinRect = cc.rect( offsetPosition.x,
                                               0,
                                               this.model.parentInverseScaleContentSize.width,
                                               this.model.parentInverseScaleContentSize.height);
            
            return randomPositions;
        },

        /**
        * Calculates the inverse scale of the content size of the views parent.
        *
        * @returns {cc.size} Inverse scale of parent view content size.
        */
        _calculateParentInverseScaleContentSize: function()
        {
            var scale = cc.PointZero();
            var parentContentSize = cc.SizeZero();
            var inverseScaleContentSize = cc.SizeZero();

            scale = cc.p( this.view.getScaleX(), this.view.getScaleY() );
            parentContentSize = this.view.getParent().getContentSize();
            inverseScaleContentSize = LumosityUtils.getInverseScaleContentSize( scale, parentContentSize );
            return inverseScaleContentSize;
        },

        /**
        * Calculates and returns a random position based on area size and
        * offset position.
        *
        * @param {cc.size} Size of the area in which the positions fall in.
        * @param {int} Offset position in the X and Y.
        *
        * @returns {cc.p} Random position.
        */
        _calculateRandomPosition: function( size, offsetPosition )
        {
            var randomPosition = cc.PointZero();

            // Generate a random position in the offset size space.
            randomPosition.x = LumosityUtils.getRandomInt( offsetPosition.x, offsetPosition.x + size.width );
            randomPosition.y = size.height - offsetPosition.y;

            return randomPosition;
        },

        /**
        * Sets the position of each sprite in the passed in array to
        * the corresponding position in the positions array.
        *
        * @param {array} Array of sprites.
        * @param {array} Array of positions.
        */
        _setSpritesPosition: function( sprites, positions )
        {
            var i;
            var sprite = null;
            var pos = cc.PointZero();

            for( i = 0; i < sprites.length; i++ )
            {
                sprite = sprites[ i ];
                pos = positions[ i ];
                sprite.setPosition( pos );
            }
        },

        /**
        * Returns the max width out of all the sprites passed in.
        *
        * @param {array} Array of sprites.
        *
        * @returns {int} Max width.
        */
        _getMaxSpriteWidth: function( sprites )
        {
            var spriteContentSize = cc.SizeZero();
            var maxWidth = 0;
            sprites.forEach( function( sprite )
            {
                spriteContentSize = sprite.getContentSize();

                if( maxWidth < spriteContentSize.width )
                {
                    maxWidth = spriteContentSize.width;
                }
            } );

            return maxWidth;
        },

        /**
        * Returns a padding value in the y position.
        *
        * @returns {int} Padding in the Y.
        */
        _getImagePaddingY: function()
        {
            var imagePaddingY = this.settings.imagePaddingPercentY * this.model.parentInverseScaleContentSize.height;
            return imagePaddingY;
        },

        onUpdate: function( delta )
        {
            this._onUpdateClouds( delta );
        },

        /**
        * Updates the cloud sprites positions based on velocity.
        */
        _onUpdateClouds: function( delta )
        {
            var cloudObjects = null;
            var sprite = null;
            var newPosition = cc.PointZero();
            var contentSize = cc.SizeZero();
            var scale = cc.PointZero();
            var self = this;

            cloudObjects = this.model.cloudObjects;
            if( cloudObjects.length === 0 )
            {
                return;
            }
            
            // Update each of the sprites position based on velocity
            cloudObjects.forEach( function( cloudObject) 
            {
                sprite = cloudObject.sprite;
                newPosition = cc.pAdd( sprite.getPosition(), cc.pMult( cloudObject.velocity, delta ) );
                sprite.setPosition( newPosition );
                contentSize = sprite.getContentSize();

                // Scale must be applied to contentSize because sprites were scaled to fix aspect skewing.
                // ( See getNodeToParentUniformScale with scaleToMin = true )
                scale.x = sprite.getScaleX();
                scale.y = sprite.getScaleY();

                if( newPosition.x < ( -contentSize.width * scale.x ) )
                {
                    sprite.setPosition( cc.p( self.model.parentInverseScaleContentSize.width, newPosition.y ) );
                }
            } );
        },

        _createSettings: function( settings )
        {
            this.settings = new CloudsSettings( settings );
        },

        _createModel: function()
        {
             this.model = new CloudsModel();
        },

        _createView: function()
        {
            this.view = new CloudsView( true );
            this.view.nodeName = "Clouds";
        }
    } );
}() );