/*global cc*/
/*global require*/
/*global Game*/
/*global LostInMigrationSettings*/
/*global LostInMigrationModel*/
/*global LostInMigrationView*/
/*global BirdFormation*/
/*global Clouds*/
/*global InteractionType*/
/*global GameComponentScoreMultiplier*/
/*global LumosityUtils*/
/*global lg*/
/*global metadataTrialKeys*/

require( "src/Game/Base/Game.js" );
require( "src/Game/LostInMigrationSettings.js" );
require( "src/Game/LostInMigrationModel.js" );
require( "src/Game/LostInMigrationView.js" );
require( "src/Game/BirdFormation.js" );
require( "src/Game/Clouds.js" );
require( "src/Utility/LumosityUtils.js" );

var TRIAL_TRANSITION_SEQUENCE_TAG = 1;

var LostInMigration;
( function()
{
    "use strict";

    LostInMigration = Game.extend(
    {
        birdFormation: null,
        clouds: null,

        /** @lends LostInMigration.prototype */
        
        /**
         *  LostInMigration is the main class for the game, Lost In Migration!
         *
         *  @constructs
         */
        ctor: function()
        {
            this._super();

            this._createBirdFormation();
            this._createClouds();

            this.init();

            this.configureMetaDataKeys();
            
            // Debug: Comment in to display the bounding box of the gameContent layer.
            //this._drawGameContentBoundingBox();
        },

        init: function()
        {
            this._initSunset();
            this._initNightfall();
            this._initPrompt();
            this._initStars();
        },

        onBegin: function()
        {
            this._super();

            // Must randomize clouds on onBegin or later because 
            this._randomizeClouds();
        },
        
        onStartGameplay: function()
        {
            this._super();

            this._randomizeBirdFormation();

            this._startSunsetAnimation();
        },

        onUpdate: function( delta )
        {
            this._super( delta );

            if( ( this.model.isGamePaused === false ) )
            {
                if( this.clouds )
                {
                    this.clouds.onUpdate( delta );
                }
            }
        },

        onComponentTimerTick: function( timeRemaining )
        {
            this._super( timeRemaining );

            if( timeRemaining === this.settings.nightfallTimeStart )
            {
                this._startNightfallAnimation();
                this._startHudAnimation();
            }

            if( timeRemaining === this.settings.starsTimeStart )
            {
                this._startStarsAnimation();
            }
        },

        onEndGameplay: function()
        {
            this._super();

            this.view.stopActionByTag( TRIAL_TRANSITION_SEQUENCE_TAG );

            this.birdFormation.onEndGameplay();
            this.clouds.onEndGameplay();
        },

        onRestart: function()
        {
            this.birdFormation.onRestart();
            this.clouds.onRestart();

            this.init();

            // Call super last because it calls onBegin.
            this._super();
        },

        interaction_received: function( interaction )
        {
            this._super( interaction, true );

            if( this.model.isGameRunning === false ||
                this.model.isGamePaused === true )
            {
                return;
            }

            if( this.model.areInteractionsBlocked === true )
            {
                return;
            }

            if ( interaction.getType() === InteractionType.Swipe )
            {
                this._blockInteractions();
                this._evaluateResponse( interaction.getSwipeDirection() );
                this._preformTrialTransitionSequence();
            }
        },

        usesInteractionType: function( type )
        {
            if( type === InteractionType.Swipe )
            {
                return true;
            }
          
            return this._super( type );
        },

        onPause: function()
        {
            this._super();

            this.view.pauseAnimations();
        },

        onResume: function()
        {
            this._super();

            this.view.resumeAnimations();
        },

        configureMetaDataKeys: function()
        {
            this._super();

            metadataTrialKeys.correct           = "correct";            // Was the trial correctly answered? (T/F)
            metadataTrialKeys.stimulusLayout    = "stimulus_layout";    // Type of flock arrangement - indexed (0=horizontal line, 1=vertical line, 2=cross, 3=V left, 4=V right, 5=V down, 6=V up)
            metadataTrialKeys.stimulusLocation  = "stimulus_location";  // x,y coordinates of the target stimulus - comma separated and wrapped in quotes
            metadataTrialKeys.targetDirection   = "target_direction";   // Pointing direction of the target (U,D,L,R)
            metadataTrialKeys.flankerDirection  = "flanker_direction";  // Pointing direction of the flankers (U,D,L,R) - Decimal angle in the flash_lim_rotation split test
            metadataTrialKeys.responseDirection = "response_direction"; // Direction input by the user (U,D,L,R)
            metadataTrialKeys.responseTime      = "response_time";      // Response time (ms) (**earlier version reported microseconds)
            metadataTrialKeys.multiplier        = "multiplier";         // Multiplier applied to the current score increment
            metadataTrialKeys.inputType         = "inputType";          // "L" if a,w,s,d was used for input, "R" if the arrows were used for input [Not applicable in the mobile version]
            metadataTrialKeys.progress          = "progress";           // Multiplier pips

            cc.log( "metadataTrialKeys: " + JSON.stringify( metadataTrialKeys ) );
        },
     
        /**
        * Evaluates the swipe gesture response and runs
        * appropriate logic based on results.
        *
        * @param {int} Swipe gesture direction.
        *
        */
        _evaluateResponse: function( swipeDirection )
        {
            var scoreMultiplierComponent = this._components[ GameComponentScoreMultiplier.prototype.getType() ];
            var centerBirdDirection = this.birdFormation.getCenterBirdDirection();
            var otherBirdsDirection = this.birdFormation.getOtherBirdsDirection();
            var feedbackWidget = this.view.getFeedbackWidget();
            var feedbackPos = this._getFeedbackPos();
            var isCorrect = false;
            var hasCongruentStimuli = false;

            // Set the feedback widget to the center birds position.
            feedbackWidget.setPosition( feedbackPos );

            this.model.currentTrial[ metadataTrialKeys.multiplier ] = this._components[ GameComponentScoreMultiplier.prototype.getType() ]._multiplier;

            isCorrect = ( centerBirdDirection === swipeDirection || lu.isDebugModeOn() );
            if( isCorrect )
            {
                scoreMultiplierComponent.increaseMultiplierProgress();
                this.view.showFeedback( true );
                this.model.numberOfFlocks++;
            }
            else
            {
                scoreMultiplierComponent.decreaseMultiplierProgress();
                this.view.showFeedback( false );

                if( this.model.numberOfFlocks > 0 )
                {
                    this.model.numberOfFlocks--;
                }
            }

            this.model.currentTrial[ metadataTrialKeys.correct ]           = ( isCorrect === true ) ? "T" : "F";
            this.model.currentTrial[ metadataTrialKeys.targetDirection ]   = this._formatDirectionForMetadata( centerBirdDirection );
            this.model.currentTrial[ metadataTrialKeys.flankerDirection ]  = this._formatDirectionForMetadata( otherBirdsDirection );
            this.model.currentTrial[ metadataTrialKeys.responseDirection ] = this._formatDirectionForMetadata( swipeDirection );

            hasCongruentStimuli = ( centerBirdDirection === otherBirdsDirection );
            this.endTrial( this.model.currentTrial, isCorrect, false, hasCongruentStimuli, false, true );
        },

        /**
        * Preforms the sequence of events between trials.
        */
        _preformTrialTransitionSequence: function()
        {
            var delayBetweenFormations = this.birdFormation.settings.delayBetweenFormations;
            var actionDelay = cc.DelayTime.create( delayBetweenFormations );
            var actionUnblockInterations = cc.CallFunc.create( this._unblockInteractions, this );
            var actionRandomizeBirdFormation = cc.CallFunc.create( this._randomizeBirdFormation, this );

            var transitionSequence = cc.Sequence.create( actionDelay, actionUnblockInterations, actionRandomizeBirdFormation );
            transitionSequence.setTag( TRIAL_TRANSITION_SEQUENCE_TAG );
            this.view.runAction( transitionSequence );
        },

        /**
        * Flags interactions to be blocked.
        */
        _blockInteractions: function()
        {
            this.model.areInteractionsBlocked = true;
        },

        /**
        * Flags interactions to be unblocked.
        */
        _unblockInteractions: function()
        {
            this.model.areInteractionsBlocked = false;
        },

        /**
        * Randomizes a bird formation.
        */
        _randomizeBirdFormation: function()
        {
            this.beginTrial();
            this.birdFormation.randomizeFormation();

            this.model.currentTrial = {};
            this.model.currentTrial[ metadataTrialKeys.stimulusLayout ] =  this.birdFormation.model.stimulusLayout;
            this.model.currentTrial[ metadataTrialKeys.stimulusLocation ] =  this._formatPointForMetadata( this._getCenterBirdWorldPos() );
            this.model.currentTrial[ metadataTrialKeys.progress ] =  this._components[ GameComponentScoreMultiplier.prototype.getType() ].getProgressMetadata();

            // Debug: Comment in to display a valid window rect in which the formation may appear
            //LumosityUtils.drawRect( this.birdFormation.model.validWinRect, 2.0, cc.c4f( 0, 1, 0, 1 ), 1, this.view.birdContainer );
        },

        /**
        * Randomizes a clouds.
        */
        _randomizeClouds: function()
        {
            this.clouds.randomizeClouds();

            // Debug: Comment in to display a valid window rect in which the clouds may appear
            //LumosityUtils.drawRect( this.clouds.model.validWinRect , 2.0, cc.c4f( 0, 0, 1, 1 ), 2, this.view.cloudContainer );
        },

        /**
        * Initializes the sunset.
        */
        _initSunset: function()
        {
            var opacity = this.view.sunsetOpacity;
            var position = this.view.sunsetStartPos;

            this.view.sunset.setOpacity( opacity );
            this.view.sunset.setPosition( position );
        },

        /**
        * Initializes the nightfall.
        */
        _initNightfall: function()
        {
            var opacity = this.view.nightfallColor1.a;
            var toColor = this.view.nightfallColor1;

            this.view.sunsetLayer.setOpacity( opacity );
            this.view.sunsetLayer.setColor( toColor );
        },

        /**
        * Initializes the prompt.
        */
        _initPrompt: function()
        {
            var opacity = this.view.promptColor.a;
            var toColor = this.view.promptColor;

            this.view.promptLabel.setOpacity( opacity );
            this.view.promptLabel.setColor( toColor );
        },

        /**
        * Initializes the stars.
        */
        _initStars: function()
        {
            var opacity = 0;

            this.view.stars.setOpacity( opacity );
        },

        /**
        * Starts the sunset animation.
        */
        _startSunsetAnimation: function()
        {
            var position = this.view.sunsetEndPos;
            var moveAction = cc.MoveTo.create( this.settings.gameLengthInSeconds, position );
            var sequence = cc.Sequence.create( moveAction );
            this.view.sunset.runAction( sequence );
        },

        /**
        * Starts the nightfall animation.
        */
        _startNightfallAnimation: function()
        {
            var toColor2 = this.view.nightfallColor2;
            var toColor3 = this.view.nightfallColor3;
            var toColor4 = this.view.nightfallColor4;

            var tintAction1 = cc.TintTo.create( this.settings.civilTwilightDuration, toColor2.r, toColor2.g, toColor2.b );
            var tintAction2 = cc.TintTo.create( this.settings.nauticalTwilightDuration, toColor3.r, toColor3.g, toColor3.b );
            var tintAction3 = cc.TintTo.create( this.settings.astronomicalTwilightDuration, toColor4.r, toColor4.g, toColor4.b );

            var sequence = cc.Sequence.create( tintAction1, tintAction2, tintAction3 );
            this.view.sunsetLayer.runAction( sequence );
        },

        /**
        * Starts the hud animation.
        */
        _startHudAnimation: function()
        {
            var toColor2 = this.view.hudColor2;
            var toColor3 = this.view.hudColor3;
            var toColor4 = this.view.hudColor4;

            var tintAction1 = cc.TintTo.create( this.settings.civilTwilightDuration, toColor2.r, toColor2.g, toColor2.b );
            var tintAction2 = cc.TintTo.create( this.settings.nauticalTwilightDuration, toColor3.r, toColor3.g, toColor3.b );
            var tintAction3 = cc.TintTo.create( this.settings.astronomicalTwilightDuration, toColor4.r, toColor4.g, toColor4.b );

            var sequence = cc.Sequence.create( tintAction1, tintAction2, tintAction3 );
            this.view.hud.setTextAnimation( sequence );
        },

        /**
        * Starts the stars animation.
        */
        _startStarsAnimation: function()
        {
            var toOpacity = 255;
            var fadeToAction = cc.FadeTo.create( this.settings.starsFadeInDuration, toOpacity );
            
            var sequence = cc.Sequence.create( fadeToAction );
            this.view.stars.runAction( sequence );
        },

        /**
        * Returns the position of the center bird in world space.
        *
        * @returns {cc.p} Center bird world position.
        */
         _getCenterBirdWorldPos: function()
        {
            var worldPos = cc.PointZero();
            var centerBirdPos = cc.PointZero();

            centerBirdPos = this.birdFormation.model.centerBirdPos;
            worldPos = this.birdFormation.view.convertToWorldSpace( centerBirdPos );

            return worldPos;
        },

        /**
        * Returns the position where the feedback should be placed.
        *
        * @returns {cc.p} Feedback position.
        */
         _getFeedbackPos: function()
        {
            var feedbackPos = cc.PointZero();
            var worldPos = cc.PointZero();

            worldPos = this._getCenterBirdWorldPos();
            feedbackPos = this.view.feedbackContainer.convertToNodeSpace( worldPos );

            return feedbackPos;
        },

        /**
        * Returns a formated cc.p for metadata.
        *
        * @returns {string} Formated cc.p.
        */
        _formatPointForMetadata: function( point )
        {
            var formatedPoint = "\"" + Math.round( point.x ) + "," + Math.round( point.y ) + "\"" ;
            return formatedPoint;
        },

        /**
        * Returns the formated cardinal direction for metadata.
        *
        * @returns {string} Formated cardinal direction.
        */
        _formatDirectionForMetadata: function( cardinalDirection )
        {
            var formatedDirection = "Invalid";

            if( cardinalDirection === 0 )
            {
                formatedDirection = "U";
            }
            else if( cardinalDirection === 1 )
            {
                formatedDirection = "R";
            }
            else if( cardinalDirection === 2 )
            {
                formatedDirection = "D";
            }
            else if( cardinalDirection === 3 )
            {
                formatedDirection = "L";
            }

            return formatedDirection;
        },

        /**
        * Debug: Renders the bounding box of the gameContent layer.
        *
        * @name Debug#showRect
        */
        _drawGameContentBoundingBox: function()
        {
            var rect = cc.RectZero();
            var gameContentContentSize = cc.SizeZero();

            gameContentContentSize = this.view.gameContent.getContentSize();
            rect = cc.rect( 0, 0, gameContentContentSize.width, gameContentContentSize.height );

            cc.log( "LostInMigration::_drawGameContentBoundingBox rect: " + JSON.stringify( rect ) );
            LumosityUtils.drawRect( rect, 2.0, cc.c4f( 1, 0, 1, 1 ), 3, this.view.gameContent );
        },

        _createSettings: function()
        {
            return new LostInMigrationSettings( "settings.json" );
        },

        _createModel: function()
        {
             return new LostInMigrationModel();
        },

        _createView: function( screenSize )
        {
            return new LostInMigrationView( screenSize, this.settings, this.model, this );
        },

        /**
        *   Creates and adds a BirdFormation to the gameContentRoot.
        */
        _createBirdFormation: function()
        {
            this.birdFormation = new BirdFormation( this.settings.birdFormationSettings );
            this.view.formationContainer.addChild( this.birdFormation.view );
        },

        /**
        *   Creates and adds clouds to the background view.
        */
        _createClouds: function()
        {
            this.clouds = new Clouds( this.settings.cloudsSettings );
            this.view.cloudContainer.addChild( this.clouds.view );
        }
    } );
}() );
