/*global cc*/
/*global require*/
/*global GameView*/
/*global GameWidgetCountdown*/
/*global GameWidgetFeedback*/
/*global GameWidgetHud*/
/*global LayoutLoader*/
/*global LayoutNodeFlags*/
/*global AnimationChanges*/
/*global LumosityUtils*/

require( "src/Game/Base/GameView.js" );
require( "src/Game/Widget/GameWidgetCountdown.js" );
require( "src/Game/Widget/GameWidgetHud.js" );
require( "src/Animation/LayoutLoader.js" );
require( "src/Animation/AnimationController.js" );
require( "src/Utility/LumosityUtils.js" );

var LostInMigrationView;
( function()
{
    "use strict";

    LostInMigrationView = GameView.extend(
    {
        nightfallColor1: cc.c4b( 255, 255, 255, 1 ),
        nightfallColor2: cc.c4b( 255, 255, 255, 1 ),
        nightfallColor3: cc.c4b( 255, 255, 255, 1 ),
        nightfallColor4: cc.c4b( 255, 255, 255, 1 ),
        hudColor1: cc.c4b( 0, 0, 0, 1 ),
        hudColor2: cc.c4b( 0, 0, 0, 1 ),
        hudColor3: cc.c4b( 0, 0, 0, 1 ),
        hudColor4: cc.c4b( 0, 0, 0, 1 ),
        promptColor: cc.c4b( 0, 0, 0, 1 ),
        sunsetStartPos: cc.PointZero(),
        sunsetEndPos: cc.PointZero(),
        sunsetOpacity: 255,

        /** @lends LostInMigrationView.prototype */
        
        /**
        * The Lost In Migration view.
        * @constructs
        * @param {cc.Size} screenSize The size of the screen that this view will be inserted into.
        * @param {Object} [settings] The settings that we can pull view-based data from.
        * @param {LostInMigrationModel} [model] The GameModel that we are linked to.
        * @param {Object} [interactionResponder] The object that we should send interaction events to. We also use this is a default animation event listener.
        * @param {boolean} [preload=true] If true, will preload all flipbooks and other assets upon creation.
        */
        ctor: function( screenSize, settings, model, interactionResponder, preload )
        {
            var hudWidget = null;
            var feedbackWidget = null;
            
            this._super( screenSize, settings, model, interactionResponder, preload );

            // Assign widgets
            hudWidget = this._widgets[ GameWidgetHud.prototype.getType() ];
            this.hud = hudWidget;  

            // Ignore scale and alpha for labels so the animation controller doesn't set them.
            AnimationChanges.addNodeIgnoredChanges( this.promptLabel, [ AnimationChanges.Scale, AnimationChanges.Alpha ] );

            // Moving the feedbackWidget into the feedbackContainer object so it
            // can get factor scale applied.
            // ( Can't be called in createViews because widgets get repositioned before createViews is called. )
            feedbackWidget = this.getFeedbackWidget();
            feedbackWidget.removeFromParent( false );
            this.feedbackContainer.addChild( feedbackWidget );
        },

        /**
        * Pauses all current running animations.
        */
        pauseAnimations: function()
        {
            this.sunset.pauseSchedulerAndActions();
            this.sunsetLayer.pauseSchedulerAndActions();
            this.stars.pauseSchedulerAndActions();
        },

        /**
        * Resumes all previous running animations.
        */
        resumeAnimations: function()
        {
            this.sunset.resumeSchedulerAndActions();
            this.sunsetLayer.resumeSchedulerAndActions();
            this.stars.resumeSchedulerAndActions();
        },

        stopAllAnimations: function( stopActionTarget )
        {
            this._super( stopActionTarget );

            this.sunset.stopAllActions();
            this.sunsetLayer.stopAllActions();
            this.stars.stopAllActions();
        },

        _createViews: function()
        {
            // Removing view from gameRoot which animates offscreen.
            // Adding it to the view which doesn't animate.
            this.backgroundContents.removeFromParent( false );
            this.addChild( this.backgroundContents );

            this.scaledContent.removeFromParent( false );
            this.backgroundContents.addChild( this.scaledContent );

            this.stars.removeFromParent( false );
            this.addChild( this.stars );

            this.sunsetLayer.removeFromParent( false );
            this.addChild( this.sunsetLayer );

            this.cloudContainer.removeFromParent( false );
            this.addChild( this.cloudContainer );

            this.formationContainer.removeFromParent( false );
            this.birdContainer.addChild( this.formationContainer );

            this.feedbackContainer.removeFromParent( false );
            this.gameContent.addChild( this.feedbackContainer );

            // Must convert sunset start and end position into scaledContent space.
            // This is because the sunset ( sun ) is a child of scaledContent inorder to
            // apply a propper scale factor from within the layout file.
            this.sunsetStartPos = this.sunStart.getPosition();
            this.sunsetStartPos = this.backgroundContents.convertToWorldSpace( this.sunsetStartPos );
            this.sunsetStartPos = this.scaledContent.convertToNodeSpace( this.sunsetStartPos );
            this.sunStart.removeFromParent();
            this.sunStart = null;

            this.sunsetEndPos = this.sunEnd.getPosition();
            this.sunsetEndPos = this.backgroundContents.convertToWorldSpace( this.sunsetEndPos );
            this.sunsetEndPos = this.scaledContent.convertToNodeSpace( this.sunsetEndPos );
            this.sunEnd.removeFromParent();
            this.sunEnd = null;
        },

        layout_creatingNode: function( nodeName, nodeData )
        {
            var result;
            var textData = null;
            var colorData = null;
            var sunsetLayer = null;
            var sunsetColor = null;

            if( nodeName === "sunsetLayer" )
            {
                sunsetLayer = cc.Sprite.create( "empty.png" );
                result =  sunsetLayer;
            }
            else if( nodeName === "sunsetColor1" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.nightfallColor1 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "sunsetColor2" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.nightfallColor2 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "sunsetColor3" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.nightfallColor3 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "sunsetColor4" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.nightfallColor4 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "sunsetAlpha" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                sunsetColor = LayoutLoader.getColor( colorData );
                this.sunsetOpacity = sunsetColor.a;
                result = false;
            }
            else if( nodeName === "hudColor1" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.hudColor1 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "hudColor2" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.hudColor2 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "hudColor3" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.hudColor3 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "hudColor4" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.hudColor4 = LayoutLoader.getColor( colorData );
                result = false;
            }
            else if( nodeName === "promptText" )
            {
                textData = nodeData[ LayoutLoader.textDataKey ];
                colorData = textData[ LayoutLoader.fontColorKey ];
                this.promptColor = LayoutLoader.getColor( colorData );
                result = false;
            }
            else
            {
                result = this._super( nodeName, nodeData );
            }

            return result;
        },

        layout_settingNodeFlags: function( nodeName, nodeData, node )
        {
            var nodeFlags = this._super( nodeName, nodeData, node );
            if( nodeName === "formationContainer" )
            {
                // formationContainer sits in the root of the layout but is programatically added
                // to the birdContainer. Because all root objects within the layout alwaysResizes,
                // we must return only the flags corresponding to the properties that we want set.
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }
            else if( nodeName === "scaledContent" )
            {
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }
            else if( nodeName === "feedbackContainer" )
            {
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }

            return nodeFlags;
        }
    } );
}() );
