/*global cc*/
/*global require*/
/*global Game*/
/*global MemoryMatchSettings*/
/*global MemoryMatchModel*/
/*global MemoryMatchView*/


require( "src/Game/Base/Game.js" );
require( "src/Game/MemoryMatchSettings.js" );
require( "src/Game/MemoryMatchModel.js" );
require( "src/Game/MemoryMatchView.js" );
require( "src/Game/MemoryMatchCard.js" );

var MemoryMatch;
( function()
{
    "use strict";

    
    MemoryMatch = Game.extend(
    {
        /** @lends MemoryMatch.prototype */
        
        /**
        *  MemoryMatch is the main class for the game, Memory Match!
        *
        *  @constructs
        */
        trialStartTime:0,
        isMaxed:false,

        trial_progress: 0,
        trial_multiplier:0,

        ctor: function()
        {
            this._super();
            this.responseTimesMatch = [];
            this.responseTimesNoMatch = [];

            // text for debug mode
            if ( lu.isDebugModeOn() )
            {
                var font = "fonts/MuseoSans-700.ttf";
                var fontSize = 25;

                var cheatText;
                if ( this._cheatMode )
                {
                    cheatText = "On";
                }
                else
                {
                    cheatText = "Off";
                }

                
                this._debugLabel1 = LumosityLabel.create( "Cheat Mode: " + cheatText.toString(), font, fontSize );
                this._debugLabel1.setPosition( 70, 105 );
                this.view._gameStage.addChild( this._debugLabel1, 200 );

                var debugMenuObject = { "labelText": "Cheat Mode",
                    "callback": this.toggleCheatMode,
                    "target": this,
                    "closeMenuOnCallback": false };

                this.view._debugMenu.addDebugButton( debugMenuObject );
            }
        },

        onBegin: function()
        {
            this._super();

            this.view.enableButtons( false );
            
            this.view.feedback.setVisible(true);

            this.view.removeAllCards();
            this.view.initCards();

            this.view.hud.setScore( 0 );
        },

        onUpdate: function ( deltaTick )
        {
            this._super( deltaTick );
            if ( lu.isDebugModeOn() )
            {
                this.debugModeText();
            }
        },

        debugModeText: function ()
        {
            var cheatText;
            if ( lu.isDebugModeOn() )
            {
                cheatText = "On";
            }
            else
            {
                cheatText = "Off";
            }

            //display values on screen
            this._debugLabel1.setString( "Cheat Mode: " + cheatText.toString() );
        },


        onRestart: function()
        {
            this._super();

            this.model.score = 0;

            this.view.hud.setScore( 0 );
        },

        //Override
        onPause: function ( instant ){
            this.view.pauseHide();
            this._super();
        },

        //Override
        onResume: function(){
            this._super();
            this.view.resumeShow();
        },
        
        _createSettings: function()
        {
            return new MemoryMatchSettings( "settings.json" );
        },
        
        _createModel: function()
        {
            return new MemoryMatchModel();
        },
        
        _createView: function( screenSize )
        {
            return new MemoryMatchView( screenSize, this.settings, this.model, this );
        },

        checkAnswer:function( yesPressed )
        {          
            var correct = false;

            var match = this.view.doCardsMatch();

            if((match && yesPressed) || (!match && !yesPressed)){
                correct = true;
            }

            if ( lu.isDebugModeOn() )
            {
                correct = true;
            }

            this.view.cardsToNextPosition(correct);

            this.updateScore(correct);
            this._recordTrialMetadata(correct, yesPressed);
            this.view.addNewCard();
        },

        updateScore: function(correct){
            //deals with the multiplier as well
            if(!this.model.hint){
                this.model.num_no_hint += 1;
            }

            this.trial_progress = this.model.numPips % this.settings.streakRequiredForBonus;
            this.trial_multiplier = this.model.scoreMultiplier;

            this.model.num_tries += 1;
            if(correct){
                this.model.score += this.settings.pointsPerCorrect * this.model.scoreMultiplier;
                this.model.num_match += 1;
                if(this.model.numPips < this.settings.maximumMultiplier * this.settings.streakRequiredForBonus / 2){
                    this.model.numPips += 1;
                }
                this.model.consecutiveIncorrects = 0;
                this.model.scoreMultiplier = Math.floor(this.model.numPips / this.settings.streakRequiredForBonus) * 2; //multiplier is just a mod of the stuff
                if(this.model.scoreMultiplier == 0){this.model.scoreMultiplier = 1;}

                if(!this.model.hint){
                    this.model.num_no_hint_correct += 1;
                }

                if(this.model.numPips % this.settings.streakRequiredForBonus == 0 && !this.isMaxed){
                    this.view.showFeedback( true, this.model.scoreMultiplier, true );
                }
                else{
                    this.view.showFeedback( true, 1, true );
                }
                this.isMaxed = this.model.scoreMultiplier >= this.settings.maximumMultiplier;
            }
            else{
                this.view.showFeedback( false, 1, true );
                                
                this.model.consecutiveIncorrects += 1;
                this.model.numPips -= this.model.consecutiveIncorrects; //pips drop by the number of consecutive errors
                if(this.model.numPips < 0){this.model.numPips = 0;}
                this.model.scoreMultiplier = Math.floor(this.model.numPips / this.settings.streakRequiredForBonus) * 2; //multiplier is just a mod of the stuff1
                if(this.model.scoreMultiplier == 0){this.model.scoreMultiplier = 1;}

                this.isMaxed = false;
            }
            this.view.hud.setScore( this.model.score );
            
        },

        _buttonPressed:function( buttonName )
        {
            this._super( buttonName );

            if ( this.model.isPlaying() && this.view.animCompleted )
            {
                if(this.view.no.getIsEnabled() && this.view.yes.getIsEnabled()){
                    if ( buttonName === "yes" )
                    {
                        this.checkAnswer( true );
                    }
                    else if ( buttonName === "no" )
                    {
                        this.checkAnswer( false );
                    }
                }
            }
        },

        onStartGameplay: function()
        {
            this._super();
            this.view.enableButtons( true );
            this.trialStartTime = lumosity.game.frametime;
            this.beginTrial();
            this.model.markTrialStartTime();
            this.view.hideInitCards();
            this.view.flipInitCards();

            this.view.playAnimation( "showPrompt" );
            this.view.playAnimation("showHighlight");
            this.view.playAnimation( "problem" );

        },

        onEndGameplay:function()
        {            
            this._super();

            this.view.flipAllCards();

            this.view.cardHighlightBottom.setVisible(false);
            this.view.cardHighlightTop.setVisible(false);

            this.view.Card1_Transition.setVisible(false);
            this.view.Card2_Transition.setVisible(false);
            this.view.Card3_Transition.setVisible(false);
            this.view.Card4_Transition.setVisible(false);

            this.view.motionTrack_StraightBottom.setVisible(false);
            this.view.motionTrack_StraightTop.setVisible(false);
            this.view.motionTrack_Curve.setVisible(false);

            this.view.playAnimation("hideHighlights");
            this.view.playAnimation("hidePrompt");
            
            this.view.enableButtons( false );
        },

        _recordTrialMetadata: function( correct, userAnsweredYes )
        {
            var delta_progress = this.model.updateResponseTime( correct, userAnsweredYes );

            var hinted = 1; //0 - no, 1 - yes, 2 - yes, on the previous trial"
            if(!this.model.hint){ //if hint is false 
                hinted = 0;
            }

            if(this.model.prevHint){ // previous trial is hinted
                hinted = 2;
            }

            var trial_stim = this.view.getSlotVal(2);
            var trial_back = this.view.getSlotVal(4);

            //TODO: FIX META DATA, MAKE SURE IT GOES THROUGH THE VERIFIER
            var trial =
            {
                stimulus : trial_stim,
                back : trial_back,
                response : userAnsweredYes ? "T" : "F",
                match : this.view.doCardsMatch() ? "T" : "F",    
                correct : correct ? "T" : "F",
                multiplier : parseInt(this.trial_multiplier),
                progress : parseInt(this.trial_progress),
                hint : hinted
            };

            
            this.model.prevHint = this.model.hint;
            this.model.hint = !correct; //if correct means next trial doesn't have hints, else next trial will be hinted

            //might need to fix  this
            this.endTrial( trial, correct, false, false, false, false, true );
            this.beginTrial();
        }
    } );
}() );