var cc = cc;
var require = require;

require( "src/Utility/MultiDimensionArray.js" );
var MultiDimensionArray = MultiDimensionArray;

require( "src/Game/Maze/MazeView.js" );
var MazeView = MazeView;

require( "src/Game/Maze/MazeGenerator.js" );
var MazeGenerator = MazeGenerator;

require( "src/AI/AStarAI.js" );
var AStarAI = AStarAI;

/**
 * The model for a Maze. Has all the data about room connections and whatnot, as well as start and goal positions.
 * If desired, it also can calculate and give a path from the start to the finish. Provides a function that will
 * use its data to tell a MazeView how it should be represented.
 * 
 * @author Eli Delventhal
 */
var Maze = cc.Class.extend(
{
    /** @lends Maze.prototype */

    /** Grid position of maze's start. Characters are placed here, and the solution starts here. */
    start: null,
    
    /** Grid position of maze's goal. When a player reaches the goal, they move on to another maze. */
    goal:null,
    
    /** The solution path to get to the end of this maze from the start. */
    _solution:null,

    /** Multi-dimensional array of Rooms. */
    _rooms:null,

    /**
     * Creates a maze of a certain size and solution distance.
     *
     * @param width     Maze's desired column count.
     * @param height    Maze's desired row count.
     * @param goalDistance  Maze's desired goal distance.
     */
    ctor:function()
    {
    },

    initRandomMaze:function( width, height, goalDistance, minTurns, maxTurns )
    {
        //var startTime = new Date().getTime();

        var mazeInfo = MazeGenerator.generateMaze( width, height, goalDistance, minTurns, maxTurns );
        this.start = mazeInfo.start;
        this.goal = mazeInfo.goal;
        this._solution = mazeInfo.solution;
        this._rooms = mazeInfo.rooms;
        this.turnCount = mazeInfo.turnCount;

        //ll.verbose( "Total maze creation took " + ( new Date().getTime() - startTime ) + "ms" );
    },

    initAuthoredMaze:function( authoredMazeData )
    {
        var mazeInfo = MazeGenerator.loadAuthoredMaze( authoredMazeData );
        this.start = mazeInfo.start;
        this.goal = mazeInfo.goal;
        this._solution = mazeInfo.solution;
        this._rooms = mazeInfo.rooms;
        this.turnCount = mazeInfo.turnCount;
    },

    /**
     * Returns maze size in x-dimension.
     * @returns {number} Width of maze.
     */
    getWidth:function()
    {
        return this._rooms.getLength( 0 );
    },

    /**
     * Returns maze size in y-dimension.
     * @returns {number} Height of maze.
     */
    getHeight:function()
    {
        return this._rooms.getLength( 1 );
    },
    
    /**
     * Returns the best (and only) path from one point to another. The
     * path contains a progressive list of Coordinates, treated like a stack.
     * If you pop the last coordinate, that will get you the next place to move.
     * Be very careful to not modify this array! Dupe it if you're going to be popping.
     * 
     * @return {Array} A list of Coordinates.
     */
    getSolution:function()
    {
        if( this._solution )
        {
            return this._solution.slice( 0 );
        }
    },
    
    /**
     *  Tells each Room to generate all the sprites that it needs to display.
     *  Explicitly called separately from the constructor in case you
     *  don't want to generate all this junk when you make a new Maze.
     *
     *  @param {number} mazeSize How big the maze should be, in design pixels.
     *  @param {mazeView} [view] The view we want to make new rooms for, or null if we want a brand new view.
     *  @return {view} New or resized maze view.
     */
    resizeMazeView:function( mazeView, mazeSize, mazeSizeScale )
    {
        mazeView.resize( mazeSize, mazeSizeScale, this.getWidth(), this.getHeight() );
        
        var x = 0;
        var y = 0;
        for ( x = 0; x < this.getWidth(); x++ )
        {
            for ( y = 0; y < this.getHeight(); y++ )
            {
                var wallDirections = this._rooms.get( x, y ).createClosedArray();

                // hide an outer wall in starting room
                if( x === this.start.x && y === this.start.y )
                {
                    if( x === 0 )
                    {
                       wallDirections.removeAll( [ MazeDirections.Left ] );
                    }
                    else if( x === this.getWidth() - 1 )
                    {
                        wallDirections.removeAll( [ MazeDirections.Right ] );
                    }
                    else if( y === 0 )
                    {
                        wallDirections.removeAll( [ MazeDirections.Down ] );
                    }
                    else if( y === this.getHeight() - 1 )
                    {
                        wallDirections.removeAll( [ MazeDirections.Up ] );
                    }
                }

                mazeView.createRoomView( x, y, wallDirections );
            }
        }

        // now that we have wall sprites in each room, create their sides
        mazeView.createWallSides();
        
        return mazeView;
    },

    /**
     * Returns the Room at the specified position.
     *
     * @param {number} x The x position.
     * @param {number} y The y position.
     * @return {Room} The Room at (x,y).
     */
    getRoom:function( x, y )
    {
        return this._rooms.get( x, y );
    }
});
