var cc = cc;

/**
*  MazeDirections is an enum representing all the directions possible to take in a maze.
*  We could increase the number to generate 3D mazes, if desired.
*  
*  @name MazeDirections
*  @class
*  @readonly
*  @enum {number}
*/
var MazeDirections =
{
    /** @lends MazeDirections */
    
    /** The up direction in a maze. */
    Up:0,
    
    /** The down direction in a maze. */
    Down:1,
    /** The left
     direction in a maze. */
    Left:2,
    
    /** The right direction in a maze. */
    Right:3,
    
    /** The total number of directions in our maze. */
    Count:4
};

/**
 *  A simple class that can be used to get constants that apply across all maze generation files.
 *  @name MazeConstants
 *  @class
 *  @readonly
 */
var MazeConstants = 
{
    /**
     *  An object that contains all the different x/y position deltas that each direction translates to.
     */
    positionDeltas:{},
    
    /**
     *  Names of the different directions, can used for file names and whatnot.
     */
    directionNames:{},
    
    /**
     *  The angles that each direction correspond to, in radians.
     */
    directionAngles:{}
};
MazeConstants.positionDeltas[ MazeDirections.Up ] = { x:0, y:1 };
MazeConstants.positionDeltas[ MazeDirections.Down ] = { x:0, y:-1 };
MazeConstants.positionDeltas[ MazeDirections.Left ] = { x:-1, y:0 };
MazeConstants.positionDeltas[ MazeDirections.Right ] = { x:1, y:0 };
MazeConstants.directionNames[ MazeDirections.Up ] = "Up";
MazeConstants.directionNames[ MazeDirections.Down ] = "Down";
MazeConstants.directionNames[ MazeDirections.Left ] = "Left";
MazeConstants.directionNames[ MazeDirections.Right ] = "Right";
MazeConstants.directionAngles[ MazeDirections.Up ] = Math.PI / 2;
MazeConstants.directionAngles[ MazeDirections.Down ] = 3 * Math.PI / 2;
MazeConstants.directionAngles[ MazeDirections.Left ] = Math.PI;
MazeConstants.directionAngles[ MazeDirections.Right ] = 0;

/**
 *  A simple class that can be used to call various static utility functions related to maze creation.
 *  @name MazeUtility
 *  @class
 *  @readonly
 */
var MazeUtility =
{
    /** @lends MazeUtility */
    /**
     * Prints out a maze to the console.
     *
     * @param {Maze} The maze to print.
     */
    printMaze: function( maze )
    {
        cc.log( MazeUtility.getMazeAsString( maze ) );
    },

    /**
     * Returns a passed Maze as a string.
     *
     * @param {Maze} The maze to convert to a string.
     */
    getMazeAsString: function( maze )
    {
        var mazeString = "\n";
        var j;
        for ( j = 0; j < maze.getHeight(); j++ )
        {
            var line = "";
            var i;
            var room;
        
            for ( i = 0; i < maze.getWidth(); i++ )
            {
                room = maze.getRoom( i, j );
                line += room.isOpen( MazeDirections.Up ) ? "* *" : "***";
            }
            mazeString += line + "\n";
            line = "";
        
            for ( i = 0; i < maze.getWidth(); i++ )
            {
                room = maze.getRoom( i, j );
                line += room.isOpen( MazeDirections.Left ) ? "  " : "* ";
                line += room.isOpen( MazeDirections.Right ) ? " " : "*";
            }
            mazeString += line + "\n";
            line = "";
        
            for ( i = 0; i < maze.getWidth(); i++ )
            {
                room = maze.getRoom( i, j );
                line += room.isOpen( MazeDirections.Down ) ? "* *" : "***";
            }
            mazeString += line + "\n";
        }
        return mazeString;
    },
    
    /**
     *  Get the direction given a delta object or an x and y pair. Returns -1 if it can't find one.
     *
     *  @param {Object|number} deltaOrX Either a delta object with an X and Y, or just the X value.
     *  @param {number} y The Y, should only use if you provided an X value.
     *  @returns {MazeDirections} The direction for this delta.
     */
    getDirection: function( deltaOrX, y )
    {
        var x = deltaOrX;
        if ( deltaOrX instanceof Object )
        {
            x = deltaOrX.x;
            y = deltaOrX.y;
        }

        var epsilon = 0.001;

        var directionIndex;
        for ( directionIndex = 0; directionIndex < MazeDirections.Count; directionIndex++ )
        {
            if ( Math.abs( MazeConstants.positionDeltas[ directionIndex ].x - x ) < epsilon &&
                Math.abs( MazeConstants.positionDeltas[ directionIndex ].y - y ) < epsilon )
            {
                return directionIndex;
            }
        }
        return -1;
    },

    getDirectionBetween: function( fromGridPos, toGridPos  )
    {
        if( fromGridPos && toGridPos )
        {
            var difference = cc.p( toGridPos.x - fromGridPos.x, toGridPos.y - fromGridPos.y );
            var magnitude = Math.sqrt( difference.x * difference.x + difference.y * difference.y );
            var normal = cc.p( difference.x / magnitude, difference.y / magnitude );
            var direction = MazeUtility.getDirection( normal );
            return direction;
        }
        return undefined;
    },
    
    /**
     *  Returns the direction opposite the passed one.
     *
     *  @param {MazeDirections} direction The direction.
     *  @returns {MazeDirections} The opposite of the passed direction, or -1 if you passed in an unknown direction.
     */
    getOppositeDirection: function( direction )
    {
        switch( direction )
        {
            case MazeDirections.Up:
                return MazeDirections.Down;
            case MazeDirections.Down:
                return MazeDirections.Up;
            case MazeDirections.Left:
                return MazeDirections.Right;
            case MazeDirections.Right:
                return MazeDirections.Left;
        }
        return -1;
    },

    /**
     * Given a current maze-relative direction, figure out what the absolute direction in world terms.
     * @param mazeDir   Maze-relative maze direction.
     * @returns {MazeDirections}    World-relative maze direction.
     */
    convertMazeDirectionToWorldDirection:function( mazeDir, mazeRotation )
    {
        var normal = MazeConstants.positionDeltas[mazeDir];
        if( normal )
        {
            var rotatedNormal = LumosityUtils.rotateVectorByAngle( normal, cc.POINT_ZERO, mazeRotation * -Math.PI / 180 );
            var rotatedDirection = MazeUtility.getDirection( rotatedNormal );
            return rotatedDirection;
        }
        return MazeDirections.Up;
    },
    
    /**
     *  Returns whether or not the passed direction is a valid one.
     *
     *  @param {MazeDirection} direction The direction.
     *  @returns {boolean} If the direction is valid, that is, it exists in the MazeDirections enum.
     */
    isValidDirection: function( direction )
    {
        return direction >= 0 && direction < MazeDirections.Count;
    },
    
    /**
     *  Creates a new array with all directions in it.
     *  
     *  @returns {Array} An array with all directions.
     */
    createArrayOfAllDirections: function()
    {
        var arr = [];
        var direction;
        for ( direction = 0; direction < MazeDirections.Count; direction++ )
        {
            arr.push( direction );
        }
        return arr;
    },

    /**
     * Given two paths, get their shared subpath from the start (if it exists ).
     * @param pathA     In-order array of grid positions.
     * @param pathB     In-order array of grid positions.
     * @returns {Array} New array of shared grid positions.
     */
    getCommonPath:function( pathA, pathB )
    {
        var commonPath = [];
        if( pathA && pathB )
        {
            var commonLength = Math.min( pathA.length, pathB.length );
            var i = 0;
            for ( i = 0; i < commonLength; i++ )
            {
                if ( pathA[ i ].x !== pathB[ i ].x || pathA[ i ].y !== pathB[ i ].y )
                {
                    // paths have now diverged, so the common path ends here
                    break;
                }
                // paths share this grid position and all others before it, so add to common path
                commonPath.push( cc.p( pathA[ i ].x, pathA[ i ].y ) );
            }
        }
        return commonPath;
    }
};