var require = require;
var cc = cc;

require( "src/Game/Maze/MazeConstants.js" );
var MazeConstants = MazeConstants;
var MazeDirections = MazeDirections;
var MazeUtility = MazeUtility;

require( "src/Utility/MultiDimensionArray.js" );
var MultiDimensionArray = MultiDimensionArray;

require( "src/Game/Maze/Room.js" );
var Room = Room;

/**
 * The MazeGenerator is used to create the mazes used by Penguin Pursuit. It is
 * relatively efficient, although very large mazes (200x200 or bigger) can take
 * some time to create. Fortunately, we don't need anything bigger than 9x9.
 * 
 * The algorithm works by creating individual groups of rooms and then charting
 * them using the Drunkard's Walk algorithm (i.e. random movement) until they
 * can't continue any further. Then, a new group is started. If two groups meet
 * one another, they are combined into one group by making an opening in a wall
 * and assigning them the same (lower) group number. Eventually, every room
 * will be in group 0, and there will be exactly one path to any point.
 * 
 * The algorithm was modified in version 2.0 to make it significantly "more
 * random," but also much slower. Version 1.0 functioned entirely without groups;
 * instead, new rooms were added iteratively, and any "island" rooms were linked
 * when they were added. Because of the iterative fashion, it was guaranteed that
 * a new room was always added adjacent to an old one, but it also meant that
 * the mazes tended to be solved via the same patterns.
 * 
 * In 2.1, I ported it to JS, and I also got rid of a lot of dumb switch statements
 * and the like for each direction, instead replacing them with for loops that
 * go through every direction, accompanied by various utility functions.
 *
 * Because this game will never used mazes larger than 9x9 or so, it's
 * by far more preferable to go with algorithm 2.0.
 * 
 * @name MazeGenerator
 * @class
 * @author Eli Delventhal
 * @version 2.0 - Ported to JS for Penguin Pursuit
 */
var MazeGenerator = {};

/**
 * Creates a maze given a width, height, and desired distance.
 * @param {number} width Number of grid squares in x-dimension.
 * @param {number} height Number of grid squares in y-dimension.
 * @param {number} idealDistance Ideal distance of goal from start.
 * @returns {MultiDimensionArray} Returns 2D array of rooms.
 */
MazeGenerator.generateMaze = function( width, height, idealDistance, minTurns, maxTurns )
{
    //Initialize the maze with no rooms at all (all set to null).
    var rooms = new MultiDimensionArray( width, height );

    //Every room belongs to a group. If a room is created adjacent to
    //a different numbered group, a connection is created and all
    //connected rooms are given the lowest group number of the two.
    var groups = new MultiDimensionArray( width, height );
    groups.fill( 100000 ); //set all default group values to a very high number
    var currentGroup = 0;

    // get a random solution path of specified length, beginning at start point
    var solution = MazeGenerator._generateSolution( width, height, idealDistance, minTurns, maxTurns );
    var distance = solution.length - 1;

    // get start and goal from solution
    var start = cc.p( solution[ 0 ].x, solution[ 0 ].y );
    var goal = cc.p( solution[ distance ].x, solution[ distance ].y );

    // This list holds all coordinates where a room has not been created.
    var emptyCoords = [];
    var x;
    var y;
    for ( x = 0; x < width; x++ )
    {
        for ( y = 0; y < height; y++ )
        {
            emptyCoords.push( { x:x, y:y } );
        }
    }

    // Create rooms for solution path.
    for( p = 0; p < solution.length; p++ )
    {
        var point = solution[p];

        //Set the group of this room.
        groups.set( currentGroup, point.x, point.y );

        //Add the new room.
        MazeGenerator._addNewRoom( rooms, point.x, point.y, groups );

        //Remove the room's coordinate from the empty list.
        MazeGenerator._removeCoordinateFromList( emptyCoords, point );
    }

    // Link together all rooms in solution path.
    for( p = 1; p < solution.length; p++ )
    {
        MazeGenerator._linkRooms( rooms, solution[p].x, solution[p].y, solution[p-1].x, solution[p-1].y );
    }

    currentGroup++;

    // For the non-solution rooms, keep adding them as long as empty ones remain.
    while ( emptyCoords.length > 0 )
    {
        //Start creating rooms at a random coordinate.
        var c = emptyCoords[ Math.floor( Math.random() * emptyCoords.length ) ];
        MazeGenerator._constructRooms( rooms, c.x, c.y, goal, MazeUtility.createArrayOfAllDirections(), emptyCoords, groups, currentGroup );

        //When construct rooms finishes, a single group of rooms will have been created.
        //So, increment the current group number for the next group of rooms.
        currentGroup++;
    }

    var turnCount = MazeGenerator._getTurnCount( solution );

    var mazeInfo =
    {
        start : start,
        goal : goal,
        solution : solution,
        rooms : rooms,
        turnCount : turnCount
    };

    return mazeInfo;
};

/**
 * Creates a maze from authored maze data in the form of rooms' open directions.
 *
 * @param {MultiDimensionArray} authoredMazeData Two-dimensional array of strings that represent the open directions in each room (e.g., "ud" for openings in Up and Down directions)
 * @returns {MultiDimensionArray} Returns 2D array of rooms.
 */
MazeGenerator.loadAuthoredMaze = function( authoredMazeData )
{
    var solution = authoredMazeData.solution.slice( 0 );
    var solutionLen = solution ? solution.length : 0;
    var start = cc.p( solution[ 0 ].x, solution[ 0 ].y );
    var goal = cc.p( solution[ solutionLen - 1 ].x, solution[ solutionLen - 1 ].y );

    var open = authoredMazeData.open;
    var width = open.length;
    var height = open[ 0 ].length;

    var rooms = new MultiDimensionArray( width, height );
    for( var y = 0; y < height; y++ )
    {
        var roomOpeningsForRow = open[ height - 1 - y ];
        for( var x = 0; x < width; x++ )
        {
            var room = MazeGenerator._addNewRoom( rooms, x, y );
            var roomOpenings = roomOpeningsForRow[ x ];
            room.setIsOpen( MazeDirections.Up, roomOpenings.indexOf( "u" ) >= 0 );
            room.setIsOpen( MazeDirections.Down, roomOpenings.indexOf( "d" ) >= 0 );
            room.setIsOpen( MazeDirections.Left, roomOpenings.indexOf( "l" ) >= 0 );
            room.setIsOpen( MazeDirections.Right, roomOpenings.indexOf( "r" ) >= 0 );
        }
    }

    var turnCount = MazeGenerator._getTurnCount( solution );

    var mazeInfo =
    {
        start : start,
        goal : goal,
        solution : solution,
        rooms : rooms,
        turnCount : turnCount
    };

    return mazeInfo;
},

MazeGenerator._getTurnCount = function( solution )
{
    var turnCount = 0;
    if( solution && solution.length )
    {
        for( var i = 2; i < solution.length; i++ )
        {
            var lastDelta = cc.p( solution[i - 1].x - solution[i - 2].x, solution[i - 1].y - solution[i - 2].y );
            var lastDir = MazeUtility.getDirection( lastDelta );

            var delta = cc.p( solution[i].x - solution[i - 1].x, solution[i].y - solution[i - 1].y );
            var dir = MazeUtility.getDirection( delta );

            if( dir !== lastDir )
            {
                turnCount++;
            }
        }
    }
    return turnCount;
},

/**
 * Generates a random solution path for given maze parameters.
 *
 * @param {number} width Width of maze.
 * @param {number} height Height of maze.
 * @param {number} idealDistance Desired solution distance, in grid spaces.
 * @param {number} minTurns Desired minimum number of turns in solution.
 * * @param {number} maxTurns Desired maximum number of turns in solution.
 */
MazeGenerator._generateSolution = function( width, height, idealDistance, minTurns, maxTurns )
{
    // cut the distance down to its maximum possible value
    idealDistance = Math.min( idealDistance, width * height - 1 );

    var start = null;
    var solution = null;
    var directions = null;
    var distanceAttempt = 0;

    var MAX_DISTANCE_ATTEMPTS = 3;
    var MIN_IDEAL_DISTANCE = 3;
    var MIN_MIN_TURNS = 1;
    var MAX_MAX_TURNS = idealDistance;

    while( true )
    {
        // try several times to generate a maze at the desired distance, since that's important
        for( distanceAttempt = 0; distanceAttempt < MAX_DISTANCE_ATTEMPTS; distanceAttempt++ )
        {
            // pick a random start at the edge of maze
            start = LumosityUtils.randomBool()
                ? cc.p( LumosityUtils.randomBool() ? 0 : width - 1, LumosityUtils.getRandomInt( 0, height - 1 ) ) // left or right
                : cc.p( LumosityUtils.getRandomInt( 0, width - 1 ), LumosityUtils.randomBool() ? 0 : height - 1 ); // top or bottom
            solution = [];
            directions = [];
            //ll.verbose("GENERATING MAZE OF DISTANCE " + idealDistance );
            if( MazeGenerator._getRandomPath( width, height, start, idealDistance, 0, minTurns, maxTurns, solution, directions ) )
            {
                return solution;
            }
        }

        ll.verbose("REDUCING MAZE COMPLEXITY");
        if( idealDistance > MIN_IDEAL_DISTANCE )
        {
            idealDistance--;
        }
        if( minTurns > MIN_MIN_TURNS )
        {
            minTurns--;
        }
        if( maxTurns < MAX_MAX_TURNS )
        {
            maxTurns++;
        }
    }
};

/**
 * Recursively returns a randomly-generated path of specified length through a maze.
 * @param {number} width Size of maze in x-dimension.
 * @param {number} height Size of maze in y-dimension.
 * @param {ccp} pathPoint Starting path point from which to search.
 * @param {number} distanceToGo Length of path
 * @param {ccp Array} pathPoints Returned list of grid points in path.
 * @param {direction} incomingDirection Direction from last point in path.
 * @returns {boolean} Returns whether recursion has completed at specified distance.
 * @private
 */
MazeGenerator._getRandomPath = function( width, height, pathPoint, distanceToGo, turns, minTurns, maxTurns, pathPoints, directions, incomingDirection )
{
    if ( !pathPoint || distanceToGo === undefined || distanceToGo < 0 || !pathPoints )
    {
        // Invalid path
        return false;
    }

    if( pathPoints.containsPoint( pathPoint ) )
    {
        // We've already included path point in results.
        return false;
    }

    pathPoints.push( pathPoint );
    directions.push( incomingDirection );

    if( distanceToGo == 0 )
    {
        // Path is complete.
        if( turns < minTurns || turns > maxTurns )
        {
            //ll.verbose("BAD PATH! " + turns + " TURNS (" + minTurns + " / " + maxTurns + ")" );
            return false;
        }

        //ll.verbose( "COMPLETE PATH! " + turns + " TURNS (" + minTurns + " / " + maxTurns + ")" );
        return true;
    }

    var moveDirs = MazeGenerator._getNextMoveDirections( distanceToGo, turns, minTurns, maxTurns, pathPoints, directions, incomingDirection );

    var dirIndex = 0;
    for ( dirIndex = 0; dirIndex < moveDirs.length; dirIndex++ )
    {
        // Try the next direction in our randomized list
        var dir = moveDirs[ dirIndex ];
        var moveDelta = MazeConstants.positionDeltas[ dir ];
        var nextPathPoint = cc.pAdd( pathPoint, moveDelta );
        if( nextPathPoint.x >= 0 && nextPathPoint.x < width && nextPathPoint.y >= 0 && nextPathPoint.y < height )
        {
            // Next direction is valid, so let's extend the path that way

            // add to number of turns if this movement is turning the path
            var nextTurns = turns;
            var isTurn = ( incomingDirection !== undefined && dir !== MazeUtility.getOppositeDirection( incomingDirection ) && dir !== incomingDirection );
            if( isTurn )
            {
                nextTurns++;
            }

            var success = MazeGenerator._getRandomPath( width, height, nextPathPoint, distanceToGo - 1, nextTurns, minTurns, maxTurns, pathPoints, directions, dir );
            if( success )
            {
                // This path ended in a valid goal, so stop looking
                return true;
            }
        }
    }

    // This point dead-ended, so pop it off the stack to backtrack
    pathPoints.pop();
    directions.pop();

    return false;
};

MazeGenerator._getNextMoveDirections = function( distanceToGo, turns, minTurns, maxTurns, pathPoints, directions, incomingDirection )
{
    // randomize order of the next directions to try
    var allDirs = [ MazeDirections.Up, MazeDirections.Down, MazeDirections.Left, MazeDirections.Right ];
    var moveDirs = allDirs.slice( 0 );
    LumosityUtils.shuffle( moveDirs );

    if ( incomingDirection !== undefined )
    {
        // we can adjust path based in incoming direction
        var i = directions.length - 1;

        // recognize if we're currently making a long, straight corridor
        var inCorridor = directions.length >= 4 &&
            directions[i]   === directions[i - 1] &&
            directions[i]   === directions[i - 2];

        // recognize if we're currently making a back-and-forth twisty path like lombard street
        var inLombardSt = directions.length >= 6 &&
            directions[i]       === MazeUtility.getOppositeDirection( directions[i - 2] ) &&
            directions[i]       === directions[i - 4] &&
            directions[i - 1]   === directions[i - 3];

        // recognize if we're currently making a stair-stepped path
        var inStairStep = directions.length >= 4 &&
            directions[i]      === directions[i - 2] &&
            directions[i - 1]  === directions[i - 3];

        // calculate relative movement options based on incoming direction
        var straightDir = incomingDirection;
        var backwardDir = MazeUtility.getOppositeDirection( incomingDirection );
        var turnDirs = allDirs.slice( 0 );
        turnDirs.remove( straightDir );
        turnDirs.remove( backwardDir );

        if( inCorridor )
        {
            //ll.verbose( "MOVING CORRIDOR DIR " + MazeConstants.directionNames[incomingDirection] + " TO END" );
            moveDirs.remove( incomingDirection );
            moveDirs[moveDirs.length] = incomingDirection;
        }
        else if( inLombardSt )
        {
            // in lombard street; take the direction that would continue the pattern and put it at the end of our options
            var nextLombardStDir = MazeUtility.getOppositeDirection( directions[i - 1] );
            //ll.verbose( "MOVING NEXT LOMBARD ST DIR " + MazeConstants.directionNames[nextLombardStDir] + " TO END" );
            moveDirs.remove( nextLombardStDir );
            moveDirs[moveDirs.length] = nextLombardStDir;
        }
        else if( inStairStep )
        {
            // in stair-step; take the direction that would continue the pattern and put it at the end of our options
            var nextStairStepDir = directions[i - 1];
            //ll.verbose( "MOVING NEXT STAIR STEP DIR " + MazeConstants.directionNames[nextStairStepDir] + " TO END" );
            moveDirs.remove( nextStairStepDir );
            moveDirs[moveDirs.length] = nextStairStepDir;
        }

        if( turns >= maxTurns )
        {
            // enough turns; remove turning directions
            moveDirs = [ straightDir ];
        }
        else if( turns + distanceToGo <= minTurns )
        {
            // need more turns; remove straight option
            moveDirs.remove( straightDir );
        }

        // never try the room we just came from
        moveDirs.remove( backwardDir );
    }

    return moveDirs;
},

/**
 * Constructs a room of the maze at the specified x and y positions.
 * It will then continue to call itself until it reaches a point
 * where it can no longer create more rooms.
 * @param {MultiDimensionArray} rooms The 2D array of rooms to work with.
 * @param {number} x The x position to start with.
 * @param {cc.p} goal The grid position of the goal.
 * @param {number} y The y position to start with.
 * @param {Array} uncheckedDirs The directions that have not yet been checked.
 * @param {Array} emptyCoords A list of all empty rooms' coordinates.
 * @param {MultiDimensionArray} groups A 2D array of all the rooms' group numbers.
 * @param {number} currentGroup The number value of the current group.
 */
MazeGenerator._constructRooms = function( rooms, x, y, goal, uncheckedDirs, emptyCoords, groups, currentGroup )
{
    //If all four directions have been tried, return from the function.
    if ( uncheckedDirs.length <= 0 )
    {
        return;
    }
        
    //If we don't already have a room here, create one.
    if ( !rooms.get( x, y ) )
    {
        //Set the group of this room.
        groups.set( currentGroup, x, y );
        
        //Add the new room.
        MazeGenerator._addNewRoom( rooms, x, y, groups );

        //If this new room is adjacent to rooms that are in any other group,
        //link them and combine their groups.
        this._randomlyCombineGroup( rooms, x, y, goal, groups );
        
        //Remove the room's coordinate from the empty list.
        MazeGenerator._removeCoordinateFromList( emptyCoords, { x:x, y:y } );
        
        //It's possible the group has changed, so be sure to reflect that.
        currentGroup = groups.get( x, y );
    }
    
    //Open up a random direction for this room. If that direction already has a room,
    //try a different direction. If there is no room there, call this function at the new position.
    var directionIndex = Math.floor( Math.random() * uncheckedDirs.length );

    //Remove the target direction from the list and assign it.
    var direction = uncheckedDirs.splice( directionIndex, 1 );
    
    //Create delta x and delta y variables to map the change.
    var delta = MazeConstants.positionDeltas[ direction ];
    
    //If the change will take us out of bounds, try again.
    if ( !MazeGenerator._isValidCoordinate( rooms, x + delta.x, y + delta.y ) )
    {
        this._constructRooms( rooms, x, y, goal, uncheckedDirs, emptyCoords, groups, currentGroup );
    }
    //Otherwise, it's okay to go in this direction.
    else
    {
        //If there isn't a room there, open us to that space and then construct a room there.
        if ( !rooms.get( x + delta.x, y + delta.y ) )
        {
            rooms.get( x, y ).setIsOpen( direction, true );
            this._constructRooms( rooms, x + delta.x, y + delta.y, goal, MazeUtility.createArrayOfAllDirections(), emptyCoords, groups, currentGroup );
        }
        //If there is already a room there, try again.
        else
        {
            this._constructRooms( rooms, x, y, goal, uncheckedDirs, emptyCoords, groups, currentGroup );
        }
    }
};

/**
 * Adds a new room to the maze, linking it to adjacent rooms correctly.
 * @param {MultiDimensionArray} rooms The maze, a 2D array of rooms.
 * @param {number} x The x position to add in.
 * @param {number} y The y position to add in.
 * @param {MultiDimensionArray} groups The groups of all the rooms.
 */
MazeGenerator._addNewRoom = function( rooms, x, y, groups )
{
    var newRoom = new Room();
    rooms.set( newRoom, x, y );

    //Search for adjacent rooms already linked to this one and link back to them.
    var direction;
    for ( direction = 0; direction < MazeDirections.Count; direction++ )
    {
        var targetX = x + MazeConstants.positionDeltas[ direction ].x;
        var targetY = y + MazeConstants.positionDeltas[ direction ].y;
        
        if ( MazeGenerator._isValidCoordinate( rooms, targetX, targetY ) )
        {
            if ( rooms.exists( targetX, targetY ) && rooms.get( targetX, targetY ).isOpen( MazeUtility.getOppositeDirection( direction ) ) )
            {
                newRoom.setIsOpen( direction, true );
            }
        }
    }

    return newRoom;
};

/**
 * Links two rooms together.
 * @param {MultiDimensionArray} rooms The maze of rooms, a 2D array of Room objects.
 * @param {number} x1 The x coordinate of the first room.
 * @param {number} y1 The y coordinate of the first room.
 * @param {number} x2 The x coordinate of the second room.
 * @param {number} y2 The y coordinate of the second room.
 */
MazeGenerator._linkRooms = function( rooms, x1, y1, x2, y2 )
{
    //If the rooms aren't adjacent, return.
    if ( Math.abs( x1 - x2 ) + Math.abs( y1 - y2 ) !== 1 )
    {
        return;
    }
    
    var xDiff = x1 - x2;
    var yDiff = y1 - y2;
    
    var direction = MazeUtility.getDirection( xDiff, yDiff );
    if ( MazeUtility.isValidDirection( direction ) )
    {
        rooms.get( x1, y1 ).setIsOpen( MazeUtility.getOppositeDirection( direction ), true );
        rooms.get( x2, y2 ).setIsOpen( direction, true );
    }
};

/**
 * Looks for adjacent rooms that aren't in the same group. If it finds one,
 * it links to it and combines their groups.
 * @param {MultiDimensionArray} rooms The 2D array maze of rooms.
 * @param {number} x The x coordinate of this room.
 * @param {number} y The y coordinate of this room.
 * @param {MultiDimensionArray} groups The groups of the rooms.
 */
MazeGenerator._randomlyCombineGroup = function( rooms, x, y, goal, groups )
{    
    //Add all the group possibilities to a list by determining
    //in what directions rooms actually exist and are of a different group.
    var possibilities = [];
    
    var direction;
    for ( direction = 0; direction < MazeDirections.Count; direction++ )
    {
        var targetX = x + MazeConstants.positionDeltas[ direction ].x;
        var targetY = y + MazeConstants.positionDeltas[ direction ].y;

        if ( MazeGenerator._isValidCoordinate( rooms, targetX, targetY ) )
        {
            if( targetX === goal.x && targetY == goal.y )
            {
                // don't break down any new walls to goal; we want it to be a dead end
                continue;
            }

            if ( rooms.exists( targetX, targetY ) && groups.get( x, y ) !== groups.get( targetX, targetY ) )
            {
                possibilities.push( direction );
            }
        }
    }
    
    //Only continue if there are adjacent rooms.
    if ( possibilities.length > 0 )
    {
        //Choose the random direction.
        var random = Math.floor( Math.random() * possibilities.length );
        
        //If we got here, we know that whatever choice was decided was a room that exists
        //and is not in the same group as this one. So, link the two and combine groups.
        direction = possibilities[ random ];
        var delta = MazeConstants.positionDeltas[ direction ];
        
        this._combineGroups( rooms, x, y, x + delta.x, y + delta.y, groups );
        this._linkRooms( rooms, x, y, x + delta.x, y + delta.y );
    }
};

/**
 * Given two rooms, sets the groups of them and all their connecting rooms
 * to the same value (the lowest of their two group numbers).
 * @param {MultiDimensionArray} rooms The 2D array maze of rooms.
 * @param {number} x1 The x coordinate of the first room.
 * @param {number} y1 The y coordinate of the first room.
 * @param {number} x2 The x coordinate of the second room.
 * @param {number} y2 The y coordinate of the second room.
 * @param {number} groups The groups of all the rooms.
 */
MazeGenerator._combineGroups = function( rooms, x1, y1, x2, y2, groups )
{
    //If the two rooms are already in the same group, return.
    if ( groups.get( x1, y1 ) === groups.get( x2, y2 ) )
    {
        return;
    }
        
    var newGroup = 0;
    //A list of rooms that need to have their group changed.
    var open = [];
    //A list of rooms that have already had their group changed.
    var closed = [];
    
    //Otherwise, set the new group number as the smallest of the two group numbers
    //and add the lower room to the closed list and the higher room to the open list.
    if ( groups.get( x1, y1 ) < groups.get( x2, y2 ) )
    {
        newGroup = groups.get( x1, y1 );
        open.push( { x:x2, y:y2 } );
        closed.push( { x:x1, y:y1 } );
    }
    else if ( groups.get( x2, y2 ) < groups.get( x1, y1 ) )
    {
        newGroup = groups.get( x2, y2 );
        open.push( { x:x1, y:y1 } );
        closed.push( { x:x2 , y:y2 } );
    }
    
    //Now loop as long as some rooms need to be changed.
    while ( open.length > 0 )
    {
        var c = open.pop();
        
        //Change the group for this room.
        groups.set( newGroup, c.x, c.y );
        
        //Add this room to the closed list and remove it from the open list.
        closed.push( c );
        
        //Add all linked rooms not already in the closed list to the open list.
        //If the room is already of the same group, add it to the closed list.
        var direction;
        for ( direction = 0; direction < MazeDirections.Count; direction++ )
        {
            var targetX = c.x + MazeConstants.positionDeltas[ direction ].x;
            var targetY = c.y + MazeConstants.positionDeltas[ direction ].y;
            var tC = { x:targetX, y:targetY };
            
            if ( MazeGenerator._isValidCoordinate( rooms, targetX, targetY ) )
            {
                if ( rooms.get( c.x, c.y ).isOpen( direction ) && rooms.exists( targetX, targetY ) && !MazeGenerator._listContainsCoordinate( closed, tC ) )
                {
                    if ( groups.get( targetX, targetY ) === newGroup )
                    {
                        closed.push( tC );
                    }
                    else
                    {
                        open.push( tC );
                    }
                }
            }
        }
    }
};

/**
 *  Returns whether or not the passed coordinate is valid (fits in the bounds).
 *
 *  @param {MultiDimensionArray} rooms The maze the check for validity for.
 *  @param {Object|number} coordinateOrX Either the coordinate object with an X and Y, or the X position.
 *  @param {number} y The Y coordinate. Don't use if you provided a coordinate object.
 *  @returns {boolean} True or false if we are in bounds.
 */
MazeGenerator._isValidCoordinate = function( rooms, coordinateOrX, y )
{
    var x = coordinateOrX;
    if ( x instanceof Object )
    {
        x = coordinateOrX.x;
        y = coordinateOrX.y;
    }
    return ( x >= 0 && y >= 0 && x < rooms.getLength( 0 ) && y < rooms.getLength( 1 ) );
};

/**
 * Checks whether or not a coordinate object inside the list has the same X and Y as the passed one.
 * @param {Array} list The list to check.
 * @param {Object} c The Coordinate to search for.
 * @returns {boolean} True or False if the list contains the Coordinate.
 */
MazeGenerator._listContainsCoordinate = function( list, c )
{
    var listIndex;
    for ( listIndex = 0; listIndex < list.length; listIndex++ )
    {
        if ( list[ listIndex ].x === c.x && list[ listIndex ].y === c.y )
        {
            return true;
        }
    }
    return false;
};

/**
 * Returns the first Coordinate that equals c from the passed list.
 * @param {Array} list The list to remove from.
 * @param {Object} c The coordinate to remove.
 */
MazeGenerator._removeCoordinateFromList = function( list, c )
{
    var listIndex;
    for ( listIndex = 0; listIndex < list.length; listIndex++ )
    {
        if ( list[ listIndex ].x === c.x && list[ listIndex ].y === c.y )
        {
            list.splice( listIndex, 1 );
            return;
        }
    }
};
