var cc = cc;
var require = require;

require( "src/Game/Maze/MazeConstants.js" );
var MazeDirections = MazeDirections;
var MazeConstants = MazeConstants;
var MazeUtility = MazeUtility;

require( "src/Utility/MultiDimensionArray.js" );
var MultiDimensionArray = MultiDimensionArray;

require( "src/Game/Maze/RoomView.js" );
var RoomView = RoomView;

require( "src/Game/Bump.js" );
var Bump = Bump;

var MazeView = cc.Node.extend(
{
   /** @lends MazeView.prototype */
   
   /** A MultiDimensionArray that represents all the room views. */
   _roomViews:null,

    /** The margin from the left and the top that the rooms start being place. Calculated using the maze size and room count. */
    _margin:null,

    /** How many pixels each room takes up in this maze view. Calculated using the full maze size. */
    _tileSize:50,

    /** The background image for the maze. */
    _background:null,

    /** A Node that holds all the rooms. */
    _roomContainer:null,

    /** A Node that holds all the maze inhabitants. */
    _inhabitantContainer:null,

    /** Compass point that rotates with maze. */
    _compass:null,

    /** The current game settings. */
    _settings:null,

    /** Whether footstep revealing is revealed. */
    _footstepsEnabled:false,

   /**
    *   A MazeView is the visual representation of a Maze object. It
    *   doesn't know anything about the actual Maze.  Instead, it is
    *   created by passing in an initial tile size and then creating
    *   RoomViews piece by piece.
    *
    *   @constructs
    *   @param {cc.Size} size How big the entire maze should be in pixels.
    *   @param {number} columns How many columns we have (rooms along the X dimension)
    *   @param {number} rows How many rows we have (rooms along the Y dimension)
    *   @param {Object} settings The current game settings.
    */
   ctor: function( mazeSize, settings )
   {
        this._super();

        this.id = "mazeView";

        this.init();

        this._settings = settings;

        this._background = LumosityUtils.createRectangularSprite( mazeSize, settings.mazeBackgroundColor );
        this._background.id = "mazeViewBackground";
        this._background.setOpacity( settings.mazeBackgroundOpacity );
        this.addChild( this._background, 0 );

        this._wallSidesRoot = cc.Node.create();
        this._wallSidesRoot.id = "mazeViewWallSidesRoot";
        this.addChild( this._wallSidesRoot );

        this._wallSidesContainer = cc.Node.create();
        this._wallSidesContainer.id = "mazeViewWallSidesContainer";
        this._wallSidesContainer.setPosition( cc.p( 0, -5 ) );
        this._wallSidesRoot.addChild( this._wallSidesContainer );

        this._roomContainer = cc.Node.create();
        this._roomContainer.id = "mazeViewRoomContainer";
        this.addChild( this._roomContainer );

        this._inhabitantContainer = cc.Node.create();
        this._inhabitantContainer.id = "mazeViewInhabitantContainer";
        this.addChild( this._inhabitantContainer );

        this._compass = cc.Sprite.create( "UI/compassRoundBackground.png" );
        this._compass.id = "mazeViewCompass";
        this._compass.setAnchorPoint( 0.5, 0 );
        this._compass.setScaleX( this.getScaleX() * 84 / this._compass.getContentSize().width );
        this._compass.setScaleY( this.getScaleY() * 42 / this._compass.getContentSize().height );
        this._compass.setColor( this._settings.wallColor );
        this.addChild( this._compass, 10 );

        this._compassPointer = cc.Sprite.create( "UI/compassRoundPointer.png" );
        this._compassPointer.id = "mazeViewCompassPointer";
        this._compassPointer.setAnchorPoint( 0.5, 0 );
        this._compassPointer.setScaleX( this.getScaleX() * 84 / this._compassPointer.getContentSize().width );
        this._compassPointer.setScaleY( this.getScaleY() * 42 / this._compassPointer.getContentSize().height );
        this.addChild( this._compassPointer, 11 );

        this._compassIndicator = cc.Sprite.create( "UI/tutorialCircle.png" );
        this._compassIndicator.id = "mazeViewCompassIndicator";
        this._compassIndicator.setAnchorPoint( 0.5, 0.5 );
        this._compassIndicator.setScaleX( this.getScaleX() * settings.compassIndicatorScale / this._compassIndicator.getContentSize().width );
        this._compassIndicator.setScaleY( this.getScaleY() * settings.compassIndicatorScale / this._compassIndicator.getContentSize().height );
        this.addChild( this._compassIndicator );
        this._compassIndicator.setOpacity( 0 );

        this._compassLabel = LumosityLabel.create( localize( "pointer"), "fonts/MuseoSans-500", 30 );
        this._compassLabel.id = "mazeViewCompassLabel";
        this._compassLabel.setColor( cc.WHITE );
        this._compassLabel.setAnchorPoint( cc.p( 0, 0.5 ) );
        this._compassLabel.setHorizontalAlignment( cc.TEXT_ALIGNMENT_LEFT );
        this._compassLabel.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
        this.addChild( this._compassLabel );
        this._compassLabel.setOpacity( 0 );

        var pointerWidth = this._compassPointer.getScaleX() * this._compassPointer.getContentSize().width;

        this.resize( mazeSize, 1, 1, 1 );
   },
   
   /**
    *   Resizes the maze. In doing this, it will empty the room container (destroying all room views).
    *   Room views must be manually created later by calling createRoomView().
    *   @param {cc.Size} size How big the entire maze should be in pixels.
    *   @param {number} columns How many columns we have (rooms along the X dimension)
    *   @param {number} rows How many rows we have (rooms along the Y dimension)
    */
   resize:function( size, sizeScale, columns, rows )
   {
       var boundsSize = cc.size( size.width * sizeScale, size.height * sizeScale );

       this._roomViews = new MultiDimensionArray( columns, rows );

       this._wallThickness = size.width * 0.01;

       //the size of each room is determined by the minimum maze dimension divided by the maximum room dimension
       var minDimension = Math.min( boundsSize.width, boundsSize.height );
       this._tileSize = Math.floor( ( minDimension - this._wallThickness ) / Math.max( columns, rows ) );

       //the margin is the difference between maze size and size needed for the rooms, divided by 2
       this._margin = cc.size( Math.floor( boundsSize.width - this._tileSize * columns ) / 2, Math.floor( boundsSize.height - this._tileSize * rows ) / 2 );

       //make sure the background is scaled to fit in the view.
       this._background.setScaleX( ( columns * this._tileSize + this._wallThickness ) / this._background.getContentSize().width );
       this._background.setScaleY( ( rows * this._tileSize  + this._wallThickness ) / this._background.getContentSize().height );

       //the containers are offset by half a tile so that their content is centered within each tile
       var containerPos = cc.p( -boundsSize.width / 2 + this._tileSize / 2 + this._wallThickness / 2, -boundsSize.height / 2 + this._tileSize / 2 + this._wallThickness / 2 );
       this._roomContainer.setPosition( containerPos );
       this._inhabitantContainer.setPosition( containerPos );
       
       //empty the room container. entities, however, are not emptied
       this._roomContainer.removeAllChildren();

       // reset wall sides
       this._wallSidesRoot.setPosition( cc.POINT_ZERO );
       this._wallSidesRoot.setRotation( 0 );
       this._wallSidesContainer.removeAllChildren();
       this._wallSidesContainer.setRotation( 0 );

       // Place compass at top of maze
       this._compass.setPosition( 0, boundsSize.height * 0.5 - this._wallThickness / 2 );
       this._compassPointer.setPosition( 0, boundsSize.height * 0.5 - this._wallThickness / 2 );
       this._compassIndicator.setPosition( 0, this._compass.getPosition().y );

       var pointer = this._compassPointer;
       var compassLabelPos = cc.p( pointer.getScaleX() * pointer.getContentSize().width * 0.7,
           boundsSize.height * 0.5 + pointer.getScaleY() * pointer.getContentSize().height * 0.55 );
       this._compassLabel.setPosition( compassLabelPos );

       // update the maze inhabitants' scales to fit in newly resized maze
       for( var i = 0; i < this._inhabitantContainer.getChildren().length; i++ )
       {
           var inhabitant = this._inhabitantContainer.getChildren()[ i ];
           this._rescaleForMaze( inhabitant );
       }
   },
   
   /**
    *   Tell the MazeView to create a RoomView and add it to itself.
    *   It will automatically adjust the wall directions based upon
    *   the position of the room in the array. This is because we don't
    *   want to generate more sprites than we need.
    *
    *   @param {number} x The X position to generate the RoomView at.
    *   @param {number} y The Y position to generate the RoomView at.
    *   @param {Array} wallDirections An array of directions where the Room has walls.
    */
   createRoomView: function( x, y, wallDirections )
   {
       // Create room views with wall information
       var roomView = new RoomView( this._tileSize, this._wallThickness, wallDirections, cc.p( x, y ), this._roomViews.getLength( 0 ), this._roomViews.getLength( 1 ), this._settings.wallColor );
       roomView.setPosition( this.getPositionFromGridPosition( cc.p( x, y ) ) );
       this._roomViews.set( roomView, x, y );
       this._roomContainer.addChild( roomView );
   },

    /**
     *   Loads sprites for each of the side walls, which provide a feeling of perspective.
     */
    createWallSides:function()
    {
        var x = 0;
        var y = 0;
        var w = 0;
        for( x = 0; x < this._roomViews.getWidth(); x++ )
        {
            for( y = 0; y < this._roomViews.getHeight(); y++ )
            {
                var roomView = this._roomViews.get( x, y );
                if( roomView )
                {
                    // have each room create sides to its walls
                    var wallSides = roomView.createWallSides( this._settings.mazeSidesColor );
                    if( wallSides )
                    {
                        for ( w = 0; w < wallSides.length; w++ )
                        {
                            // add each wall side to the container, which we can then manipulate when the maze rotates
                            var wallSide = wallSides[w];

                            var wallSidePos = this._wallSidesRoot.convertToNodeSpace( roomView.convertToWorldSpace( wallSide.getPosition() ) );

                            this._wallSidesContainer.addChild( wallSide );
                            wallSide.setPosition( wallSidePos );
                        }
                    }
                }
            }
        }

        this.rotate( 0, true );
    },

    /**
     * Rotates maze to new rotation in degrees.
     * @param {number} degrees New maze rotation.
     * @param {boolean} instant Whether or not to snap maze rotation.
     */
    rotate:function( degrees, instant, tutorialMode, finishedCallback, finishedTarget )
    {
        if ( instant )
        {
            // instantly rotate maze to specified angle
            this.setRotation( degrees );
            // counter-rotate wall sides to maintain perspective effect
            this._wallSidesRoot.setRotation( -degrees );
            this._wallSidesContainer.setRotation( degrees );
            
            if ( finishedCallback )
            {
                finishedCallback.call( finishedTarget );
            }
        }
        else
        {
            // smoothly rotate maze to specified angle
            this.runAction( this._createRotationAction( degrees, tutorialMode, finishedCallback, finishedTarget ) );
            // counter-rotate wall sides to maintain perspective effect
            this._wallSidesRoot.runAction( this._createRotationAction( -degrees, tutorialMode ) );
            this._wallSidesContainer.runAction( this._createRotationAction( degrees, tutorialMode ) );
        }

        var i = 0;
        for( i = 0; i < this._inhabitantContainer.getChildrenCount(); i++ )
        {
            var node = this._inhabitantContainer.getChildren()[i];
            if( !node.skipMazeRotation )
            {
                if( instant )
                {
                    // instantly counter-rotate entities to opposite angle
                    node.setRotation( -degrees );
                }
                else
                {
                    // smoothly counter-rotate entities to opposite angle
                    node.runAction( this._createRotationAction( -degrees, tutorialMode ) );
                }
            }
        }
    },

    /**
     * Enables the revealing of footsteps.
     *
     * @param enable Whether to enable footsteps.
     */
    enableFootsteps:function( enable )
    {
        this._footstepsEnabled = enable;
    },

    /**
     * Creates an action that can be used to smoothly rotate maze and its objects.
     * @param {number} degrees New rotation.
     * @returns {Action} Returns new rotation action.
     * @private
     */
    _createRotationAction:function( degrees, tutorialMode, finishedCallback, finishedTarget )
    {
        var angleDiff = Math.abs( degrees - this.getRotation() );
        var shortestDistance = ( angleDiff + 180 ) % 360 - 180;

        var rotateTime = tutorialMode ? this._settings.rotationAnimationTime90Tutorial : this._settings.rotationAnimationTime90;
        var rotateRate = this._settings.rotationAnimationRate90;
        if( Math.abs( shortestDistance >= 180 ) )
        {
            rotateTime = tutorialMode ? this._settings.rotationAnimationTime180Tutorial : this._settings.rotationAnimationTime180;
            rotateRate = this._settings.rotationAnimationRate180;
        }

        var rotAction = cc.EaseInOut.create( cc.RotateTo.create( rotateTime, degrees ), rotateRate );
        if ( finishedCallback )
        {
            return cc.Sequence.create( rotAction, cc.CallFunc.create( finishedCallback, finishedTarget ) );
        }
        return rotAction;
    },

    showCompass:function( show )
    {
        this._compass.setVisible( show );
        this._compassPointer.setVisible( show );
        this._compassIndicator.setVisible( show );
    },

    indicateCompass:function( indicate )
    {
        this._compassIndicator.setVisible( indicate );
    },

    /**
     * Reveals footsteps as penguin moves from one grid position to another.
     *
     * @param fromGridPos   Starting grid position.
     * @param toGridPos     Destination grid position.
     * @param time          Time it takes to arrive.
     */
    revealFootsteps:function( fromGridPos, toGridPos, time )
    {
        // delay between each footstep reveal
        var footstepDelay1 = time * 0.25;
        var footstepDelay2 = time * 0.42;
        var footstepDelay3 = time * 0.25;

        var from = cc.p( Math.max( 0, Math.round( fromGridPos.x ) ), Math.max( 0, Math.round( fromGridPos.y ) ) );
        var to = cc.p( Math.max( 0, Math.round( toGridPos.x ) ), Math.max( 0, Math.round( toGridPos.y ) ));
        var delta = cc.pSub( to, from );

        // figure out direction the penguin is heading, to reveal footsteps in the correct order
        var headingDir = null;
        var epsilon = 0.001;
        if( delta.x > epsilon )
        {
            headingDir = MazeDirections.Right;
        }
        else if( delta.x < -epsilon )
        {
            headingDir = MazeDirections.Left;
        }
        else if( delta.y > epsilon )
        {
            headingDir = MazeDirections.Up;
        }
        else if( delta.y < -epsilon )
        {
            headingDir = MazeDirections.Down;
        }

        // create sequence to reveal each footstep sprite over time
        var footstepsSequence = cc.Sequence.create(
            cc.CallFunc.create( this._revealFootstep.bind( this, from, null ), this ),
            cc.DelayTime.create( footstepDelay1 ),
            cc.CallFunc.create( this._revealFootstep.bind( this, from, headingDir ), this ),
            cc.DelayTime.create( footstepDelay2 ),
            cc.CallFunc.create( this._revealFootstep.bind( this, to, MazeUtility.getOppositeDirection( headingDir ) ), this ),
            cc.DelayTime.create( footstepDelay3 ),
            cc.CallFunc.create( this._revealFootstep.bind( this, to, null ), this )
        );

        this.runAction( footstepsSequence );
    },

    /**
     * Reveals footstep at a given grid position.
     *
     * @param gridPos   Maze grid position where the footstep should be revealed.
     * @param positionDirection     The footstep's offset direction from the center of the grid space. Use null for the center footstep.
     * @private
     */
    _revealFootstep:function( gridPos, positionDirection )
    {
        // find the room view that contains the footstep, and reveal it there
        if( this._footstepsEnabled )
        {
            if( this._roomViews )
            {
                var roomView = this._roomViews.get( gridPos.x, gridPos.y );
                if ( roomView )
                {
                    roomView.revealFootstep( positionDirection );
                }
            }
        }
    },

   /**
    *   Use to find the real (local) position of a given grid position.
    *   A grid position is considered a row / column pair in the maze.
    *
    *   @param {cc.Point} gridPosition  Grid position in question.
    *   @param {cc.Node} node   Optional child node of maze that can specify additional positioning behavior.
    *   @returns {cc.Point} The real position in pixels.
    */
   getPositionFromGridPosition:function( gridPosition, node )
   {
       var position = cc.p( this._margin.width + gridPosition.x * this._tileSize, this._margin.height + gridPosition.y * this._tileSize );
       if( node )
       {
            if ( node.accountForWalls )
            {
                position = cc.pSub( position, cc.p( this._wallThickness * 0.5, this._wallThickness * 0.5 ) );
            }
            if ( node.mazeOffset )
            {
                position = cc.pAdd( position, cc.pMult( node.mazeOffset, this._tileSize ) );
            }
       }
       return position;
   },
   
   /**
    *   Use to find the grid position of a given real (local) position.
    *   A grid position is considered a row / column pair in the maze.
    *
    *   @param {cc.Point} position Position in real space.
    *   @param {cc.Node} node   Optional child node of maze that can specify additional positioning behavior.
    *   @returns {cc.Point} The grid position, can be partial non-integer values.
    */
   getGridPositionFromPosition:function( position, node )
   {
       var adjustedPosition = cc.p( position.x, position.y );
       if( node )
       {
            if( node.accountForWalls )
            {
               adjustedPosition = cc.pAdd( adjustedPosition, cc.p( this._wallThickness * 0.5, this._wallThickness * 0.5 ) );
            }
            if ( node.mazeOffset )
            {
                adjustedPosition = cc.pSub( adjustedPosition, cc.pMult( node.mazeOffset, this._tileSize ) );
            }
       }
       return cc.p( ( ( adjustedPosition.x - this._margin.width ) / this._tileSize ), ( ( adjustedPosition.y - this._margin.height ) / this._tileSize ) );
   },
   
   /**
    *   Adds an inhabitant to the maze. It will be rendered within the inhabitant container.
    *   Also, we add closures for it to automatically find and set its grid position.
    *   These functions are getGridPosition(), setGridPosition(), and getPositionFromGridPosition().
    *   If the functions already exist, then they will not be added.
    *
    *   @param {cc.Node} node Any node that we want to add to this view.
    *   @param {cc.Point} gridPosition Grid position where we should add the node.
    */
    addInhabitant:function( node, gridPosition )
    {
        // add setGridPosition() function to node
        if ( !node.setGridPosition )
        {
           node.setGridPosition = function( mazeView, gridPos )
           {
               this.setPosition( mazeView.getPositionFromGridPosition( gridPos, node ) );
           }.bind( node, this );
        }

        // add getGridPosition() function to node
        if ( !node.getGridPosition )
        {
           node.getGridPosition = function( mazeView )
           {
               return mazeView.getGridPositionFromPosition( this.getPosition(), node );
           }.bind( node, this );
        }

        // add getPositionFromGridPosition() function to node
        if ( !node.getPositionFromGridPosition )
        {
           node.getPositionFromGridPosition = function( mazeView, position )
           {
               return mazeView.getPositionFromGridPosition( position, node );
           }.bind( node, this );
        }

        // set the node's starting grid position and add it to maze
        node.setGridPosition( gridPosition );
                              
        // move node to inhabitant container
        if ( node.getParent() != this._inhabitantContainer )
        {
           node.removeFromParent( false );
           this._inhabitantContainer.addChild( node );
        }

        this._rescaleForMaze( node );
   },

    _rescaleForMaze:function( node )
    {
        // apply additional scaling
        if( node )
        {
            var rescalingSize = 1;
            if ( node.getVisibleSize )
            {
                rescalingSize = node.getVisibleSize();
            }
            else
            {
                rescalingSize = node.getContentSize();
            }
            var rescalingScale = Math.min( this._tileSize / rescalingSize.width, this._tileSize / rescalingSize.height );
            var mazeScale = node.mazeScale || 1;

            var newScale = rescalingScale * mazeScale;
            node.setScale( newScale );
        }
    },


   /**
    *   Debug function, draws a bunch of red tiles in each RoomView for an A* solution.
    *
    *   @name Debug#__displaySolution
    *   @param {Array} solution An array of {x,y} coordinates that shows the solution.
    */
   __displaySolution:function( solution )
   {
       while ( solution.length > 0 )
       {
           var coordinate = solution.pop();
           var roomView = this._roomViews.get( coordinate.x, coordinate.y );
           if ( roomView )
           {
               roomView.__createFloorSprite();
           }
       }
   }
});