var cc = cc;
var require = require;

require( "src/Game/Maze/MazeConstants.js" );
var MazeDirections = MazeDirections;
var MazeConstants = MazeConstants;
var MazeUtility = MazeUtility;

var RoomView = cc.Node.extend(
{
    /** @lends RoomView.prototype */
    
    /** The size of the walls of this room along the thin dimension. */
    _thinSize:10,
    
    /** An array of sprites for the wall directions. Note that any direction can be undefined if it has no wall there. */
    _wallSprites:[],

    /** Debug sprite used to visualize room boundaries. */
    __floorSprite:null,
    
    /**
     *  A RoomView is the visible representation of a Room object. It does
     *  not retain a reference to the Room object, however. Instead, you
     *  create it telling it what to show, and it does so.
     *  
     *  @constructs
     *  @param {number} tileSize The size in pixels that this Room takes up.
     *  @param {Array} wallDirections An array of MazeDirections where this RoomView has walls.
     */
    ctor:function( tileSize, wallThickness, wallDirections, gridPos, columns, rows, wallColor )
    {
        this._super();
        this.init();

        this._wallSprites = new Array( MazeDirections.Count );

        this._thinSize = wallThickness;
        this._tileSize = tileSize;

        this._footstepSprites = new Array( MazeDirections.Count );

        var directionIndex;
        for ( directionIndex = 0; directionIndex < wallDirections.length; directionIndex++ )
        {
            // create a wall for each wall direction in array
            var direction = wallDirections[ directionIndex ];

            if ( ( gridPos.x < columns - 1 ) && direction === MazeDirections.Right )
            {
                //all rooms except for the right-most should not draw the right wall
                continue;
            }
            else if ( ( gridPos.y < rows - 1 ) && direction === MazeDirections.Up )
            {
                //all rooms except for the bottom-most should not draw the bottom wall
                continue;
            }

            this._createWallSprite( direction, gridPos, rows, columns, wallColor );
        }

        // create center footstep and one footstep in each direction
        this._centerFootstepSprite = this._createFootstepSprite();
        var footstepDir = 0;
        for ( footstepDir = 0; footstepDir < MazeDirections.Count; footstepDir++ )
        {
            if( !wallDirections.contains( footstepDir ) )
            {
                this._footstepSprites[ footstepDir ] = this._createFootstepSprite( footstepDir );
            }
        }
    },

    /**
     * Shows an individual footstep sprite.
     *
     * @param positionDirection MazeDirection offset relative to center of grid space. Use null for center footstep.
     */
    revealFootstep:function( positionDirection )
    {
        var footstepSprite = this._centerFootstepSprite;
        if( positionDirection !== null )
        {
            footstepSprite = this._footstepSprites[ positionDirection ];
        }
        if ( footstepSprite )
        {
            footstepSprite.setVisible( true );
        }
    },

    /**
     * Creates the individual grey wall side sprites for each wall sprite.
     *
     * @returns {Array} Returns array of wall side sprites.
     */
    createWallSides:function( wallSidesColor )
    {
        var wallSides = [];
        var i = 0;
        for ( i = 0; i < this._wallSprites.length; i++ )
        {
            // create a wall side sprite for each existing wall sprite
            var wallSprite = this._wallSprites[ i ];
            if( wallSprite )
            {
                var wallSide = cc.Sprite.create( "wall.png" );
                wallSide.setScaleX( wallSprite.getScaleX() );
                wallSide.setScaleY( wallSprite.getScaleY() );
                wallSide.setPosition( wallSprite.getPosition() );
                wallSide.setColor( wallSidesColor );
                wallSides.push( wallSide );
            }
        }
        return wallSides;
    },
    
    /**
     *  Creates an individual wall sprite.
     *
     *  @param {WallDirections} direction The direction to create the wall in, relative to the center of the room.
     */
    _createWallSprite: function( direction, gridPos, columns, rows, wallColor )
    {
        var delta = MazeConstants.positionDeltas[ direction ];
        var longSize = this._tileSize;
        var pos = cc.p( delta.x * ( this._tileSize + this._thinSize ) / 2, delta.y * ( this._tileSize + this._thinSize ) / 2 );

        if ( direction === MazeDirections.Up || direction === MazeDirections.Down )
        {
            // top and bottom walls are short and wide
            longSize += this._thinSize;
            pos.x -= this._thinSize / 2;
        }

        if ( direction === MazeDirections.Right )
        {
            // pull the rightmost wall inward, since it's an 'edge case'!
            pos.x -= this._thinSize;
        }

        if ( direction === MazeDirections.Up )
        {
            // pull the topmost wall inward, since it's an 'edge case'!
            pos.y -= this._thinSize;
        }
        
        var spriteSize = cc.p( delta.x === 0 ? longSize : this._thinSize, delta.y === 0 ? longSize : this._thinSize );

        // make outside walls twice as thick
        var extraThickness = this._thinSize * 0.5;
        var extremeX = ( direction === MazeDirections.Right && gridPos.x === columns - 1 ) ||
            ( direction === MazeDirections.Left && gridPos.x === 0 );
        var extremeY = ( direction === MazeDirections.Up && gridPos.y === rows - 1 ) ||
            ( direction === MazeDirections.Down && gridPos.y === 0 );
        if( extremeX )
        {
            // thicken up right and left edges
            spriteSize.x += extraThickness;
            pos.x += ( gridPos.x === 0 ? extraThickness / 2 : -extraThickness / 2 );
        }
        if( extremeY )
        {
            // thicken up top and bottom edges
            spriteSize.y += extraThickness;
            pos.y += ( gridPos.y === 0 ? extraThickness / 2 : -extraThickness / 2 );
        }

        // create and add sprites for each wall
        var wallSprite = cc.Sprite.create( "wall.png" );
        this._wallSprites[ direction ] = wallSprite;
        wallSprite.setScaleX( spriteSize.x / wallSprite.getContentSize().width );
        wallSprite.setScaleY( spriteSize.y / wallSprite.getContentSize().height );
        wallSprite.setPosition( pos );
        wallSprite.setColor( wallColor );
        this.addChild( wallSprite );
    },

    /**
     * Creates an indivdual footstep sprite in the room.
     *
     * @param direction MazeDirection relative to room's center.
     * @returns {cc.Sprite} Returns sprite for footstep.
     * @private
     */
    _createFootstepSprite:function( direction )
    {
        var footstepSprite = cc.Sprite.create( "penguin/penguin_footprintDot.png" );

        // create center or offset-to-the-side footstep, depending on the direction
        var position = cc.p( -this._thinSize / 2, -this._thinSize / 2 );
        if( direction !== undefined )
        {
            var delta = MazeConstants.positionDeltas[ direction ];
            position = cc.pAdd( position, cc.pMult( delta, this._tileSize * 0.33 ) );
        }
        footstepSprite.setPosition( position );

        // scale sprite to fit nicely in room
        footstepSprite.setScale( this._tileSize / footstepSprite.getContentSize().width * 0.09 );
        footstepSprite.setAnchorPoint( 0.5, 0.5 );
        footstepSprite.setVisible( false );

        this.addChild( footstepSprite );
        return footstepSprite;
    },
    
    /**
     *  Creates a floor sprite for debugging purposes. Draws this at the room's full size underneath the walls.
     *
     *  @name Debug#__createFloorSprite
     */
    __createFloorSprite:function()
    {
        this.__floorSprite = cc.Sprite.create( "wall.png" );
        this.__floorSprite.setColor( cc.c3b( 255, 0, 0 ) );
        this.__floorSprite.setScaleX( this._tileSize / this.__floorSprite.getContentSize().width );
        this.__floorSprite.setScaleY( this._tileSize / this.__floorSprite.getContentSize().height );
        if ( this.__floorSprite )
        {
            this.__floorSprite.setPosition( 0, 0 );
            this.addChild( this.__floorSprite, 0 );
        }
    }
});