var cc = cc;
var require = require;

require( "src/Entity/AnimatedEntity.js" );
var AnimatedEntity = AnimatedEntity;

require( "src/Game/Maze/MazeConstants.js" );
var MazeUtility = MazeUtility;

require( "src/Game/PenguinConstants.js" );
var PenguinConstants = PenguinConstants;

var PenguinCharacter = AnimatedEntity.extend(
{
    /** @lends PenguinCharacter.prototype */

    /** The game model */
    _model:null,

    /** The penguin's current state */
    _state:null,

    /** Whether the penguin is player- or computer-controlled */
    _isPlayer:false,

    /** Delegate to receive movement events */
    _movementResponder:null,

    /** Whether the penguin can move **/
    _movementEnabled:true,

    /** Whether the penguin is currently moving */
    _moving:false,

    /** The penguin's current maze-relative move direction */
    _moveDirection:null,

    _timeOfLastMove:0,

    /** The penguin's current grid postion in the maze */
    _gridPosition:cc.POINT_ZERO,

    /** The penguin's list of visited grid spaces */
    _traveledPath:[],
    
    /**
     *  A PenguinCharacter is the logic behind a penguin character moving.
     *  Can take a PenguinCharacterView that it knows how to move around when told.
     *  Also can be subclassed for player or AI-controlled penguin behavior.
     *  Subclass of AnimatedEntity.
     *
     *  @constructs
     *  @param {PenguinCharacterView} view The character view that we'll move around.
     *  @param {boolean} isPlayer   Whether the penguin is the player or computer.
     *  @param {GameModel} model  The game model.
     */
    ctor:function( view, isPlayer, model, movementSounds )
    {
        this._super( view )

        this._isPlayer = isPlayer;
        this._model = model;
        this._movementSounds = movementSounds;

        this.setState( PenguinStates.IdleDown );
    },

    /**
     *
     * @param gridPosition  Destination grid position.
     * @param finishedFunction  Callback to invoke when movement is finished.
     * @param finishedTarget    Object of finished callback.
     * @returns {boolean}   Whether the penguin successfully moved.
     */
    moveToGridPosition:function( gridPosition, forceMove, overrideTime, direction )
    {
        //get the direction that we're moving in in order to draw the correct sprite
        var curGridPos = cc.p( this._gridPosition.x, this._gridPosition.y );
        curGridPos = cc.p( Math.floor( curGridPos.x ), Math.floor( curGridPos.y ) );
        var difference = cc.p( gridPosition.x - curGridPos.x, gridPosition.y - curGridPos.y );
        var magnitude = Math.sqrt( difference.x * difference.x + difference.y * difference.y );
        var normal = cc.p( difference.x / magnitude, difference.y / magnitude );

        var direction = MazeUtility.getDirection( normal );
        this._moveDirection = direction;
        this.updateWalkDirection();

        if( !this._validMove( curGridPos, gridPosition ) && !forceMove )
        {
            // nope, invalid move!
            return false;
        }

        // start moving penguin to destination
        this._goalGridPosition = cc.p( gridPosition.x, gridPosition.y );
        var time = this._moveTime * magnitude;
        if( overrideTime !== undefined )
        {
            time = overrideTime;
        }
        this.view.moveToGridPosition( gridPosition, time, this.movementFinished.bind( this, gridPosition, direction ), this, this._isPlayer );
        this._moving = true;
        this._timeOfLastMove = lumosity.game.frametime;
        this._curMoveRatio = 0;

        return true;
    },

    /**
     * Stops the penguin's current movement.
     */
    stopMoving:function()
    {
        this._moving = false;

        if( this._timeOfLastMove > 0 )
        {
            // after penguin stops, update the ratio of movement he accomplished during move (0 to 1)
            //  this is used to include partial distances in scoring calculation
            this._curMoveRatio = ( lumosity.game.frametime - this._timeOfLastMove ) / this._moveTime;
            this._curMoveRatio = LumosityUtils.clamp( this._curMoveRatio, 0, 1 );
        }
        else
        {
            this._curMoveRatio = 0;
        }
        this._timeOfLastMove = 0;

        this.view.stopMoving();

        this.updateIdleDirection();

        this.updateMovementSounds();
    },

    /**
     * Overridable function to validate a potential move.
     *
     * @param fromGridPos   Starting grid position.
     * @param toGridPos     Destination grid position.
     * @returns {boolean} Whether the move is valid.
     * @private
     */
    _validMove:function( fromGridPos, toGridPos )
    {
        return true;
    },

    /**
     * Sets the movement responder; used for movement feedback.
     *
     * @param movementResponder     Delegate object that implements playerMoved().
     */
    setMovementResponder:function( movementResponder )
    {
        this._movementResponder = movementResponder;
    },

    reportStartedMove:function( toGridPos, direction )
    {
        if( this._movementResponder && this._movementResponder.penguinStartedMove )
        {
            this._movementResponder.penguinStartedMove( this._isPlayer, toGridPos, direction );
        }
    },

    reportFinishedMove:function( towardGoal, toGridPos, direction )
    {
        if( this._movementResponder && this._movementResponder.penguinFinishedMove )
        {
            this._movementResponder.penguinFinishedMove( this._isPlayer, towardGoal, toGridPos, direction );
        }
    },

    reportAttemptedInvalidMove:function()
    {
        if( this._movementResponder && this._movementResponder.penguinAttemptedInvalidMove )
        {
            this._movementResponder.penguinAttemptedInvalidMove( this._isPlayer );
        }
    },

    /**
     * Called on penguin at the start of pre-trial phase.
     */
    startPreTrial:function()
    {
        this._movementEnabled = false;

        this.stopMoving();

        this.setState( PenguinStates.IdleDown );
    },

    /**
     * Called on penguin at the start of trial phase.
     *
     * @param trial     Current trial data.
     */
    startTrial:function( trial )
    {
        this._movementEnabled = true;
        this._moving = false;

        // reset traveled path to only include starting position
        var startingGridPos = cc.p( this._gridPosition.x, this._gridPosition.y );
        this._traveledPath = [ startingGridPos ];
    },

    /**
     * Called on penguin at the end of the trial phase.
     */
    endTrial:function()
    {
        this._movementEnabled = false;

        this.stopMoving();

        this.setState( PenguinStates.IdleDown );
    },

    /**
     * Starts the penguin's win or lose behavior.
     *
     * @param won   Whether the penguin won or lost.
     */
    setWon:function( won )
    {
        if( won )
        {
            this.setState( PenguinStates.Won );
        }
        else
        {
            this.setState( PenguinStates.Lost );
        }
    },

    /**
     * Pauses penguin's behavior.
     */
    pause:function()
    {
        this.view.pause();
    },

    /**
     * Resumes penguin's paused behavior.
     */
    resume:function()
    {
        this.view.resume();
    },

    /**
     * Sets the penguin's current PenguinState, which is reflected in the view.
     * @param state     The penguin's new PenguinState.
     */
    setState:function( state )
    {
        this._state = state;
        this.view.setState( state );
    },

    /**
     * Sets the corresponding idle state, given a direction.
     *
     * @param direction     The direction of the idle state.
     */
    setIdleState:function( direction )
    {
        switch( direction )
        {
            case MazeDirections.Up:
                this.setState( PenguinStates.IdleUp );
                break;
            case MazeDirections.Down:
                this.setState( PenguinStates.IdleDown );
                break;
            case MazeDirections.Left:
                this.setState( PenguinStates.IdleLeft );
                break;
            case MazeDirections.Right:
                this.setState( PenguinStates.IdleRight );
                break;
        }
    },

    /**
     * Sets the corresponding walk state, given a direction.
     *
     * @param direction     The direction of the walk state.
     */
    setWalkState:function( direction )
    {
        switch( direction )
        {
            case MazeDirections.Up:
                this.setState( PenguinStates.WalkingUp );
                break;
            case MazeDirections.Down:
                this.setState( PenguinStates.WalkingDown );
                break;
            case MazeDirections.Left:
                this.setState( PenguinStates.WalkingLeft );
                break;
            case MazeDirections.Right:
                this.setState( PenguinStates.WalkingRight );
                break;
        }
    },

    /**
     * Adjusts the penguin's walk state after taking the maze's rotation into account.
     */
    updateWalkDirection:function()
    {
        if( this._moveDirection != null )
        {
            // figure out the absolute direction of the rotated maze, and use that to get penguin's walk direction.
            var rotatedDirection = MazeUtility.convertMazeDirectionToWorldDirection( this._moveDirection, this._model.mazeRotation );
            this.setWalkState( rotatedDirection );
        }
    },

    /**
     * Adjusts the penguin's idle state after taking the maze's rotation into account.
     */
    updateIdleDirection:function()
    {
        if( this._moveDirection != null )
        {
            // figure out the absolute direction of the rotated maze, and use that to get penguin's idle direction.
            var rotatedDirection = MazeUtility.convertMazeDirectionToWorldDirection( this._moveDirection, this._model.mazeRotation );
            this.setIdleState( rotatedDirection );
        }
    },

    /**
     * Makes penguin wave.
     */
    wave:function()
    {
        this.setState( PenguinStates.Waving );
    },

    /**
     * Sets the penguin's maze grid position.
     *
     * @param gridPos   New grid position for penguin.
     */
    setGridPosition:function( gridPos )
    {
        if( gridPos )
        {
            this._gridPosition = cc.p( gridPos.x, gridPos.y );
        }
        else
        {
            ll.verbose( "INVALID GRID POS" );
            this._gridPosition = cc.POINT_ZERO;
        }
    },

    /**
     * Gets the penguin's maze grid position.
     */
    getGridPosition:function()
    {
        return LumosityUtils.copyPoint( this._gridPosition );
    },

    /**
     * Overridable callback invoked after penguin has started moving to his destination.
     */
    movementStarted:function( toGridPos, direction )
    {
        this.updateMovementSounds();

        this.reportStartedMove( toGridPos, direction );
    },

    /**
     * Overridable callback invoked after penguin has completed moving to his destination.
     */
    movementFinished:function( toGridPos, direction )
    {
        this._moving = false;

        // update new position to match its goal
        this.setGridPosition( this._goalGridPosition );

        var movingTowardGoal = this.isMoveTowardGoal( this._goalGridPosition );

        this._updateTraveledPath();

        this.reportFinishedMove( movingTowardGoal, toGridPos, direction );
    },

    updateMovementSounds:function()
    {
        if( this._moving )
        {
            if( this._movementSounds && this._movementSounds.length )
            {
                var delayTime = this._moveTime / 2;
                delayTime = Math.min( delayTime, 0.3 );

                this.view.playMovementSounds( this._movementSounds, delayTime );
            }
        }
        else
        {
            this.view.stopMovementSounds();
        }
    },

    /**
     * Sets the penguin's move time per grid space.
     *
     * @param moveTime  Time it takes to move one grid space.
     */
    setMoveTime:function( moveTime )
    {
        this._moveTime = moveTime;
    },

    /**
     * Updates penguin's traveled path with current position.
     * @private
     */
    _updateTraveledPath:function()
    {
        var gridPos = cc.p( this._gridPosition.x, this._gridPosition.y );

        // check backward along path from current position
        var i = 0;
        for ( i = this._traveledPath.length - 1; i >= 0; i-- )
        {
            if ( this._traveledPath[i].x === gridPos.x && this._traveledPath[i].y === gridPos.y )
            {
                // penguin has been here before, so reduce path by lopping off everything since visiting this grid space
                this._traveledPath.splice( i + 1, this._traveledPath.length - i );
                return;
            }
        }
        // penguin has not yet been to current grid space, so add it to his traveled path
        this._traveledPath.push( gridPos );
    },

    /**
     * Gets the penguin's traveled path of in-order, distinct grid spaces.
     *
     * @returns {*}     List of grid spaces.
     */
    getTraveledPath:function()
    {
        return this._traveledPath;
    },

    isOnSolutionPath:function()
    {
        var solutionPath = this._model.trial.maze.getSolution();
        var traveledSolutionPath = MazeUtility.getCommonPath( this._traveledPath, solutionPath );
        return ( this._traveledPath.length === traveledSolutionPath.length );
    },

    /**
     * Returns the length of the solution path that the penguin has followed. Includes partial distances between grid spaces.
     *
     * @returns {number} Distance traveled along solution path.
     */
    getTraveledSolutionPathDistance:function()
    {
        var solutionPath = this._model.trial.maze.getSolution();
        var traveledSolutionPath = MazeUtility.getCommonPath( this._traveledPath, solutionPath );
        var traveledSolutionSpaces = traveledSolutionPath.length;

        // initial distance is the number of solution grid spaces traversed
        var dist = traveledSolutionSpaces - 1;

        // if penguin hasn't deviated from solution path, check for partial movement
        if( this._traveledPath.length === traveledSolutionPath.length )
        {
            // calculate current, forward, and backward grid spaces along solution path
            var fwdSolutionGridPos = solutionPath[traveledSolutionSpaces];
            var curSolutionGridPos = solutionPath[traveledSolutionSpaces - 1];
            var bwdSolutionGridPos = solutionPath[traveledSolutionSpaces - 2];

            // get directions from current grid space to spaces of forward and backward progress
            var fwdDir = MazeUtility.getDirectionBetween( curSolutionGridPos, fwdSolutionGridPos );
            var bwdDir = MazeUtility.getDirectionBetween( curSolutionGridPos, bwdSolutionGridPos );

            //ll.verbose("CURMOVERATIO: " + this._curMoveRatio);
            //ll.verbose("CURDIR: " + MazeConstants.directionNames[this._moveDirection] +
            //    ", FORWARDDIR: " + MazeConstants.directionNames[fwdDir] +
            //    ", BACKWARD: " + MazeConstants.directionNames[bwdDir] );

            if( fwdDir === this._moveDirection )
            {
                // penguin is currently making forward progress, so add this move's partial distance
                dist += this._curMoveRatio;
            }
            else if( bwdDir === this._moveDirection )
            {
                // penguin is currently making backward progress, so subtract this move's partial distance
                dist -= this._curMoveRatio;
            }
        }
        dist = Math.max( dist, 0 );
        return dist;
    },

    isMoveTowardGoal:function( newGridPos )
    {
        if( newGridPos && this._traveledPath && this._model.trial && this._model.trial.maze )
        {
            var traveledSolutionPath = MazeUtility.getCommonPath( this._traveledPath, this._model.trial.maze.getSolution() );

            var newPath = this._traveledPath.slice( 0 );
            newPath.push( newGridPos );
            var newSolutionPath = MazeUtility.getCommonPath( newPath, this._model.trial.maze.getSolution() );

            if( traveledSolutionPath && newSolutionPath && traveledSolutionPath.length < newSolutionPath.length )
            {
                return true;
            }
        }
        return false;
    },

    /**
     *  Plays the walk animation in a given direction. Plays it only once.
     *
     *  @param {MazeDirections} direction The direction to play the walk animation.
     *  @param {boolean} [force=false] If true, will force this animation to play.
     */
    playWalkAnimation:function( direction, force )
    {
        this._direction = direction;
        var played = this.playAnimation( this.view._getAnimationNameForDirection( direction ), force );
        if ( played )
        {
            //need to flip ourselves if we are going left
            this.view.setScaleX( Math.abs( this.view.getScaleX() ) * ( direction === MazeDirections.Left ? -1.0 : 1.0 ) );
        }
    },

    /**
     *  If you want to implement animation priority, like for example if animation A is playing
     *  then animation B should never play, then override this function. You can also use it
     *  to implement any game state checks you might want.
     *
     *  @override
     *  @param {string} animationName Check to see if this animation should be played.
     *  @returns {boolean} If true, then playAnimation() will be able to play this animation.
     */
    shouldPalyAnimation:function( animationName )
    {
        //we only care if we're currently playing an animation
        if ( this.view.isPlaying() )
        {
            var currentAnimName = this.view.getPlayingAnimationName();

            //if we're trying to play a walk animation, then we only want to interrupt different walk animations
            if ( this.view.isWalkAnimation( animationName ) )
            {
                if( this.view.isWalkAnimation( currentAnimName ) )
                {
                    if( animationName !== currentAnimName || this._direction !== this._lastDirection )
                    {
                        this._lastDirection = this._direction;
                        return true;
                    }
                }
                return false;
            }
            //otherwise, we only want to interrupt walk animations... probably
            else
            {
                return this.view.isWalkAnimation( currentAnimName );
            }
        }
        else
        {
            return this._super( animationName );
        }
    },

    /**
     *  Called by the AnimatedEntityView when an animation has been finished because we're
     *  a listener for our view.
     *
     *  @override
     *  @param {string} animationName The name of the animation that just finished.
     */
    animatedEntity_finishedAnimation:function( animationName )
    {
        if ( this.view.isWalkAnimation( animationName ) )
        {
            //automatically resume any walk animations
            this.view.play( animationName );
        }
        else if( this.view.getAnimNameState( animationName ) === PenguinStates.Waving )
        {
            // go to idle state when finished waving
            this.setState( PenguinStates.IdleDown );
        }
    }
});