var cc = cc;
var require = require;

require( "src/Game/PenguinCharacter.js" );
var PenguinCharacter = PenguinCharacter;

require( "src/Game/Maze/MazeConstants.js" );
var MazeDirections = MazeDirections;
var MazeConstants = MazeConstants;

require( "src/Game/PenguinCharacterView.js" );
var PenguinCharacterView = PenguinCharacterView;

var PenguinCharacterPlayer = PenguinCharacter.extend(
{
    /** @lends PenguinCharacterPlayer.prototype */

    /** The current maze, used for navigation */
    _maze:null,

    _bump:null,

    /** Whether we're showing the invalid move feedback atm */
    _showingInvalidMove:false,

    /** Player's currently pressed input directions */
    _inputDirections:[ false, false, false, false ],

    /**
     *  A subclass of PenguinCharacter that handles player input.
     *
     *  @constructs
     *  @param model    The game model.
     *  @param settings The game settings.
     */
    ctor: function( model, settings )
    {
        var view = new PenguinCharacterView( settings.penguinTintPlayer );
        view.id = "playerPenguinView";
        this._super( view, true, model, settings.sounds.movementPlayer );

        this._bump = model.bump;
    },

    /**
     * Called on penguin at the start of trial phase.
     *
     * @param trial     Current trial data.
     */
    startTrial:function( trial )
    {
        this._super( trial );

        this._showingInvalidMove = false;

        this._updateMovement();
    },

    /**
     * Sets the current maze; used for navigation.
     *
     * @param maze
     */
    setMaze:function( maze )
    {
        this._maze = maze;
    },

    /**
     * Updates penguin movement in the current direction. Displays invalid move feedback
     * if he can't move there.
     *
     *  @returns {boolean}True if the player actually will be moving to a new grid space.
     */
    _updateMovement:function()
    {
        var direction = this._inputDirections.indexOf( true );
        if( direction < 0 || this._moving )
        {
            return false;
        }

        var currentGridPos = cc.p( this._gridPosition.x, this._gridPosition.y );
        var currentGridPosFloored = cc.p( Math.floor( currentGridPos.x ), Math.floor( currentGridPos.y ) );
        var delta = MazeConstants.positionDeltas[ direction ];
        var targetGridPos = cc.p( Math.floor( currentGridPosFloored.x + delta.x ), Math.floor( currentGridPosFloored.y + delta.y ) );

        var moved = this.moveToGridPosition( targetGridPos, undefined, undefined, direction );
        if( moved )
        {
            this.movementStarted( targetGridPos, direction );
        }

        if( !moved && !this._showingInvalidMove )
        {
            // can't go here, so show the invalid move feedback
            this._showingInvalidMove = true;
            this.view.showInvalidMove( direction, this.invalidMovementFinished, this );

            var worldDirection = MazeUtility.convertMazeDirectionToWorldDirection( direction, this._model.mazeRotation );

            var mazeDownDirection = MazeDirections.Down;

            if( Math.abs( LumosityUtils.smallestAngleBetween( this._model.mazeRotation, 90 ) ) <= 45 )
            {
                mazeDownDirection = MazeDirections.Right;
            }
            else if( Math.abs( LumosityUtils.smallestAngleBetween( this._model.mazeRotation, 180 ) ) <= 45 )
            {
                mazeDownDirection = MazeDirections.Up;
            }
            else if( Math.abs( LumosityUtils.smallestAngleBetween( this._model.mazeRotation, 270 ) ) <= 45 )
            {
                mazeDownDirection = MazeDirections.Left;
            }

            this._bump.play( currentGridPos, direction, worldDirection, mazeDownDirection, this.view.mazeOffset );

            this.reportAttemptedInvalidMove();
        }

        return true;
    },

    /**
     * Overridable function to validate a potential move.
     *
     * @param fromGridPos   Starting grid position.
     * @param toGridPos     Destination grid position.
     * @returns {boolean}   Whether the move is valid.
     * @private
     */
    _validMove:function( fromGridPos, toGridPos )
    {
        if( fromGridPos.x === toGridPos.x && fromGridPos.y === toGridPos.y )
        {
            // no change in position
            return false;
        }

        var direction = MazeUtility.getDirectionBetween( fromGridPos, toGridPos );
        var fromRoom = this._maze.getRoom( fromGridPos.x, fromGridPos.y );
        var toRoom = this._maze.getRoom( toGridPos.x, toGridPos.y );
        if ( !toRoom || ( fromRoom && !fromRoom.isOpen( direction ) ) )
        {
            // destination room doesn't exist or rooms aren't connected
            return false;
        }

        return true;
    },

    /**
     * Update the penguin's current movment directions with player's current input.
     *
     * @param direction     Input maze direciton.
     * @param active        Whether direction input is active.
     */
    setInputDirection:function( direction, active )
    {
        var inputChanged = this._inputDirections[ direction ] !== active;
        this._inputDirections[ direction ] = active;
        if( this._movementEnabled && inputChanged )
        {
            // set new status for specified input direction
            if( !this._updateMovement() && !this._moving )
            {
                // no valid movement input, so stop
                this.stopMoving();
            }
        }
    },

    /**
     * Overridable callback invoked after penguin has completed moving to his destination.
     */
    movementFinished:function( actionIndex, toGridPos, direction )
    {
        this._super( actionIndex, toGridPos, direction );

        if( this._movementEnabled && !this._updateMovement() )
        {
            this.stopMoving();
        }

        this.updateMovementSounds();
    },

    /**
     *  Overridable callback invoked when penguin's invalid movement feedback has finished.
     */
    invalidMovementFinished:function()
    {
        this._showingInvalidMove = false;
    }
});