var cc = cc;
var require = require;
var lumosity = lumosity;

require( "src/Entity/AnimatedEntityView.js" );
var AnimatedEntityView = AnimatedEntityView;

require( "src/Game/Maze/MazeConstants.js" );
var MazeDirections = MazeDirections;
var MazeUtility = MazeUtility;
var MazeConstants = MazeConstants;

require( "src/Game/PenguinCharacter.js" );
var PenguinCharacter = PenguinCharacter;
var PenguinStates = PenguinStates;

var PenguinCharacterView = AnimatedEntityView.extend(
{
    /** @lends PenguinCharacterView.prototype */

    /** The penguin's current PenguinState */
    _state:null,

    /** The action being used to move us */
    _movementAction:null,

    /** The action being used to repeatedly play the movement sound. */
    _movementSoundAction:null,

    /** The current MazeView */
    _mazeView:null,

    /**
     *  A PenguinCharacterView shows an individual penguin. Controls walking
     *  animations and orientation and the like. Subclass of AnimatedEntityView.
     *
     *  @constructs
     *  @param {object} Object containing color values for penguin tinting.
     */
    ctor:function( tintValues )
    {
        this._super( "penguin/penguin_sprites", this._createAnimationNames(), null, true );

        if( tintValues )
        {
            this._sprite.setColor( cc.c3b( tintValues.r, tintValues.g, tintValues.b ) );
        }
    },

    /**
     * Sets the view's current PenguinState, which it uses to update its animation state.
     *
     * @param state     Penguin's current PenguinState.
     */
    setState:function( state )
    {
        if( this._state !== state )
        {
            var animName = this.getStateAnimName( state );
            if( animName )
            {
                this.play( animName );
            }

            var facingLeft = ( state === PenguinStates.IdleLeft || state === PenguinStates.WalkingLeft );
            var facingScale = Math.abs( this.getScaleX() ) * ( facingLeft ? -1.0 : 1.0 );
            this.setScaleX( facingScale );

            this._state = state;
        }
    },

    /**
     *  Move to the passed grid position over the given time. If a finishedFunction and
     *  target are provided, they will be called once moving is finished.
     *
     *  @param {cc.Point} gridPosition The grid position to move to.
     *  @param {number} time The time to take to move there.
     *  @param {function} [finishedFunction] The function to call when finished.
     *  @param {Object} [finishedTarget] The object to call the finishedFunction on.
     */
    moveToGridPosition:function( gridPosition, time, finishedFunction, finishedTarget, revealFootsteps )
    {
        this.moveToPosition( this.getPositionFromGridPosition( gridPosition ), time, finishedFunction, finishedTarget );

        if( revealFootsteps && this._mazeView && this._mazeView.revealFootsteps )
        {
            // reveal footsteps over time as penguin moves
            this._mazeView.revealFootsteps( this.getGridPosition(), gridPosition, time );
        }
    },
    
    /**
     *  Move to the passed position over the given time. If a finishedFunction and
     *  target are provided, they will be called once moving is finished.
     *
     *  @param {cc.Point} position The position to move to.
     *  @param {number} time The time to take to move there.
     *  @param {function} [finishedFunction] The function to call when finished.
     *  @param {Object} [finishedTarget] The object to call the finishedFunction on.
     */
    moveToPosition:function( position, time, finishedFunction, finishedTarget )
    {
        //if we already have a movement action, stop it
        this.stopMoving();

        // create view's movement sequence
        var moveAction = cc.MoveTo.create( time, position );
        var finishedAction = cc.CallFunc.create( this._movementFinished.bind( this, finishedFunction, finishedTarget ), this );
        var sequence = cc.Sequence.create( moveAction, finishedAction );
        
        this._movementAction = sequence;
        this.runAction( sequence );
    },

    stopActions:function()
    {
        this.stopAllActions();

        this._movementAction = null;
        this._movementSoundAction = null;
    },

    /**
     * Stops the penguin if he's currently moving.
     */
    stopMoving:function()
    {
        if ( this._movementAction )
        {
            this.stopAction( this._movementAction );
            this._movementAction = null;
        }
    },

    stopMovementSounds:function()
    {
        if( this._movementSoundAction )
        {
            this.stopAction( this._movementSoundAction );
            this._movementSoundAction = null;
        }
    },

    playMovementSounds:function( soundNames, delayTime )
    {
        if( !this._movementSoundAction )
        {
            var playFunc = function( soundNames )
            {
                var soundName = soundNames.getRandomElement();
                audio.playEffect( soundName );
            }.bind( undefined, soundNames );

            var playAction = cc.CallFunc.create( playFunc, this );
            var delayAction = cc.DelayTime.create( delayTime );
            var sequence = cc.Sequence.create( playAction, delayAction );
            var repeatAction = cc.RepeatForever.create( sequence );
            this.runAction( repeatAction );

            this._movementSoundAction = repeatAction;
        }
    },

    /**
     * Displays feedback for attempting to move in an invalid direction.
     * @param direction     Direction the penguin is trying to move.
     * @param finishedFunction  Callback invoked when feedback is finished showing.
     * @param finishedTarget    Object that finishedFunction callback is invoked with.
     */
    showInvalidMove:function( direction, finishedFunction, finishedTarget )
    {
        // create sequence to show the penguin move forward a little and bounce back
        var delta = MazeConstants.positionDeltas[direction];
        var time = 0.05;
        var position = this.getPositionFromGridPosition( this.getGridPosition() );
        var toPosition = cc.pAdd( position, cc.pMult( delta, 15 ) );
        var moveActionTo = cc.MoveTo.create( time, toPosition );
        var moveActionBack = cc.MoveTo.create( time, position );
        var finishedAction = cc.CallFunc.create( this._movementFinished.bind( this, finishedFunction, finishedTarget ), this );

        var sequence = cc.Sequence.create(
            moveActionTo,
            moveActionBack,
            finishedAction );
        this.runAction( sequence );
    },

    /**
     * Pauses view behavior.
     */
    pause:function()
    {
        this._super();
        this.pauseSchedulerAndActions();
    },

    /**
     * Resumes paused view behavior.
     */
    resume:function()
    {
        this._super();
        this.resumeSchedulerAndActions();
    },

    /**
     * Sets the current maze view.
     *
     * @param view  The current maze view.
     */
    setMazeView:function( view )
    {
        this._mazeView = view;
    },

    /**
     *  Check to see if the passed animation is a walk animation.
     *
     *  @returns {boolean}  Whether the animation is a walk animation.
     */
    isWalkAnimation:function( animationName )
    {
        return animationName.indexOf( "walkAnim" ) >= 0;
    },

    /**
     * Looks up a PenguinState for an animation name.
     * @param animName  Name of animation in question.
     * @returns {*}     The associated PenguinState.
     */
    getAnimNameState:function( animName )
    {
        if( animName )
        {
            var i = 0;
            for( i = 0; i < PenguinStates.Count; i++ )
            {
                if( this.getStateAnimName( i ) === animName )
                {
                    return i;
                }
            }
        }
        return null;
    },

    /**
     * Looks up an animation name for a PenguinState.
     * @param state     PenguinState in question.
     * @returns {*}     The associated animation name.
     */
    getStateAnimName:function( state )
    {
        switch( state )
        {
            case PenguinStates.IdleUp:
                return "redpenguin_backView_idleAnim";

            case PenguinStates.IdleDown:
                return "redpenguin_frontView_idleAnim";

            case PenguinStates.IdleLeft:
            case PenguinStates.IdleRight:
                return "redpenguin_sideView_idleAnim";

            case PenguinStates.WalkingUp:
                return "redpenguin_backView_walkAnim";

            case PenguinStates.WalkingDown:
                return "redpenguin_frontView_walkAnim";

            case PenguinStates.WalkingLeft:
            case PenguinStates.WalkingRight:
                return "redpenguin_sideView_walkAnim";

            case PenguinStates.Won:
                return "redpenguin_frontView_eatFishAnim";

            case PenguinStates.Lost:
                return "redpenguin_frontView_loseCryAnim";

            case PenguinStates.Waving:
                return "redpenguin_frontView_waveAnim";
        }
    },

    /**
     *  Creates a new array with all the animation names that we care about.
     *
     *  @returns {Array} The array of names. Pass these into the super constructor.
     */
    _createAnimationNames:function()
    {
        var animations = [];
        animations.push( this.getStateAnimName( PenguinStates.IdleUp ) );
        animations.push( this.getStateAnimName( PenguinStates.IdleDown ) );
        animations.push( this.getStateAnimName( PenguinStates.IdleLeft ) );
        animations.push( this.getStateAnimName( PenguinStates.IdleRight ) );
        animations.push( this.getStateAnimName( PenguinStates.WalkingUp ) );
        animations.push( this.getStateAnimName( PenguinStates.WalkingDown ) );
        animations.push( this.getStateAnimName( PenguinStates.WalkingLeft ) );
        animations.push( this.getStateAnimName( PenguinStates.WalkingRight ) );
        animations.push( this.getStateAnimName( PenguinStates.Won ) );
        animations.push( this.getStateAnimName( PenguinStates.Lost ) );
        animations.push( this.getStateAnimName( PenguinStates.Waving ) );
        return animations;
    },

    /**
     * Called when movement has finished, which forwards event to optional delegate.
     *
     * @param finishedCallback  Callback to invoke when movement is finished.
     * @param finishedTarget    Object to invoke finishedCallback with.
     * @private
     */
    _movementFinished:function( finishedCallback, finishedTarget )
    {
        this._movementAction = null;

        if ( finishedCallback )
        {
            if ( finishedTarget )
            {
                finishedCallback.call( finishedTarget, this );
            }
            else
            {
                finishedCallback( this );
            }
        }
    }
});