var cc = cc;
var require = require;
/*global Interaction*/

require( "src/Game/Base/Game.js" );
var Game = Game;

require( "src/Game/PenguinPursuitSettings.js" );
var PenguinPursuitSettings = PenguinPursuitSettings;

require( "src/Game/PenguinPursuitModel.js" );
var PenguinPursuitModel = PenguinPursuitModel;

require( "src/Game/PenguinPursuitView.js" );
var PenguinPursuitView = PenguinPursuitView;

require( "src/Game/Maze/MazeConstants.js" );
var MazeConstants = MazeConstants;
var MazeDirections = MazeDirections;

var PenguinPursuit = Game.extend(
{
    /** @lends PenguinPursuit.prototype */

    /** Current number of valid moves in a row. */
    _validMovesInARow:0,

    /** Current number of invalid moves in a row. */
    _invalidMovesInARow:0,

    /** Whether the buttons are being highlighted to get the player's attention. */
    _highlightingButtons:false,

    /** Current number of button presses since last button highlight (used to decide whether to highlight). */
    _buttonsPressedSinceLastButtonHighlight:0,

    /** Keys for game-specific metadata. */
    _metadataKeys:{},

    /** Keys for action metadata. */
    _metadataActionKeys:{},

    /**
     *  PenguinPursuit is the main class for the game, Penguin Pursuit!
     *
     *  @constructs
     */
    ctor:function()
    {
        this._super();

        this.view.addDebugMenuButton( { labelText: "Win", callback: this.endTrialButtonPressed.bind( this, true ), target: this, closeMenuOnCallback: true } );
        this.view.addDebugMenuButton( { labelText: "Lose", callback: this.endTrialButtonPressed.bind( this, false ), target: this, closeMenuOnCallback: true } );
        this.view.addDebugMenuButton( { labelText: "Easy Mode", callback: this.setEasyMode.bind( this ), target: this, closeMenuOnCallback: true } );
        this.view.addDebugMenuButton( { labelText: "Reached Level 1", callback: this.setStartingLevel.bind( this, 1 ), target: this, closeMenuOnCallback: true } );
        this.view.addDebugMenuButton( { labelText: "Reached Level 10", callback: this.setStartingLevel.bind( this, 10 ), target: this, closeMenuOnCallback: true } );
        this.view.addDebugMenuButton( { labelText: "Reached Level 20", callback: this.setStartingLevel.bind( this, 20 ), target: this, closeMenuOnCallback: true } );
        this.view.addDebugMenuButton( { labelText: "Reached Level 30", callback: this.setStartingLevel.bind( this, 30 ), target: this, closeMenuOnCallback: true } );
        this.view.addDebugMenuButton( { labelText: "Reached Level 40", callback: this.setStartingLevel.bind( this, 40 ), target: this, closeMenuOnCallback: true } );

        // set responder for movement events
        this.model.playerPenguin.setMovementResponder( this );
        this.model.computerPenguin.setMovementResponder( this );
    },

    _createSettings:function()
    {
        return new PenguinPursuitSettings( "settings.json" );
    },

    _createModel:function()
    {
        if( this.model )
        {
            // don't recreate model if we already have one
            this.model.reset();
            return this.model;
        }
        else
        {
            return new PenguinPursuitModel( this.settings );
        }
    },

    _createView:function( screenSize )
    {
        return new PenguinPursuitView( screenSize, this.settings, this.model, this );
    },

    onBegin:function()
    {
        this._super();

        this.view.setScore( 0 );
        this.view.setLevel( this.model.levelIndex + 1 );
        this.view.setTrial( 1, this.settings.getGameLengthInTrials() );

        this._createNewTrial();

        this.startPreTrialPhase();
    },

    onStartGameplay:function()
    {
        this._super();
    },

    onEndGameplay:function()
    {
        this._super();
    },

    onFinish:function( wasQuit )
    {
        this._super( wasQuit );
    },

    onRestart:function()
    {
        this._super();
    },

    /**
     * Pauses game behavior.
     */
    onPause:function()
    {
        this._super();

        this.model.playerPenguin.pause();
        this.model.computerPenguin.pause();
    },

     /**
     * Resumes paused game behavior.
     */
    onResume:function()
    {
        this._super();

        this.model.playerPenguin.resume();
        this.model.computerPenguin.resume();
    },

    playStartingAnimation:function()
    {
        this.view.playAnimation( this.model.persistentData.isPracticeMode ? "startTutorial" : "startGame", this.onStartGameplay, this );
    },

    /**
     * Cue for when maze enters stage.
     */
    cue_mazeEnter:function()
    {
        this.view.mazeEnter();
    },

    /**
     * Cue for when we should go to the next trial.
     */
    cue_createNextTrial:function()
    {
        this.model.trialIndex++;
        this.view.setTrial( this.model.trialIndex + 1, this.settings.getGameLengthInTrials() );

        this._createNewTrial();
    },

    /**
     * Cue for when the player penguin should wave adorably at the player.
     */
    cue_penguinWave:function()
    {
        this.model.playerPenguin.wave();
    },

    /**
     * Cue for when the countdown should begin.
     */
    cue_countdownPlay:function()
    {
        this.view.playReadyGo();
    },

    cue_setGoLabel:function()
    {
        this.view.setGoLabel();
    },

    cue_tutorialPlay:function()
    {
        this.startTrialPhase();
    },

    cue_endNextTutorial:function()
    {
        this.startTrialPhase();
    },

    cue_setTutorialText:function()
    {
        if( this.model.persistentData.isPracticeMode )
        {
            var trialSettings = this.model.getCurTrialSettings();
            if( trialSettings )
            {
                this.view.showTutorialPrompt( localize( trialSettings.tutorialMessage ) );
            }
        }
    },

    cue_showEndTutorialPopup:function()
    {
       this.view.showEndTutorialPopup( this.model.tutorialSkipped );
    },

    /**                                                           `
     * Cue for when the countdown is over.
     */
    cue_endCountdown:function()
    {
        this.startTrialPhase();
    },

    /**
     * Cue for when the winning penguin should celebrate.
     */
    cue_playWinAnim:function()
    {
        var winningPenguin = this.model.trial.playerWon ? this.model.playerPenguin : this.model.computerPenguin;
        if( winningPenguin )
        {
            winningPenguin.setWon( true );
        }
    },

    /**
     * Cue for when the losing penguin should hate himself.
     */
    cue_playLoseAnim:function()
    {
        if( !this.model.persistentData.isPracticeMode )
        {
            var losingPenguin = !this.model.trial.playerWon ? this.model.playerPenguin : this.model.computerPenguin;
            if( losingPenguin )
            {
                losingPenguin.setWon( false );
            }
        }
    },

    /**
     * Cue for when the winning penguin should cheer.
     */
    cue_playWinSound:function()
    {
        audio.playEffect( this.settings.sounds.playerWinVocal );
    },

    /**
     * Cue for when the winning penguin should cheer.
     */
    cue_playLoseSound:function()
    {
        if( !this.model.persistentData.isPracticeMode )
        {
            audio.playEffect( this.settings.sounds.playerLoseVocal );
        }
    },

    endTrialButtonPressed:function( win )
    {
        if( win )
        {
            this.model.playerPenguin.moveToGridPosition( this.model.trial.maze.goal, true, 0.5 );
        }
        else
        {
            this.model.computerPenguin.moveToGridPosition( this.model.trial.maze.goal, true, 0.5 );
        }
    },

    setStartingLevel:function( level )
    {
        this.model.levelIndex = LumosityUtils.clamp( level - 1, 0, this.settings.levels.length - 1 );
        this.onEndGameplay();
    },

    setEasyMode:function()
    {
       this.model.easyMode = true;
       this._updateForEasyMode();
    },

    /**
     * Set whether the player is requesting a move in a given direction.
     * @param direction     Maze direction.
     * @param active        Whether the player wants to move in given direction.
     */
    setInputDirection:function( direction, active )
    {
        this.model.playerPenguin.setInputDirection( direction, active );

        this.updateMoveHint();

        if( active && this._highlightingButtons )
        {
            this._buttonsPressedSinceLastButtonHighlight++;
        }
    },

    /**
     * Called when a penguin starts a move.
     * Member of movement responder delegate.
     */
    penguinStartedMove:function( isPlayer, toGridPos, direction )
    {
        if( isPlayer )
        {
            // the player has moved, so update whether the maze should rotate
            this.model.trial.timesMovedSinceRotation++;
            this._tryMazeRotation();

            this.updateMoveHint();

            var timeOffset = lu.convertSecondsToMilliseconds( lumosity.game.frametime - this.model.trial.startTime );
            this.model.beginRecordingAction( "movement", direction, toGridPos, timeOffset, this.model.trialIndex );
        }
    },

    /**
     * Called when a penguin ends a move.
     * Member of movement responder delegate.
     */
    penguinFinishedMove:function( isPlayer, towardGoal, toGridPos, direction )
    {
        if( isPlayer )
        {
            var onPath = this.model.playerPenguin.isOnSolutionPath();
            var timeOffsetRelease = lu.convertSecondsToMilliseconds( lumosity.game.frametime - this.model.trial.startTime );
            var traveledDistance = this.model.playerPenguin.getTraveledPath().length;
            this.model.finishRecordingAction( this.model.getLastActionIndexOfType( "movement" ), onPath, timeOffsetRelease, traveledDistance );
            
            if( this.model.trial )
            {
                this.model.trial.traveledDistance++;
            }

            this._recordValidMove( towardGoal );
        }

        this._checkForWin();
    },

    /**
     * Called when a penguin attempts an inavlid move.
     * Member of movement responder delegate.
     */
    penguinAttemptedInvalidMove:function( isPlayer )
    {
        if( isPlayer )
        {
            audio.playEffect( this.settings.sounds.bump );

            this._recordValidMove( false );
        }
    },

    /**
     * Updates the tutorial feedback for the player's next direction.
     */
    updateMoveHint:function()
    {
        if( this.model.trial.showMoveHint )
        {
            if( this._showingMoveHint && this._validMovesInARow >= this.model.trial.hideHintMoves )
            {
                this._showingMoveHint = false;
            }
            else if( !this._showingMoveHint && this._invalidMovesInARow >= this.model.trial.showHintMoves )
            {
                this._showingMoveHint = true;
            }

            if( this._showingMoveHint )
            {
                var playerPos = this.model.playerPenguin.getGridPosition();
                var nextPos = null;

                var solution = this.model.trial.maze.getSolution();
                if( solution && solution.length )
                {
                    var solutionIndex = solution.indexOfPoint( playerPos );
                    if( solutionIndex >= 0 )
                    {
                        var nextPosIndex = solutionIndex + 1;
                        if( nextPosIndex < solution.length )
                        {
                            nextPos = solution[ nextPosIndex ];
                        }
                    }
                    else
                    {
                        var playerPath = this.model.playerPenguin.getTraveledPath();
                        if( playerPath && playerPath.length >= 2 )
                        {
                            nextPos = playerPath[ playerPath.length - 2 ];
                        }
                    }
                }

                if( nextPos )
                {
                    var deltaPos = cc.p( nextPos.x - playerPos.x, nextPos.y - playerPos.y );
                    var nextDir = MazeUtility.getDirection( deltaPos.x, deltaPos.y );
                    this.view.setMoveHint( nextDir, this.settings.hintButtonTint, this.settings.hintButtonOpacity );
                    return;
                }
            }
        }

        this.view.setMoveHint( -1, this.settings.hintButtonTint, this.settings.hintButtonOpacity );
    },

    /**
     * Sets up metadata keys.
     */
    configureMetaDataKeys: function()
    {
        this._super();

        this._metadataKeys.action_csv       = "action_csv";     // Actions csv.

        metadataTrialKeys.level             = "level";          // Trial level.
        metadataTrialKeys.travel            = "travel";         // Distance traveled, in grid squares.
        metadataTrialKeys.layout            = "layout";         // Maze layout.
        metadataTrialKeys.length            = "length";         // Manhattan distance to goal.
        metadataTrialKeys.side              = "side";           // Number of maze grid squares on a side.
        metadataTrialKeys.rotations         = "rotations";      // Number of maze rotations.
        metadataTrialKeys.turns             = "turns";          // Number of solution path turns.
        metadataTrialKeys.correct           = "correct";        // Whether the player reached the goal before the computer (T/F).
        metadataTrialKeys.points            = "points";         // Points earned in trial.
        metadataTrialKeys.difference        = "difference";     // Grid space distance between player and computer.
        //metadataTrialKeys.time              = "time";           // Time it took to complete trial.

        // set up metadata keys for action csv section
        this._metadataActionKeys.direction  = "direction";
        this._metadataActionKeys.location_x  = "location_x";
        this._metadataActionKeys.location_y  = "location_y";
        this._metadataActionKeys.on_path  = "on_path";
        this._metadataActionKeys.time_offset  = "time_offset";
        this._metadataActionKeys.time_offset_release  = "time_offset_release";
        this._metadataActionKeys.travel  = "travel";
        this._metadataActionKeys.trial  = "trial";
    },

    /**
     * Creates metadata for a given trial.
     *
     * @param trial     Model trial data.
     * @param visitedMazeSpaces    Array of player's visited grid spaces in maze.
     * @returns {Object}    Returns object containing trial metadata.
     */
    createTrialMetadata:function( trial, visitedMazeSpaces )
    {
        var trialMetadata = {};
        if( trial )
        {
            trialMetadata[ metadataTrialKeys.level ] = trial.levelIndex + 1;
            trialMetadata[ metadataTrialKeys.travel ] = trial.traveledDistance;
            trialMetadata[ metadataTrialKeys.layout ] = this.createMazeLayoutMetadata( trial.maze, visitedMazeSpaces );
            trialMetadata[ metadataTrialKeys.length ] = trial.maze.getSolution().length;
            trialMetadata[ metadataTrialKeys.side ] = trial.maze.getWidth();
            trialMetadata[ metadataTrialKeys.rotations ] = trial.rotationCount;
            trialMetadata[ metadataTrialKeys.turns ] = trial.maze.turnCount;
            trialMetadata[ metadataTrialKeys.correct ] = trial.playerWon ? "T" : "F";
            trialMetadata[ metadataTrialKeys.points ] = trial.score;
            trialMetadata[ metadataTrialKeys.difference ] = trial.positionDifference;
            //trialMetadata[ metadataTrialKeys.time ] = Math.floor( Math.max( trial.endTime - trial.startTime, 0 ) * 1000 );
        }
        return trialMetadata;
    },

    /**
     * Creates maze layout metadata representation for a given maze.
     *
     * @param maze      The maze in question.
     * @param visitedMazeSpaces   Array of grid spaces visited by player.
     * @returns {string} Returns base-32 encoded maze metadata string.
     */
    createMazeLayoutMetadata:function( maze, visitedMazeSpaces )
    {
        var layoutMetadata = "";
        if( maze )
        {
            for( var y = maze.getHeight() - 1; y >= 0; y-- )
            {
                for( var x = 0; x < maze.getWidth(); x++ )
                {
                    var encodingNumber = 0;

                    var room = maze.getRoom( x, y );
                    if( room )
                    {
                        var closed = room.createClosedArray();
                        if( closed )
                        {
                            // set bits for encoded wall directions
                            if( closed.contains( MazeDirections.Up ) )
                            {
                                encodingNumber += Math.pow( 2, 0 );
                            }
                            if( closed.contains( MazeDirections.Right ) )
                            {
                                encodingNumber += Math.pow( 2, 1 );
                            }
                            if( closed.contains( MazeDirections.Down ) )
                            {
                                encodingNumber += Math.pow( 2, 2 );
                            }
                            if( closed.contains( MazeDirections.Left ) )
                            {
                                encodingNumber += Math.pow( 2, 3 );
                            }

                            // encode bit for visited rooms
                            if( visitedMazeSpaces && visitedMazeSpaces.containsPoint( cc.p( x, y ) ) )
                            {
                                encodingNumber += Math.pow( 2, 4 );
                            }
                        }
                    }

                    var encodingString = encodingNumber.toString( 32 ).toUpperCase();;
                    layoutMetadata += encodingString;
                }
            }
        }
        return layoutMetadata;
    },

    createActionMetadata:function()
    {
        var metadataActions = [];

        for( var i = 0; i < this.model.actions.length; i++ )
        {
            var action = this.model.actions[i];

            var metadataAction = {};

            var directionLetter = "";
            if( action.type === "movement" )
            {
                switch( action.direction )
                {
                    case MazeDirections.Up:
                        directionLetter = "U";
                        break;
                    case MazeDirections.Down:
                        directionLetter = "D";
                        break;
                    case MazeDirections.Left:
                        directionLetter = "L";
                        break;
                    case MazeDirections.Right:
                        directionLetter = "R";
                        break;
                }
            }
            else if( action.type === "rotation" )
            {
                directionLetter = action.direction > 0 ? ">" : "<";
            }

            metadataAction[ this._metadataActionKeys.direction ] = directionLetter;
            metadataAction[ this._metadataActionKeys.location_x ] = action.gridPos.x;
            metadataAction[ this._metadataActionKeys.location_y ] = action.gridPos.y;
            metadataAction[ this._metadataActionKeys.on_path ] = action.onPath ? "T" : "F";
            metadataAction[ this._metadataActionKeys.time_offset ] = action.timeOffset;
            metadataAction[ this._metadataActionKeys.time_offset_release ] = action.timeOffsetRelease;
            metadataAction[ this._metadataActionKeys.travel ] = action.travelDistance;
            metadataAction[ this._metadataActionKeys.trial ] = action.trialIndex + 1;

            metadataActions.push( metadataAction );
        }

        return metadataActions;
    },

    _buttonPressed:function( buttonName )
    {
        this._super(buttonName);
        
        this._doKeyboardButton( buttonName, true );
    },

    _buttonReleased:function( buttonName )
    {
        this._super(buttonName);
        
        this._doKeyboardButton( buttonName, false );
    },

    _doKeyboardButton:function( buttonName, pressed )
    {
        if( cc.KEY == undefined ){ return; }
        if ( buttonName === Interaction.keyboardPrefix + String( cc.KEY.up ) )
        {
            this.setInputDirection( MazeDirections.Up, pressed );
        }
        else if ( buttonName === Interaction.keyboardPrefix + String( cc.KEY.down ) )
        {
            this.setInputDirection( MazeDirections.Down, pressed );
        }
        else if ( buttonName === Interaction.keyboardPrefix + String( cc.KEY.left ) )
        {
            this.setInputDirection( MazeDirections.Left, pressed );
        }
        else if ( buttonName === Interaction.keyboardPrefix + String( cc.KEY.right ) )
        {
            this.setInputDirection( MazeDirections.Right, pressed );
        }
    },

    _updateForEasyMode:function()
    {
        if( this.model.easyMode )
        {
            if( this.model.computerPenguin )
            {
                // slow down other penguin
                this.model.computerPenguin.setMoveTime( 8 );
            }
        }
    },

    /**
     * Checks to see if the trial has been won by a penguin.
     *
     * @private
     */
    _checkForWin:function()
    {
        //check to see if either penguin is at the goal
        if ( this.model.playerPenguin && this.model.computerPenguin && this.model.phase === PenguinPursuitPhases.Trial )
        {
            var playerPos = this.model.playerPenguin._gridPosition;
            var computerPos = this.model.computerPenguin._gridPosition;

            if ( playerPos.x === this.model.trial.maze.goal.x && playerPos.y === this.model.trial.maze.goal.y )
            {
                //the player won! :-D
                this.endTrialPhase( true );
            }
            else if ( computerPos.x === this.model.trial.maze.goal.x && computerPos.y === this.model.trial.maze.goal.y )
            {
                //the computer won. :'-(
                this.endTrialPhase( false );
            }
        }
    },

    /**
     * Moves the game to the next trial, possibly changing levels if required.
     * @private
     */
    _moveToNextTrial:function( levelDelta )
    {
        // based on past performance, update the current level
        if( this.model.persistentData.isPracticeMode )
        {
            if( levelDelta > 0 )
            {
                this.view.playAnimation( "nextTutorial", this._movedToNextTrial, this );
                this.startPreTrialPhase();
            }
            else
            {
                this.onCompleteTutorial( false );
            }
        }
        else
        {
            this.view.stopActions();
            this.view.setLevel( this.model.levelIndex + 1 );

            // play specific animations for different changes in level
            var nextLevelAnim = "slideLeft";
            if( levelDelta > 0 )
            {
                nextLevelAnim = "levelUp";
            }
            else if( levelDelta < 0 )
            {
                nextLevelAnim = "levelDown";
            }
            else if( !this.model.trial.playerWon )
            {
                nextLevelAnim = "slideRight";
            }
            this.view.playAnimation( nextLevelAnim, this._movedToNextTrial, this );

            this.startPreTrialPhase();
        }
    },

    /**
     * Called when the Flash movie has finished changing trials.
     * @private
     */
    _movedToNextTrial:function()
    {
        // start next countdown
        if( !this.model.persistentData.isPracticeMode )
        {
            this.view.playReadyGo();
        }
    },

    /**
     * Callback invoked when an individual path marker has been revealed.
     *
     * @param markerIndex
     * @param totalMarkers
     */
    pathMarkerRevealed:function( isPlayerPath, index )
    {
        // adjust adding score as each path marker is revealed
        if( !this.model.persistentData.isPracticeMode )
        {
            // don't update adding score for first player space; it's not worth anything
            if( !( isPlayerPath && index === 0 ) )
            {
                var addingScoreDelta = this.settings.levelScoreMultiplier * ( this.model.levelIndex + 1 );
                this.view.updateAddingScore( addingScoreDelta );
            }
        }
    },

    applyTrialScore:function()
    {
        if( !this.model.persistentData.isPracticeMode )
        {
            this.view.endAddingScore( this.model.trial.score );
            this.view.setScore( this.model.score );
        }
    },

    /**
     * Periodically updates button highlights on a timer, indicating buttons if player doesn't move.
     *
     * @param highlight Whether to start or end highlighting buttons.
     */
    setHighlightingButtons:function( highlight )
    {
        if( highlight && !this._highlightingButtons )
        {
            var sequence = cc.Sequence.create(
                cc.DelayTime.create( this.settings.highlightButtonsInterval ),
                cc.CallFunc.create( this.updateButtonHighlight.bind( this ) )
            );
            var repeatAction = cc.RepeatForever.create( sequence );
            this.view.highlightButtonsNode.runAction( repeatAction );
        }
        else if( !highlight && this._highlightingButtons )
        {
            this.view.highlightButtonsNode.stopAllActions();
        }
        this._highlightingButtons = highlight;
    },

    /**
     * If player is in trial and he hasn't touched the buttons in a while, highlight them.
     */
    updateButtonHighlight:function()
    {
        if( this.model.phase === PenguinPursuitPhases.Trial && !this._buttonsPressedSinceLastButtonHighlight )
        {
            this.view.highlightButtons( this.settings.highlightButtonsScale, this.settings.highlightButtonsTime );
        }
        this._buttonsPressedSinceLastButtonHighlight = 0;
    },

    /**
     * Sets whether the game should indicate the compass and up button.
     *
     * @param compass Whether to indicate the compass.
     * @param upButton Whether to indicate the up button.
     * @param compassLabel Whether to indicate the compass label.
     * @param upButtonLabel Whether to indicate the up button label.
     */
    setIndicating:function( compass, upButton, compassLabel, upButtonLabel )
    {
        this.view.showIndicators( compass, upButton, compassLabel, upButtonLabel );
    },

    /**
     * Sets game to the pre-trial phase (right before gameplay).
     */
    startPreTrialPhase:function()
    {
        this.model.phase = PenguinPursuitPhases.PreTrial;
        this.model.maxLevel = this.model.levelIndex + 1;

        this.model.playerPenguin.startPreTrial();
        this.model.computerPenguin.startPreTrial();
        this.model.bump.startPreTrial();

        this.view.startPreTrial();

        this._validMovesInARow = 0;
        this._invalidMovesInARow = 0;
        this._showingMoveHint = true;

        this.updateMoveHint();

        this.setHighlightingButtons( false );
        this.setIndicating( false, false, false, false );

        this._buttonsPressedSinceLastButtonHighlight = 0;
    },

    /**
     * Sets game to the trial phase (normal gameplay).
     */
    startTrialPhase:function()
    {
        this.model.phase = PenguinPursuitPhases.Trial;

        this.model.playerPenguin.startTrial( this.model.trial );
        this.model.computerPenguin.startTrial( this.model.trial );
        this._beginHeadStartTimer( this.settings.playerHeadStartTime );

        this.model.trial.startTime = lumosity.game.frametime;

        this.view.startTrial();

        this.beginTrial();

        this._tryMazeRotation( true );

        this.setHighlightingButtons( this.model.trial.highlightButtons );
        this.setIndicating( this.model.trial.indicateCompass, this.model.trial.indicateUpButton,
            this.model.trial.indicateCompassLabel, this.model.trial.indicateUpButtonLabel );
    },

    /**
     * Sets game to the post-trial phase (right after gameplay).
     */
    startPostTrialPhase:function()
    {
        this.model.phase = PenguinPursuitPhases.PostTrial;
    },

    /**
     * Ends the current trial, depending on if player won or not.
     *
     * @param playerWon     Whether the player won.
     */
    endTrialPhase:function( playerWon )
    {
        this.startPostTrialPhase();

        this.view.endTrial();

        // update model to reflect win or loss
        this.model.handleTrialComplete( playerWon );

        this.model.trial.endTime = lumosity.game.frametime;

        this.model.playerPenguin.endTrial();
        this.model.computerPenguin.endTrial();
        this.model.bump.endTrial();
        this.model.fish.show( false );

        if( playerWon )
        {
            audio.playEffect( this.settings.sounds.playerWinMusic );
        }
        else
        {
            audio.playEffect( this.settings.sounds.playerLoseMusic );
        }

        this.setHighlightingButtons( false );
        this.setIndicating( false, false, false, false );

        // play end trial sequence in Flash file, which calls win/lose animation cues
        this.view.playAnimation( "endTrial" );

        if( !this.model.persistentData.isPracticeMode )
        {
            this.view.startAddingScore();
        }

        // calculate trial score based on player and computer distances
        var playerDist = this.model.playerPenguin.getTraveledSolutionPathDistance();
        var computerDist = this.model.computerPenguin.getTraveledSolutionPathDistance();
        if( playerWon )
        {
            this.model.trial.score = ( playerDist * 2 - computerDist ) * ( this.model.levelIndex + 1 ) * this.settings.levelScoreMultiplier;
        }
        else
        {
            this.model.trial.score = playerDist * ( this.model.levelIndex + 1 ) * this.settings.levelScoreMultiplier;
        }
        // round score to the nearest 10
        this.model.trial.score = Math.round( this.model.trial.score / 10 ) * 10;

        this.model.score += this.model.trial.score;

        this.model.trial.positionDifference = playerDist - computerDist;

        // reveal penguins' paths at end of trial
        var playerSolutionPath = MazeUtility.getCommonPath( this.model.playerPenguin.getTraveledPath(), this.model.trial.maze.getSolution() );
        var computerRemainingPath = this.model.persistentData.isPracticeMode ? null : this.model.computerPenguin.getRemainingPath();
        this.view.revealPathMarkers( playerSolutionPath, computerRemainingPath, this.settings.pathMarkerRevealTime,
            this.settings.pathMarkerRevealMaxDelay, this.pathMarkerRevealed.bind( this ), this.finalizeTrial.bind( this ) );
    },

    /**
     * Wraps up a trial, collecting relevant info for metadata and moving to the next trial.
     */
    finalizeTrial:function()
    {
        var trialMetadata = this.createTrialMetadata( this.model.trial, this.model.playerPenguin.getTraveledPath() );
        this.endTrial( trialMetadata, this.model.trial.playerWon, false, false, false, false );

        var levelDelta = this.adjustLevelForNextTrial();

        // fold the trial score into the player's main score element, after a delay
        LumosityUtils.executeDelayedFunction( this.applyTrialScore, this.settings.applyTrialScoreDelay, this, this.view );

        this.queueTrialTransition( levelDelta );
    },

    /**
     * Updates current level index based on player's performance and leveling criteria.
     */
    adjustLevelForNextTrial:function()
    {
        var levelDelta = 0;

        if( this.model.persistentData.isPracticeMode )
        {
            var oldTutorialIndex = this.model.tutorialIndex;
            if( this.settings.tutorialLevels.length > this.model.tutorialIndex + 1 )
            {
                this.model.tutorialIndex++;
            }
            levelDelta = this.model.tutorialIndex - oldTutorialIndex;
        }
        else
        {
            var oldLevelIndex = this.model.levelIndex;

            var barelyLost = this.model.trial.positionDifference >= this.settings.barelyLostDistance * -1;

            // adjust level based on recent wins and losses
            if( this.model.trial.playerWon &&
                ( this.model.trial.authoredMaze || this.model.consecutiveWinsAtLevel >= this.settings.consecutiveWinsToRaiseLevel ) )
            {
                this.model.levelIndex++;
            }
            else if( !this.model.trial.playerWon &&
                ( this.model.trial.authoredMaze || ( this.model.consecutiveLossesAtLevel >= this.settings.consecutiveLossesToLowerLevel && !barelyLost ) ) )
            {
                this.model.levelIndex--;
            }

            this.model.levelIndex = LumosityUtils.clamp( this.model.levelIndex, 0, this.settings.levels.length - 1 );

            if( this.model.levelIndex != oldLevelIndex )
            {
                // level changed, so reset how player has done at this new level
                this.model.consecutiveWinsAtLevel = 0;
                this.model.consecutiveLossesAtLevel = 0;
            }
            else
            {
                if( this.model.trial.playerWon )
                {
                    this.model.consecutiveLossesAtLevel = 0;
                }
                else
                {
                    this.model.consecutiveWinsAtLevel = 0;
                }
            }

            levelDelta = this.model.levelIndex - oldLevelIndex;
        }
        return levelDelta;
    },

    /**
     * Moves on to the next trial or ends the game if it's over.
     */
    queueTrialTransition:function( levelDelta )
    {
        var transitionTime = this.model.persistentData.isPracticeMode ? this.settings.endTrialTransitionTimeTutorial : this.settings.endTrialTransitionTime;
        if ( this.model.trialIndex >= this.settings.getGameLengthInTrials() - 1 )
        {
            // all trials are complete; end game

            var actionMetadata = this.createActionMetadata( this.model.trial );
            this.model.metadata[ this._metadataKeys.action_csv ] = lu.jsToCsv( actionMetadata );

            LumosityUtils.executeDelayedFunction( this.onEndGameplay, transitionTime, this, this.view );
        }
        else
        {
            // still trials left, so move on to the next
            LumosityUtils.executeDelayedFunction( this._moveToNextTrial.bind( this, levelDelta ), transitionTime, this, this.view );
        }
    },

    /**
     * Rotates the maze.
     *
     * @param degrees   Target rotation, in degrees.
     * @param instant   Whether the rotation should happen instantly or over time.
     */
    rotateMaze:function( degrees, instant )
    {
        this.model.mazeRotation = degrees;
        
        if ( instant )
        {
            //for instant rotations, since these get called willy-nilly, we don't want to record an action end
            this.view.rotateMaze( degrees, instant, this.model.persistentData.isPracticeMode );
        }
        else
        {
            this.view.rotateMaze( degrees, instant, this.model.persistentData.isPracticeMode, this.finishedRotatingMaze, this );
        }

        if( this.model.phase === PenguinPursuitPhases.Trial )
        {
            // update penguins' walk directions to reflect new rotation
            
            //
            //NOTE: This code was changed to fix a bug where the penguin
            //      is walking as the maze rotates in practice mode. See bug:
            //      https://lumoslabs.atlassian.net/browse/MOBGAMES-1861
            //
            this.model.playerPenguin.updateIdleDirection();
            this.model.computerPenguin.updateIdleDirection();
            // this.model.playerPenguin.updateWalkDirection();
            // this.model.computerPenguin.updateWalkDirection();
        }
        this.updateMoveHint();

     this.updateMoveHint()
    },
    
    finishedRotatingMaze:function()
    {
        //instant maze rotations will happen at first, and not have a matching action
        var actionIndex = this.model.getLastActionIndexOfType( "rotation" );
        if ( actionIndex < 0 )
        {
            return;
        }
        
        var onPath = this.model.playerPenguin.isOnSolutionPath();
        var timeOffsetRelease = lu.convertSecondsToMilliseconds( lumosity.game.frametime - this.model.trial.startTime );
        var traveledDistance = this.model.playerPenguin.getTraveledPath().length;
        
        this.model.finishRecordingAction( actionIndex, onPath, timeOffsetRelease, traveledDistance );
    },

    /**
     * Generates a new trial, including a new maze.
     * @private
     */
    _createNewTrial:function()
    {
        var trialSettings = this.model.getCurTrialSettings();

        this.model.createNewTrial( this.settings, trialSettings, this.model.levelIndex, this.model.trialIndex );
        this.model.assignNewRotationStepCount( trialSettings.minRotateMoves, trialSettings.maxRotateMoves );

        // create new maze for trial
        var mazeSizeScale = this.model.persistentData.isPracticeMode ? this.settings.tutorialMazeSizeScale : 1;
        var mazeOffsetScale = this.model.persistentData.isPracticeMode ? this.settings.tutorialMazeOffsetScale : 0;
        this.view.prepareNewMaze( mazeSizeScale, mazeOffsetScale, !this.model.trial.hideCompass );

        this.rotateMaze( 0, true );
        this.view.mazeView.setVisible( true );
        this.model.fish.show( true );

        this._updateForEasyMode();
    },

    /**
     * Give player a small head start before allowing AI penguin to move.
     * @param time Head start delay, in seconds.
     * @private
     */
    _beginHeadStartTimer:function( time )
    {
        LumosityUtils.executeDelayedFunction( this._headStartTimerComplete, time, this, this.view );
    },

    /**
     * Start AI penguin's movement after head start delay has elapsed.
     * @private
     */
    _headStartTimerComplete:function()
    {
        if( this.model.phase == PenguinPursuitPhases.Trial && this.model.computerPenguin )
        {
            this.model.computerPenguin.beginMoving();
        }
    },

    _recordValidMove:function( moveValid )
    {
        this._validMovesInARow = moveValid ? this._validMovesInARow + 1 : 0;
        this._invalidMovesInARow = !moveValid ? this._invalidMovesInARow + 1 : 0;
        this.updateMoveHint();
    },

    /**
     * Attempts to rotate the maze if conditions are right.
     * @private
     */
    _tryMazeRotation:function( starting )
    {
        var trialSettings = this.model.getCurTrialSettings();
        if ( trialSettings.maxRotations !== undefined && this.model.trial.rotationCount >= trialSettings.maxRotations )
        {
            // no more rotations are allowed
            return;
        }

        if( this.model.easyMode )
        {
            return;
        }

        var startRotated = this.model.trial.startsRotated && starting;
        var oldRotation = this.model.mazeRotation;

        if ( this.model.shouldRotationHappen() || startRotated )
        {
            // pick a random rotation angle and spin maze
            var possibleRotations = trialSettings.rotationAngles || this.settings.defaultRotationAngles;
            var possibleWeights =  trialSettings.rotationWeights;

            var nextRotation = this._getNextRotationAngle( this.model.mazeRotation, possibleRotations, possibleWeights );
            if( nextRotation !== undefined )
            {
                // valid rotation found, so spin dat maze
                this.rotateMaze( nextRotation );
                this.model.assignNewRotationStepCount( trialSettings.minRotateMoves, trialSettings.maxRotateMoves );
                this.model.trial.rotationCount++;

                var direction = nextRotation - oldRotation;
                var gridPos = this.model.playerPenguin.getGridPosition();
                var timeOffset = lu.convertSecondsToMilliseconds( lumosity.game.frametime - this.model.trial.startTime );
                
                this.model.beginRecordingAction( "rotation", direction, gridPos, timeOffset, this.model.trialIndex );
            }
        }

        this.updateMoveHint();
    },

    /**
     * Using state-of-the-art technology, considers current rotation and weighted angle probabilities to get the next maze rotation angle.
     *
     * @param curAngle  Current maze rotation angle.
     * @param possibleAngles    Array of allowed rotation angles.
     * @param possibleWeights   Array of weights for the above rotation angles.
     * @returns {*} Next rotation angle.
     * @private
     */
    _getNextRotationAngle:function( curAngle, possibleAngles, possibleWeights )
    {
        var w = 0;
        var weights = [];
        var curAngleIndex = -1;
        var totalWeight = 0;
        var curWeight = 0;
        var weightProbability = Math.random();

        if( possibleAngles && possibleAngles.length )
        {
            // build safe list of weights for each possible angle
            for( w = 0; w < possibleAngles.length; w++ )
            {
                weights[ w ] = ( possibleWeights && possibleWeights.length > w ) ? possibleWeights[ w ] : 1.0;
            }

            // zero the probability that our current angle will be our next
            curAngleIndex = possibleAngles.indexOf( curAngle );
            if( curAngleIndex >= 0 )
            {
                weights[ curAngleIndex ] = 0;
            }

            // get total weight for relative weights normalization
            for( w = 0; w < weights.length; w++ )
            {
                totalWeight += weights[ w ];
            }

            if( totalWeight > 0 )
            {
                // normalize weights
                for( w = 0; w < weights.length; w++ )
                {
                    weights[ w ] /= totalWeight;
                }
                for( w = 0; w < weights.length; w++ )
                {
                    // add this weight to the current sum of weights, so we can use the probability to choose one
                    curWeight += weights[ w ];

                    // check the new weight sum against the current weight
                    if( weightProbability < curWeight )
                    {
                        // random probability is less than current weight, so we have a winner!
                        // return the corresponding rotation angle
                        return possibleAngles[ w ];
                    }
                }
            }
        }

        return undefined;
    }
});
