var require = require;

require( "src/Game/Base/GameModel.js" );
var GameModel = GameModel;

require( "src/Game/PenguinCharacterAI.js" );
var PenguinCharacterAI = PenguinCharacterAI;

require( "src/Game/PenguinCharacterPlayer.js" );
var PenguinCharacterPlayer = PenguinCharacterPlayer;

require( "src/Game/Maze/Maze.js" );
var Maze = Maze;

require( "src/Game/Fish.js" );
var Fish = Fish;

require( "src/Game/Bump.js" );
var Bump = Bump;

/**
 * Discrete enumerated states that the model can be in. Model can only be in one state at a time.
 *
 * @type {{None: number, PreTrial: number, Trial: number, PostTrial: number}}
 */
var PenguinPursuitPhases =
{
    None        : 0,
    PreTrial    : 1,
    Trial       : 2,
    PostTrial   : 3
};

var PenguinPursuitModel = GameModel.extend(
{
    /** @lends PenguinPursuitModel.prototype */

    /** Current game settings. */
    settings:null,

    /** Game's current phase. */
    phase:PenguinPursuitPhases.None,

    /** Current trial index. */
    trialIndex:0,

    /** Current level index. */
    levelIndex:0,

    /** Maximum level reached by the player, this playthrough. */
    maxLevel:0,

    /** Current trial index in tutorial trials. */
    tutorialIndex:0,

    /** Current number of trials won/lost at the current level; used in raising/lowering level. */
    consecutiveWinsAtLevel:0,

    /** Current number of trials won/lost at the current level; used in raising/lowering level. */
    consecutiveLossesAtLevel:0,

    /** Number of trials the player has won. */
    correctTrials:0,

    /** Current maze rotation. */
    mazeRotation:0,

    /** Queue of actions for metadata. */
    actions:[],

    /** Current trial data. */
    trial:null,

    /** Computer penguin character. */
    computerPenguin:null,

    /** Player penguin character. */
    playerPenguin:null,

    /** Fish at goal. */
    fish:null,

    /** Bump feedback animation. */
    bump:null,
                                           
    endingLevelIndexName: "endingLevelIndex",
    
    /**
     * The Penguin Pursuit model.
     * @constructs
     */
    ctor:function( settings )
    {
        this.settings = settings;

        this.fish = new Fish( settings );
        this.bump = new Bump( settings );
        this.computerPenguin = new PenguinCharacterAI( this, settings );
        this.playerPenguin = new PenguinCharacterPlayer( this, settings );
        this.reset();
    },

    /**
     * Clears model's state.
     */
    reset:function()
    {
        this.phase = PenguinPursuitPhases.None;
        this.trialIndex = 0;
        this.levelIndex = this.settings.startingLevelIndex;
        this.maxLevel = 0;
        this.tutorialIndex = 0;
        this.consecutiveWinsAtLevel = 0;
        this.consecutiveLossesAtLevel = 0;
        this.correctTrials = 0;
        this.mazeRotation = 0;
        this.actions = [];
        this.trial = null;
        this.score = 0;

        if( this.settings.loadFromLastResults )
        {
            this._loadFromLastResults( this.settings );
        }
    },

    /**
     * Exports game results to app.
     *
     * @returns {*} Current game results.
     */
    exportResults:function()
    {
        var stats = {};

        stats[ this.StatsKeys.LevelStat ] = this.maxLevel;
        stats[ this.endingLevelIndexName ] = this.levelIndex;
                                           
        return this._super( stats, this.StatsKeys.LevelStat );
    },

    /**
     * Called when each trial is complete.
     *
     * @param playerWon Whether the player won.
     */
    handleTrialComplete:function( playerWon )
    {
        this.trial.playerWon = playerWon;

        // adjust number of won/lost trials at current level
        if( playerWon )
        {
            this.correctTrials++;

            this.consecutiveWinsAtLevel++;
            this.consecutiveLossesAtLevel = 0;
        }
        else
        {
            this.consecutiveWinsAtLevel = 0;
            this.consecutiveLossesAtLevel++;
        }
    },

    getCurTrialSettings:function()
    {
        if( this.persistentData.isPracticeMode )
        {
            return this.settings.tutorialLevels[ this.tutorialIndex ];
        }
        else
        {
            return this.settings.levels[ this.levelIndex ];
        }
    },

    /**
     *  Creates a new trial from the level data passed in, including a new maze. Resets penguins.
     *
     *  @param {Object} Current game settings.
     *  @param {Object} levelSettings Data from the settings on how this level should be created.
     */
    createNewTrial:function( settings, levelSettings, levelIndex, trialIndex )
    {
        this.phase = PenguinPursuitPhases.None;

        var maze = new Maze();
        if( levelSettings.authoredMaze )
        {
            maze.initAuthoredMaze( settings.authoredMazes[ levelSettings.authoredMaze ] );
        }
        else
        {
            maze.initRandomMaze( levelSettings.size, levelSettings.size, levelSettings.goalDistance, settings.getMinPathTurns( levelIndex ), settings.getMaxPathTurns( levelIndex ) );
        }

        this.trial =
        {
            levelIndex: levelIndex,
            trialIndex: trialIndex,
            maze: maze,
            playerWon: false,
            traveledDistance: 0,
            timesMovedSinceRotation: 0,
            rotationStepsRequired: 0,
            rotationCount: 0,
            authoredMaze: !!levelSettings.authoredMaze,
            startsRotated: !!levelSettings.startsRotated,
            showMoveHint: !!levelSettings.showMoveHint,
            showHintMoves: levelSettings.showHintMoves || 3,
            hideHintMoves: levelSettings.hideHintMoves || 3,
            highlightButtons: !!levelSettings.highlightButtons,
            hideCompass: !!levelSettings.hideCompass,
            indicateCompass: !!levelSettings.indicateCompass,
            indicateUpButton: !!levelSettings.indicateUpButton,
            indicateCompassLabel: !!levelSettings.indicateCompassLabel,
            indicateUpButtonLabel: !!levelSettings.indicateUpButtonLabel,
            startTime: 0,
            endTime: 0,
            score: 0
        };

        this.playerPenguin.setGridPosition( this.trial.maze.start );
        this.playerPenguin.setMoveTime( levelSettings.playerMoveTime );
        this.playerPenguin.setMaze( this.trial.maze );

        this.computerPenguin.setGridPosition( this.trial.maze.start );
        if( this.persistentData.isPracticeMode )
        {
            this.computerPenguin.setActiveInTrial( false );
        }
        else
        {
            this.computerPenguin.setActiveInTrial( true );
            this.computerPenguin.setMoveTime( levelSettings.computerMoveTime );
        }
    },

    beginRecordingAction:function( type, direction, gridPos, timeOffset, trialIndex )
    {
        var action =
        {
            type: type,
            direction: direction,
            gridPos: LumosityUtils.copyPoint( gridPos ),
            timeOffset: timeOffset,
            trialIndex: trialIndex
        }

        this.actions.push( action );
    },

    finishRecordingAction:function( actionIndex, onPath, timeOffsetRelease, travelDistance )
    {
        this.actions[ actionIndex ].onPath = onPath;
        this.actions[ actionIndex ].timeOffsetRelease = timeOffsetRelease;
        this.actions[ actionIndex ].travelDistance = travelDistance;
    },
    
    getActionCount:function()
    {
        return this.actions.length;
    },
    
    getLastActionIndexOfType:function( type )
    {
        var index;
        for ( index = this.actions.length - 1; index >= 0; index-- )
        {
            if ( this.actions[ index ].type === type )
            {
                return index;
            }
        }
        return -1;
    },

    /**
     *  Assigns a new value to the step count required for a rotation.
     *  Also clears the timesMovedSinceRotation variable to 0.
     *
     *  @param {number} min The min step count.
     *  @param {number} max The max step count.
     */
    assignNewRotationStepCount:function( min, max )
    {
        this.trial.rotationStepsRequired = Math.floor( Math.random() * ( max - min ) + min );
        this.trial.timesMovedSinceRotation = 0;
    },

    /**
     * Returns whether a rotation should happen, based on how long it's been since last rotation.
     *
     * @returns {boolean}   Whether we should rotate.
     */
    shouldRotationHappen:function()
    {
        return this.trial.timesMovedSinceRotation >= this.trial.rotationStepsRequired;
    },

    /**
     * Adjusts model state based on last game results (passed in from app).
     *
     * @param settings  Current game settings.
     * @private
     */
    _loadFromLastResults:function( settings )
    {
        // get level info from last results
        if( lumosity.game.params &&
            lumosity.game.params.last_results &&
            lumosity.game.params.last_results.gameStats )
        {
            var endingLevelIndex = lumosity.game.params.last_results.gameStats[ this.endingLevelIndexName ];
            if( LumosityUtils.isNumeric( endingLevelIndex ) )
            {
                // start each playthrough at a level offset from the player's last ending level
                this.levelIndex = Math.max( endingLevelIndex + settings.startingLevelOffsetFromEndingLevel, 0 );
                //ll.verbose( "SETTING LEVEL INDEX FROM STATS: " + this.levelIndex );
            }
        }
    }
});
