var cc = cc;
var require = require;

require( "src/Game/Base/GameView.js" );
var GameView = GameView;

require( "src/Game/Widget/GameWidgetHud.js" );
var GameWidgetHud = GameWidgetHud;

require( "src/Game/PenguinCharacterView.js" );
var PenguinCharacterView = PenguinCharacterView;

require( "src/Game/Maze/MazeConstants.js" );
var MazeDirections = MazeDirections;
var MazeConstants = MazeConstants;

require( "src/View/LumosityButtonLayer.js" );
var LumosityButtonLayer = LumosityButtonLayer;

require( "src/View/LumosityButton.js" );
var LumosityButton = LumosityButton;

require( "src/View/LumosityJoystickLayer.js" );
var LumosityJoystickLayer = LumosityJoystickLayer;

require( "src/View/LumosityJoystick.js" );
var LumosityJoystick = LumosityJoystick;

var PenguinPursuitView = GameView.extend(
{
    /** @lends PenguinPursuitView.prototype */

    /** Name of the countdown animation to play before timer starts. */
    _READY_SET_GO_WAIT_NAME:"UI_readyGoSlideAnim",

    /** Name of the countdown animation to player as the timer counts down. */
    _READY_SET_GO_PLAY_NAME:"UI_readyGoSlidePlay",

    /** The PenguinCharacterView for the penguin controlled by the player. */
    playerPenguinView:null,
    
    /** The PenguinCharacterView for the penguin controlled by the computer. */
    computerPenguinView:null,

    /** The fish's AnimatedEntityView. */
    fishView:null,

    /** The bump's AnimatedEntityView. */
    bumpView:null,

    /** The MazeView that represents the latest maze. */
    mazeView:null,

    /** Size of the maze view, specified by Flash file. */
    mazeSize:null,

    /** Node used only to run an action that continually highlights buttons. */
    highlightButtonsNode:null,

    /** Container node for maze, from Flash file. */
    _mazeContainer:null,

    /** Container node for direction buttons, from Flash file. */
    _buttonContainer:null,

    /** Array of player's path markers revealed at end of trial, to visualize scoring. */
    _playerPathMarkers:[],

    /** Array of computer's path markers revealed at end of trial, to visualize scoring. */
    _computerPathMarkers:[],

    /** The value of the adding score displayed in upper right. Displayed when tablulating scores for each grid space. */
    _curAddingScore:0,

    /**
     * The Penguin Pursuit view.
     * @constructs
     * @param {cc.Size} screenSize The size of the screen that this view will be inserted into.
     * @param {Object} [settings] The settings that we can pull view-based data from.
     * @param {PenguinPursuitModel} [model] The GameModel that we are linked to.
     * @param {Object} [interactionResponder] The object that we should send interaction events to. We also use this is a default animation event listener.
     * @param {boolean} [preload=true] If true, will preload all flipbooks and other assets upon creation.
     */
    ctor:function( screenSize, settings, model, interactionResponder, preload )
    {
        this._super( screenSize, settings, model, interactionResponder, preload );

        this.id = "penguinPursuitView";
        this._mazeContainer.id = "mazeContainer";

        // set up maze view
        var mazeWidth = Math.min( this._mazeContainer.getContentSize().width, this._mazeContainer.getContentSize().height );
        this.mazeSize = cc.size( mazeWidth, mazeWidth );
        this.mazeView = new MazeView( this.mazeSize, settings );
        this._mazeContainer.addChild( this.mazeView );

        this.fishView = this.model.fish.view;
        this.fishView.mazeScale = settings.fishScale;
        this.mazeView.addInhabitant( model.fish.view, cc.POINT_ZERO );
        this.fishView.setZOrder( settings.zOrders.fish );

        this.bumpView = this.model.bump.view;
        this.bumpView.mazeScale = 1.2;
        this.bumpView.skipMazeRotation = true;
        this.bumpView.accountForWalls = true;
        this.mazeView.addInhabitant( this.bumpView, cc.POINT_ZERO );
        this.bumpView.setZOrder( settings.zOrders.bump );
        this.model.bump.setMazeView( this.mazeView );

        // figure out max number of grid spaces and create path markers
        var maxPathMarkers = 64;
        var pathMarkerIndex = 0;
        for ( pathMarkerIndex = 0; pathMarkerIndex < maxPathMarkers; pathMarkerIndex++ )
        {
            var playerMarker = cc.Sprite.create( "penguin/penguin_footprintDotRed.png" );
            playerMarker.id = "playerMarker" + pathMarkerIndex;
            playerMarker.mazeScale = settings.pathMarkerScale;
            playerMarker.accountForWalls = true;
            this._playerPathMarkers.push( playerMarker );
            this.mazeView.addInhabitant( playerMarker, cc.POINT_ZERO );
            playerMarker.setZOrder( settings.zOrders.pathMarkerPlayer );

            var computerMarker = cc.Sprite.create( "penguin/penguin_footprintDotBlue.png" );
            computerMarker.id = "computerMarker" + pathMarkerIndex;
            computerMarker.mazeScale = settings.pathMarkerScale;
            computerMarker.accountForWalls = true;
            this._computerPathMarkers.push( computerMarker );
            this.mazeView.addInhabitant( computerMarker, cc.POINT_ZERO );
            computerMarker.setZOrder( settings.zOrders.pathMarkerComputer );
        }
        this.hidePathMarkers();

        this.computerPenguinView = this.model.computerPenguin.view;
        this.computerPenguinView.accountForWalls = true;
        this.computerPenguinView.mazeOffset = cc.p( settings.penguinOffsetComputer.x, settings.penguinOffsetComputer.y );
        this.computerPenguinView.mazeScale = settings.penguinScale;
        this.computerPenguinView.setMazeView( this.mazeView );
        this.mazeView.addInhabitant( this.computerPenguinView, cc.POINT_ZERO );
        this.computerPenguinView.setZOrder( settings.zOrders.penguinComputer );

        this.playerPenguinView = this.model.playerPenguin.view;
        this.playerPenguinView.accountForWalls = true;
        this.playerPenguinView.mazeOffset = cc.p( settings.penguinOffsetPlayer.x, settings.penguinOffsetPlayer.y );
        this.playerPenguinView.mazeScale = settings.penguinScale;
        this.playerPenguinView.setMazeView( this.mazeView );
        this.mazeView.addInhabitant( this.playerPenguinView, cc.POINT_ZERO );
        this.playerPenguinView.setZOrder( settings.zOrders.penguinPlayer );

        // set up ui buttons
        var selectedTopDirectionName = "UI/UI_pieButton_topselected.png";
        var unselectedTopDirectionName = "UI/UI_pieButton_topdeselected.png"
        var selectedBottomDirectionName = "UI/UI_pieButton_bottomselected.png";
        var unselectedBottomDirectionName = "UI/UI_pieButton_bottomdeselected.png"

        var xButton = this._buttonContainer.getContentSize().width * 0.5;
        var yButton = this._buttonContainer.getContentSize().height * 0.5;
        var dirButtonPos = cc.p( xButton, yButton );

        var EXTRA_BUTTON_SCALE = 1.2;
        var referenceButtonSprite = cc.Sprite.create( selectedTopDirectionName );
        var buttonScale = cc.p( this._buttonContainer.getScaleX() * 150 * EXTRA_BUTTON_SCALE / referenceButtonSprite.getContentSize().width,
                                this._buttonContainer.getScaleY() * 84 * EXTRA_BUTTON_SCALE / referenceButtonSprite.getContentSize().height );

        this._directionButtons = [];

        var direction;
        for ( direction = 0; direction < MazeDirections.Count; direction++ )
        {
            // create a button for each MazeDirection and add it to the button container
            var selectedImage = cc.Sprite.create( direction === MazeDirections.Up ? selectedTopDirectionName : selectedBottomDirectionName );
            var unselectedImage = cc.Sprite.create( direction === MazeDirections.Up ? unselectedTopDirectionName : unselectedBottomDirectionName );
            var dirAngle = MazeConstants.directionAngles[ direction ];

            var button = new LumosityButton();

            button.setBackgroundImageForState( unselectedImage, LumosityButtonState.ButtonStateNormal );
            button.setBackgroundImageForState( selectedImage, LumosityButtonState.ButtonStateSelected );

            button.setPosition( dirButtonPos );

            // offset anchor points because buttons are uncentered trapezoids
            button.setAnchorPoint( cc.p( 0.5, -0.25 ) );

            button.setScaleX( buttonScale.x );
            button.setScaleY( buttonScale.y );

            // set up trapezoidal hit regions for direction buttons
            var hitRegion =
            [
                cc.p( button.getContentSize().width * 0.7,     button.getContentSize().height * 0.13 ),
                cc.p( button.getContentSize().width * 0.3,     button.getContentSize().height * 0.13 ),
                cc.p( button.getContentSize().width * -0.65,   button.getContentSize().height * 1.8 ),
                cc.p( button.getContentSize().width * 1.65,    button.getContentSize().height * 1.8 )
            ];
            button.setHitRegion( hitRegion );
            //button.__debugCreateHitRegionSprite();

            button.setRotation( -dirAngle * 180 / Math.PI + 90 );

            button.setPressedCallback( this._buttonPressed.bind( this, direction ) );
            button.setReleasedCallback( this._buttonReleased.bind( this, direction ) );
            button.setCancelledCallback( this._buttonReleased.bind( this, direction ) );
            button.setTarget( this );

            button.dragBeginsTouch = true;

            if( direction === MazeDirections.Up )
            {
                this._upButtonIndicator = cc.Sprite.create( "UI/tutorialCircle.png" );
                this._upButtonIndicator.setAnchorPoint( 0.5, 0.5 );
                this._upButtonIndicator.setPosition( button.getContentSize().width / 2, button.getContentSize().height * 0.53 );
                var upButtonScale = button.getContentSize().width / this._upButtonIndicator.getContentSize().width * settings.upButtonIndicatorScale;
                this._upButtonIndicator.setScaleX( upButtonScale );
                this._upButtonIndicator.setScaleY( upButtonScale );
                button.addChild( this._upButtonIndicator, 10 );
                this._upButtonIndicator.setOpacity( 0 );

                var labelPos = cc.p( button.getContentSize().width * 0.5, button.getContentSize().height * 1.25 );

                var labelFontSize = 30 / ( button.getScaleX() * this._buttonContainer.getScaleX() );
                this._upButtonLabel = LumosityLabel.create( localize( "upArrow" ), "fonts/MuseoSans-500", labelFontSize );
                this._upButtonLabel.setColor( cc.WHITE );
                this._upButtonLabel.setAnchorPoint( cc.p( 0.5, 0.5 ) );
                this._upButtonLabel.setHorizontalAlignment( cc.TEXT_ALIGNMENT_LEFT );
                this._upButtonLabel.setVerticalAlignment( cc.VERTICAL_TEXT_ALIGNMENT_CENTER );
                this._upButtonLabel.setPosition( labelPos );
                button.addChild( this._upButtonLabel );
                this._upButtonLabel.setOpacity( 0 );
            }

            var buttonZOrder = direction === MazeDirections.Up ? 10 : 0;
            this._buttonContainer.addChild( button, buttonZOrder );
            this.addButton( button );

            this._directionButtons.push( button );
        }

        if( this.levelCard )
        {
            this.levelCard.setLevelLabel( localize( "LevelCard_Level" ) );
        }

        this._scorePointsPos = LumosityUtils.copyPoint( this.scorePoints.getPosition() );

        // create dummy node for running the button highlight action
        this.highlightButtonsNode = cc.Node.create();
        this.addChild( this.highlightButtonsNode );
    },

    setMoveHint:function( direction, hintButtonTint, hintButtonOpacity )
    {
        // remove hint color on all buttons
        for( var d = 0; d < MazeDirections.Count; d++ )
        {
            var tintButton = ( direction >= 0 && d !== direction );
            var color = ( tintButton ? hintButtonTint : cc.WHITE );
            var opacity = ( tintButton ? hintButtonOpacity : 255 );

            this._directionButtons[ d ].backgroundImageForState( LumosityButtonState.ButtonStateNormal ).setColor( color );
            this._directionButtons[ d ].backgroundImageForState( LumosityButtonState.ButtonStateSelected ).setColor( color );

            this._directionButtons[ d ].backgroundImageForState( LumosityButtonState.ButtonStateNormal ).setOpacity( opacity );
            this._directionButtons[ d ].backgroundImageForState( LumosityButtonState.ButtonStateSelected ).setOpacity( opacity );
        }
    },

    /**
     * Pauses view behavior.
     */
    onPause:function()
    {
        this._super();
        if( this.mazeView )
        {
            this.mazeView.pauseSchedulerAndActions();
        }
    },

    /**
     * Resumes paused view behavior.
     */
    onResume:function()
    {
        this._super();
        if( this.mazeView )
        {
            this.mazeView.resumeSchedulerAndActions();
        }
    },

    /**
     * Called when maze is revealed before a trial.
     */
    mazeEnter:function()
    {
        // put the countdown in its waiting state
        this.readyGoLabelChild.setString( localize( "countdownReady" ) );
    },

    /**
     * Starts timer counting down and plays the associated sequence in the Flash file.
     */
    playReadyGo:function()
    {
        this.playAnimation( "readySetGo" );
    },

    setGoLabel:function()
    {
        this.readyGoLabelChild.setString( localize( "countdownGo" ) );
    },

    /**
     * Highlights buttons by scaling them.
     */
    highlightButtons:function( scale, time )
    {
        var scale1 = this._buttonContainer.getScaleX() * scale;
        var scale2 = this._buttonContainer.getScaleX();
        var sequence = cc.Sequence.create(
            cc.EaseIn.create( cc.ScaleTo.create( time, scale1, scale1 ), 2 ),
            cc.EaseOut.create( cc.ScaleTo.create( time, scale2, scale2 ), 2 )
        );
        this._buttonContainer.runAction( sequence );
    },

    /**
     * Hides or displays indicator elements for compass and up button.
     *
     * @param compass Whether to indicate the compass.
     * @param upButton Whether to indicate the up button.
     */
    showIndicators:function( compass, upButton, compassLabel, upButtonLabel )
    {
        var fadeTime = 0.3;
        var compassRevealDelay = 0.75;
        var upButtonRevealDelay = 1.5;

        this.mazeView._compassIndicator.stopAllActions();
        this.mazeView._compassLabel.stopAllActions();

        this._upButtonIndicator.stopAllActions();
        this._upButtonLabel.stopAllActions();

        if( compass )
        {
            this.mazeView._compassIndicator.runAction( cc.RepeatForever.create( cc.RotateBy.create( 30, 360 ) ) );

            this.mazeView._compassIndicator.runAction(
                cc.Sequence.create(
                    cc.DelayTime.create( compassRevealDelay ),
                    cc.FadeTo.create( fadeTime, 255 )
                )
            );
        }
        else
        {
            this.mazeView._compassIndicator.runAction( cc.FadeTo.create( fadeTime, 0 ) );
        }

        if( upButton )
        {
            this._upButtonIndicator.runAction( cc.RepeatForever.create( cc.RotateBy.create( 30, 360 ) ) );

            this._upButtonIndicator.runAction(
                cc.Sequence.create(
                    cc.DelayTime.create( upButtonRevealDelay ),
                    cc.FadeTo.create( fadeTime, 255 )
                )
            );
        }
        else
        {
            this._upButtonIndicator.runAction( cc.FadeTo.create( fadeTime, 0 ) );
        }

        if( compassLabel )
        {
            this.mazeView._compassLabel.runAction(
                cc.Sequence.create(
                    cc.DelayTime.create( compassRevealDelay ),
                    cc.FadeTo.create( fadeTime, 255 )
                )
            );
        }
        else
        {
            this.mazeView._compassLabel.runAction( cc.FadeTo.create( fadeTime, 0 ) );
        }

        if( upButtonLabel )
        {
            this._upButtonLabel.runAction(
                cc.Sequence.create(
                    cc.DelayTime.create( upButtonRevealDelay ),
                    cc.FadeTo.create( fadeTime, 255 )
                )
            );
        }
        else
        {
            this._upButtonLabel.runAction( cc.FadeTo.create( fadeTime, 0 ) );
        }
    },

    /**
     * Sets score value of adding score element.
     *
     * @param score Score to display.
     */
    setAddingScore:function( score )
    {
        this._curAddingScore = score;
        this.scorePoints.setString( "+" + score );
    },

    /**
     * Starts showing additive score element under hud.
     */
    startAddingScore:function()
    {
        this.scorePointContainer.setVisible( true );
        this.scorePoints.setVisible( true );
        this.scorePoints.setOpacity( 0 );
        this.scorePoints.runAction( cc.FadeIn.create( 0.15 ) );
        this.scorePoints.setPosition( this._scorePointsPos );
        this.setAddingScore( 0 );
    },

    /**
     * Updates additive score beneath hud.
     *
     * @param addingScore   Score to set in additive score element.
     */
    updateAddingScore:function( scoreDelta )
    {
        this.setAddingScore( this._curAddingScore + scoreDelta );
    },

    /**
     * Stops showing additive score element under hud.
     */
    endAddingScore:function( finalScore )
    {
        this.setAddingScore( finalScore );
        this.scorePoints.runAction( cc.FadeOut.create( 0.15 ) );
        this.scorePoints.runAction( cc.MoveBy.create( 0.15, cc.p( 0, 40 ) ) );
        if ( this._widgets )
        {
            var hudWidget = this._widgets[ GameWidgetHud.prototype.getType() ];
            if ( hudWidget )
            {
                hudWidget.setScoreHighlighted( true );
            }
        }
    },

    /**
     *  Sets tutorial prompt message.
     *
     * @param message Message string to set.
     */
    showTutorialPrompt:function( message )
    {
        this.tutorialText.setString( message );
    },

    /**
     *  Creates a new MazeView based upon the passed Maze object. If there was already a MazeView,
     *  will remove that from its parent first. The new MazeView is added to the maze container.
     */
     prepareNewMaze:function( mazeSizeScale, mazeOffsetScale, showCompass )
     {
        // build a new maze

        this.model.trial.maze.resizeMazeView( this.mazeView, this.mazeSize, mazeSizeScale );

        var mazePos = cc.p( this.mazeSize.width * 0.5, this.mazeSize.height * 0.56 );
        mazePos.y += this.mazeSize.height * mazeOffsetScale;
        this.mazeView.setPosition( mazePos );

        var mazeStart = this.model.trial.maze.start;
        var mazeGoal = this.model.trial.maze.goal;

        this.playerPenguinView.setGridPosition( mazeStart );
        this.computerPenguinView.setGridPosition( mazeStart );
        this.fishView.setGridPosition( mazeGoal );
        this.bumpView.setGridPosition( mazeStart );

        this.mazeView.showCompass( showCompass );
     },

    /**
     * Rotate the maze around its center to a destination angle.
     *
     * @param {number} degrees      The degrees to rotate to.
     * @param {boolean} instant     Whether or not to instantly rotate maze.
     */
     rotateMaze:function( degrees, instant, tutorialMode, finishedFunction, finishedTarget )
     {
         this.mazeView.rotate( degrees, instant, tutorialMode, finishedFunction, finishedTarget );
     },

    /**
     * Clears any running actions on nodes in view.
     */
    stopActions:function()
    {
        this.stopAllActions();

        if( this.mazeView )
        {
            this.mazeView.stopAllActions();
        }
        if( this.fishView )
        {
            this.fishView.stopAllActions();
        }
        if( this.bumpView )
        {
            this.bumpView.stopAllActions();
        }
        if( this.playerPenguinView )
        {
            this.playerPenguinView.stopActions();
        }
        if( this.computerPenguinView )
        {
            this.computerPenguinView.stopActions();
        }
    },

    /**
     * Called when view is beginning the pre-trial phase.
     */
    startPreTrial:function()
    {
        if( this._buttonLayer )
        {
            this._buttonLayer.clearRadioButtonPresses();
        }
        this.hidePathMarkers();

        var hudWidget = this._widgets[ GameWidgetHud.prototype.getType() ];
        hudWidget.setScoreHighlighted( false );
    },

    /**
     * Called when view is beginning the trial phase.
     */
    startTrial:function()
    {
        this.mazeView.enableFootsteps( true );
    },

    /**
     * Called when view is ending the trial phase.
     */
    endTrial:function()
    {
        this.mazeView.enableFootsteps( false );
    },

    /**
     *  Visually sets current level in level card.
     *
     * @param level     Zero-based level index.
     */
    setLevel:function( level )
    {
        if( this.levelCard )
        {
            this.levelCard.setLevelValue( "" + level );
        }
    },

    /**
     * Visually sets current trial and max trial numbers in hud.
     *
     * @param trialIndex    Zero-based trial index.
     * @param maxTrials     Total number of trials.
     */
    setTrial:function( trialIndex, maxTrials )
    {
        // find hud widget and update its trial labels
        if ( this._widgets )
        {
            var hudWidget = this._widgets[ GameWidgetHud.prototype.getType() ];
            if ( hudWidget )
            {
                hudWidget.setTrial( trialIndex, maxTrials );
            }
        }
    },

    /**
     * Visually sets current score.
     *
     * @param score     Current score number.
     */
    setScore:function( score )
    {
        // find hud widget and update its score label
        if ( this._widgets )
        {
            var hudWidget = this._widgets[ GameWidgetHud.prototype.getType() ];
            if ( hudWidget )
            {
                hudWidget.setScore( score );
            }
        }
    },

    /**
     * Reveals a path marker by index.
     *
     * @param {cc.Point} pathGridPos Grid position of marker in maze.
     * @param {number} index Zero-based index of marker within its array.
     * @param {boolean} isPlayerPath Whether the marker represents the player's path through maze.
     * @private
     */
    _revealPathMarker:function( pathGridPos, index, isPlayerPath )
    {
        if( pathGridPos )
        {
            var pathMarker = isPlayerPath ? this._playerPathMarkers[index] : this._computerPathMarkers[index];
            if( pathMarker )
            {
                // path marker's grid position is valid, so build reveal sequence
                pathMarker.setGridPosition( pathGridPos );
                pathMarker.setVisible( true );
                pathMarker.setOpacity( 255 );
                pathMarker.setScale( 1 );
                var sequence = cc.Sequence.create(
                    cc.ScaleTo.create( 0.1, 1.2, 1.2 ),
                    cc.ScaleTo.create( 0.3, 1, 1 ),
                    cc.DelayTime.create( 0.4 ),
                    cc.EaseOut.create( cc.FadeTo.create( 0.25, 0 ), 0.5 )
                    //cc.ScaleTo.create( 0.5, 0.01, 0.01 )
                );
                pathMarker.runAction( sequence );
            }
        }
    },

    /**
     * Reveals the appropriate path markers for a given path.
     *
     * @param {Array} playerPath In-order array of grid spaces that represents player's path through maze.
     * @param {Array} computerPath In-order array of grid spaces that represents computer's remaining path through maze.
     * @param {number} time Total time over which all path markers should be revealed.
     * @param {function} markerRevealedCallback Callback invoked when each marker is revealed.
     * @param {function} allMarkersRevealedCallback Callback invoked when all markers have been revealed.
     */
    revealPathMarkers:function( playerPath, computerPath, time, maxDelay, markerRevealedCallback, allMarkersRevealedCallback )
    {
        var markerCount = 0;
        if( playerPath )
        {
            markerCount += playerPath.length;
        }
        if( computerPath )
        {
            markerCount += computerPath.length;
        }

        var markerDelay = 0;
        if( markerCount > 0 )
        {
            markerDelay = Math.min( time / markerCount, maxDelay );
        }

        var actions = [];
        var i = 0;
        if( playerPath )
        {
            for( i = 0; i < playerPath.length; i++ )
            {
                this._addRevealMarkerActions( actions, i, markerDelay, playerPath[ i ], true, markerRevealedCallback );
            }
        }
        if( computerPath )
        {
            for( i = 0; i < computerPath.length; i++ )
            {
                this._addRevealMarkerActions( actions, i, markerDelay, computerPath[ i ], false, markerRevealedCallback );
            }
        }

        if( actions.length > 0 )
        {
            var revealSequence = LumosityUtils.createSequenceWithArray( actions );
            this.runAction( revealSequence );
        }

        if( allMarkersRevealedCallback )
        {
            var allRevealedSequence = cc.Sequence.create(
                cc.DelayTime.create( time ),
                cc.CallFunc.create( allMarkersRevealedCallback.bind( this ), this )
            );
            this.runAction( allRevealedSequence );
        }
    },

    /**
     * Creates the actions needed to reveal a single path marker, and adds it them to the reveal sequence.
     *
     * @param {Array} actions Array of actions that comprise the reveal sequence.
     * @param {number} index Index of path marker in its path marker array.
     * @param {number} markerDelay Delay between each marker reveal, in seconds.
     * @param {cc.Poin} pathGridPos Grid position of path point where marker should be revealed.
     * @param {boolean} isPlayerPath Whether the marker represents the player or the computer.
     * @param {function} markerRevealedCallback Callback invoked each time a path marker is reavealed.
     * @private
     */
    _addRevealMarkerActions:function( actions, index, markerDelay, pathGridPos, isPlayerPath, markerRevealedCallback )
    {
        actions.push( cc.CallFunc.create( this._revealPathMarker.bind( this, pathGridPos, index, isPlayerPath ), this ) );
        actions.push( cc.DelayTime.create( markerDelay ) );
        if( markerRevealedCallback )
        {
            actions.push( cc.CallFunc.create( markerRevealedCallback.bind( this, isPlayerPath, index ), this ) );
        }
    },

    /**
     * Makes all path markers invisible.
     */
    hidePathMarkers:function()
    {
        var i = 0;
        for ( i = 0; i < this._playerPathMarkers.length; i++ )
        {
            this._playerPathMarkers[i].setVisible( false );
        }
        for ( i = 0; i < this._computerPathMarkers.length; i++ )
        {
            this._computerPathMarkers[i].setVisible( false );
        }
    },

    /**
     * Overridden method that is called each time the game view creates a node from the Flash file.
     *
     * @param nodeName      Name of node's layer in Flash file.
     * @param nodeData      Node-specific data for creation.
     * @returns {*}         Whether the node was created.
     * @private
     */
    _layoutCreatingNode:function( nodeName, nodeData )
    {
        switch( nodeName )
        {
            case "settings":
                return false;

            case "readyGoLabel":
                var labelChild = LayoutLoader.createLabel( nodeData.textData );
                labelChild.setAnchorPoint( 0.5, 0.5 );
                labelChild.setPosition( nodeData.size.width / 2, nodeData.size.height / 2 );
                this.readyGoLabelChild = labelChild;

                var labelParent = cc.Node.create();
                labelParent.addChild( labelChild );
                return labelParent;

        }
        return this._super( nodeName, nodeData );
    },

    /**
     * Overridden method that is called when we should create our views.
     * @private
     */
     _createViews:function()
     {
         this._mazeContainer = this.maze_gridTemplate_8x8;
         this._buttonContainer = this.UI_animation;
     },

    /**
     * Delegate function used to react to pressed button input.
     *
     * @param direction     Maze direction of input button.
     * @private
     */
     _buttonPressed:function( direction )
     {
         if ( this.interactionResponder )
         {
             this.interactionResponder.setInputDirection( direction, true );
         }
     },

    /**
     * Delegate function used to react to released button input.
     *
     * @param direction     Maze direction of input button.
     * @private
     */
    _buttonReleased:function( direction )
    {
        if ( this.interactionResponder )
        {
            this.interactionResponder.setInputDirection( direction, false );
        }
    },

    _endTrialButtonPressed:function( win )
    {
        if( this.interactionResponder )
        {
            this.interactionResponder.endTrialButtonPressed( win );
        }
    },

     /**
     *  Displays a path solution in the maze view.
     *
     *  @name Debug#__displayMazeSolution
     */
    __displayMazeSolution:function( solution )
    {
        if ( this.mazeView )
        {
            this.mazeView.__displaySolution( solution );
        }
    }
});
