/*global cc*/
/*global require*/
/*global LumosityUtils*/

require( "src/Utility/LumosityUtils.js" );

function LevelsData()
{ 
    "use strict";
}

var LevelsData = new LevelsData();

LevelsData.pickupCharacters = "ABCDEFGHIJKLMNOPQ".split("");
LevelsData.dropoffCharacters = "abcdefghijklmnopq".split("");
LevelsData.vehicleCharacters = "1234".split("");

/**
 * Parses levels text to return levels array.
 * 
 * @param  {string} levelsText The levels text to parse.
 * @return {array}             An array of levels data.
 */
LevelsData.parse = function( levelsText )
{
    "use strict";

    var stages = null;
    var i;
    var j;
    var lines = null;
    var skewedVariance = 0;
    var shortest = 0;
    var radix = 10;

    stages = LevelsData.toParagraphs( levelsText );
    for( i = 0; i < stages.length; i++ )
    {
        for( j = 0; j < stages[ i ].length; j++ )
        {
            lines = stages[ i ][ j ].split( "\n" );
            skewedVariance = parseFloat( lines.pop(), radix );
            shortest = parseInt( lines.pop(), radix );
            stages[ i ][ j ] =
            {
                "levelText": lines.join( "\n" ),
                "shortest": shortest,
                "skewedVariance" : skewedVariance
            };
        }
    }

    return stages;
};

/**
 * Returns an array of paragraphs. A paragraph represents a level.
 * 
 * @param  {string} levelsText The levels text to parse.
 * @return {array}             The paragraph within the text block.
 */
LevelsData.toParagraphs = function( levelsText )
{
    "use strict";

    var paragraphs = null;
    var i;

    paragraphs = LevelsData.trim( levelsText );
    paragraphs = paragraphs.split( "\n\n\n" );

    for( i = 0; i < paragraphs.length; i++ )
    {
        paragraphs[ i ] = paragraphs[ i ].split( "\n\n" );
    }

    return paragraphs;
};

/**
 * Returns an array of rows.
 * 
 * @param  {string} levelText The level text to split into rows.
 * @return {array}            The array of rows.
 */
LevelsData.toRows = function( levelText )
{
    "use strict";

    var rows = null;
    rows = levelText.split("\n");
    return rows;
};

/**
 * Returns a table. A table consists of split rows and columns.
 * 
 * @param  {string} levelText The level text to split into a table.
 * @return {array}            The table.
 */
LevelsData.toTable = function( levelText )
{
    "use strict";

    var table = null;
    var r;

    table = LevelsData.toRows( levelText );
    for( r = 0; r < table.length; r++ )
    {
        table[ r ] = table[ r ].split("");
    }

    return table;
};

/**
 * Turns a table back into text.
 * 
 * @param  {array} table The table.
 * @return {string}      The text.
 */
LevelsData.toText = function( table )
{
    "use strict";

    var newline = "\n";
    var clone = null;
    var r;

    clone = table.slice();
    for( r = 0; r < clone.length; r++ )
    {
        clone[ r ] = clone[ r ].join("");
    }

    return clone.join( newline );
};

/**
 * Returns the row count within the level text.
 * 
 * @param  {string} levelText The level text.
 * @return {int}              The row count.
 */
LevelsData.rowCount = function( levelText )
{
    "use strict";

    var rows = null;
    rows = LevelsData.toRows( levelText );
    return rows.length;
};

/**
 * Returns the column count within the level text.
 * 
 * @param  {string} levelText The level text.
 * @return {int}              The column count.
 */
LevelsData.columnCount = function( levelText )
{
    "use strict";

    var rows = null;
    rows = LevelsData.toRows( levelText );
    return rows[0].length;
};

/**
 * Applies variances to the level text.
 * Horizontal/Vertical flipping, Shuffling of interest.
 * 
 * @param  {[type]} levelText The level text to apply the variances to.
 * @return {[type]}           The variance applied level text.
 */
LevelsData.vary = function( levelText )
{
    "use strict";

    var varyLevelText = "";

    varyLevelText = LevelsData.trim( levelText );
    if( Math.random() < 0.5 )
    {
        varyLevelText = LevelsData.flipHorizontal( varyLevelText );
    }
    if( Math.random() < 0.5 )
    {
        varyLevelText = LevelsData.flipVertical( varyLevelText );
    }

    varyLevelText = LevelsData.shuffleInterests( varyLevelText );
    return varyLevelText;
};

/**
 * Flips the level text horizontally.
 * 
 * @param  {string} levelText The level text.
 * @return {string}           Horizontally flipped level text.
 */
LevelsData.flipHorizontal = function( levelText )
{
    "use strict";

    var table = null;
    var r;
    var c;
    var right = 0;
    var middle = 0;
    var temp = "";
    var c1 = 0;
    var flippedLevelText = "";

    table = LevelsData.toTable( levelText );
    for( r = 0; r < table.length; r++ )
    {
        right = table[ r ].length -1;
        middle = right >> 1;
        for( c = 0; c <= middle; c++ )
        {
            temp = table[ r ][ c ];
            c1 = right - c;
            table[ r ][ c ] = table[ r ][ c1 ];
            table[ r ][ c1 ] = temp;
        }
    }

    flippedLevelText = LevelsData.toText( table );
    return flippedLevelText;
};

/**
 * Flips the level text vertically.
 * 
 * @param  {string} levelText The level text.
 * @return {string}           Vertically flipped level text.
 */
LevelsData.flipVertical = function( levelText )
{
    "use strict";

    var table = null;
    var r;
    var c;
    var bottom = 0;
    var middle = 0;
    var temp = "";
    var r1 = 0;
    var flippedLevelText = "";

    table = LevelsData.toTable( levelText );
    bottom = table.length - 1;
    middle = bottom >> 1;
    for( r = 0; r <= middle; r++ )
    {
        r1 = bottom - r;
        for( c = 0; c < table[ r ].length ; c++ )
        {
            temp = table[ r ][ c ];
            table[ r ][ c ] = table[ r1 ][ c ];
            table[ r1 ][ c ] = temp;
        }
    }

    flippedLevelText = LevelsData.toText( table );
    return flippedLevelText;
};
/**
 * Shuffles the interests in the level text.
 * 
 * @param  {string} levelText The level text.
 * @return {string}           The shuffled level text.
 */
LevelsData.shuffleInterests = function( levelText )
{
    "use strict";

    var characters = null;
    var c;
    var index = 0;
    var maxIndex = 0;
    var count = 0;
    var shuffledIndexes = null;
    var p = 0;
    var d = 0;
    var shuffledLevelText = null;

    characters = levelText.split("");
    for( c = 0; c < characters.length; c++ )
    {
        index = LevelsData.pickupCharacters.indexOf( characters[ c ] );
        if( index >= 0 )
        {
            count++;
            if( maxIndex < index )
            {
                maxIndex = index;
            }
        }
    }

    shuffledIndexes = [];
    for( index = 0; index < count; index++ )
    {
        shuffledIndexes.push( index );
    }

    LumosityUtils.shuffle( shuffledIndexes );
    for( c = 0; c < characters.length; c++ )
    {
        p = LevelsData.pickupCharacters.indexOf( characters[ c ] );
        d = LevelsData.dropoffCharacters.indexOf( characters[ c ] );
        if( p >= 0 )
        {
            characters[ c ] = LevelsData.pickupCharacters[ shuffledIndexes[ p ] ];
        }
        else if( d >= 0 )
        {
            characters[ c ] = LevelsData.dropoffCharacters[ shuffledIndexes[ d ] ];
        }
    }

    shuffledLevelText = characters.join("");
    return shuffledLevelText;
};

/**
 * Trims text.
 * 
 * @param  {string} text The text to trim.
 * @return {array}       The trimmed text.
 */
LevelsData.trim = function( text )
{
    "use strict";

    var lines = [];
    var columns = [];
    var left;
    var i;

    lines = LevelsData.normalize( text );
    lines = lines.split( "\n" );
    while( "" === lines[ 0 ] )
    {
        lines.shift();
    }

    columns = lines[ 0 ].split("");
    for( left = 0; left < columns.length; left++ )
    {
        if( " " !== columns[ left ] )
        {
            break;
        }
    }

    for( i = 0; i < lines.length; i++ )
    {
        lines[ i ] = lines[ i ].substr( left );
    }

    while( "" === lines[ lines.length - 1 ] )
    {
        lines.pop();
    }

    return lines.join( "\n" );
};

/**
 * Normalize Windows and Mac line-endings and trim empty lines and indentation.
 * 
 * @param  {string} text The text to be normalized.
 * @return {string}      The normalized text.
 */
LevelsData.normalize = function( text )
{
    "use strict";

    var normalizedText = null;

    normalizedText = text.replace( /\r\n/g, "\n" ).replace( /\r/g, "\n" );
    return normalizedText;
};

