/*global cc*/
/*global require*/
/*global localize*/
/*global GameView*/
/*global GameWidgetLevelCard*/
/*global GameWidgetHud*/
/*global GameWidgetPopup*/
/*global StageSelect*/
/*global Dashboard*/
/*global LayoutLoader*/
/*global LayoutNodeFlags*/
/*global LumosityButtonState*/
/*global LumosityButton*/
/*global LumosityUtils*/

require( "src/Game/Base/GameView.js" );
require( "src/Game/Widget/GameWidgetLevelCard.js" );
require( "src/Game/Widget/GameWidgetHud.js" );
require( "src/Game/Widget/GameWidgetPopup.js" );
require( "src/Game/StageSelect.js" );
require( "src/Game/Dashboard.js" );
require( "src/Animation/LayoutLoader.js" );
require( "src/View/LumosityButton.js" );

var PetDetectiveView;
( function()
{
    "use strict";
    
    PetDetectiveView = GameView.extend(
    {
        clippingMask: null,
        clippingMaskSquare: null,
        containerNode: null,

        effects: 
        {
            "carDrive.mp3" : "sounds/carDrive.mp3",
            "carIdleLoop.mp3" : "sounds/carIdleLoop.mp3",
            "carPickupPet.mp3" : "sounds/carPickupPet.mp3",
            "complete.mp3" : "sounds/complete.mp3",
            "deliverPet.mp3" : "sounds/deliverPet.mp3",
            "outOfGas.mp3" : "sounds/outOfGas.mp3",
            "screenWipe.mp3" : "sounds/screenWipe.mp3",
            "select.mp3" : "sounds/select.mp3"
        },

        /** @lends PetDetectiveView.prototype */
        
        /**
        * The Pet Detective view.
        * @constructs
        * @param {cc.Size} screenSize The size of the screen that this view will be inserted into.
        * @param {Object} [settings] The settings that we can pull view-based data from.
        * @param {PetDetectiveModel} [model] The GameModel that we are linked to.
        * @param {Object} [interactionResponder] The object that we should send interaction events to. We also use this is a default animation event listener.
        * @param {boolean} [preload=true] If true, will preload all flipbooks and other assets upon creation.
        */
        ctor:function( screenSize, settings, model, interactionResponder, preload )
        {
            // Load sprite sheet first to account for any sprites created within the super.
            // cc.SpriteFrameCache.getInstance().addSpriteFrames( "interests.plist", "interests.png" );

            this.preloadSoundEffects();

            this._super( screenSize, settings, model, interactionResponder, preload );
        },

        preloadSoundEffects :function()
        {
            LumosityUtils.forEachIn( this.effects, function( name, value )
            {
                cc.AudioEngine.getInstance().preloadEffect( value );
            } );
        },

        _createViews: function()
        {
            var contentSize = cc.SizeZero();
            var menuPromptTextContentSize = cc.SizeZero();
            var menuUnlockTextContentSize = cc.SizeZero();

            this._super();

            // Setting ClippingNode to be like _gameStage.
            // This is because all game objects will be added to clipping node.
            contentSize = this._gameStage.getContentSize();

            this.containerNode = cc.Node.create();
            this.containerNode.setAnchorPoint( cc.p( 0.5, 0.5 ) );
            this.containerNode.setContentSize( contentSize );
            this.containerNode.setPosition( cc.p( contentSize.width >> 1, contentSize.height >> 1 ) );

            // Mask in Masks to achieve a circle mask inside a square mask because
            // it is too difficult to control the alpha threshold at run time to create
            // this desired affect.
            this.clippingNodeSquare.setContentSize( contentSize );
            this.clippingMaskSquare.setPosition( cc.p( contentSize.width >> 1, contentSize.height >> 1 ) );
            this.clippingNodeSquare.setAlphaThreshold( 1 );

            this.clippingNode.setContentSize( contentSize );
            this.clippingMask.setPosition( cc.p( contentSize.width >> 1, contentSize.height >> 1 ) );
            this.clippingNode.setAlphaThreshold( 0 );

            this.clippingNodeSquare.addChild( this.containerNode );

            this.centeredContent.removeFromParent();
            this.containerNode.addChild( this.centeredContent );

            // Remove the stage select from it's parent and add it to the clipping node.
            this.menu.removeFromParent();
            this.containerNode.addChild( this.menu );

            //this.bottomUI.removeFromParent();
            //this.clippingNode.addChild( this.bottomUI );

            this.menuText.removeFromParent();
            this.containerNode.addChild( this.menuText );

            //this.trialRestartNode.removeFromParent();
            //this.clippingNode.addChild( this.trialRestartNode );

            menuPromptTextContentSize = this.menuPromptText.getContentSize();
            menuUnlockTextContentSize = this.menuUnlockText.getContentSize();

            // It seems you must setDeimensions again ( LayoutLoader previously set it ) if you switch parent nodes.
            this.menuPromptText.setDimensions( cc.size( menuPromptTextContentSize.width, 0 ) );
            this.menuUnlockText.setDimensions( cc.size( menuUnlockTextContentSize.width, 0 ) );

            //LumosityUtils.printNodeHierarchy( this.clippingNodeSquare );
        },

        /**
        * Pauses all current running animations.
        */
        pauseAnimations: function()
        {
            var buttons = null;

            this.clippingMask.pauseSchedulerAndActions();
            this.dashboard.pauseSchedulerAndActions();
            this.stageSelect.pauseSchedulerAndActions();
            
            buttons = this.stageSelect.buttons;
            buttons.forEach( function( button )
            {
                button.pauseSchedulerAndActions();
            } );
        },

         /**
        * Resumes all previous running animations.
        */
        resumeAnimations: function()
        {
            var buttons = null;

            this.clippingMask.resumeSchedulerAndActions();
            this.dashboard.resumeSchedulerAndActions();
            this.stageSelect.resumeSchedulerAndActions();

            buttons = this.stageSelect.buttons;
            buttons.forEach( function( button )
            {
                button.resumeSchedulerAndActions();
            } );
        },

        setLevel:function( level )
        {
            var levelCard = null;

            if( !this._widgets )
            {
                return;
            }

            levelCard = this._widgets[ GameWidgetLevelCard.prototype.getType() ];
            if( levelCard )
            {
                levelCard.setLevelValue( level );
            }
        },

        setTrial:function( trial, maxTrials )
        {
            var hudWidget = null;

            if( !this._widgets )
            {
                return;
            }

            hudWidget = this._widgets[ GameWidgetHud.prototype.getType() ];
            if( hudWidget )
            {
                hudWidget.setTrial( trial, maxTrials );
            }
        },

        setLevelCardText: function( text )
        {
            var levelCard = null;

            if( !this._widgets )
            {
                return;
            }

            levelCard = this._widgets[ GameWidgetLevelCard.prototype.getType() ];
            if( levelCard )
            {
                levelCard.setLevelLabel( text );
            }
        },

        setTutorialPromptText: function( text )
        {
            this.tutorialPrompt.setString( text );
        },

        enableClippingNodeCircle: function()
        {
            this.containerNode.retain();
            this.containerNode.removeFromParent( false );
            this.clippingNode.addChild( this.containerNode );
            this.containerNode.release();
        },

        enableClippingNodeSquare: function()
        {
            this.containerNode.retain();
            this.containerNode.removeFromParent( false );
            this.clippingNodeSquare.addChild( this.containerNode );
            this.containerNode.release();
        },

        _createTrialRestartButton: function( bgUpImageName, bgDownImageName, bgDisabledImageName, designSize, target, callback )
        {
            var button = null;
            var bgUpImage = null;
            var bgDownImage = null;
            var bgDisabledImage = null;

            bgUpImage = cc.Sprite.create( bgUpImageName );
            bgDownImage = cc.Sprite.create( bgDownImageName );
            bgDisabledImage = cc.Sprite.create( bgDisabledImageName );

            button = this._createButton( bgUpImage,
                                         bgDownImage,
                                         bgDisabledImage );

            // Buttons created within the layout loader do not need to set their contentSize.
            //button.setContentSize( designSize, false );
            button.setTarget( target );
            button.setReleasedCallback( callback );

            return button;
        },

        _createButton: function( bgUpImage, bgDownImage, bgDisabledImage )
        {   
            var button = null;

            button = new LumosityButton();
            button.setBackgroundImageForState( bgUpImage, LumosityButtonState.ButtonStateNormal );
            button.setBackgroundImageForState( bgDownImage, LumosityButtonState.ButtonStateSelected );
            button.setBackgroundImageForState( bgDisabledImage, LumosityButtonState.ButtonStateDisabled );

            return button;
        },

        enableTrialRestartButton: function( isEnabled )
        {
            this.trialRestartButton.setEnabled( isEnabled );
        },

        layout_creatingNode: function( nodeName, nodeData )
        {
            var result;
            var contentSize = cc.SizeZero();
            var stageSelect = null;
            var dashboard = null;
            var trialRestartButton = null;

            if( nodeName === "clippingNode" )
            {
                // Create the clipping mask
                this.clippingMask = cc.Sprite.create( "mask.png" );
                this.clippingMask.nodeName = "ClippingMask";

                result = cc.ClippingNode.create( this.clippingMask );
            }
            else if( nodeName === "clippingNodeSquare" )
            {
                // Create the clipping mask
                this.clippingMaskSquare = cc.Sprite.create( "mask.png" );
                this.clippingMaskSquare.nodeName = "ClippingMaskSquare";

                result = cc.ClippingNode.create( this.clippingMaskSquare );
            }
            else if( nodeName === "stageSelect" )
            {
                contentSize = nodeData[ LayoutLoader.sizeKey ];
                stageSelect = new StageSelect( this.interactionResponder, contentSize, this.settings.stageSelectSettings, this.model.currentStage, this.settings.stageCount );
                result =  stageSelect;
            }
            else if( nodeName === "dashboard" )
            {
                contentSize = nodeData[ LayoutLoader.sizeKey ];
                dashboard = new Dashboard( this.interactionResponder, contentSize );
                result =  dashboard;
            }
            else if( nodeName === "trialRestartButton" )
            {
                contentSize = nodeData[ LayoutLoader.sizeKey ];
                trialRestartButton = this._createTrialRestartButton( "trialRestartButtonUp.png", "trialRestartButtonDown.png", "trialRestartButtonUp.png", contentSize, this.interactionResponder, this.interactionResponder.trialRestartButtonPressed );
                this.addButton( trialRestartButton );
                result = trialRestartButton;
            }
            else
            {
                result = this._super( nodeName, nodeData );
            }

            return result;
        },

        layout_settingNodeFlags: function( nodeName, nodeData, node )
        {
            var nodeFlags = this._super( nodeName, nodeData, node );
            if( nodeName === "centeredContent" )
            {
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }
            // else if( nodeName === "clippingNode" )
            // {
            //     nodeFlags &= ~( LayoutNodeFlags.RootScale );
            // }
            else if( nodeName === "menuText" )
            {
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }
            else if( nodeName === "menu" )
            {
                nodeFlags &= ~( LayoutNodeFlags.RootScale );
            }
            // else if( nodeName === "bottomUI" )
            // {
            //     nodeFlags &= ~( LayoutNodeFlags.RootScale );
            // }
            // else if( nodeName === "trialRestartNode" )
            // {
            //     nodeFlags &= ~( LayoutNodeFlags.RootScale );
            // }

            return nodeFlags;
        },

        _findTouchedButtons: function( touch )
        {
            var i;
            var location = touch.getLocation();
            var button = null;
            var touchedButtons = [];

            for( i = 0; i < this._buttons.length; i++ )
            {
                button = this._buttons[ i ];
                // Similar to the LumosityButtonLayer, but don't ignore buttons who are disabled.
                // Would make change in LumosityButtonLayer, but don't want to break all the games if
                // this change doesn't go well.
                if( !button.isHidden && button.containsWorldPoint( location ) )
                {
                    this._addButtonSortedByZ( touchedButtons, button );
                }
            }

            return touchedButtons;
        }
    } );
}() );
