/*global cc*/
/*global require*/
/*global lumosity*/
/*global Street*/
/*global PracticeStreetSettings*/
/*global PracticeStreetView*/
/*global LevelsData*/
/*global LumosityUtils*/

require( "src/Game/Street.js" );
require( "src/Game/PracticeStreetSettings.js" );
require( "src/Game/PracticeStreetView.js" );
require( "src/Game/LevelsData.js" );
require( "src/Utility/LumosityUtils.js" );

var PracticeStreet;
( function()
{
    "use strict";    
    PracticeStreet = Street.extend(
    { 
        practiceSettings: null,

        /** @lends PracticeStreet.prototype */
        
        /**
        *  Street is the main class for a Street!
        *
        *  @constructs
        *  @param {array} Settings that define the Street.
        */
        ctor: function( settings, practiceSettings )
        {
            this._createPracticeSettings( practiceSettings );
            this._super( settings );
        },

         _cleanup: function()
        {
            this._super();
            this.view.removeTutorialCircle();
        },

        _startVehicle: function()
        {
            var practiceTrialInfo = null;

            this.vehicle.setOpacity( 255 );
            this.vehicle.startEngineRunningAnimation();
            this.vehicle.startEngineRunningAudio();

            practiceTrialInfo = this.getCurrentPracticeTrialInfo();
            if( practiceTrialInfo.showFuel )
            {
                this.vehicle.startFuelShowAnimation();
            }
        },

        /**
         * Reads in the levels file data and creates an array of stages.
         *
         * @return {Array}     The stages.
         */
        _createStages: function()
        {
            var fileData = "";
            var stages = null;

            fileData = lumosity.io_read( "pet_detective_practice_levels.txt" );
            stages = LevelsData.parse( fileData );
            return stages;
        },

        _initNextTrial: function( stage, trial, trialCount )
        { 
            var practiceTrialInfo = null;
            var tutorialCircle = null;

            this._super( stage, trial, trialCount );

            practiceTrialInfo = this.getCurrentPracticeTrialInfo();
            if( practiceTrialInfo.showTutorialCircle )
            {
                // Create, Add, and hide the tutorial circle.
                tutorialCircle = this._createTutorialCircle( this.practiceSettings.tutorialCircleImageName, cc.PointZero() );
                this.view.addTutorialCircle( tutorialCircle );
                this.view.hideTutorialCircle();
            }
        },

        _initMetaData: function()
        {
           
        },

        _setTrialCompleteMetadata: function()
        {

        },

        _setRoundCompleteMetadata: function()
        {

        },

        _initLevel: function( stage, trial /*, trialCount */ )
        {
            var stages = null;
            var levels = null;
            var level = null;

            stages = this.model.stages;
            levels = stages[ stage - 1 ];
            level = levels[ trial - 1 ];

            // No variations for practice mode.
            this.model.level = level;
            this.model.level.varyLevelText = this.model.level.levelText;
            this.model.level.endLevelText = this.model.level.varyLevelText;
        },

        onStartGameplay: function()
        {
            var practiceTrialInfo = null;
            var character = 0;
            var index = 0;
            var pickup = null;
            var position = cc.PointZero();

            this._super();

            practiceTrialInfo = this.getCurrentPracticeTrialInfo();
            if( practiceTrialInfo.showTutorialCircle )
            {
                character = practiceTrialInfo.tutorialCirclePickupCharacter;
                index = LevelsData.pickupCharacters.indexOf( character );
                pickup = this.pickups[ index ];
            
                position = this.view.calculateTutorialCirclePosition( pickup );
                this.view.setTutorialCirclePosition( position );

                this.view.playTutorialCircleRevealAnimation();
            }
        },

        /**
         * Process the data for receiving a pickup.
         * 
         * @param  {int} address The tile address.
         */
        _processPickup: function( address )
        {
            var practiceTrialInfo = null;
            var character = 0;
            var index = 0;
            var dropoff = null;
            var didPickup = false;
            var position = cc.PointZero();

            didPickup = this._super( address );
        
            if( didPickup )
            {
                practiceTrialInfo = this.getCurrentPracticeTrialInfo();
                if( practiceTrialInfo.showTutorialCircle )
                {
                    character = practiceTrialInfo.tutorialCircleDropoffCharacter;
                    index = LevelsData.dropoffCharacters.indexOf( character );
                    dropoff = this.dropoffs[ index ];

                    // Hide and play reveal animation for the dropoff tutorial circle.
                    this.view.hideTutorialCircle();

                    position = this.view.calculateTutorialCirclePosition( dropoff );
                    this.view.setTutorialCirclePosition( position );
                   
                    this.view.playTutorialCircleRevealAnimation();
                }
            }

            return didPickup;
        },

         /**
         * Process the data for a dropoff.
         * 
         * @param  {int} address The tile address.
         */
        _processDropoff: function( address )
        {
            var practiceTrialInfo = null;
            var didDropoff = false;

            didDropoff = this._super( address );

            if( didDropoff )
            {
                practiceTrialInfo = this.getCurrentPracticeTrialInfo();
                if( practiceTrialInfo.showTutorialCircle )
                {
                    this.view.playTutorialCircleConcealAnimation();
                }
            }

            return didDropoff;
        },

         _createVehicleFuelSucceedAnimation: function( callback, target, data )
        {
            //var startFuelHideAnimFunc = null;
            var callbackFunc = null;
            var callbackDelayAction = null;
            var sequence = null;
            var practiceTrialInfo = null;

            practiceTrialInfo = this.getCurrentPracticeTrialInfo();
            // Sara doesn't want the animation to hide upon success.
            //startFuelHideAnimFunc = cc.CallFunc.create( this.vehicle.startFuelHideAnimation, this.vehicle );
            callbackFunc = cc.CallFunc.create( callback, target, data );
            callbackDelayAction = cc.DelayTime.create( this.vehicle.fuelHideAnimationDuration() );

            if( practiceTrialInfo.showFuel )
            {
                sequence = cc.Sequence.create( /*startFuelHideAnimFunc,*/ callbackDelayAction, callbackFunc );
            }
            else
            {
                sequence = cc.Sequence.create( callbackDelayAction, callbackFunc );
            }

            return sequence; 
        },

        getCurrentPracticeTrialInfo: function()
        {
            var practiceTrialInfo = this.practiceSettings.getPracticeTrialInfo( this.model.trial );
            return practiceTrialInfo;
        },

        _getBoundDesignSize: function()
        {
            return this.practiceSettings.boundDesignSize;
        },

        _createTutorialCircle: function( imageName, position )
        {
            var tutorialCircle = null;

            tutorialCircle = cc.Sprite.create( imageName );
            tutorialCircle.setPosition( position );
            LumosityUtils.scaleForContentSize( tutorialCircle, this.practiceSettings.tutorialCircleContentSize, true );
            tutorialCircle.nodeName = "TutorialCircle";

            return tutorialCircle;
        },

        _createPracticeSettings: function( practiceSettings )
        {
            this.practiceSettings = new PracticeStreetSettings( practiceSettings );
        },

         _createView: function()
        {
            this.view = new PracticeStreetView();
            this.view.nodeName = "PracticeStreet";
            this.view.setAnchorPoint( 0.5, 0.5 );
        }
    } );
}() );
