/*global cc*/
/*global require*/
/*global lumosity*/
/*global StreetSettings*/
/*global StreetModel*/
/*global StreetView*/
/*global LevelsData*/
/*global TiledMap*/
/*global Vehicle*/
/*global VehicleDirection*/
/*global PathPlannerBestFirst*/
/*global LumosityButton*/
/*global LumosityButtonState*/
/*global LumosityUtils*/
/*global metadataTrialKeys*/

require( "src/Game/StreetSettings.js" );
require( "src/Game/StreetModel.js" );
require( "src/Game/StreetView.js" );
require( "src/Game/LevelsData.js" );
require( "src/Tile/TiledMap.js" );
require( "src/Game/Vehicle.js" );
require( "src/AI/PathPlannerBestFirst.js" );
require( "src/View/LumosityButton.js" );
require( "src/Utility/LumosityUtils.js" );

var Street;
( function()
{
    "use strict";    
    Street = cc.Class.extend(
    {
        pathPlanner: null,
        vehicle: null,
        pickups: null,
        dropoffs: null,

        settings: null,
        model: null,
        view: null,

        _target: null,
        _trialCompleteCallback: null,
        _trialCompleteAnimationFinishCallback: null,
        _maxVehicleCapacityButtonPressedCallback: null,

        _pickupCallback: null,
        _dropoffCallback: null,
        _beginMetadataTrialCallback: null,
        _endMetadataTrialCallback: null,
        
        /** @lends Street.prototype */
        
        /**
        *  Street is the main class for a Street!
        *
        *  @constructs
        *  @param {array} Settings that define the Street.
        */
        ctor: function( settings )
        {
            this._createSettings( settings );
            this._createModel();
            this._createView();

            this._init();

            cc.log( "Street::ctor" );
        },

        _init: function()
        { 
            // Create the path planner.
            this.pathPlanner = new PathPlannerBestFirst( this );

            // Create all the level info.
            this.model.stages = this._createStages();

            this.pickups = [];
            this.dropoffs = [];
        },

        _initNextTrial: function( stage, trial, trialCount )
        {   
            this._initLevel( stage, trial, trialCount );

            // Create the grid info.
            this.model.grid = this._createGrid();

            // Create and display the tiledMap.
            this.view.tiledMap = this._createTiledMap();
            this.view.tiledMap.setRotation( 90 );

            // Scale the tiledMap if it is too large.
            var newScale = cc.p( 1, 1 );
            var areaPercent = this._calculateTiledMapToScreenAreaPercent( true );

            cc.log( "areaPercent: " + JSON.stringify( areaPercent ) );
            if( areaPercent.x >= 100 || areaPercent.y >= 100 )
            {
                newScale = 1.0 / ( Math.max( areaPercent.x, areaPercent.y ) / 100.0 );
                this.view.tiledMap.setScale( newScale );
            }
            this.view.addChild( this.view.tiledMap );
        },

        _initMetadata: function()
        {
            this.model.roundData[ this.model.trial - 1 ] = {};
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_diagram, this._formatLevelTextForMetadata( this.model.level.levelText ) );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_vary_diagram, this._formatLevelTextForMetadata( this.model.level.varyLevelText ) );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_fuel, this.model.level.shortest );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_shortest, this.model.level.shortest );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_skewedVariance, this.model.level.skewedVariance );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_targets, this.pickups.length );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_width, this.model.rowCount );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_height, this.model.columnCount );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_timeOffset, this._target.getMetadataComponent().getRunningTime( true ) );
        },

        _initLevel: function( stage, trial, trialCount )
        {
            this.model.level = this._createRandomLevel( stage, trial, trialCount );
            this.model.level.varyLevelText = LevelsData.vary( this.model.level.levelText );
            this.model.level.endLevelText = this.model.level.varyLevelText;
        },

        setRoundDataForKey: function( key, info )
        {
            this.model.roundData[ this.model.trial - 1 ][ key ] = info;
        },

        setTrialDataForKey: function( key, info )
        {
            this.model.trialData[ key ] = info;
        },

        _createViews: function()
        {
            var self = this;
            var interests = null;

            // Add all the interests to the view.
            this.pickups = this._createPickups();
            this.dropoffs = this._createDropoffs();
            interests = this.pickups.concat( this.dropoffs );
            interests.forEach( function( interest )
            {
                self.view.addChild( interest );
            } );

            // Order matters on enabling first then showing.
            // This is because enabling or disabling images
            // will add a child to the node which will
            // not be affected by previous opacity settings.
            // 
            // Disable the interests.
            this._enableInterests( false );

            // Hide the interests.
            this._showInterests( false );

            // Add the vehicle to the view.
            this.vehicle = this._createVehicle();
            self.view.addChild( this.vehicle.view );

            // Hide the vehicle.
            this.vehicle.setOpacity( 0 );

            cc.log( "Street::_createViews" );
        },

        onStartGameplay: function()
        {
            // Show the interests.
            this._showInterests( true );

            // Enable the pickups and disable dropoffs.
            this._enablePickups( true );
            this._enableDropoffs( false );

            this._startVehicle();

            // Start new Metadata trial.
            this.beginMetadataTrial();

            cc.log( "Street::onStartGameplay" );
        },

        beginMetadataTrial: function()
        {
            if( !this._beginMetadataTrialCallback )
            {
                return;
            }

            this._setTrialBeginMetadata();
            this._beginMetadataTrialCallback.call( this._target, this );
        },

        endMetadataTrial: function( isReset )
        {
            if( !this._endMetadataTrialCallback )
            {
                return;
            }

            this._setTrialCompleteMetadata( isReset );
            this._endMetadataTrialCallback.call( this._target, this );
        },

        _startVehicle: function()
        {
            this.vehicle.setOpacity( 255 );
            this.vehicle.startEngineRunningAnimation();
            this.vehicle.startEngineRunningAudio();
            this.vehicle.startFuelShowAnimation();
        },

        _stopVehicle: function()
        {
            // Stop Audio
            this.vehicle.stopEngineRunningAudio();
            this.vehicle.stopDriveAudio();
            this.vehicle.onStopAnimations();
            this.model.areInteractionsBlocked = false;
        },

        onUpdate: function( delta )
        {
            this.model.timeToCompleteRound += delta;
        },

        onPause: function()
        {
            this.view.pauseAnimations();
            if( this.vehicle )
            {
                this.vehicle.onPause();

                if( this.vehicle.isPlayingAnimation() )
                {
                    this.vehicle.stopEngineRunningAudio();
                }
            }
        },

        onResume: function()
        {
            this.view.resumeAnimations();
            if( this.vehicle )
            {
                this.vehicle.onResume();

                if( this.vehicle.isPlayingAnimation() )
                {
                    this.vehicle.startEngineRunningAudio();
                }
            }
        },

        onRestart: function()
        {
            this._cleanup();
        },

        onResetTrial: function()
        {
            this.model.resetCount = this.model.resetCount + 1;

            this._setup();
            this._resetVehicle();
            this._restartInterests();

            this.endMetadataTrial( true );

            // Reset Metadata
            this.model.steps = 0;
            this.model.pickupCount = 0;
            this.model.dropoffCount = 0;
        },

        _resetVehicle: function()
        {
            var position = cc.PointZero();
            var rect = cc.RectZero();
            var direction = VehicleDirection.Right;

            if( !this.vehicle )
            {
                return;
            }

            // Stop Vehicle
            this._stopVehicle();

            // Reset position.
            rect = this.view.tiledMap.tileRectFromTileCoords( cc.p( this.model.vehicleColumn, this.model.vehicleRow ) );
            position = cc.p( rect.x + ( rect.width >> 1 ), rect.y + ( rect.height >> 1 ) );
            position = this._convertTileToViewSpace( position );
            this.vehicle.view.setPosition( position );

            // Reset direction.
            direction = this._calculateVehicleTileFacing();
            this.vehicle.setDirection( direction );

            // Reset fuel.
            this.vehicle.setFuel( this.model.level.shortest );
            this.vehicle.setFuelText( this.vehicle.getFuel().toString() + "/" + this.vehicle.getFuelCount().toString() );

            // Reset passengers.
            this.vehicle.initPassengers( this.model.vehicleCapacity );

            // Start all vehicles start animations.
            this._startVehicle();
        },

        _restartInterests: function()
        {
            // Hide all the interests highlighted image.
            this._hideAllInterestsHighlightedImage();

            // Show the interests.
            this._showInterests( true );

            // Enable the pickups and disable dropoffs.
            this._enablePickups( true );
            this._enableDropoffs( false );
        },

        setupNextTrial: function( stage, trial, trialCount )
        {
            this._cleanup();

            this.model.stage = stage;
            this.model.trial = trial;
            // Init the next trial.
            this._initNextTrial( this.model.stage, this.model.trial, trialCount );

            this._setup();
            this._createViews();

            // Reset Metadata
            this.model.resetCount = 0;
            this.model.steps = 0;
            this.model.pickupCount = 0;
            this.model.dropoffCount = 0;
            this.model.timeToCompleteRound = 0;

            // Metadata
            this._initMetadata();
        },

        _cleanup: function()
        {
            // Remove all views.
            this._removeAllInterests();

            if( this.vehicle )
            {
                this.vehicle.view.removeFromParent( true );
            }

            if( this.view.tiledMap )
            {
                this.view.tiledMap.removeFromParent( true );
            }
        },

        /**
         * Sets the target.
         * 
         * @param {object} target
         */
        setTarget: function( target )
        {
            this._target = target;
        },

        /**
         * Sets the trial complete callback for the target.
         * 
         * @param {Function} trialCompleteCallback
         */
        setTrialCompleteCallback: function( trialCompleteCallback )
        {
            this._trialCompleteCallback = trialCompleteCallback;
        },

        /**
         * Sets the trial complete animation finish callback for the target.
         * 
         * @param {Function} trialCompleteAnimationFinishCallback
         */
        setTrialCompleteAnimationFinishCallback: function( trialCompleteAnimationFinishCallback )
        {
            this._trialCompleteAnimationFinishCallback = trialCompleteAnimationFinishCallback;
        },

        /**
         * Sets the max vehicle capacity callback for the target.
         * 
         * @param {Function} maxVehicleCapacityButtonPressedCallback
         */        
        setMaxVehicleCapacityButtonPressedCallback: function( maxVehicleCapacityButtonPressedCallback )
        {
            this._maxVehicleCapacityButtonPressedCallback = maxVehicleCapacityButtonPressedCallback;
        },

        /**
         * Sets the pickup callback for the target.
         * 
         * @param {Function} pickupCallback
         */
        setPickupCallback: function( pickupCallback )
        {
            this._pickupCallback = pickupCallback;
        },

        /**
         * Sets the dropoff callback for the target.
         * 
         * @param {Function} dropoffCallback
         */
        setDropoffCallback: function( dropoffCallback )
        {
            this._dropoffCallback = dropoffCallback;
        },

        /**
         * Sets the begin metadata trial callback for the target.
         * 
         * @param {Function} beginMetadataTrialCallback
         */
        setBeginMetadataTrialCallback: function( beginMetadataTrialCallback )
        {
            this._beginMetadataTrialCallback = beginMetadataTrialCallback;
        },

        /**
         * Sets the end metadata trial callback for the target.
         * 
         * @param {Function} endMetadataTrialCallback
         */
        setEndMetadataTrialCallback: function( endMetadataTrialCallback )
        {
            this._endMetadataTrialCallback = endMetadataTrialCallback;
        },

        /**
         * Returns all interests.
         * 
         * @return {Array} The interests.
         */
        interests: function()
        {
            var interests = null;
            interests = this.pickups.concat( this.dropoffs );
            return interests;
        },

        /**
         * Returns the available pickups.
         * 
         * @return {array} The available pickups.
         */
        _availablePickups: function()
        {
            var availablePickups = [];
            var bitmask = 0;
            var pickup = null;
            var self = this;

            LumosityUtils.forEachIn( this.settings.pickups, function( value /*, imageName*/ )
            {
                bitmask = parseInt( ( 1 << value ), 10 );
                if( bitmask & self.model.pickupsBitmask )
                {
                    pickup = self.pickups[ value ];
                    availablePickups.push( pickup );
                }
            } );

            return availablePickups;
        },

        /**
         * Removes all the interests from the view.
         */
        _removeAllInterests: function()
        {
            var self = this;
            var interests = null;

            interests = this.interests();
            interests.forEach( function( interest )
            {
                self.view.removeChild( interest );
            } );

            this.pickups.length = 0;
            this.dropoffs.length = 0;
        },

        /**
         * Show or hides interests.
         * 
         * @param  {Boolean} isEnabled Whether to show or hide.
         */
        _showInterests: function( isVisible )
        {
            var interests = null;
            var opacity = isVisible ? 255 : 0;

            interests = this.interests();
            interests.forEach( function( interest )
            {
                interest.setOpacity( opacity );
            } );
        },

        /**
         * Hides all interests highlighted image.
         */
        _hideAllInterestsHighlightedImage: function()
        {
            var interests = null;

            interests = this.interests();
            interests.forEach( function( interest )
            {
                if( interest.highlightImage )
                {
                    interest.highlightImage.setVisible( false );
                }
            } );
        },

        /**
         * Enables or disables pickups's touch.
         * 
         * @param  {Boolean} isEnabled Whether to enable or disable.
         */
        _enablePickups: function( isEnabled )
        {
            this.pickups.forEach( function( interest )
            {
                interest.setEnabled( isEnabled );
                interest.isHidden = false;
            } );
        },

        /**
         * Enables or disables available pickups's touch.
         * 
         * @param  {Boolean} isEnabled Whether to enable or disable.
         */
        _enableAvailablePickups: function( isEnabled )
        {
            var availablePickups = this._availablePickups();
            availablePickups.forEach( function( pickup )
            {
                pickup.setEnabled( isEnabled );
            } );
        },

        /**
         * Enables or disables dropoffs's touch.
         * 
         * @param  {Boolean} isEnabled Whether to enable or disable.
         */
        _enableDropoffs: function( isEnabled )
        {
            this.dropoffs.forEach( function( interest )
            {
                interest.setEnabled( isEnabled );
            } );
        },

        /**
         * Enables or disables interest's touch.
         * 
         * @param  {Boolean} isEnabled Whether to enable or disable.
         */
        _enableInterests: function( isEnabled )
        {
            var interests = null;

            interests = this.interests();
            interests.forEach( function( interest )
            {
                interest.setEnabled( isEnabled );
            } );
        },

        /**
         * Highlights or unhighlights interests.
         * 
         * @param  {Boolean} isEnabled Whether to ighlight or unhighlight.
         */
        _highlightInterests: function( isHighlighted )
        {
            var interests = null;

            interests = this.interests();
            interests.forEach( function( interest )
            {
                interest.setHighlighted( isHighlighted );
            } );
        },

        /**
         * Highlights or unhighlights pickups.
         * 
         * @param  {Boolean} isEnabled Whether to ighlight or unhighlight.
         */
        _highlightPickups: function( isHighlighted )
        {
            this.pickups.forEach( function( interest )
            {
                interest.setHighlighted( isHighlighted );
            } );
        },

        /**
         * Highlights or unhighlights dropoffs.
         * 
         * @param  {Boolean} isEnabled Whether to ighlight or unhighlight.
         */
        _highlightDropoffs: function( isHighlighted )
        {
            this.dropoffs.forEach( function( interest )
            {
                interest.setHighlighted( isHighlighted );
            } );
        },

        /**
         * Reads in the levels file data and creates an array of stages.
         *
         * @return {Array}     The stages.
         */
        _createStages: function()
        {
            var fileData = "";
            var stages = null;

            fileData = lumosity.io_read( "pet_detective_levels.txt" );
            stages = LevelsData.parse( fileData );
            return stages;
        },

        /**
         * Creates and returns a random level.
         * 
         * @param  {[type]} stage      The stage.
         * @param  {[type]} trial      The trial.
         * @param  {[type]} trialCount The total amount of trials.
         * 
         * @return {[type]}            The random level.
         */
        _createRandomLevel: function( stage, trial, trialCount )
        {
            var stages = null;
            var index = -1;
            var levels = null;
            var i;
            var sortedLevels = null;
            var start = 0;
            var end = 0;
            var level = null;

            stages = this.model.stages;
            // Get an index into the stage's array to retrieve an array of levels.
            index = Math.max( 1, Math.min( stages.length, stage ) ) - 1;
            levels = stages[ index ];

            sortedLevels = levels.slice();
            for( i = 0; i < sortedLevels.length; i++ )
            {
                sortedLevels[ i ].normalVariance = sortedLevels[ i ].skewedVariance / Math.pow( sortedLevels[ i ].shortest, 2 );
            }
            sortedLevels.sort( this._numericallyDecendingSort );

            start = sortedLevels.length * ( trial - 1 ) / trialCount;
            end = sortedLevels.length * trial / trialCount;
            level = sortedLevels.slice( start, end ).getRandomElement();

            cc.log( "stage: " + stage + " trial: " + trial + " trialCount: " + trialCount + " level: " + sortedLevels.indexOf( level ) );
            return level;
        },

        /**
         * Creates an A* Grid.
         * 
         * @return {object} The A* Grid.
         */
        _createGrid: function()
        {
            var rows = null;
            var levelText = "";
            var r;
            var c;
            var address = 0;
            var isTraversable = false;
            var grid = [];

            levelText = this.model.level.varyLevelText;
            rows = LevelsData.toRows( levelText );
            this.model.rowCount = LevelsData.rowCount( levelText );
            this.model.columnCount = LevelsData.columnCount( levelText );
            this.model.gridRowCount = this.model.rowCount + 2;
            this.model.gridColumnCount = this.model.columnCount + 2;

            for( r = 0; r < this.model.gridRowCount; r++ )
            {
                for( c = 0; c < this.model.gridColumnCount; c++ )
                {
                    address = c + this.model.gridColumnCount * r;
                    grid[ address ] = { "isTraversable": false };

                    isTraversable = ( 0 < r && r < this.model.gridRowCount - 1 ) &&
                                    ( 0 < c && c < this.model.gridColumnCount - 1 );

                    if( isTraversable )
                    {
                        isTraversable = " " !== rows[r - 1].substr( c - 1, 1 );
                        grid[ address ].isTraversable = isTraversable;
                    }
                }
            }

            return grid;
        },

        /**
         * Creates a TiledMap.
         * 
         * @return {TiledMap} The TiledMap.
         */
        _createTiledMap: function()
        {
            var tiledMap = null;
            var r;
            var c;
            var gid = 0;

            this.model.tiledRowCount = Math.floor( ( ( this.model.rowCount + 1 ) / 2 ) );
            this.model.tiledColumnCount = Math.floor( ( ( this.model.columnCount + 1 ) / 2 ) );
            cc.log( "TileMapSize: " + this.model.tiledRowCount + " " + this.model.tiledColumnCount );
            tiledMap = TiledMap.create( "TileMap9x9.tmx" );
            tiledMap.setAnchorPoint( this._calculateTiledMapAnchorPoint() );
            for( r = 0; r < this.model.tiledRowCount; r++ )
            {
                for( c = 0; c < this.model.tiledColumnCount; c++ )
                {
                    gid = this._getBitmask( this.g( r ), this.g( c ) );
                    tiledMap.setTileCoordWithGID( "Background", cc.p( c, r ), gid );
                }
            }

            return tiledMap;
        },

        /**
         * Sets up the model data for objects that reside on the TiledMap.
         */
        _setup: function()
        {
            var rows = null;
            var r;
            var c;
            var character = "";
            var address = 0;
            var index = -1;

            // Reset the model data.
            this.model.pickups = [];
            this.model.pickupsBitmask = 0;
            this.model.dropoffs = [];
            this.model.dropoffsBitmask = 0;


            rows = LevelsData.toRows( this.model.level.varyLevelText );
            for( r = 0; r < this.model.rowCount; r++ )
            {
                for( c = 0; c < this.model.columnCount; c++ )
                {
                    character = rows[ r ].substr( c, 1 );
                    address = this.t( c ) + this.model.tiledColumnCount * this.t( r );

                    // Pickup
                    if( LevelsData.pickupCharacters.indexOf( character ) >= 0 )
                    {
                        index = LevelsData.pickupCharacters.indexOf( character );
                        this.model.pickups[ index ] = address;
                        this.model.pickupsBitmask |=  ( 1 << index );
                    }
                    // Dropoff
                    else if( LevelsData.dropoffCharacters.indexOf( character ) >= 0 )
                    {
                        index = LevelsData.dropoffCharacters.indexOf( character );
                        this.model.dropoffs[ index ] = address;
                        this.model.dropoffsBitmask |=  ( 1 << index );
                    }
                    // Vehicle
                    else if( LevelsData.vehicleCharacters.indexOf( character ) >= 0 )
                    {
                        this.model.vehicleCapacity = LevelsData.vehicleCharacters.indexOf( character ) + 1;
                        this.model.vehicleRow = this.t( r );
                        this.model.vehicleColumn = this.t( c );
                    }
                }
            }
        },

        /**
         * Creates the pickupable interests.
         * 
         * @return {Array} The array of pickupable interests.
         */
        _createPickups: function()
        {
            var self = this;
            var bitmask = -1;
            var address = 0;
            var r = 0;
            var c = 0;
            var pickup = null;
            var pickups = [];

            LumosityUtils.forEachIn( this.settings.pickups, function( value, imageName )
            {
                bitmask = parseInt( ( 1 << value ), 10 );
                if( bitmask & self.model.pickupsBitmask )
                {
                    address = self.model.pickups[ value ];
                    r = Math.floor( address / self.model.tiledColumnCount );
                    c = Math.floor( address % self.model.tiledColumnCount );
                    pickup = self._createInterest( imageName, r, c, self.settings.pickupDesignSize );
                    pickup.handleIsPressedSuper = pickup.handleIsPressed;
                    pickup.handleIsPressed = self._pickupHandleIsPressed.bind( pickup );
                    pickup.disabledCallback = self._maxVehicleCapacityButtonPressedCallback.bind( self._target );         
                    pickups[ value ] = pickup;
                }
            } );

            return pickups;
        },


        /**
         * Creates the dropoffable interests.
         * 
         * @return {Array} The array of dropoffable interests.
         */
        _createDropoffs: function()
        {
            var self = this;
            var bitmask = -1;
            var address = 0;
            var r = 0;
            var c = 0;
            var dropoff = null;
            var dropoffs = [];

            LumosityUtils.forEachIn( this.settings.dropoffs, function( value, info )
            {
                bitmask = parseInt( ( 1 << value ), 10 );
                if( bitmask & self.model.dropoffsBitmask )
                {
                    address = self.model.dropoffs[ value ];
                    r = Math.floor( address / self.model.tiledColumnCount );
                    c = Math.floor( address % self.model.tiledColumnCount );
                    dropoff = self._createInterest( info, r, c, self.settings.dropoffDesignSize );
                    dropoffs[ value ] = dropoff;
                }
            } );

            return dropoffs;
        },


        /**
         * Creates a generic interest at a tiledMap location.
         * 
         * @param  {object} info        The interest's information.
         * @param  {int} row            The row.
         * @param  {int} column         The column.
         * 
         * @return {cc.Sprite}          The interest.
         */
        _createInterest: function( info, row, column, iconDesignSize )
        {
            var rect = cc.RectZero();
            var position = cc.PointZero();
            var iconPosition = cc.PointZero();
            var backgroundImageNormal = null;
            var backgroundImageSelected = null;
            var backgroundImageDisabled = null;
            var backgroundImageHighlighted = null;
            var iconImageNormal = null;
            var iconImageSelected = null;
            var iconImageDisabled = null;
            var iconImageHighlighted = null;
            var highlightImage = null;
            var interest = null;
            var designSize = cc.SizeZero();


            designSize = cc.size( iconDesignSize.width, iconDesignSize.height );

            // Convert row and column to tile space.
            rect = this.view.tiledMap.tileRectFromTileCoords( cc.p( column, row ) );
            position = cc.p( rect.x + ( rect.width >> 1 ), rect.y + ( rect.height >> 1 ) );
            position = this._convertTileToViewSpace( position );

            if( info.backgroundImageNormal )
            {
                backgroundImageNormal = cc.Sprite.create( "interests/" + info.backgroundImageNormal );
            }

            if( info.backgroundImageSelected )
            {
               backgroundImageSelected = cc.Sprite.create(  "interests/" + info.backgroundImageSelected );
            }

            if( info.backgroundImageDisabled )
            {
               backgroundImageDisabled = cc.Sprite.create(  "interests/" + info.backgroundImageDisabled );
            }

            if( info.backgroundImageHighlighted )
            {
               backgroundImageHighlighted = cc.Sprite.create(  "interests/" + info.backgroundImageHighlighted );
            }

            if( info.iconImageNormal )
            {
                iconImageNormal = cc.Sprite.create(  "interests/" + info.iconImageNormal );
            }

            if( info.iconImageSelected )
            {
                iconImageSelected = cc.Sprite.create(  "interests/" + info.iconImageSelected );
            }

            if( info.iconImageDisabled )
            {
                iconImageDisabled = cc.Sprite.create(  "interests/" + info.iconImageDisabled );
            }

            if( info.iconImageHighlighted )
            {
                iconImageHighlighted = cc.Sprite.create(  "interests/" + info.iconImageHighlighted );
            }

            // Create the button and set the position.
            interest = this._createButton( backgroundImageNormal, backgroundImageSelected, backgroundImageDisabled, backgroundImageHighlighted, iconImageNormal, iconImageSelected, iconImageDisabled, iconImageHighlighted );
            interest.setAnchorPoint( cc.p( 0.5, 0.25 ) );
            interest.setPosition( position );
            interest.setTarget( this );
            interest.setReleasedCallback( this._onButtonRelease );
            interest.setContentSize( designSize, false );
            
            // Center the icons
            iconPosition = cc.p( interest.getContentSize().width >> 1, interest.getContentSize().height >> 1 );
            if( iconImageNormal )
            {
                iconImageNormal.setPosition( iconPosition );
                LumosityUtils.scaleForContentSize( iconImageNormal, designSize );
            }

            if( iconImageSelected )
            {
                iconImageSelected.setPosition( iconPosition );
                LumosityUtils.scaleForContentSize( iconImageSelected, designSize );
            }

            if( iconImageDisabled )
            {
                iconImageDisabled.setPosition( iconPosition );
                LumosityUtils.scaleForContentSize( iconImageDisabled, designSize );
            }

            if( iconImageHighlighted )
            {
                iconImageHighlighted.setPosition( iconPosition );
                LumosityUtils.scaleForContentSize( iconImageHighlighted, designSize );
            }

            // Highlight image.
            if( info.backgroundImageSelected )
            {
                highlightImage = cc.Sprite.create(  "interests/" + info.backgroundImageSelected );
                highlightImage.setVisible( false );
                highlightImage.setAnchorPoint( cc.p( 0, 0 ) );

                LumosityUtils.scaleForContentSize( highlightImage, designSize );
                interest.highlightImage = highlightImage;
                interest.addChild( highlightImage );
            }

            return interest;
        },

        /**
         * Creates and returns a button.
         * 
         * @param  {cc.Sprite} bgUpImage     The background normal image.
         * @param  {cc.Sprite} bgDownImage   The background selected image.
         * @param  {cc.Sprite} iconUpImage   The icon normal image.
         * @param  {cc.Sprite} iconDownImage The icon selected image.
         * 
         * @return {LumosityButton}          The button.
         */
        _createButton: function( bgUpImage, bgDownImage, bgDisabledImage, bgHighlightedImage, iconUpImage, iconDownImage, iconDisabledImage, iconHighlightedImage )
        {   
            var button = null;

            button = new LumosityButton();
            button.setBackgroundImageForState( bgUpImage, LumosityButtonState.ButtonStateNormal );
            button.setBackgroundImageForState( bgDownImage, LumosityButtonState.ButtonStateSelected );
            button.setBackgroundImageForState( bgDisabledImage, LumosityButtonState.ButtonStateDisabled );
            button.setBackgroundImageForState( bgHighlightedImage, LumosityButtonState.ButtonStateHighlighted );
            button.setImageForState( iconUpImage, LumosityButtonState.ButtonStateNormal );
            button.setImageForState( iconDownImage, LumosityButtonState.ButtonStateSelected );
            button.setImageForState( iconDisabledImage, LumosityButtonState.ButtonStateDisabled );
            button.setImageForState( iconHighlightedImage, LumosityButtonState.ButtonStateHighlighted );

            return button;
        },

        _pickupHandleIsPressed: function( isPressed, cancelled )
        {
            var callback = null;
            this.handleIsPressedSuper( isPressed, cancelled );

            if( isPressed === true && this._isEnabled === false )
            {
                callback = this.disabledCallback;

                if( this.willHandleCallbacks )
                {
                    this._sendCallback( callback );
                }
            }
        },

        /**
         * Creates a vehicle.
         * 
         * @return {Vehicle} The vehicle.
         */
        _createVehicle: function()
        {
            var rect = cc.RectZero();
            var position = cc.PointZero();
            var vehicle = null;
            var direction = VehicleDirection.Right;

            rect = this.view.tiledMap.tileRectFromTileCoords( cc.p( this.model.vehicleColumn, this.model.vehicleRow ) );
            position = cc.p( rect.x + ( rect.width >> 1 ), rect.y + ( rect.height >> 1 ) );
            position = this._convertTileToViewSpace( position );

            vehicle = new Vehicle( this.settings.vehicleSettings, this.model.vehicleCapacity, this.model.level.shortest );
            vehicle.view.setPosition( position );

            direction = this._calculateVehicleTileFacing();
            vehicle.setDirection( direction );

            return vehicle;
        },

        /**
         * Calculates the vehicles direction based on gid of the tile it is on.
         * 
         * @return {VehicleDirection} The direction of the vehicle.
         */
        _calculateVehicleTileFacing: function()
        {
             var gid = 0;
             var tileFacings = null;
             var direction = VehicleDirection.Right;

             gid = this._getBitmask( this.g( this.model.vehicleRow), this.g( this.model.vehicleColumn ) );
             tileFacings = this.settings.tileFacings;

             if( tileFacings.left.contains( gid ) )
             {
                direction = VehicleDirection.Left;
             }
             else if( tileFacings.right.contains( gid) )
             {
                direction = VehicleDirection.Right;
             }
             else if( tileFacings.up.contains( gid) )
             {
                direction = VehicleDirection.Up;
             }
             else if( tileFacings.down.contains( gid) )
             {
                direction = VehicleDirection.Down;
             }

             return direction;
        },
        
        /**
         * Converts an index within the TiledMap space into an index into
         * the A* grid space.
         *
         * @param  {int} tiledIndex The index into the TiledMap.
         * @return {int}            The index into the A* grid.
         */
        g: function( tiledIndex )
        {
            var gridIndex = 0;
            gridIndex = Math.floor( ( ( tiledIndex * 2 ) + 1 ) );
            return gridIndex;
        },

        /**
         * Converts an index within the LevelsData into an index into
         * the TiledMap space.
         *
         * @param  {int} gridIndex The index into the LevelsData.
         * @return {int}           The index into the TiledMap.
         */
        t: function( gridIndex )
        {
            var tiledIndex = 0;
            tiledIndex = Math.floor( ( ( gridIndex + 1 ) / 2 ) );
            return tiledIndex;
        },

        /**
         * Returns a bitmask representing the gid of the tile in which to display.
         * http://www.angryfishstudios.com/2011/04/adventures-in-bitmasking/
         * 
         * @param  {int} row    The A* grid row.
         * @param  {int} column The A* grid column.
         * 
         * @return {int}        The bitmask
         */
        _getBitmask: function( row, column )
        {
            var bitmask = 0;

            bitmask = ( this._isTraversable( row - 1, column ) ? 1 : 0 ) +
                      ( this._isTraversable( row, column + 1 ) ? 2 : 0 ) +
                      ( this._isTraversable( row + 1, column ) ? 4 : 0 ) +
                      ( this._isTraversable( row, column - 1 ) ? 8 : 0 );

            return bitmask;
        },

        /**
         * Returns whether the tile at the given row and column are traversable.
         * 
         * @param  {int} row    The A* grid row.
         * @param  {int} column The A* grid column.
         * 
         * @return {boolean} Whether the tile can be traversed.
         */
        _isTraversable: function( row, column )
        {
            var isTraversable = false;
   
            isTraversable = this.model.grid[ column + this.model.gridColumnCount * row ].isTraversable;
            return isTraversable;
        },

        /**
         * Calculates the TiledMap's anchorpoint
         * 
         * @return {cc.p} The anchorpoint.
         */
        _calculateTiledMapAnchorPoint: function()
        {
            var anchorPoint = cc.PointZero();

            anchorPoint = cc.p( ( this.model.tiledColumnCount / 2 ) / 9, 1 - ( ( this.model.tiledRowCount / 2 ) / 9 ) );
            return anchorPoint;
        },

        /**
         * Calculates the percentage of the screen the tileMap occupies.
         */
        _calculateTiledMapToScreenAreaPercent: function( isRotated )
        {
            var tiledMapContentSize = cc.SizeZero();
            var streetBounds = cc.SizeZero();
            var hierarchyScale = cc.PointZero();
            var areaPercent = cc.PointZero();

            streetBounds = this._getBoundDesignSize();

            // Scale the design size by the change is size of the device.
            hierarchyScale = LumosityUtils.getNodeHierarchicalScale( this.view.tiledMap );
            streetBounds.width *= hierarchyScale.x;
            streetBounds.height *= hierarchyScale.y;

            tiledMapContentSize = this.getTiledMapContentSize( isRotated );

            areaPercent = cc.p( ( tiledMapContentSize.width / streetBounds.width ) * 100, ( tiledMapContentSize.height / streetBounds.height ) * 100 );

            cc.log( "Street::_calculateTiledMapToScreenAreaPercent " + "tiledMapContentSize: " + JSON.stringify( tiledMapContentSize ) + "streetBounds: " + JSON.stringify( streetBounds ) );
            return areaPercent;
        },

        getTiledMapContentSize: function( isRotated )
        {
            var tiledMapContentSize = cc.SizeZero();
            var tileSize = cc.SizeZero();

            tileSize = this.view.tiledMap.getTileSize();
            tiledMapContentSize = cc.size( this.model.tiledColumnCount * tileSize.width, this.model.tiledRowCount * tileSize.height );

            if( isRotated )
            {
                tiledMapContentSize = cc.size( tiledMapContentSize.height, tiledMapContentSize.width );
            }

            return tiledMapContentSize;
        },

        _getBoundDesignSize: function()
        {
            return this.settings.boundDesignSize;
        },

        /**
         * Converts point from tile space to view space.
         * 
         * @param  {cc.p} point The point in tile space.
         * 
         * @return {cc.p}       The point in view space.
         */
        _convertTileToViewSpace: function( point )
        {
            var layer = null;
            var newPoint = cc.PointZero();

            layer = this.view.tiledMap.getLayer( "Background" );
            newPoint = layer.convertToWorldSpace( point );
            newPoint = this.view.convertToNodeSpace( newPoint );
            return newPoint;
        },

        /**
         * Returns the direction the vehicle should face when
         * heading towards the provided next position.
         * 
         * @param  {int} prevPos    The previous position in view space.
         * @param  {int} nextPos    The next position in view space.
         * 
         * @return {boolean} Direction of the vehicle.
         */
        _getVehicleDirection: function( prevPos, nextPos )
        {
            var direction = VehicleDirection.Left;
            var deltaY = Math.abs( prevPos.y - nextPos.y );
            var deltaX = Math.abs( prevPos.x - nextPos.x );
            var epsilon = 0.001;

            if( ( deltaY > epsilon ) && prevPos.y > nextPos.y )
            {
                direction = VehicleDirection.Down;
            }
            else if( ( deltaY > epsilon ) && prevPos.y < nextPos.y )
            {
                direction = VehicleDirection.Up;
            }
            else if( ( deltaX > epsilon ) && prevPos.x > nextPos.x )
            {
                direction = VehicleDirection.Left;
            }
            else if( ( deltaX > epsilon ) && prevPos.x < nextPos.x )
            {
                direction = VehicleDirection.Right;
            }

            return direction;
        },

        /**
         * Sorts an array of levels numerically decending by variance.
         * 
         * @param  {object} level1
         * @param  {object} level2
         * 
         * @return {int}
         */
        _numericallyDecendingSort: function( level1, level2 )
        {
            var normalVariance1 = level1.normalVariance;
            var normalVariance2 = level2.normalVariance;

            if( normalVariance1 < normalVariance2 )
            {
                return 1;
            }

            if( normalVariance1 > normalVariance2 )
            {
                return -1;
            }

            return 0;
        },

        /**
         * Release callback for interest buttons.
         * 
         * @param  {LumosityButton} button The button that was released.
         */
        _onButtonRelease: function( button )
        {
            var worldPoint = this.view.convertToWorldSpace( button.getPosition() );
            var localPoint = this.view.tiledMap.convertToNodeSpace( worldPoint );
            var tileCoord = this.view.tiledMap.tileCoordForPosition( localPoint );
            var startPosX = this.g( this.model.vehicleColumn );
            var startPosY = this.g( this.model.vehicleRow );
            var endPosX = this.g( tileCoord.x );
            var endPosY = this.g( tileCoord.y );
            var bestFirstPath = null;
            var i;

            if( this.model.areInteractionsBlocked === true )
            {
                return;
            }

            // Trim the list to be processed in tile space.
            this.model.bestFirstPath.length = 0;
            bestFirstPath = this._bestFirstPath( startPosX, startPosY, endPosX, endPosY );
            for( i = 0; i <  bestFirstPath.length; i = i + 2 )
            {
                this.model.bestFirstPath.push( bestFirstPath[ i ] );
            }

            cc.log( "Solution: " + JSON.stringify( this.model.bestFirstPath ) );


            // Start audio.
            this.vehicle.startDriveAudio();

            this._blockInteractions();
            this.model.hasProcessedDropoff = false;
            if( button.highlightImage )
            {
                button.highlightImage.setVisible( true );
            }

            this._handleMovement();
        },

        /**
         * Calculates and returns a best first path between two
         * sets of points.
         * 
         * @param  {int} startX The start position x.
         * @param  {int} startY The start position y.
         * @param  {int} endX   The end position x.
         * @param  {int} endY   The end position y.
         * 
         * @return {Array}        The array of position representing the best first path.
         */
        _bestFirstPath: function( startX, startY, endX, endY )
        {
            var columnSize = this.model.gridColumnCount;
            var rowSize =  this.model.gridRowCount;
            var bestFirstPath = null;

            // Plan the path and run.
            this.pathPlanner.planPath( startX, startY, endX, endY, columnSize, rowSize );
            this.pathPlanner.run();

            bestFirstPath = this.pathPlanner.solution().slice();
            this.pathPlanner.cleanup();

            return bestFirstPath;
        },

        /**
         * Handles the movement between tiles.
         */
        _handleMovement: function()
        {
            var node = null;
            var rect = cc.RectZero();
            var position = cc.PointZero();
            var sequence = null;
            var moveAction = null;
            var callbackAction = null;
            var tileCoordX = 0;
            var tileCoordY = 0;
            var newFuel = 0;
            var direction = VehicleDirection.Left;

            // Completed path.
            if( !this.model.bestFirstPath.length || ( this.vehicle.getFuel() <= 0 ) )
            {
                this._finishMovement();
                return;
            }

            // Convert grid based node coord into tile space coord.
            node = this.model.bestFirstPath.pop();
            tileCoordX = Math.floor( ( node.x - 1 ) / 2 );
            tileCoordY = Math.floor( ( node.y - 1 ) / 2 );

            // Convert tile space coord into view space position.
            rect = this.view.tiledMap.tileRectFromTileCoords( cc.p( tileCoordX, tileCoordY ) );
            position = cc.p( rect.x + ( rect.width >> 1 ), rect.y + ( rect.height >> 1 ) );
            position = this._convertTileToViewSpace( position );

            // Create and run a cc.MoveTo action on the calculated position.
            moveAction = cc.MoveTo.create( this.vehicle.settings.movementDurationPerTile, position );
            callbackAction = cc.CallFunc.create( this._handleMovement, this );
            sequence = cc.Sequence.create( moveAction, callbackAction );
            this.vehicle.view.runAction( sequence );

            // Set the vehicle direction towards the desired tile coordinates.
            direction = this._getVehicleDirection( this.vehicle.view.getPosition(), position );
            this.vehicle.setDirection( direction );

            // Store the new tile coord into the model.
            this.model.vehicleRow = tileCoordY;
            this.model.vehicleColumn = tileCoordX;
            this.model.steps = this.model.steps + 1;
            
            if( !this.model.freeVehicleMovement )
            {
                // Adjust fuel cost
                newFuel = this.vehicle.getFuel();
                newFuel -= this.vehicle.settings.fuelCostPerTile;
                this.vehicle.setFuel( newFuel );
                this.vehicle.setFuelText( this.vehicle.getFuel().toString() + "/" + this.vehicle.getFuelCount().toString() );
            }
        },

        /**
         * Handles finished movement.
         */
        _finishMovement: function()
        {
            // Stop Audio
            this.vehicle.stopDriveAudio();

            // Modify the model data to reflect the pickup/dropoff.
            this._trade();

            // Update the visuals of the interests.
            this._updateInterests();

            // Un-block the input.
            this._unblockInteractions();

            // End Metadata trial.
            this.endMetadataTrial( false );

            // Start new Metadata trial.
            this.beginMetadataTrial();

            // Evaluate whether the trial has been succeeded or failed.
            if( this._hasSucceededTrial() )
            {
                this._onTrialComplete( true );
            }
            else if( this._hasFailedTrial() )
            {
                this._onTrialComplete( false );
            }
        },

        /**
         * Processes transactions for pickups/dropoffs.
         */
        _trade: function()
        {
            var address = 0;

            address = this.model.vehicleColumn + this.model.tiledColumnCount * this.model.vehicleRow;
            this._processPickup( address );
            this._processDropoff( address );
        },

        /**
         * Process the data for receiving a pickup.
         * 
         * @param  {int} address The tile address.
         * @return { bool } Whether there was a pickup.
         */
        _processPickup: function( address )
        {
            var passengerIndex = -1;
            var pickupIndex = -1;

            passengerIndex = this.vehicle.model.passengers.indexOf( -1 );
            if( passengerIndex < 0 )
            {
                return false;
            }

            pickupIndex = this.model.pickups.indexOf( address );
            if( pickupIndex < 0 )
            {
                return false;
            }

            // Flag that we have received a pickup.
            this.vehicle.model.passengers[ passengerIndex ] = pickupIndex;
            this.model.pickupsBitmask &= ~( 1 << pickupIndex );

            // Broadcast pickup to target.
            this._pickupCallback.call( this._target, this, passengerIndex, pickupIndex );

            return true;
        },

        /**
         * Process the data for a dropoff.
         * 
         * @param  {int} address The tile address.
         * @return { bool } Whether there was a dropoff.
         */
        _processDropoff: function( address )
        {
            var passengerIndex = -1;
            var dropoffIndex = -1;

            dropoffIndex = this.model.dropoffs.indexOf( address );
            if( dropoffIndex < 0 )
            {
                return false;
            }

            passengerIndex = this.vehicle.model.passengers.indexOf( dropoffIndex );
            if( passengerIndex < 0 )
            {
                return false;
            }

            // Flag that we have done a dropoff.
            this.vehicle.model.passengers[ passengerIndex ] = -1;
            this.model.dropoffsBitmask &= ~( 1 << dropoffIndex );
            this.model.hasProcessedDropoff = true;

            // Broadcast dropoff to target.
            this._dropoffCallback.call( this._target, this, passengerIndex );
            return true;
        },

        /**
         * Updates the visuals of the interests.
         */
        _updateInterests: function()
        {
            var address = 0;

            address = this.model.vehicleColumn + this.model.tiledColumnCount * this.model.vehicleRow;
            if( this.model.hasProcessedDropoff )
            {
                this._updateDropoff( address );
            }
            else
            {
                this._updatePickup( address );
            }
        },

        /**
         * Updates the visuals of a received pickup as well
         * as its corresponding dropoff.
         * 
         * @param  {int} address The tile address.
         */
        _updatePickup: function( address )
        {
            var pickupIndex = -1;
            var pickup = null;
            var dropoff = null;
            var passengerIndex = 0;

            pickupIndex = this.model.pickups.indexOf( address );
            if( pickupIndex < 0 )
            {
                return;
            }

            // Disable and hide pickup
            pickup = this.pickups[ pickupIndex ];
            pickup.setEnabled( false );
            pickup.setOpacity( 0 );
            // Used to filter buttons from being pressed when disabled.
            pickup.isHidden = true;

            // Set pickup's highlight image to invisible.
            if( pickup.highlightImage )
            {
                pickup.highlightImage.setVisible( false );
            }

            // Enable corresponding dropoff.
            dropoff = this.dropoffs[ pickupIndex ];
            dropoff.setEnabled( true );
            dropoff.setHighlighted( true );

            // Check if you have reached max vehicle capacity.
            passengerIndex = this.vehicle.model.passengers.indexOf( -1 );
            if( passengerIndex < 0 )
            {
                // Disable available pickups.
                this.model.hasReachedMaxVehicleCapacity = true;
                this._enableAvailablePickups( false );
            }

            // Set metadata character at vehicle address.
            this.model.characterAtVehicleAddress = LevelsData.pickupCharacters[ pickupIndex ];
            this.model.pickupCount = this.model.pickupCount + 1;
            this._removeCharacterFromEndLevelText( this.model.characterAtVehicleAddress );
        },

        /**
         * Updates the visuals of a dropoff.
         * 
         * @param  {int} address The tile address.
         */
        _updateDropoff: function( address )
        {
            var dropoffIndex = -1;
            var dropoff = null;

            dropoffIndex = this.model.dropoffs.indexOf( address );
            if( dropoffIndex < 0 )
            {
                return;
            }

            // Disable dropoff.
            dropoff = this.dropoffs[ dropoffIndex ];
            dropoff.setEnabled( false );
            dropoff.setHighlighted( false );
            dropoff.setOpacity( 0 );

            // Set dropoff's highlight image to invisible.
            if( dropoff.highlightImage )
            {
                dropoff.highlightImage.setVisible( false );
            }

            // Check if you were previously aat max vehicle capacity.
            if( this.model.hasReachedMaxVehicleCapacity )
            {
                // Re-enable available pickups.
                this.model.hasReachedMaxVehicleCapacity = false;
                this._enableAvailablePickups( true );
            }

            // Set metadata character at vehicle address.
            this.model.characterAtVehicleAddress = LevelsData.dropoffCharacters[ dropoffIndex ];
            this.model.dropoffCount = this.model.dropoffCount + 1;
            this._removeCharacterFromEndLevelText( this.model.characterAtVehicleAddress );
        },

        /**
         * Handles trial complete.
         */
        _onTrialComplete: function( hasSucceeded )
        {
            var vehicleAnimation = null;

            this._enableInterests( false );
            this._highlightDropoffs( false );

            // Set the round complete metadata.
            this._setRoundCompleteMetadata();

            // Notify trial completion.
            this._trialCompleteCallback.call( this._target, this, hasSucceeded );

            if( hasSucceeded )
            {
                vehicleAnimation = this._createVehicleFuelSucceedAnimation( this._trialCompleteAnimationFinishCallback, this._target, { isCorrect: hasSucceeded } );
            }
            else
            {
                this.vehicle.setFuelText( "0" );
                this.vehicle.stopEngineRunningAnimation();
                this.vehicle.stopEngineRunningAudio();

                vehicleAnimation = this._createVehicleFuelFailAnimation( this._trialCompleteAnimationFinishCallback, this._target, { isCorrect: hasSucceeded } );
            }

            this.vehicle.view.runAction( vehicleAnimation );
        },

        _setRoundCompleteMetadata: function()
        {
            this._removeCharacterFromEndLevelText( '4', '.' );
            this._addCharacterToEndLevelText( '4', this.model.vehicleRow, this.model.vehicleColumn );

            // Round Metadata
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_end_diagram, this._formatLevelTextForMetadata( this.model.level.endLevelText ) );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_reset, this.model.resetCount );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_steps, this.model.steps );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_pickups, this.model.pickupCount );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_dropoffs, this.model.dropoffCount );
            this.setRoundDataForKey( metadataTrialKeys.roundCSV_time, LumosityUtils.convertSecondsToMilliseconds( this.model.timeToCompleteRound ) );
        },

        _setTrialBeginMetadata: function()
        {
            this.model.trialData = {};
        },

        _setTrialCompleteMetadata: function( isReset )
        {
            var address = isReset ? -1 : this._formatVehicleAddressForMetadata();
            var target = isReset ? '' : this.model.characterAtVehicleAddress;

            var vehicleAddress = this.model.vehicleColumn + this.model.tiledColumnCount * this.model.vehicleRow;
            var dropoff = target === '' ? '' : ( this._pickupCharacterAtAddress( vehicleAddress ) !== -1 ? 'F' : this._dropoffCharacterAtAddress( vehicleAddress ) !== -1 ? 'T' : '' );

            // Trial Metadata
            this.setTrialDataForKey( metadataTrialKeys.trialCSV_round, this.model.trial );
            this.setTrialDataForKey( metadataTrialKeys.trialCSV_address, address );
            this.setTrialDataForKey( metadataTrialKeys.trialCSV_target, target );
            this.setTrialDataForKey( metadataTrialKeys.trialCSV_dropoff, dropoff );
            this.setTrialDataForKey( metadataTrialKeys.trialCSV_passengers, this._formatPassengersForMetadata() );
        },

        _createVehicleFuelSucceedAnimation: function( callback, target, data )
        {
            //var startFuelHideAnimFunc = null;
            var callbackFunc = null;
            var callbackDelayAction = null;
            var sequence = null;

            // Sara doesn't want the animation to hide upon success.
            //startFuelHideAnimFunc = cc.CallFunc.create( this.vehicle.startFuelHideAnimation, this.vehicle );
            callbackFunc = cc.CallFunc.create( callback, target, data );
            callbackDelayAction = cc.DelayTime.create( this.vehicle.fuelHideAnimationDuration() );
            sequence = cc.Sequence.create( /* startFuelHideAnimFunc, */ callbackDelayAction, callbackFunc );

            return sequence; 
        },

        _createVehicleFuelFailAnimation: function( callback, target, data )
        {
            var startFuelFlickerAnimFunc = null;
            var startFuelHideAnimFunc = null;
            var stopFuelFlickerAnimFunc = null;
            var callbackFunc = null;
            var callbackDelayAction = null;
            var delayAction = null;
            var sequence = null;

            startFuelFlickerAnimFunc = cc.CallFunc.create( this.vehicle.startFuelFlickerAnimation, this.vehicle );
            delayAction = cc.DelayTime.create( this.settings.vehicleFuelFailAnimation.delayDuration );
            stopFuelFlickerAnimFunc = cc.CallFunc.create( this.vehicle.stopFuelFlickerAnimation, this.vehicle );
            startFuelHideAnimFunc = cc.CallFunc.create( this.vehicle.startFuelHideAnimation, this.vehicle );
            callbackDelayAction = cc.DelayTime.create( this.vehicle.fuelHideAnimationDuration() );
            callbackFunc = cc.CallFunc.create( callback, target, data );
            sequence = cc.Sequence.create( startFuelFlickerAnimFunc, delayAction, stopFuelFlickerAnimFunc, startFuelHideAnimFunc, callbackDelayAction, callbackFunc );
            
            return sequence;
        },

        _hasSucceededTrial: function()
        {
            var hasSucceededTrial = ( this.model.dropoffsBitmask === 0 );
            return hasSucceededTrial;
        },

        _hasFailedTrial: function()
        {
            var hasFailedTrial = ( ( this.vehicle.getFuel() <= 0 ) && ( this.model.dropoffsBitmask !== 0 ) );
            return hasFailedTrial;
        },

        /**
        * Flags interactions to be blocked.
        */
        _blockInteractions: function()
        {
            this.model.areInteractionsBlocked = true;
        },

        /**
        * Flags interactions to be unblocked.
        */
        _unblockInteractions: function()
        {
            this.model.areInteractionsBlocked = false;
        },

        _formatLevelTextForMetadata: function( levelText )
        {
            return levelText.replace( /\n/g, ";" );
        },

        _formatPassengersForMetadata: function()
        {
            var i;
            var passengers = [];
            var index = 0;

            // Filter list of -1s
            var filteredPassengers = this.vehicle.model.passengers.filter( function( element )
            {
                return element !== -1;
            } );

            // Loop through the passengers and convert them into pickup letters.
            for( i = 0; i < filteredPassengers.length; i++ )
            {
                index = filteredPassengers[ i ];
                passengers.push( LevelsData.pickupCharacters[ index ] );
            }


            return passengers.length > 0 ? passengers.sort().join("").toString() : "";
        },

        _formatVehicleAddressForMetadata: function()
        {
            var address = this.model.vehicleColumn.toString() + this.model.vehicleRow.toString();
            return parseInt( address, 10 );
        },

        _addCharacterToEndLevelText: function( character, row, column )
        {
            var rows = LevelsData.toRows( this.model.level.endLevelText );
            rows[ row ] = rows[ row ].substr( 0, column ) + character + rows[ row ].substr( column + character.length );
            this.model.level.endLevelText = rows.join( ';' );
        },

        _removeCharacterFromEndLevelText: function( character )
        {
            var re = new RegExp( character, "g" );
            this.model.level.endLevelText = this.model.level.endLevelText.replace( re, '.' );
        },

        _pickupCharacterAtAddress: function( address )
        {
            var character = "";
            var pickupIndex = 0;

            pickupIndex = this.model.pickups.indexOf( address );
            if( pickupIndex === -1 )
            {
                return -1;
            }

            character = LevelsData.pickupCharacters[ pickupIndex ];
            return character;
        },

        _dropoffCharacterAtAddress: function( address )
        {
            var character = "";
            var dropoffIndex = 0;

            dropoffIndex = this.model.dropoffs.indexOf( address );
            if( dropoffIndex === -1 )
            {
                return -1;
            }

            character = LevelsData.dropoffCharacters[ dropoffIndex ];
            return character;
        },

        /**
         * Returns true or false on whether the tile at the 
         * desired location is valid.
         * 
         * @param  {int}  column The column.
         * @param  {int}  row    The row.
         * @return {Boolean}     Whether or not the tile is valid.
         */
        isValidTile: function( column, row )
        {
            var isValidTile = true;

            isValidTile = this._isTraversable( row, column );
            return isValidTile;
        },

        _createSettings: function( settings )
        {
            this.settings = new StreetSettings( settings );
        },
        
        _createModel: function()
        {
            this.model = new StreetModel();
        },
        
        _createView: function()
        {
            this.view = new StreetView();
            this.view.nodeName = "Street";
            this.view.setAnchorPoint( 0.5, 0.5 );
        }
    } );
}() );
