/*global cc*/
/*global require*/
/*global VehicleSettings*/
/*global VehicleModel*/
/*global VehicleView*/
/*global LumositySprite*/
/*global LumosityUtils*/

require( "src/Game/VehicleSettings.js" );
require( "src/Game/VehicleModel.js" );
require( "src/Game/VehicleView.js" );
require( "src/View/LumositySprite.js" );
require( "src/Utility/LumosityUtils.js" );

var ENGINE_RUNNING_ANIMATION_TAG =  1;
var FUEL_FLICKER_ANIMATION_TAG = 2;

var Vehicle;
( function()
{
    "use strict";    
    Vehicle = cc.Class.extend(
    {
        settings: null,
        model: null,
        view: null,
        
        /** @lends Vehicle.prototype */
        
        /**
        *  Vehicle is the main class for a Vehicle!
        *
        *  @constructs
        *  @param {array} Settings that define the Vehicle.
        */
        ctor: function( settings, capacity, fuel )
        {
            // var position = cc.PointZero();

            this._createSettings( settings );
            this._createModel();
            this._createView();

            // Special order required.
            this._createVehicleImages();
            this.setImage( 0 );
            this.view.createViews();

            this.initPassengers( capacity );
            
            this.setFuelCount( fuel );
            this.setFuel( fuel );
            this.setFuelText( this.getFuel().toString() + "/" + this.getFuelCount().toString() );
            this.setFuelOpacity( 0 );
        },

        onPause: function()
        {
            this.view.pauseAnimations();
        },

        onResume: function()
        {
            this.view.resumeAnimations();
        },

        onStopAnimations: function()
        {
            this.view.stopAnimations();
        },

        /**
        * Creates all the vehicle images.
        */
        _createVehicleImages: function()
        {
            this._createVehicleImage( "car/carLeft.png", this.settings.vehicleDesignSize, 0 );
            this._createVehicleImage( "car/carRight.png", this.settings.vehicleDesignSize, 0 );
            this._createVehicleImage( "car/carUp.png", this.settings.vehicleDesignSize, 0 );
            this._createVehicleImage( "car/carDown.png", this.settings.vehicleDesignSize, 0 );
        },

        /**
        * Creates and adds a vehicle image sprite to the view.
        *
        * @param {string} Image Name.
        * @param {cc.p} Position of the image in the view.
        *
        * @returns {cc.Sprite} Vehicle sprite.
        */
        _createVehicleImage: function( imageName, contentSize, opacity )
        {
            var cloud = null;

            cloud = new LumositySprite( imageName );
            cloud.setAnchorPoint( cc.PointZero() );
            cloud.setContentSize( contentSize, true );
            cloud.setOpacity( opacity );

            this.view.addVehicleImage( cloud );
            return cloud;
        },

        /**
         * Initializes the passengers.
         * 
         * @param  {int} capacity The capacity of passengers.
         */
        initPassengers: function( capacity )
        {
            var i;

            for( i = 0; i < capacity; i++ )
            {
                this.model.passengers[ i ] = -1;
            }
        },

        /**
         * Sets the fuel count.
         * 
         * @param {int} fuelCount The fuel count.
         */
        setFuelCount: function( fuelCount )
        {
            this.model.fuelCount = fuelCount;
        },

        /**
         * Gets the fuel count.
         * 
         * @return {int} The fuel count.
         */
        getFuelCount: function()
        {
            return this.model.fuelCount;
        },

        /**
         * Sets the fuel.
         * 
         * @param {int} fuel The fuel.
         */
        setFuel: function( fuel )
        {
            this.model.fuel = fuel;
        },

        /**
         * Gets the fuel.
         * 
         * @return {int} The fuel.
         */
        getFuel: function()
        {
            return this.model.fuel;
        },

        /**
         * Sets the fuel text.
         * 
         * @param {string} text The fuel text.
         */
        setFuelText: function( text )
        {
            this.view.text.setString( text );
        },

        /**
         * Sets the fuel anchorPoint based on direction of vehicle.
         * 
         * @param {VehicleDirection} direction The direction of the vehicle.
         */
        _setFuelAnchorPoint: function( direction )
        {
            var anchorPoint = this.settings.fuelAnchorPoints[ direction ];
            this.view.fuel.setAnchorPoint( anchorPoint );
        },

        /**
         * Sets the image.
         * 
         * @param {int} The index of the image.
         */
        setImage: function( index )
        {
            this.view.setImage( index );
        },

        /**
         * Sets the opacity.
         * 
         * @param {int} The opacity.
         */
        setOpacity: function( opacity )
        {
            this.view.setOpacity( opacity );
        },

        /**
         * Sets the fuels opacity.
         * 
         * @param {int} The opacity.
         */
        setFuelOpacity: function( opacity )
        {
            LumosityUtils.setOpacityOnNode( this.view.fuel, opacity );
        },

        /**
         * Sets the direction.
         * 
         * @param {VehicleDirection} The direction.
         */
        setDirection: function( direction )
        {
            if( direction !== this.model.direction )
            {
                this.setImage( direction );
                this.model.direction = direction;
                this._setFuelAnchorPoint( direction );
            }
        },

        /**
         * Gets the direction.
         * 
         * @return {VehicleDirection} The direction.
         */
        getDirection: function()
        {
            return this.model.direction;
        },

        /**
         * Starts the engine running animation.
         */
        startEngineRunningAnimation: function()
        {
            var defaultPosition = cc.PointZero();
            var topPosition = cc.PointZero();
            var bottomPosition = cc.PointZero();
            var engineRunningAnimation = null;

            topPosition = cc.p( defaultPosition.x, defaultPosition.y + this.settings.engineRunningAnimation.topOffsetY );
            bottomPosition = cc.p( defaultPosition.x, defaultPosition.y - this.settings.engineRunningAnimation.bottomOffsetY );

            engineRunningAnimation = this._createEngineRunningAnimation( this.settings.engineRunningAnimation.moveToTopDuration, 
                                                                         this.settings.engineRunningAnimation.moveToBottomDuration, 
                                                                         topPosition,
                                                                         bottomPosition );
            engineRunningAnimation.setTag( ENGINE_RUNNING_ANIMATION_TAG );

            
            this.view.animationView.runAction( engineRunningAnimation );
        },

        /**
         * Stops the engine running animation.
         */
        stopEngineRunningAnimation: function()
        {
            this.view.animationView.stopActionByTag( ENGINE_RUNNING_ANIMATION_TAG );
        },

        /**
         * Start the engine running audio.
         */
        startEngineRunningAudio: function()
        {
            var soundID = cc.AudioEngine.getInstance().playEffect( "sounds/carIdleLoop.mp3", true );
            this.model.carIdleSoundID = soundID;
        },

        /**
         * Stops the engine running audio.
         */
        stopEngineRunningAudio: function()
        {
            cc.AudioEngine.getInstance().stopEffect( this.model.carIdleSoundID );
        },

        isPlayingAnimation: function()
        {
            return ( this.view.animationView.numberOfRunningActions() > 0 );
        },

        /**
         * Start the drive audio.
         */
        startDriveAudio: function()
        {
            var soundID = cc.AudioEngine.getInstance().playEffect( "sounds/carDrive.mp3", true );
            this.model.carDriveSoundID = soundID;
        },

        /**
         * Stops the drive audio.
         */
        stopDriveAudio: function()
        {
            cc.AudioEngine.getInstance().stopEffect( this.model.carDriveSoundID );
        },

        /**
         * Starts the show fuel animation.
         */
        startFuelShowAnimation: function()
        {
            var defaultPosition = this.view.getFuelDefaultPosition();
            var startPosition = cc.PointZero();
            var topPosition = cc.PointZero();
            var endPosition = cc.PointZero();
            var fuelRiseAnimation = null;
            var fadeInAction = null;

            startPosition = cc.p( defaultPosition.x, defaultPosition.y - this.settings.fuelShowAnimation.startOffsetY );
            topPosition = cc.p( defaultPosition.x, defaultPosition.y + this.settings.fuelShowAnimation.topOffsetY );
            endPosition = defaultPosition;

            // Initialize starting values for animation.
            LumosityUtils.setOpacityOnNode( this.view.fuel, 0 );
            this.view.fuel.setPosition( startPosition );

            fuelRiseAnimation = this._createFuelRiseAnimation( this.settings.fuelShowAnimation.moveToTopDuration,
                                                               this.settings.fuelShowAnimation.moveToEndDuration,
                                                               topPosition,
                                                               endPosition );

            fadeInAction = cc.FadeTo.create( this.settings.fuelShowAnimation.fadeInDuration, this.settings.fuelShowAnimation.fadeInOpacity );

            // Translation actions can be run directly on fuel, but alpha actions must use
            // LumosityUtility method because fuel is just a cc.Node containing no opacity property.
            this.view.fuel.runAction( fuelRiseAnimation );
            LumosityUtils.runActionOnNode( this.view.fuel, fadeInAction );
        },

        /**
         * Starts the hide fuel animation.
         */
        startFuelHideAnimation: function()
        {
            var defaultPosition = this.view.getFuelDefaultPosition();
            var startPosition = cc.PointZero();
            var endPosition = cc.PointZero();
            var fuelFallAnimation = null;
            var fadeOutAction = null;

            startPosition = defaultPosition;
            endPosition = cc.p( defaultPosition.x, defaultPosition.y - this.settings.fuelHideAnimation.endOffsetY );

            // Initialize starting values for animation.
            LumosityUtils.setOpacityOnNode( this.view.fuel, 255 );
            this.view.fuel.setPosition( startPosition );

            fuelFallAnimation = this._createFuelFallAnimation( this.settings.fuelHideAnimation.moveToEndDuration, endPosition );
            fadeOutAction = cc.FadeTo.create( this.settings.fuelHideAnimation.fadeOutDuration, this.settings.fuelHideAnimation.fadeOutOpacity );

            this.view.fuel.runAction( fuelFallAnimation );
            LumosityUtils.runActionOnNode( this.view.fuel, fadeOutAction );
        },

        /**
         * Returns the hide animation duration.
         */
        fuelHideAnimationDuration: function()
        {
            var duration = 0;
            duration = Math.max( this.settings.fuelHideAnimation.moveToEndDuration,  this.settings.fuelHideAnimation.fadeOutDuration );
            return duration;
        },

        /**
         * Starts the fuel flicker animation.
         */
        startFuelFlickerAnimation: function()
        {
            var fuelFlickerAnimation = null;

            fuelFlickerAnimation = this._createFuelFlickerAnimation( this.settings.fuelFlickerAnimation.fadeOutDuration,
                                                                     this.settings.fuelFlickerAnimation.fadeInDuration,
                                                                     this.settings.fuelFlickerAnimation.fadeOutOpacity,
                                                                     this.settings.fuelFlickerAnimation.fadeInOpacity );
            fuelFlickerAnimation.setTag( FUEL_FLICKER_ANIMATION_TAG );

            LumosityUtils.runActionOnNode( this.view.fuel, fuelFlickerAnimation );
        },

        /**
         * Stops the fuel flicker animation.
         */
        stopFuelFlickerAnimation: function()
        {
            LumosityUtils.stopActionOnNodeByTag( this.view.fuel, FUEL_FLICKER_ANIMATION_TAG );
        },

        _createEngineRunningAnimation: function( moveToTopDuration, moveToBottomDuration, topPosition, bottomPosition )
        {
            var moveToTopAction = cc.MoveTo.create( moveToTopDuration, topPosition );
            var moveToBottomAction = cc.MoveTo.create( moveToBottomDuration, bottomPosition );
            var sequence = cc.Sequence.create( moveToTopAction, moveToBottomAction );
            var repeatAction = cc.RepeatForever.create( sequence );

            return repeatAction;
        },

        _createFuelRiseAnimation: function( moveToTopDuration, moveToEndDuration, topPosition, endPosition )
        {
            var moveToTopAction = cc.MoveTo.create( moveToTopDuration, topPosition );
            var moveToEndAction = cc.MoveTo.create( moveToEndDuration, endPosition );
            var sequence = cc.Sequence.create( moveToTopAction, moveToEndAction );

            return sequence;
        },

        _createFuelFallAnimation: function( moveToEndDuration, endPosition )
        {
            var moveToEndAction = cc.MoveTo.create( moveToEndDuration, endPosition );

            return moveToEndAction;
        },

        _createFuelFlickerAnimation: function( fadeOutDuration, fadeInDuration, fadeOutOpacity, fadeInOpacity )
        {
            var fadeOutAction = cc.FadeTo.create( fadeOutDuration, fadeOutOpacity );
            var fadeInAction = cc.FadeTo.create( fadeInDuration, fadeInOpacity );
            var sequence = cc.Sequence.create( fadeOutAction, fadeInAction );
            var repeatAction = cc.RepeatForever.create( sequence );

            return repeatAction;
        },
        
        _createSettings: function( settings )
        {
            this.settings = new VehicleSettings( settings );
        },
        
        _createModel: function()
        {
            this.model = new VehicleModel();
        },
        
        _createView: function()
        {
            this.view = new VehicleView();
            this.view.nodeName = "Vehicle";
            this.view.setAnchorPoint( 0.5, 0.5 );
        }
    } );
}() );
