/*global audio*/
/*global cc*/
/*global require*/

require( "src/Game/Base/Game.js" );
/*global Game*/
require( "src/Game/PinballRecallModel.js" );
/*global BoardCellType*/
/*global Dir*/
/*global PinballRecallModel*/
require( "src/Game/PinballRecallSettings.js" );
/*global PinballRecallSettings*/
require( "src/Game/PinballRecallUtils.js" );
/*global Point*/
require( "src/Game/PinballRecallView.js" );
/*global PinballRecallState*/
/*global PinballRecallView*/
require( "src/Utility/LumosityUtils.js" );
/*global lu*/

var PinballRecall;
( function()
{
    "use strict";

    PinballRecall = Game.extend(
    {
        _presentPauseTimer: 0.0,
        _buttonActiveDelayTimer: 0.0,
        _bumperBounceDelayTimer: 0.0,

        // Needed to prevent us from responding to the feedback animation finished callback,
        // which we'll get if the game is restarted during the feedback animation.
        _waitingOnFeedback: false,

        ctor: function()
        {
            this._super();
        },
        
        _createSettings: function()
        {
            return new PinballRecallSettings( "settings.json" );
        },
        
        _createModel: function()
        {
            return new PinballRecallModel( this.settings );
        },
        
        _createView: function( screenSize )
        {
            return new PinballRecallView( screenSize, this.settings, this.model, this );
        },

        onBegin: function()
        {
            this._super();

            this.view.getHudWidget().setTrial( this.model.trial, this.settings.getMaxTrials() );
            this.view.getHudWidget().setScore( this.model.score );

            this.view.setLevelCardValue( this.model.getGameLevel() );

            this.model.initHighestLevel();
        },

        onStartGameplay: function()
        {
            this._super();

            this.startNextTrial();
        },

        onUpdate: function( delta )
        {
            this._super( delta );

            // Update things even when the game isn't running.

            if ( this.model.isPlaying() )
            {
                if ( this._presentPauseTimer > 0.0 )
                {
                    this._presentPauseTimer -= delta;

                    if ( this._presentPauseTimer <= 0.0 )
                    {
                        this._presentPauseTimer = 0.0;

                        this.stateComplete( PinballRecallState.Present_Pause );
                    }
                }

                if ( this._buttonActiveDelayTimer > 0.0 )
                {
                    this._buttonActiveDelayTimer -= delta;

                    if ( this._buttonActiveDelayTimer <= 0.0 )
                    {
                        this._buttonActiveDelayTimer = 0.0;

                        this.stateComplete( PinballRecallState.Present_ButtonActiveDelay );
                    }
                }

                if ( this._bumperBounceDelayTimer > 0.0 )
                {
                    this._bumperBounceDelayTimer -= delta;

                    if ( this._bumperBounceDelayTimer <= 0.0 )
                    {
                        this._bumperBounceDelayTimer = 0.0;

                        this.onBumperBounceTimerFinished();
                    }
                }
            }
        },

        initTrialMetadata: function()
        {
            this.beginTrial();
        },

        finalizeTrialMetadata: function()
        {
            // Subtract the move from the last tile on the board to the target position.
            this.model.getTrialMetadata().path -= 1;

            this.endTrial( this.model.getTrialMetadata(), this.model.isSucceeded(), false, false, false, false );
        },

        stateComplete: function( state )
        {
            if ( state === PinballRecallState.Present_ShowBumpers )
            {
                if ( this.model.getHideBumpers() )
                {
                    this.goToState( PinballRecallState.Present_Pause );
                }
                else
                {
                    this.goToState( PinballRecallState.Present_ButtonActiveDelay );
                }
            }
            else if ( state === PinballRecallState.Present_Pause )
            {
                this.goToState( PinballRecallState.Present_HideBumpers );
            }
            else if ( state === PinballRecallState.Present_HideBumpers )
            {
                this.goToState( PinballRecallState.Present_ButtonActiveDelay );
            }
            else if ( state === PinballRecallState.Present_ButtonActiveDelay )
            {
                this.goToState( PinballRecallState.Input );
            }
            else if ( state === PinballRecallState.Input )
            {
                this.goToState( PinballRecallState.Simulate_GrowBall );
            }
            else if ( state === PinballRecallState.Simulate_GrowBall )
            {
                this.view.deactivateStartArrow();
                this.goToState( PinballRecallState.Simulate );
            }
            else if ( state === PinballRecallState.Simulate )
            {
                if ( !this.model.isFinished() )
                {
                    this.goToState( PinballRecallState.Simulate );
                }
                else
                {
                    this.finalizeTrialMetadata();

                    if ( this.view.getNumHiddenBumpers() > 0 )
                    {
                        this.goToState( PinballRecallState.Simulate_RevealBumpers );
                    }
                    else
                    {
                        this.goToShrinkBallOrFeedback();
                    }
                }
            }
            else if ( state === PinballRecallState.Simulate_RevealBumpers )
            {
                this.goToShrinkBallOrFeedback();
            }
            else if ( state === PinballRecallState.Simulate_ShrinkBall )
            {
                this.goToState( PinballRecallState.Simulate_Feedback );
            }
            else if ( state === PinballRecallState.Simulate_Feedback )
            {
                if ( this.model.isSucceeded() )
                {
                    this.goToState( PinballRecallState.Simulate_SuccessSequence );
                }
                else
                {
                    this.goToNextTrial();
                }
            }
            else if ( state === PinballRecallState.Simulate_SuccessSequence )
            {
                this.goToNextTrial();
            }
        },

        goToState: function( state )
        {
            if ( state === PinballRecallState.Present_ShowBumpers )
            {
                this.view.showAllBumpers( this );

                audio.playEffect( this.settings.bumperAppearSound );
            }
            else if ( state === PinballRecallState.Present_Pause )
            {
                // this.view.playAnimation( "presentPause", this.onPresentPauseDone );
                this._presentPauseTimer = this.settings.pauseTimeBetweenShowHideSec;
            }
            else if ( state === PinballRecallState.Present_HideBumpers )
            {
                this.view.hideAllBumpers( this );

                audio.playEffect( this.settings.bumperDisappearSound );
            }
            else if ( state === PinballRecallState.Present_ButtonActiveDelay )
            {
                this._buttonActiveDelayTimer = this.settings.pauseTimeBeforeButtonsActiveSec;
            }
            else if ( state === PinballRecallState.Input )
            {
                this.initTrialMetadata();

                this.view.enableInput();
                // this.model.setTrialStartTime( new Date() );
                this.view.turnOnStartArrow();
            }
            else if ( state === PinballRecallState.Simulate_GrowBall )
            {
                this.view.disableInput();
                this.view.makeGuessButtonStick();
                this.view.makeBall();
                this.view.growBall( this );
            }
            else if ( state === PinballRecallState.Simulate )
            {
                this.moveNext();
            }
            else if ( state === PinballRecallState.Simulate_RevealBumpers )
            {
                this.view.revealHiddenBumpers( this );

                audio.playEffect( this.settings.bumperAppearSound );
            }
            else if ( state === PinballRecallState.Simulate_ShrinkBall )
            {
                this.view.shrinkBall( this );
            }
            else if ( state === PinballRecallState.Simulate_Feedback )
            {
                this._waitingOnFeedback = true;
                this.view.showFeedback( this.model.isSucceeded(), this.model.getGuessPosition() );
                this.view.updateGuessButtonImage( this.model.isSucceeded() );
            }
            else if ( state === PinballRecallState.Simulate_SuccessSequence )
            {
                this.view.startSuccessSequence();
            }
        },

        goToShrinkBallOrFeedback: function()
        {
            if ( this.model.isSucceeded() )
            {
                this.goToState( PinballRecallState.Simulate_ShrinkBall );
            }
            else
            {
                this.goToState( PinballRecallState.Simulate_Feedback );
            }
        },

        onPresentPauseDone: function()
        {
            this.stateComplete( PinballRecallState.Present_Pause );
        },

        onButtonPressed: function( button )
        {
            var answer = button.boardPos;

            // var endPos = this.model.getEndPosition();
            // cc.log( "button pressed at " + answer.x + ", " + answer.y );
            // cc.log( "end is " + endPos.x + ", " + endPos.y );
            // if ( Point.equals( answer, endPos ) )
            // {
            //     // Correct answer.
            //     cc.log( ">> CORRECT" );
            // }
            // else
            // {
            //     // Incorrect answer.
            //     cc.log( ">> wrong" );
            // }

            // AUTOMATIC:
            // this.model.getTrialMetadata().response_time = (new Date()).getTime() - this.model.getTrialStartTime().getTime();

            this.model.setGuessPosition( answer, true );
            this.model.setCurrentPosition( this.model.getStartPosition() );

            this.stateComplete( PinballRecallState.Input );

            // this.view.deactivateStartArrow();
            // this.preparePinball();

            audio.playEffect( this.settings.ballRollSound );
        },

        moveNext: function()
        {
            lu.assert( !this.model.isFinished() );
            
            this.model.getTrialMetadata().path += 1;
            
            var fromX = this.model.getCurrentPosition().x;
            var fromY = this.model.getCurrentPosition().y;

            var direction = this.model.getPinballDirection();

            var toX = fromX + direction.x;
            var toY = fromY + direction.y;
            
            if ( this.model.getBoardCellType( cc.p( toX, toY ) ) === BoardCellType.BumperLeft )
            {
                this.model.setPinballDirection( cc.p( direction.y, direction.x ) );

                this.model.setBumperBouncePosition( cc.p( toX, toY ) );
                this.model.setBumperBounceDirection( direction );
                this.startBumperBounceTimer( this.settings.timeNeededForPinballToMoveToAnotherCell / 2 );
                
                this.model.getTrialMetadata().bounces += 1;
            }
            else if ( this.model.getBoardCellType( cc.p( toX, toY ) ) === BoardCellType.BumperRight )
            {
                this.model.setPinballDirection( cc.p( -1 * direction.y, -1 * direction.x ) );

                this.model.setBumperBouncePosition( cc.p( toX, toY ) );
                this.model.setBumperBounceDirection( direction );
                this.startBumperBounceTimer( this.settings.timeNeededForPinballToMoveToAnotherCell / 2 );
                
                this.model.getTrialMetadata().bounces += 1;
            }

            var maxX = this.model.getBoardSize().x + 1;
            var maxY = this.model.getBoardSize().y + 1;

            if ( toX === 0 || toX === maxX || toY === 0 || toY === maxY )
            {
                this.model.setFinished( true );

                this.model.setEndPosition( cc.p( toX, toY ), true );
                
                var correct = Point.equals( this.model.getGuessPosition(), this.model.getEndPosition() );

                if ( correct )
                {
                    var levelNumber = this.model.getGameLevel();
                    var gainedPoints = levelNumber * levelNumber * 10;
                    this.model.setSucceeded( true, true );
                    this.model.score += gainedPoints;

                    this.view.awardPoints( gainedPoints );
                }
                else
                {
                    this.model.setSucceeded( false, true );
                    this.model.setFailed( true );
                }
                
                audio.playEffect( this.settings.ballDropSound );
            }
            
            this.model.setCurrentPosition( cc.p( toX, toY ) );
            this.view.firePinball( fromX, fromY, toX, toY, this );
        },

        startBumperBounceTimer: function( duration )
        {
            this._bumperBounceDelayTimer = duration;
        },

        onBumperBounceTimerFinished: function()
        {
            this.view.showBumperBounce(
                this.model.getBumperBouncePosition(),
                this.model.getBumperBounceDirection()
            );
            
            audio.playEffect( this.settings.bumperBounceSound );
        },

        // onUpdateEachSecond: function()
        // {
        //     this._super();
        // },

        // onEndGameplay: function()
        // {
        //     this._super();
        // },

        // onFinish: function( wasQuit )
        // {
        //     this._super( wasQuit );
        // },

        // onPause: function()
        // {
        //     this._super();
        // },

        // onResume: function()
        // {
        //     this._super();
        // },

        onRestart: function()
        {
            this._presentPauseTimer = 0.0;
            this._buttonActiveDelayTimer = 0.0; 
            this._bumperBounceDelayTimer = 0.0;

            this._super();

            this._waitingOnFeedback = false;
        },

        feedback_isFinished: function( feedbackWidget )
        {
            // Needed to prevent us from responding to the feedback animation finished callback,
            // which we'll get if the game is restarted during the feedback animation.
            if ( this._waitingOnFeedback )
            {
                this._waitingOnFeedback = false;
                this.stateComplete( PinballRecallState.Simulate_Feedback );
            }
        },

        goToNextTrial: function()
        {
            // if ( _pinballRecallInstance->isRunning() )
            // {
            //     if ( _trials && _trials->getCurrentTrial() == _pinballRecallSettings->getTrialCount() - 1 )
            //     {
            //         // NOTE: trial numbers are 0 to max - 1
            //         _trials->goToNextTrial();
            //     }
            //     else
            //     {
                    // this.view.slideBoard( this.model.isSucceeded(), ( this.levelUpDue() || this.levelDownDue() ), this.onSlideFinished );
            //     }
            // }

            if ( this.model.trial < this.settings.getMaxTrials() )
            {
                this.model.trial += 1;
                this.view.getHudWidget().setTrial( this.model.trial, this.settings.getMaxTrials() );

                var currentLevelNumber = this.model.getGameLevel();
                var nextLevelNumber = currentLevelNumber;
                if ( this.levelUpDue() )
                {
                    nextLevelNumber += 1;
                }
                else if ( this.levelDownDue() )
                {
                    nextLevelNumber -= 1;
                }

                this.view.slideBoard( this.model.isSucceeded(), ( this.levelUpDue() || this.levelDownDue() ), nextLevelNumber, this.onSlideFinished );
            }
            else
            {
                // this.model.saveLevelInfoForStats( this.settings );
                
                this.onEndGameplay();
            }
        },

        onSlideFinished: function()
        {
            this.startNextTrial();
        },

        startNextTrial: function()
        {
            this.goToState( PinballRecallState.Present_ShowBumpers );
        },

        cue_generateBoardView: function()
        {
            this.advanceToNextTrial();
        },

        advanceToNextTrial: function()
        {
            if ( this.levelUpDue() )
            {
                // cc.log( "level up" );

                this.model.setGameLevel( this.model.getGameLevel() + 1 );
                
                this.model.metadata.levels_up += 1;
            }
            else if ( this.levelDownDue() )
            {
                // cc.log( "level down" );

                this.model.setGameLevel( this.model.getGameLevel() - 1 );
                
                this.model.metadata.levels_down += 1;
            }
            
            this.setupNextTrial();

            // cc.log( "this.model.getGameLevel() = " + this.model.getGameLevel() );
            // cc.log( "this.settings.successfulGamesToAdvanceLevel = " + this.settings.successfulGamesToAdvanceLevel );
            // cc.log( "this.model.getSuccessfulTrialsAtThisLevel() = " + this.model.getSuccessfulTrialsAtThisLevel() );
        },

        levelUpDue: function()
        {
            if ( this.model.getSuccessfulTrialsAtThisLevel() === this.settings.successfulGamesToAdvanceLevel )
            {
                if ( this.settings.getLevel( this.model.getGameLevel() + 1 ) )
                {
                    return true;
                }
            }
            
            return false;
        },

        levelDownDue: function()
        {
            if ( this.model.getGameLevel() > 1 )
            {
                if ( this.model.getFailedTrialsAtThisLevel() === this.settings.successfulGamesToAdvanceLevel )
                {
                    return true;
                }
            }
            
            return false;
        },

        setupNextTrial: function()
        {
            var level = this.settings.getLevel( this.model.getGameLevel() );

            // DEBUG:
            // level = this.settings.__getRandomLevel();

            lu.assert( level !== null && level !== undefined );

            var width = level.width;
            var height = level.height;
            var bumperCount = level.bumperCount;
            var bumperDecoyMin = level.bumperDecoyMin;
            var bumperDecoyMax = level.bumperDecoyMax;
            var hideBumpers = level.hideBumpers;

            var bumperDecoyCount = 0;
            
            // Figure out how many bumpers should be in the path.
            this.model.setBumpersTotal( bumperCount );

            if ( bumperCount > 0 )
            {
                if ( bumperDecoyMax > 0 && bumperDecoyMin <= bumperDecoyMax )
                {
                    bumperDecoyCount = lu.getRandomInt( bumperDecoyMin, bumperDecoyMax );
                    
                    if ( bumperDecoyCount <= bumperCount )
                    {
                        this.model.setBumpersOnPath( bumperCount - bumperDecoyCount );
                    }
                }
                else
                {
                    this.model.setBumpersOnPath( bumperCount );
                }
            }
            
            // Set up the model.
            this.model.startNextTrial();
            
            this.model.setAllowInput( false );
            this.model.setFinished( false );
            this.model.setSucceeded( false );
            this.model.setFailed( false );
            this.model.setUnusedBumpersRevealed( false );
            // Don't do this now, board generation will do it.
            // this.model.setBoardSize( cc.p( width, height ) );
            this.model.setHideBumpers( hideBumpers );

            this.model.setEndPosition( cc.p( 0, 0 ) );
            this.model.setGuessPosition( cc.p( 0, 0 ) );
            
            this.model.setCurrentPosition( cc.p( 0, 0 ) );

            // Don't do this now, path generation will do it.
            // this.model.setStartPosition( cc.p( 0, 0 ) );
            // this.model.setPinballDirection( cc.p( 0, 0 ) );

            this.model.generateBoard( width, height );
            this.generatePath( width, height, bumperCount, this.model.getBumpersOnPath() );
            this.placeDecoyBumpers( bumperDecoyCount );

            this.view.generateBoard();

            // this.model.getTrialMetadata().index = this.model.trial;
            this.model.getTrialMetadata().level = this.model.getGameLevel();
            this.model.getTrialMetadata().grid = this.model.getBoardAsString();
            this.model.getTrialMetadata().mirror_count = bumperCount;
            this.model.getTrialMetadata().decoy_count = bumperDecoyCount;
            this.model.getTrialMetadata().bounces = 0;
            // this.model.getTrialMetadata().startPosition = this.model.getStartPosition().toString();
            // this.model.getTrialMetadata().endPosition = this.model.getEndPosition().toString();
            this.model.getTrialMetadata().width = width;
            this.model.getTrialMetadata().height = height;
            this.model.getTrialMetadata().path = 0;

            this.view.disableInput();

            // cc.log( "trial is ready" );
        },

        generatePath: function( width, height, bumperCount, bumperQuota )
        {
            this.model.setBumpersPlaced( 0 );
            this.model.setBumpersNeeded( bumperQuota );
            this.model.setBumperPathGenerationFailures( 0 );
            this.model.setBumperPathGenerationSuccesses( 0 );
            
            // double startTime = ofxGenericDate::now()->getTime();
            // var startTime = new Date();

            this.model.__path = [];
            this.model.__path.push( this.model.getStartPosition() );
            
            var nextPos = Point.add( this.model.getStartPosition(), this.model.getPinballDirection() );
            var result = this.generatePathRecursive( nextPos, this.model.getPinballDirection() );
            lu.assert( result, "Failed to generate a path." );

            // if ( this.generatePathRecursive( nextPos, this.model.getPinballDirection() ) )
            // {
            //     // LumosLogVerbose( "PinballRecall::generatePath2 : Path is successfully generated." );
            // }
            // else
            // {
            //     // LumosLogError( "PinballRecall::generatePath2 : Failed to generate path." );
            // }
            
            // LumosLogVerbose( "PinballRecall::generatePath2 : Successes: " + ofxGToString( this.model.getBumperPathGenerationSuccesses() ) + ", Failures: " + ofxGToString( this.model.getBumperPathGenerationFailures() ) );
            // LumosLogVerbose( "PinballRecall::generatePath2 : Path generation took " + ofxGToString( ofxGenericDate::now()->getTime() - startTime ) + " seconds." );
        },

        generatePathRecursive: function( pos, dir )
        {
            // LumosLogError( "Dir: " + ofxGToString( dir.x ) + ", " + ofxGToString( dir.y ) );
            // LumosLogError( "Pos: " + ofxGToString( pos.x ) + ", " + ofxGToString( pos.y ) );
            
            var nextPos = Point.add( pos, dir );
            var newDirRight = this.getNewDirection( dir, BoardCellType.BumperRight );
            var nextPosRight = Point.add( pos, newDirRight );
            var newDirLeft = this.getNewDirection( dir, BoardCellType.BumperLeft );
            var nextPosLeft = Point.add( pos, newDirLeft );
            
            var cellType = this.model.getBoardCellType( pos );
            var onPath = this.model.isBoardCellOnPath( pos );

            this.model.__path.push( pos );

            var result = false;

            // RECURSION TERMINATION CONDITION: Off the edge of the board.
            if ( cellType === BoardCellType.Drain )
            {
                // Haven't placed all the bumpers we require. Fail.
                if ( this.model.getBumpersPlaced() < this.model.getBumpersNeeded() )
                {
                    this.model.incrementBumperPathGenerationFailures();
                    result = false;//return false;
                }
                // All required bumpers are down. Success.
                else
                {
                    this.model.setEndPosition( pos, true );
                    this.model.incrementBumperPathGenerationSuccesses();
                    result = true;//return true;
                }
            }
            // Path is self-intersecting.
            else if ( onPath )
            {
                // Continue straight.
                if ( cellType === BoardCellType.Empty )
                {
                    if ( this.generatePathRecursive( nextPos, dir ) )
                    {
                        this.model.incrementBumperPathGenerationSuccesses();
                        result = true;//return true;
                    }
                    else
                    {
                        this.model.incrementBumperPathGenerationFailures();
                        result = false;//return false;
                    }
                }
                // Continue, turn right.
                else if ( cellType === BoardCellType.BumperRight )
                {
                    if ( this.generatePathRecursive( nextPosRight, newDirRight ) )
                    {
                        this.model.incrementBumperPathGenerationSuccesses();
                        result = true;//return true;
                    }
                    else
                    {
                        this.model.incrementBumperPathGenerationFailures();
                        result = false;//return false;
                    }
                }
                // Continue, turn left.
                else if ( cellType === BoardCellType.BumperLeft )
                {
                    if ( this.generatePathRecursive( nextPosLeft, newDirLeft ) )
                    {
                        this.model.incrementBumperPathGenerationSuccesses();
                        result = true;//return true;
                    }
                    else
                    {
                        this.model.incrementBumperPathGenerationFailures();
                        result = false;//return false;
                    }
                }
                else
                {
                    this.model.incrementBumperPathGenerationFailures();
                    result = false;//return false;
                }
            }
            // Traverse through this cell, favor leaving it blank.
            else if ( lu.getRandomInt( 0, 100 ) < this.settings.blankProbability )
            {
                this.model.setBoardCellOnPath( pos, true );

                if ( this.generatePathRecursive( nextPos, dir ) )
                {
                    this.model.incrementBumperPathGenerationSuccesses();
                    result = true;//return true;
                }
                // Didn't complete the path leaving this cell blank,
                // and we have bumpers to place anyway, so try putting one down.
                else if ( this.model.getBumpersPlaced() < this.model.getBumpersNeeded() )
                {
                    // Place a right bumper.
                    if ( lu.getRandomInt( 0, 100 ) > 50 )
                    {
                        this.model.setBoardCellType( pos, BoardCellType.BumperRight );
                        this.model.incrementBumpersPlaced();
                        
                        // Turn right and continue.
                        if ( this.generatePathRecursive( nextPosRight, newDirRight ) )
                        {
                            this.model.incrementBumperPathGenerationSuccesses();
                            result = true;//return true;
                        }
                        // Didn't work. Change directions and go left.
                        else
                        {
                            this.model.setBoardCellType( pos, BoardCellType.BumperLeft );

                            if ( this.generatePathRecursive ( nextPosLeft, newDirLeft ) )
                            {
                                this.model.incrementBumperPathGenerationSuccesses();
                                result = true;//return true;
                            }
                            else
                            {
                                this.model.setBoardCellType( pos, BoardCellType.None );
                                this.model.setBoardCellOnPath( pos, false );
                                this.model.decrementBumpersPlaced();
                                this.model.incrementBumperPathGenerationFailures();
                                result = false;//return false;
                            }
                        }
                    }
                    // Place a left bumper.
                    else
                    {
                        this.model.setBoardCellType( pos, BoardCellType.BumperLeft );
                        this.model.setBoardCellOnPath( pos, true );
                        this.model.incrementBumpersPlaced();
                        
                        // Turn left and continue.
                        if ( this.generatePathRecursive( nextPosLeft, newDirLeft ) )
                        {
                            this.model.incrementBumperPathGenerationSuccesses();
                            result = true;//return true;
                        }
                        // Didn't work. Change directions and go right.
                        else
                        {
                            this.model.setBoardCellType( pos, BoardCellType.BumperRight );

                            if ( this.generatePathRecursive( nextPosRight, newDirRight ) )
                            {
                                this.model.incrementBumperPathGenerationSuccesses();
                                result = true;//return true;
                            }
                            else
                            {
                                this.model.setBoardCellType( pos, BoardCellType.None );
                                this.model.setBoardCellOnPath( pos, false );
                                this.model.decrementBumpersPlaced();
                                this.model.incrementBumperPathGenerationFailures();
                                result = false;//return false;
                            }
                        }
                    }
                }
                // Failed going straight, and no bumpers remaining to place. Bummer.
                else
                {
                    this.model.setBoardCellOnPath( pos, onPath );
                    this.model.incrementBumperPathGenerationFailures();
                    result = false;//return false;
                }
            }
            // We have more bumpers to place.
            else if ( this.model.getBumpersPlaced() < this.model.getBumpersNeeded() )
            {
                // Place a right bumper.
                if ( lu.getRandomInt( 0, 100 ) > 50 )
                {
                    this.model.setBoardCellType( pos, BoardCellType.BumperRight );
                    this.model.setBoardCellOnPath( pos, true );
                    this.model.incrementBumpersPlaced();
                    
                    // Success.
                    if ( this.generatePathRecursive( nextPosRight, newDirRight ) )
                    {
                        this.model.incrementBumperPathGenerationSuccesses();
                        result = true;//return true;
                    }
                    // Did not complete.
                    else
                    {
                        // Change facing.
                        this.model.setBoardCellType( pos, BoardCellType.BumperLeft );

                        // Try again.
                        if ( this.generatePathRecursive( nextPosLeft, newDirLeft ) )
                        {
                            this.model.incrementBumperPathGenerationSuccesses();
                            result = true;//return true;
                        }
                        else
                        {
                            // Fuck it, don't place a bumper.
                            this.model.setBoardCellType( pos, BoardCellType.None );
                            this.model.decrementBumpersPlaced();

                            // Last try.
                            if ( this.generatePathRecursive( nextPos, dir ) )
                            {
                                this.model.incrementBumperPathGenerationSuccesses();
                                result = true;//return true;
                            }
                            else
                            {
                                this.model.setBoardCellOnPath( pos, false );
                                this.model.incrementBumperPathGenerationFailures();
                                result = false;//return false;
                            }
                        }
                    }
                }
                else
                {
                    // Place a left bumper.
                    this.model.setBoardCellType( pos, BoardCellType.BumperLeft );
                    this.model.setBoardCellOnPath( pos, true );
                    this.model.incrementBumpersPlaced();
                    
                    // Success.
                    if ( this.generatePathRecursive( nextPosLeft, newDirLeft ) )
                    {
                        this.model.incrementBumperPathGenerationSuccesses();
                        result = true;//return true;
                    }
                    // Did not complete.
                    else
                    {
                        // Change facing.
                        this.model.setBoardCellType( pos, BoardCellType.BumperRight );

                        // Try again.
                        if ( this.generatePathRecursive( nextPosRight, newDirRight ) )
                        {
                            this.model.incrementBumperPathGenerationSuccesses();
                            result = true;//return true;
                        }
                        else
                        {
                            // Fuck it, don't place a bumper.
                            this.model.setBoardCellType( pos, BoardCellType.None );
                            this.model.decrementBumpersPlaced();

                            // Last try.
                            if ( this.generatePathRecursive( nextPos, dir ) )
                            {
                                this.model.incrementBumperPathGenerationSuccesses();
                                result = true;//return true;
                            }
                            else
                            {
                                this.model.setBoardCellOnPath( pos, false );
                                this.model.incrementBumperPathGenerationFailures();
                                result = false;//return false;
                            }
                        }
                    }
                }
            }
            // No bumpers left to place. Just keep going.
            else
            {
                this.model.setBoardCellOnPath( pos, true );

                // Keep going.
                if ( this.generatePathRecursive( nextPos, dir ) )
                {
                    this.model.incrementBumperPathGenerationSuccesses();
                    result = true;//return true;
                }
                else
                {
                    this.model.setBoardCellOnPath( pos, onPath );
                    this.model.incrementBumperPathGenerationFailures();
                    result = false;
                }
            }

            if ( !result )
            {
                this.model.__path.pop();
            }

            return result;//return false;
        },

        getNewDirection: function( dir, bumperType )
        {
            var retval = dir;
            
            if ( bumperType === BoardCellType.BumperRight )
            {
                if ( Point.equals( dir, Dir.Up ) )
                {
                    retval = Dir.Left;
                }
                else if ( Point.equals( dir, Dir.Down ) )
                {
                    retval = Dir.Right;
                }
                else if ( Point.equals( dir, Dir.Left ) )
                {
                    retval = Dir.Up;
                }
                else if ( Point.equals( dir, Dir.Right ) )
                {
                    retval = Dir.Down;
                }
            }
            else if ( bumperType === BoardCellType.BumperLeft )
            {
                if ( Point.equals( dir, Dir.Up ) )
                {
                    retval = Dir.Right;
                }
                else if ( Point.equals( dir, Dir.Down ) )
                {
                    retval = Dir.Left;
                }
                else if ( Point.equals( dir, Dir.Left ) )
                {
                    retval = Dir.Down;
                }
                else if ( Point.equals( dir, Dir.Right ) )
                {
                    retval = Dir.Up;
                }
            }

            lu.assert(
                Point.equals( retval, Dir.Down ) ||
                Point.equals( retval, Dir.Up ) ||
                Point.equals( retval, Dir.Right ) ||
                Point.equals( retval, Dir.Left ),
                "bad retval: " + JSON.stringify( retval ) );
            
            return retval;
        },

        placeDecoyBumpers: function( decoyBumperCount )
        {
            var i;
            var pos;
            for ( i = 0; i < decoyBumperCount; i += 1 )
            {
                pos = this.model.getRandomBoardPositionNotOnPath();

                if ( !Point.equals( pos, cc.p( -1, -1 ) ) && !Point.equals( pos, cc.p( 0, 0 ) ) )
                {
                    if ( lu.getRandomInt( 0, 100 ) < 50 )
                    {
                        this.model.setBoardCellType( pos, BoardCellType.BumperRight );
                        
                        // LumosLogError( "Placed decoy right bumper at " + ofxGToString( (int)pos.x ) + "," + ofxGToString( (int)pos.y ) );
                    }
                    else
                    {
                        this.model.setBoardCellType( pos, BoardCellType.BumperLeft );
                        
                        // LumosLogError( "Placed decoy left bumper at " + ofxGToString( (int)pos.x ) + "," + ofxGToString( (int)pos.y ) );
                    }
                }
                // else
                // {
                //     LumosLogError( "PinballRecall::placeDecoyBumers : Could not find location to place decoy bumper." );
                // }
            }
        }
    } );
}() );
