/*global audio*/
/*global cc*/
/*global ll*/
/*global require*/

require( "src/Game/Base/Game.js" );
/*global Game*/
require( "src/Game/RaindropsModel.js" );
/*global RaindropsModel*/
/*global RaindropsResponseType*/
/*global RaindropsTrialType*/
require( "src/Game/RaindropsSettings.js" );
/*global RaindropsSettings*/
require( "src/Game/RaindropsSun.js" );
/*global RaindropsSun*/
require( "src/Game/RaindropsUtils.js" );
/*global RaindropsUtils*/
require( "src/Game/RaindropsView.js" );
/*global RaindropsView*/
require( "src/Game/Widget/GameWidgetHud.js" );
/*global GameWidgetHud*/
require( "src/Utility/LumosityUtils.js" );
/*global lu*/

var Raindrops;
( function()
{
    "use strict";

    Raindrops = Game.extend(
    {
        /** @lends Raindrops.prototype */

        __autoplay: false,

        _spawnDropletsAndClouds: false,
        
        /**
        *  Raindrops is the main class for the game, Raindrops!
        *
        *  @constructs
        */
        ctor: function()
        {
            this._super();
        },
        
        _createSettings: function()
        {
            return new RaindropsSettings( "settings.json" );
        },
        
        _createModel: function()
        {
            return new RaindropsModel();
        },
        
        _createView: function( screenSize )
        {
            return new RaindropsView( screenSize, this.settings, this.model, this );
        },

        onBegin: function()
        {
            this._super();

            this.model.raindropsSettings = this.settings;
            this.model.raindropsView = this.view;

            this.view.game = this;

            this.view.setWaterLevel( 0, false );

            this.updateHudScore();

            this.clearValue();

            this.view.onBegin();
        },

        onStartGameplay: function()
        {
            this._super();
            
            this.model.beginTrial();
            this.beginTrial();

            this.view.onStartGameplay();
        },

        onUpdate: function( delta )
        {
            this._super( delta );

            var deltaTime = delta;

            // Always update, even before onStartGameplay(), so the surface animates with the water rise animation.
            this.view.onUpdateAlways( delta );

            if ( this.model.isGameRunning )
            {
                if ( !this.model.isGamePaused )
                {
                    this.model.updateDroplets( deltaTime );
                    
                    // This bursts droplets if any are beyond the screen
                    this.checkDropletsStates();

                    if ( this._spawnDropletsAndClouds )
                    {
                        if ( this.model.getWaterLevel() <= this.settings.getMaximumWaterLevel() )
                        {
                            this.spawnNewDroplets( deltaTime );
                        }
                    }
                }
                
                if ( this._spawnDropletsAndClouds )
                {
                    // Check if there are enough clouds
                    var cloudsNeeded = ( this.model.getWaterLevel() + 1 ) * this.settings.getCloudsPerWaterLevel();
                    var cloudCount = this.view.getCloudCount();
                    var i;
                    for ( i = cloudCount; i < cloudsNeeded; i += 1 )
                    {
                        this.view.addCloud();
                    }
                }
                
                this.view.onUpdate( deltaTime, this.model.isGamePaused );
            }
        },

        gameOn: function()
        {
            this._spawnDropletsAndClouds = true;
        },

        onUpdateEachSecond: function()
        {
            this._super();
        },

        gameOver: function()
        {
            this.onEndGameplay();
        },

        /**
         *  Called when the main gameplay in a game has ended. At this point, results will be sent to the C++ layer
         *  so that they can be sent to the server, and the gameEnd animation will begin playing.
         *  No user input will be accepted after this event.
         */
        onEndGameplay: function()
        {
            this.view.onEndGameplay();
            
            this._super();
        },

        /**
         *  Called when a Game has completely finished. This could be because the gameplay has naturally ended
         *  and the gameEnd animation has finished, or it could be because the user prematurely exited the game.
         *  Regardless, a Game will always go through this function when it's being exited.
         *
         *  @param {boolean} wasQuit If true, the game was prematurely ended.
         */
        onFinish: function( wasQuit )
        {
            this._super( wasQuit );
        },

        /**
         * Called when a Game has been paused.
         */
        onPause: function()
        {
            this._super();

            ll.verbose("Raindrops", "onPause");
        },

        /**
         * Called when a Game has been unpaused.
         */
        onResume: function()
        {
            this._super();

            ll.verbose("Raindrops", "onResume");
        },

        backspace: function()
        {
            this.model.getMetadata().num_corrections += 1;
            this.model.getTrialMetadata().num_corrections += 1;

            this.clearValue();
        },

        clearValue: function()
        {
            this.model.setInputtedValue( 0 );
            this.view.clearValue();
        },

        checkDropletsStates: function()
        {
            // We need to splash droplets at the water level
            var splashHeight = this.view.getDropletSplashHeight();
            var dropletsToSplash = this.model.getDropletsBelow( splashHeight );

            // @todo assert( dropletsToSplash.length === 0 || dropletsToSplash.length === 1 );

            if ( dropletsToSplash.length > 0 )
            {
                if ( this.__autoplay )
                {
                    var dropletToBurst = this.model.getLowestDroplet();

                    if ( dropletToBurst !== null )
                    {
                        this.burstDroplets( [ dropletToBurst ], dropletToBurst.getBurstSFX() );
                    }

                    return;
                }

                this.splashDroplets( dropletsToSplash );
            }
        },

        generateTrialMetaData: function( success, clearedByWater, droplet )
        {
            var clearedBySun = false;

            if ( success && droplet )
            {
                if ( droplet instanceof RaindropsSun )
                {
                    clearedBySun = true;
                }
            }

            var trial = this.model.getTrialMetadata();

            // Will be null if:
            // * in debug mode and 555 is entered
            // * an incorrect answer is entered
            if ( droplet )
            {
                // Can use bitwise OR to force cast to signed integer, but JSLint rules forbid it.
                trial.horizontal = Math.floor( droplet._dropletView.getPositionX() ); // | 0
                trial.vertical = Math.floor( droplet._dropletView.getPositionY() ); // | 0
            }
            // else
            // {
            //     trial.horizontal = 0;
            //     trial.vertical = 0;
            // }

            if ( success )
            {
                if ( !clearedBySun )
                {
                    trial.type = RaindropsTrialType.RegularProblem;
                }
                else
                {
                    trial.type = RaindropsTrialType.Sun;
                }
            }
            else
            {
                trial.type = RaindropsTrialType.IncorrectResponse;
            }
            
            if ( success )
            {
                if ( !clearedBySun )
                {
                    trial.response = RaindropsResponseType.Correct;
                }
                else
                {
                    trial.response = RaindropsResponseType.ClearedBySun;
                }
            }
            else
            {
                if ( droplet )
                {
                    trial.response = RaindropsResponseType.Missed;
                }
                else
                {
                    trial.response = RaindropsResponseType.ClearedByWaterOrEndOfGame;
                }
            }

            trial.correct = success ? 'T' : 'F';
            
            // equation - equation (blank for incorrect response)
            if ( success )
            {
                if ( droplet )
                {
                    trial.equation = droplet.getEquation().toString();
                }
                else
                {
                    trial.equation = "";
                }
            }
            else
            {
                trial.equation = "";
            }
            
            // time_offset - number of milliseconds since start of game, appearance time for problems or response time for incorrect responses
            // var now = new Date();
            // var now_ms = now.getTime();
            // var creationTime_ms = trial.getCreationTimeMs();
            // var timeOffset_ms = now_ms - creationTime_ms;
            // trial.time_offset = timeOffset_ms;
            
            // response_time - number milliseconds since problem appearance when problem cleared by answer, sun, or water hit; blank for incorrect response
            // if ( !droplet )
            // {
            //     trial.response_time = creationTime_ms;
            // }
            // else
            // {
            //     var spawnTime = droplet.getCreationTime();
            //     var spawnTime_ms = spawnTime * 1000;
            //     var responseTime_ms = now_ms - spawnTime_ms;
            //     trial.response_time = responseTime_ms;
            // }
            
            // speed - Speed of raindrop (or current speed for incorrect responses)
            if ( success && droplet )
            {
                trial.speed = droplet.getFallRate();
            }
            else
            {
                trial.speed = this.model.getDropletFallSpeed();
            }
            
            // water_level - Current water level
            trial.water_level = this.model.getWaterLevel();
            
            // num_drops - Number of drops on-screen
            trial.num_drops = this.model.getDropletCount();

            if ( trial.num_corrections > 0 )
            {
                this.model.getMetadata().num_tries_corrections += 1;
            }
        },

        splashDroplets: function( dropletsToSplash )
        {
            // Grab the first droplet and generate trial meta data based on it
            this.generateTrialMetaData( false, true, dropletsToSplash[ 0 ] );
            
            var i;
            var droplet;
            for ( i = 0; i < dropletsToSplash.length; i += 1 )
            {
                droplet = dropletsToSplash[ i ];

                if ( droplet )
                {
                    this.splashDroplet( droplet );
                }
            }

            // Once the droplets hits the water, all others need to be burst
            this.burstAllDroplets();
            
            // Raise the water level
            this.increaseWaterLevel();
            
            // Reset input box
            this.clearValue();
            
            // Update trial state
            this.updateTrialState();
        },

        splashDroplet: function( droplet )
        {
            if ( droplet )
            {
                this.model.removeDroplet( droplet );
                droplet.splash();

                if ( droplet instanceof RaindropsSun )
                {
                    this.view.setSunIsOut( false );
                }
            }
        },

        burstAllDroplets: function()
        {
            while ( this.model.getDropletFront() )
            {
                this.burstDroplet( this.model.getDropletFront() );
            }
        },

        burstDroplet: function( droplet, playSound )
        {
            if ( droplet )
            {
                this.model.removeDroplet( droplet );

                droplet.burst();

                if ( droplet instanceof RaindropsSun )
                {
                    this.view.setSunIsOut( false );
                }
            }
        },

        increaseWaterLevel: function()
        {
            this._spawnDropletsAndClouds = false;

            this.model.incrementWaterLevel();
            var level = this.model.getWaterLevel();
            this.view.setWaterLevel( level, true );

            if ( level >= this.settings.getDisablePauseAtWaterLevel() )
            {
                this.view.disablePauseButton();
            }
        },

        updateTrialState: function()
        {
            var trial = this.model.getTrialMetadata();
            var correct = trial.correct === 'T';
            this.endTrial( trial, correct, false, false, false, false );
            this.model.endTrial( correct );

            this.model.beginTrial();
            this.beginTrial();
        },

        spawnNewDroplets: function( deltaTime )
        {
            var currTime = new Date();
            var now_ms = currTime.getTime();
            var now_sec = now_ms / 1000.0;
            
            var spawnTime = this.model.getDropletCreateFrequency();
            var spawnTimerPassed = now_sec > this.model.getLastDropletCreateTime() + spawnTime;
            
            var dropletCount = this.model.getDropletCount();
            var maximumDropletCount = this.model.getAllowedDropletCount();
            var maxDropletCountOut = dropletCount >= maximumDropletCount;
            
            if ( ( spawnTimerPassed && !maxDropletCountOut ) || dropletCount === 0 )
            {
                var dropletFallSpeed = this.model.getDropletFallSpeed();

                var droplet = null;

                var equation = this.model.createEquation();
                
                // TODO: droplet pool
                if ( this.model.shouldSpawnSun() )
                {
                    //@todo clean up creation interface
                    droplet = this.view.constructSun( now_sec, equation, dropletFallSpeed );
                }
                else
                {
                    //@todo clean up creation interface
                    droplet = this.view.constructDroplet( now_sec, equation, dropletFallSpeed );
                }

                this.view.addDropletAtRandomX( droplet );
                this.model.addDroplet( droplet );

                this.model.setLastDropletCreateTime( now_sec );

                //@portme
                // if ( gameResultMetadata )
                // {
                //     gameResultMetadata->incrementNumberOfProblems();
                //     gameResultMetadata->setFinalSpeed( dropletFallSpeed );
                // }
            }
        },

        testValue: function()
        {
            var inputtedValue = this.model.getInputtedValue();

            if ( lu.isDebugModeOn() )
            {
                if ( inputtedValue === 555 )
                {
                    if ( this.__autoplay )
                    {
                        var dropletToSplash = this.model.getLowestDroplet();

                        this.splashDroplets( [ dropletToSplash ] );
                    }
                    else
                    {
                        this.generateTrialMetaData( false, false, null );

                        // Once the droplets hits the water, all others need to be burst
                        this.burstAllDroplets();
                        
                        // Raise the water level
                        this.increaseWaterLevel();
                        
                        // Reset input box
                        this.clearValue();
                        
                        // Update trial state
                        this.updateTrialState();
                    }
                }
                else
                {
                    var dropletToBurst = this.model.getLowestDroplet();

                    if ( dropletToBurst !== null )
                    {
                        this.burstDroplets( [ dropletToBurst ], dropletToBurst.getBurstSFX() );
                    }
                }
            }
            else
            {
                var answers = this.model.getDropletsWithAnswer( inputtedValue );

                var hasCorrectAnswers = answers.length > 0;
                
                if ( hasCorrectAnswers )
                {
                    var burstSFX = this.settings.dropletBurstSFX;

                    if ( RaindropsUtils.containsSun( answers ) )
                    {
                        answers = this.model.getDroplets();
                        burstSFX = this.settings.sunBurstSFX;
                    }

                    this.burstDroplets( answers, burstSFX );
                }
                else
                {
                    this.generateTrialMetaData( false, false, null );
                    
                    this.view.setIncorrectAnswerEntered();

                    audio.playEffect( "sounds/answerWrong.mp3" );
                }
            }
            
            this.updateTrialState();
        },

        burstDroplets: function( droplets, burstSFX )
        {
            this.generateTrialMetaData( true, false, droplets[ 0 ] );

            audio.playEffect( burstSFX );
            
            var droplet;
            var scorePointsToAdd;
            while ( droplets.length > 0 )
            {
                droplet = droplets[ 0 ];
                droplets.splice( 0, 1 );
                
                this.burstDroplet( droplet );
                
                scorePointsToAdd = this.settings.getRaindropScoreValue();
                // this.model.getTrialMetadata().addPoints( scorePointsToAdd );
                this.model.score = this.model.score + scorePointsToAdd;
                this.updateHudScore();
            }
            
            this.clearValue();

            audio.playEffect( "sounds/answerCorrect.mp3" );
        },

        updateHudScore: function()
        {
            if ( this.view._widgets )
            {
                var hudWidget = this.view._widgets[ GameWidgetHud.prototype.getType() ];
                if ( hudWidget )
                {
                    hudWidget.setScore( this.model.score );
                    hudWidget.setScoreHighlighted( true );
                }
            }
        },

        inputValue: function( value )
        {
            var currValue = this.model.getInputtedValue();
            var newValue = currValue.toString() + value.toString();
            if ( newValue.length <= 3 )
            {
                this.setInputtedValue( parseInt( newValue, 10 ) );
            }
        },

        setInputtedValue: function( value )
        {
            this.model.setInputtedValue( value );
            this.view.setValue( this.model.getInputtedValue() );
        },

        /**
         * Called when the Game has been restarted within JS.
         * By default, this will re-create the model, stop all
         * animations, then set everything as in the layout and
         * call gameBegin. Not at all guaranteed to work out of the box.
         */
        onRestart: function()
        {
            this.model.destroyAllDroplets();
            this._super();
        }
    } );
}() );
