/*global audio*/
/*global cc*/
/*global ll*/
/*global require*/

require( "src/Game/RaindropsEquation.js" );
/*global Operator*/
require( "src/Game/RaindropsUtils.js" );
/*global RaindropsUtils*/
require( "src/Utility/LumosityUtils.js" );
/*global lu*/

var RaindropsDroplet;
(function()
{
    "use strict";

    RaindropsDroplet = cc.Class.extend(
    {
        /** @lends RaindropsDroplet.prototype */

        _dropletView: null,
        _sprite: null,
        _operandTop: null,
        _operandBottom: null,
        _operator: null,

        _fallRate: 0.0,
        _creationTime: null,
        _burstSFX: null,

        _equation: null,

        _operatorPadding: 0,

        // Can probably retire these two:
        _splashed: false,
        _burst: false,
        
        /**
         *  RaindropsDroplet is the main class for a droplet!
         *
         *  @constructs
         */
        ctor: function( operatorPadding )
        {
            this._operatorPadding = operatorPadding;
        },

        getIdleAnimName: function()
        {
            return "";
        },

        getBurstAnimName: function()
        {
            return "droplet_burstAnim";
        },

        getSplashAnimName: function()
        {
            return "droplet_splashAnim";
        },

        init: function( equation, fallRate )
        {
            // Do nothing.
        },

        begin: function()
        {
            // Do nothing.
        },

        getCreationTime: function()
        {
            return this._creationTime;
        },

        getFallRate: function()
        {
            return this._fallRate;
        },

        getBurstSFX: function()
        {
            return this._burstSFX;
        },

        update: function( deltaTime )
        {
            var oldY = this._dropletView.getPositionY();
            var delta = this._fallRate * deltaTime;
            //@todo might not want the minus in here, might want it coming from the speed/rate instead
            var newY =  oldY - delta;
            this._dropletView.setPositionY( newY );
        },

        getViewBottomY: function()
        {
            var spriteFrame = RaindropsUtils.getFrame( this._dropletView );
            var heightBelowPosition = this._dropletView.getAnchorPoint().y * spriteFrame.height;
            var bottomY = spriteFrame.y - heightBelowPosition;

            return bottomY;
        },

        getLifeDuration: function()
        {
            return new Date() - this._creationTime;
        },

        burst: function()
        {
            if ( this._burst || this._splashed )
            {
                return;
            }

            this._burst = true;

            this._operandTop.setVisible( false );
            this._operandBottom.setVisible( false );
            this._operator.setVisible( false );

            this.ae.addListener( this );
            this.ae.play( this.getBurstAnimName() );
        },

        animatedEntity_finishedAnimation: function( animation )
        {
            this.removeView();
        },

        removeView: function()
        {
            this._dropletView.getParent().removeChild( this._dropletView );
        },

        getTextArea: function()
        {
            var topBounds = RaindropsUtils.getTextBoundsUsingLabel( this._operandTop );
            var bottomBounds = RaindropsUtils.getTextBoundsUsingLabel( this._operandBottom );
            var textBounds = RaindropsUtils.bbUnion( topBounds, bottomBounds );
            // var textBounds = RaindropsUtils.bbUnion( this._operandTop.getBoundingBox(), this._operandBottom.getBoundingBox() );

            // ll.verbose( "RaindropsDroplet", "getTextArea", "textBounds.x = " + textBounds.x );

            var showDropletTextArea = false;
            if ( showDropletTextArea )
            {
                RaindropsUtils.addBoundingBoxToNode( this._operandTop.getParent(), textBounds, cc.c3b( 0, 255, 0 ), 128 );
            }

            var textBoundsRainSpace = RaindropsUtils.bbConvertFromTo(
                textBounds,
                this._operandTop.getParent(),
                this._operandTop.getParent().getParent()
            );

            // ll.verbose( "RaindropsDroplet", "getTextArea", "textBoundsRainSpace.x = " + textBoundsRainSpace.x );

            return textBoundsRainSpace;
        },

        setValues: function( leftOperand, rightOperand, operator )
        {
            var topValueSize = lu.measureText( this._operandTop.getString(), this._operandTop.getFontName(), this._operandTop.getFontSize() );
            topValueSize *= this._operandTop.getScaleX();

            var bottomValueSize = lu.measureText( this._operandBottom.getString(), this._operandBottom.getFontName(), this._operandBottom.getFontSize() );
            bottomValueSize *= this._operandBottom.getScaleX();

            var wider = this._operandTop;
            var shorter = this._operandBottom;

            var widerWidth = topValueSize;
            var shorterWidth = bottomValueSize;

            if ( bottomValueSize > topValueSize )
            {
                wider = this._operandBottom;
                shorter = this._operandTop;

                widerWidth = bottomValueSize;
                shorterWidth = topValueSize;
            }

            var dropletWidth = this._dropletView.getContentSize().width;

            // Remember that all three labels (top operand, bottom operand, and operator) have a centered anchor point.

            // Wider operand is centered in the droplet.
            var widerX = dropletWidth/ 2;
            wider.setPositionX( widerX );

            // Shorter operand is positioned such that its right edge is aligned with the right edge of the wider operand.
            var shorterX = dropletWidth / 2 + widerWidth / 2 - shorterWidth / 2;
            shorter.setPositionX( shorterX );

            var operatorValueSize = lu.measureText( this._operator.getString(), this._operator.getFontName(), this._operator.getFontSize() );
            operatorValueSize *= this._operator.getScaleX();

            // Operator is positioned to be to the left of the bottom operand.
            var operatorX = this._operandBottom.getPositionX() - bottomValueSize / 2 - operatorValueSize / 2 - this._operatorPadding;
            this._operator.setPositionX( operatorX );

            // lu.printNode( this._dropletView );
            // lu.printNode( this._operandTop );
            // lu.printNode( this._operandBottom );
            // lu.printNode( this._operator );
            // lu.printNode( this._sprite );
        },

        __printOperatorMeasurements: function()
        {
            var opValueSize = RaindropsUtils.getTextSizeUsingLabelFont( this._operator.getString(), this._operator ).width;
            ll.log( "opValueSize = " + opValueSize );

            var opBounds = RaindropsUtils.getTextBoundsUsingLabel( this._operator );
            ll.log( "opBounds.width = " + opBounds.width );

            var opClone = RaindropsUtils.cloneLabel( this._operator );

            var opTexPixels = opClone.getTexture().getPixelsWide();
            ll.log( "opTexPixels = " + opTexPixels );

            var opDims = opClone.getDimensions();
            ll.log( "opDims = " + opDims.width );
        },

        splash: function()
        {
            if ( this._splashed || this._burst )
            {
                return;
            }

            this._splashed = true;

            this._operandTop.setVisible( false );
            this._operandBottom.setVisible( false );
            this._operator.setVisible( false );

            if ( this.getSplashAnimName() === "" )
            {
                this.removeView();
            }
            else
            {
                this.ae.addListener( this );
                this.ae.play( this.getSplashAnimName() );
            }

            // Unnecessary, handled by water rise animations.
            //audio.playEffect( "sounds/dropletSplash.mp3" );
        },

        getEquation: function()
        {
            return this._equation;
        },

        getAnswer: function()
        {
            var retval = 0;
            
            if( this._equation )
            {
                retval = this._equation.getAnswer();
            }
            
            return retval;
        }
    } );
}() );
