/*global cc*/
/*global require*/

require( "src/Game/Base/GameModel.js" );
/*global GameModel*/
require( "src/Game/RaindropsEquation.js" );
/*global RaindropsEquation*/
require( "src/Game/RaindropsSun.js" );
/*global RaindropsSun*/
require( "src/Game/RaindropsUtils.js" );
/*global RaindropsUtils*/
require( "src/Utility/LumosityUtils.js" );
/*global lu*/

// type - "trial" type [P - regular problem, S - sun, X - incorrect response]
var RaindropsTrialType =
{
    RegularProblem: 'P',
    Sun: 'S',
    IncorrectResponse: 'X'
};

// response - response type [T - correct, F - missed, C - cleared by sun, blank - cleared by water/end of game]
var RaindropsResponseType =
{
    Correct: 'T',
    Missed: 'F',
    ClearedBySun: 'C',
    ClearedByWaterOrEndOfGame: ''
};

var RaindropsTrialMetadata;
( function()
{
    "use strict";

    RaindropsTrialMetadata = cc.Class.extend(
    {
        // type - "trial" type [P - regular problem, S - sun, X - incorrect response, B - input blocked]
        type: null,
        // response - response type [T - correct, F - missed, C - cleared by sun, blank - cleared by water/end of game]
        response: null,
        // equation - equation (blank for incorrect response)
        equation: null,
        // speed - Speed of raindrop (or current speed for incorrect responses)
        speed: null,
        // water_level - Current water level
        water_level: null,
        // num_drops - Number of drops on-screen
        num_drops: null,
        // *input_type - [L - number row, R - number pad]
        input_type: null

        // AUTOMATIC:

        // time_offset - number of milliseconds since start of game, appearance time for problems or response time for incorrect responses
        // time_offset: null,
        // response_time - number milliseconds since problem appearance when problem cleared by answer, sun, or water hit; blank for incorrect response
        // response_time: null,

        // ctor: function()
        // {
        //     this._points = 0;

        //     this._creationTimeMs = new Date();
        //     this._creationTimeMs = this._creationTimeMs.getTime();
        // },

        // getCreationTimeMs: function()
        // {
        //     return this._creationTimeMs;
        // },

        // addPoints: function( points )
        // {
        //     this._points = points;
        // }
    } );
}() );

var RaindropsModel;
( function()
{
    "use strict";
    
    RaindropsModel = GameModel.extend(
    {
        /** @lends RaindropsModel.prototype */
                                      
        bestStatKey: "NumberOfTrialsCorrect",

        raindropsSettings: null,
        raindropsView: null,

        _droplets: null,

        _inputtedValue: 0,
        _waterLevel: 0,
        _lastDropletCreateTime: 0,

        _initialWaterHeight: -1,

        hackMetadataNumTrials: 0,
        hackMetadataNumTrialsCorrect: 0,
        
        /**
        * The Raindrops model.
        * @constructs
        */
        ctor: function()
        {
            this._super();

            this._inputtedValue = 0;
            this._waterLevel = 0;
            this._lastDropletCreateTime = 0.0;

            this._droplets = [];

            // NOTES:

            // (^ currently only stored on mobile)
            // * web only

            // MANUAL: must track ourselves

            // ^num_problems - Number of raindrops
            // ^speed - Final speed (if different from num_correct in new version)
            // *keyboard_percent - Percentage of responses on keyboard vs. num pad

            this.metadata.num_problems = 0;
            this.metadata.speed = 0;
            this.metadata.keyboard_percent = 0;

            this.metadata.num_corrections = 0;
            this.metadata.num_tries_corrections = 0;

            // AUTOMATIC: nothing to do here

            // time (ms) - Total time in-game
            // num_tries - Number of problem attempts (Number of times the user hit Enter)
            // num_correct - Number of problems correct
            // ^response_time_correct - Average response time for correct trials in milliseconds
        },

        beginTrial: function()
        {
            this._currentTrial = new RaindropsTrialMetadata();

            this._currentTrial.num_corrections = 0;
        },

        getMetadata: function()
        {
            return this.metadata;
        },

        getTrialMetadata: function()
        {
            return this._currentTrial;
        },

        endTrial: function( wasLastTrialCorrect )
        {
            this.hackMetadataNumTrials += 1;

            if ( wasLastTrialCorrect )
            {
                this.hackMetadataNumTrialsCorrect += 1;
            }
        },

        exportResults: function( /* stats, bestStatKey */ )
        {
            var stats = {};
            stats[ this.bestStatKey ] = this.hackMetadataNumTrialsCorrect;

            return this._super( stats, this.bestStatKey );
        },

        setInputtedValue: function( value )
        {
            this._inputtedValue = value;
        },

        updateDroplets: function( deltaTime )
        {
            var i;
            var droplet;
            for ( i = 0; i < this._droplets.length; i += 1 )
            {
                droplet = this._droplets[ i ];

                if ( droplet )
                {
                    droplet.update( deltaTime );
                }
            }
        },

        getDropletsBelow: function( height )
        {
            var result = [];

            var i;
            var droplet;
            var dropletBottomY;
            for ( i = 0; i < this._droplets.length; i += 1 )
            {
                droplet = this._droplets[ i ];
                dropletBottomY = droplet.getViewBottomY();

                if ( dropletBottomY < height )
                {
                    result.push( droplet );
                }
            }

            return result;
        },

        getDropletFront: function()
        {
            if ( this._droplets.length > 0 )
            {
                return this._droplets[ 0 ];
            }

            return null;
        },

        removeDroplet: function( dropletToRemove )
        {
            // @todo assert that this element exists in the array
            this._droplets.splice( this._droplets.indexOf( dropletToRemove ), 1 );
        },

        destroyAllDroplets: function()
        {
            var i;
            var droplet;
            for ( i = 0; i < this._droplets.length; i += 1 )
            {
                droplet = this._droplets[ i ];
                droplet.removeView();
            }

            this._droplets = [];
        },

        incrementWaterLevel: function()
        {
            // Was:
            // std::numeric_limits< unsigned int >::max()
            // Should probably be:
            // var limit = Number.MAX_VALUE;
            // Or maybe:
            // var limit = this.raindropsSettings.getMaximumWaterLevel() + 1;
            var limit = 1000000;

            if ( this._waterLevel < limit )
            {
                this._waterLevel += 1;
            }
        },

        getWaterLevel: function()
        {
            return this._waterLevel;
        },

        getLastDropletCreateTime: function()
        {
            return this._lastDropletCreateTime;
        },

        getDropletCount: function()
        {
            return this._droplets.length;
        },

        getLastDroplet: function()
        {
            if ( this._droplets.length > 0 )
            {
                return this._droplets[ this._droplets.length - 1 ];
            }

            return null;
        },

        getDropletCreateFrequency: function()
        {
            var createFrequency = 1.9;
            
            var numberOfTrials = this.hackMetadataNumTrials;
            
            createFrequency *= Math.pow(0.99, numberOfTrials);
                                      
            /* Ben: old system for this
            if( numberOfTrials <= 40 )
            {
                createFrequency -= numberOfTrials * 0.08;
            }
            else
            {
                createFrequency -= 40 * 0.08;
                createFrequency *= Math.pow(0.98, numberOfTrials - 40);
            }*/
            
            return createFrequency;
        },

        getAllowedDropletCount: function()
        {
            var totalCorrect = this.hackMetadataNumTrialsCorrect;
            var count = this.raindropsSettings.getMaxAllowedDropletCount( totalCorrect );
            count -= this.getWaterLevel();

            return Math.max(count, 1);
        },

        getDropletFallSpeed: function()
        {
            var visibleTime = 11;
            
            // Each time the water level increases, slow down future raindrops a little more
            visibleTime += 3.5 * this.getWaterLevel();
            
            if ( this._initialWaterHeight === -1 )
            {
                this._initialWaterHeight = this.raindropsView.getContentSize().height - this.raindropsView.getDropletSplashHeight();
            }
            
            var numberOfTrials = this.hackMetadataNumTrials;
            
            if ( numberOfTrials <= 40 )
            {
                visibleTime *= Math.pow(0.985, numberOfTrials);
            }
            else if ( numberOfTrials > 40 && numberOfTrials <= 60)
            {
                visibleTime *= Math.pow(0.985, 40);
                visibleTime *= Math.pow(0.992, numberOfTrials - 40);
            }
            else
            {
                visibleTime *= Math.pow(0.985, 40);
                visibleTime *= Math.pow(0.992, 20);
                visibleTime *= Math.pow(0.996, numberOfTrials - 60);
            }
                                      
            /* Ben: old system for this
            var countCorrect;
            for ( countCorrect = 1; countCorrect <= numberOfTrials; countCorrect += 1 )
            {
                if ( countCorrect < 40)
                {
                    visibleTime -= 0.2;
                }
                else
                {
                    visibleTime *= 0.99;
                }
            }*/
                                      
            visibleTime += 1.5;
            
            return this._initialWaterHeight / visibleTime;
        },

        createEquation: function()
        {
            var numberAnswersCorrect = 0;
            numberAnswersCorrect = this.hackMetadataNumTrialsCorrect;
            
            var settings = this.getRaindropsSettings();

            var addition = settings.getAdditionRange( numberAnswersCorrect );
            var subtraction = settings.getSubtratctionRange( numberAnswersCorrect );
            var multiplication = settings.getMultiplicationRange( numberAnswersCorrect );
            var division = settings.getDivisionRange( numberAnswersCorrect );
            
            return new RaindropsEquation( addition, subtraction, multiplication, division );
        },

        shouldSpawnSun: function()
        {
            var shouldSpawnSun = false;
            
            if ( !this.isSunOut() )
            {    
                var numberOfTrials = this.hackMetadataNumTrials;
                
                var sunFrequency = 30;

                if ( numberOfTrials > 0 )
                {
                    if ( numberOfTrials <= 160 )
                    {
                        shouldSpawnSun = ( numberOfTrials % sunFrequency === 0 );
                    }
                    else
                    {
                        shouldSpawnSun = ( numberOfTrials % ( 2 * sunFrequency ) === 0 );
                    }
                }
            }
            
            // DEBUG
            //return lu.randomBool();

            //return (numberOfTrials % 2) !== 0;
            return shouldSpawnSun;
        },

        addDroplet: function( droplet )
        {
            this.metadata.num_problems += 1;
            
            this._droplets.push( droplet );
        },

        setLastDropletCreateTime: function( time )
        {
            this._lastDropletCreateTime = time;
        },

        getRaindropsSettings: function()
        {
            return this.raindropsSettings;
        },

        getInputtedValue: function()
        {
            return this._inputtedValue;
        },

        getLowestDroplet: function()
        {
            var result = null;
            var lowest = Number.MAX_VALUE;

            var i;
            var droplet;
            var dropletBottomY;
            for ( i = 0; i < this._droplets.length; i += 1 )
            {
                droplet = this._droplets[ i ];
                dropletBottomY = droplet.getViewBottomY();

                if ( dropletBottomY < lowest )
                {
                    lowest = dropletBottomY;
                    result = droplet;
                }
            }

            return result;
        },

        getDropletsWithAnswer: function( answer )
        {
            var result = [];

            var i;
            var droplet;
            for ( i = 0; i < this._droplets.length; i += 1 )
            {
                droplet = this._droplets[ i ];

                if ( droplet.getAnswer() === answer )
                {
                    result.push( droplet );
                }
            }
            
            return result;
        },

        getDroplets: function()
        {
            return this._droplets.slice(0);
        },

        isSunOut: function()
        {
            return RaindropsUtils.containsSun( this._droplets );
        }
    } );
}() );
