/*global cc*/
/*global require*/
/*global Game*/
/*global SpeedMatchOverdriveSettings*/
/*global SpeedMatchOverdriveModel*/
/*global SpeedMatchOverdriveView*/

require( "src/Game/Base/Game.js" );
require( "src/Game/SpeedMatchOverdriveSettings.js" );
require( "src/Game/SpeedMatchOverdriveModel.js" );
require( "src/Game/SpeedMatchOverdriveView.js" );

var SpeedMatchOverdrive;
( function()
{
    "use strict";
    
    SpeedMatchOverdrive = Game.extend(
    {
        /** @lends SpeedMatchOverdrive.prototype */
        
        /**
        *  SpeedMatchOverdrive is the main class for the game, Speed Match Overdrive!
        *
        *  @constructs
        */
        ctor: function()
        {
            this._super();
        },
        
        _createSettings: function()
        {
            return new SpeedMatchOverdriveSettings( "settings.json" );
        },
        
        _createModel: function()
        {
            return new SpeedMatchOverdriveModel(this.settings);
        },
        
        _createView: function( screenSize )
        {
            return new SpeedMatchOverdriveView( screenSize, this.settings, this.model, this );
        },
        
        onBegin: function()
        {
            this._super();
            this.feedback = this.view.getFeedbackWidget();
            this.scoring = this.getScoringComponent();
            this.allowButtonInput = false;
            this.view.enableButtons( false );
            this.model.chooseNextSymbol();
            
            this.view.updateSymbols();
            this.view.updateText( true );
        },
        
        onStartGameplay: function()
        {
            this._super();
            this.view.enableButtons( true );
            this.model.markTrialStartTime();
            
            this._showNewStimulus();
        },
        
        onEndGameplay: function()
        {
            //we need to make sure button mashing etc won't carry on after the timer runs out
            this.view.enableButtons( false );
            this.view.stopAnimation( "problem" );

            this.getMetadataComponent()._trialCount = this.model.totalResponses;            

            return this._super();
        },
        
        _buttonPressed: function( name, interaction )
        {
            if ( !this.allowButtonInput || !this.model.isGameRunning )
            {
                return;
            }
                        
            this._super( name );
                        
            switch ( name )
            {
                case "btnYes":
                    return this.handleResponse( this.model.doSymbolsMatch(), 2, interaction );
                case "btnNo":
                    return this.handleResponse( !this.model.doSymbolsMatch() && !this.model.doPartialMatch(), 0, interaction );
                case "btnPartial":
                    return this.handleResponse( this.model.doPartialMatch(), 1, interaction ); 
            }
        },
        
        handleResponse: function( correct, userResponse, sourceInteraction )
        {            
            if ( LumosityUtils.isDebugModeOn() )
            {
                correct = true;
            }
             
            //this must be called before the multiplier etc are increased
            this._recordTrialMetadata( correct, userResponse, sourceInteraction );
            
            this.model.markResponse( correct );
            this.view.showPartialHint = this.model.forcePartialTrialCount > 0;
            this.view.updateText( false, false, correct);
 
            if ( correct )
            {
                this.scoring.increaseMultiplierProgress();
                
                //update the text in practice mode, but don't do it on the last correct since the text will fade out
                if ( this.model.persistentData.isPracticeMode && !this.model.hasFinishedPractice() )
                {
                    this.view.updateText( false, false, correct );
                }
            }
            else
            {
                this.scoring.decreaseMultiplierProgress();
                
                //if in practice mode, we want to show feedback about what they did wrong
                if ( this.model.persistentData.isPracticeMode )
                {
                    this.view.updateText( false, true, correct );
                    this.model.maintainString = true;
                    
                    //don't advance to the new card until they get it right
                    this.feedback.play( correct);
                    return;
                }
 
            }
            
            this.feedback.play( correct );
            
            this.allowButtonInput = false;
            
            //don't show anything else if we have finished gameplay in practice mode
            if ( this.model.hasFinishedPractice() )
            {
                return;
            }
            
            this.model.chooseNextSymbol();
            this.view.updateSymbols();
            this.model.markTrialStartTime();
            
            //we can't use stopAnimation because that stops all animations on all the cards, so we do this instead
            this.view.card1.stopAllActions();
            this.view.card2.stopAllActions();
            
            this._showNewStimulus();
        },
        
        cue_allowInput: function()
        {
            if ( this.model.hasFinishedPractice() )
            {
                return;
            }
            
            this.model.hasAllowedInput = true;
            
            //bleh kinda hacky
            if ( !this.model.maintainString )
            {
                this.view.updateText( false, false, false, this.model.forcePartialTrialCount > 0);
            }
            this.model.maintainString = false;
            
            this.allowButtonInput = true;
            this.model.trialStartTime = lumosity.game.frametime;
            this.beginTrial();
        },
        
        _showNewStimulus: function()
        {
            this.view.playAnimation( "problem" );
        },
        
        _recordTrialMetadata: function( correct, userResponse, sourceInteraction )
        {
            var typeCalc = this.model.getCurrentShapeIndex() === this.model.getLastShapeIndex()? 1 : 0;
            typeCalc += this.model.getCurrentColorIndex() === this.model.getLastColorIndex() ? 2 : 0;
            //trial metadata
            var trial =
            {
                hint : this.view.showPartialHint ? "T" : "F",
                color : this.model.getCurrentColorIndex(),
                color_back: this.model.getLastColorIndex(),
                type : typeCalc, 
                back : this.settings.stimulusMetadataIndices[ this.model.lastSymbolId ],
                stimulus : this.settings.stimulusMetadataIndices[ this.model.currentSymbolId ],
                response : userResponse,
                correct : correct ? "T" : "F",
                multiplier : this._components[ GameComponentScoreMultiplier.prototype.getType() ]._multiplier,
                progress : this._components[ GameComponentScoreMultiplier.prototype.getType() ].getProgressMetadata()
            };
            
            var inputType = "T";
            if ( sourceInteraction && sourceInteraction.getSource() === InteractionSource.Keyboard )
            {
                if ( sourceInteraction.isKeyboardButton( cc.KEY.w, true ) || sourceInteraction.isKeyboardButton( cc.KEY.a, true ) ||
                     sourceInteraction.isKeyboardButton( cc.KEY.s, true ) || sourceInteraction.isKeyboardButton( cc.KEY.d, true ) )
                {
                    inputType = "L";
                }
                else
                {
                    inputType = "R";
                }
            }
            trial.input_type = inputType;
            
            this.endTrial( trial, correct, false, false, false, false, true );
        },
        
        feedback_isFinished: function()
        {
            //end the gameplay if we finished practice
            if ( this.model.hasFinishedPractice() )
            {
                this.onEndGameplay();
            }
        },
        
        interaction_alias:function( alias, data )
        {
            this._super( alias, data );
            
            if ( alias === "correct" )
            {
                this.handleResponse( true, this.model.doSymbolsMatch(), null );
            }
            else if ( alias === "incorrect" )
            {
                this.handleResponse( false, !this.model.doSymbolsMatch(), null );
            }
        }
    } );
}() );
