/*global require*/
/*global GameModel*/

require( "src/Game/Base/GameModel.js" );

var SpeedMatchOverdriveModel;
( function()
{
    "use strict";
    
    SpeedMatchOverdriveModel = GameModel.extend(
    {
        /** @lends SpeedMatchOverdriveModel.prototype */
        
        /**
        * The Speed Match Overdrive model.
        * @constructs
        */
        ctor: function( settings )
        {
            this._super();
            
            this.totalResponses = 0;
            this.totalResponsesCorrect = 0;
            this.consecutiveResponsesCorrect = 0;
            this.trialStartTime = -1;
            this.reactionTimes = [];
            this.currentTrialSet = settings.trials[ "Shape" ];
            this.currentColorSet = settings.trials[ "Color" ];
            this.scriptedTrials = settings.tutorialScriptedTrials;
            this.currentSymbolId = this.currentTrialSet.symbolIds.getRandomElement();
            this.currentColorId = this.currentColorSet.symbolIds.getRandomElement();
            this.lastSymbolId = this.currentSymbolId;
            this.lastColorId = this.currentColorId;
            this.lastTrialWasColorMatch = false;
            this.lastTrialWasShapeMatch = false;
            this.trialStartTime = 0; //this may be redundant but whatever
            this.practiceProgress = 0;
            this.hasAllowedInput = false;
            
            //just put these in a separate thing to keep them a little more organized
            this.responseTimesMatch = [];
            this.responseTimesNoMatch = [];
            this.responseTimesShapeMatch = [];
            this.responseTimesColorMatch = [];
            this.responseTimesMatchCorrect = [];
            this.responseTimesNoMatchCorrect = [];
            this.responseTimesShapeMatchCorrect = [];
            this.responseTimesColorMatchCorrect = [];
            this.responseTimesNoHintCorrect = [];

            //partial trial probability
            this.recentPartialResults = [];
            this.recentWrongPartialAnswers = 0;
            this.forcePartialTrialCount = 0


            this.metadata =
            {
                num_full_match: 0,
                num_shape_match: 0,
                num_color_match: 0,
                num_no_match: 0,
                num_correct_full_match: 0,
                num_correct_shape_match: 0,
                num_correct_color_match: 0,
                num_correct_no_match: 0,
                response_time_full_match: 0,
                response_time_shape_match: 0,
                response_time_color_match: 0,
                response_time_correct_full_match: 0,
                response_time_correct_shape_match: 0,
                response_time_correct_color_match: 0,
                response_time_correct_no_match: 0,
                num_no_hint: 0,
                num_no_hint_correct: 0,
                response_time_no_hint_correct: 0,
                response_time_no_match: 0
            };
        },
        
        exportResults: function()
        {
            var stats =
            {
                cards: this.totalResponsesCorrect,
                total: this.totalResponses,
                netCorrect: Math.max( this.totalResponsesCorrect - ( this.totalResponses - this.totalResponsesCorrect ), 0 )
            };
            
            return this._super( stats, "netCorrect" );
        },

        doSymbolsMatch: function()
        {
            return this.lastSymbolId === this.currentSymbolId && this.lastColorId === this.currentColorId;
        },
       
        doPartialMatch: function()
        {
            return (this.lastSymbolId === this.currentSymbolId && this.lastColorId !== this.currentColorId) || (this.lastSymbolId !== this.currentSymbolId && this.lastColorId === this.currentColorId);
        },


        setSymbolId: function( symbolId, colorId )
        {
            this.lastTrialWasColorMatch = ( colorId === this.currentColorId );
            this.lastTrialWasShapeMatch = ( symbolId === this.currentSymbolId );
            this.lastSymbolId = this.currentSymbolId;
            this.lastColorId = this.currentColorId;
            this.currentSymbolId = symbolId;
            this.currentColorId = colorId;
        },

        chooseNextSymbol: function()
        {
            var trialIds = this.currentTrialSet.getSymbolIds();
            var colorIds = this.currentColorSet.getSymbolIds();
            
            if ( this.currentSymbolId === "" )
            {
                this.setSymbolId( trialIds.getRandomElement(), colorIds.getRandomElement() );
            }
            else
            {
                var showMatch = Math.random() < 0.5;
                var showPartial = Math.random() < 1/3;
                //in practice mode, show a scripted trial (matching / not matching in a fixed order)
                if ( this.persistentData.isPracticeMode && !this.hasFinishedPractice() )
                {
                    showMatch = !!this.scriptedTrials[ this.practiceProgress ];
                }
                if ( this.forcePartialTrialCount > 0 || showPartial ) 
                {
                    this.forcePartialTrialCount = Math.max(this.forcePartialTrialCount - 1, 0);
                    var currentIndex = trialIds.indexOf( this.currentSymbolId );
                    var colorIndex = colorIds.indexOf( this.currentColorId );
    
                    var sameShape = Math.random() < 0.5;
                    var nextShape = this.currentSymbolId;
                    var nextColor = this.currentColorId;
                    if( sameShape )
                    {
                        var nextTrialIndex = lu.getRandomInt( 0, trialIds.length - 1, false, true );
    
                        if ( nextTrialIndex >= currentIndex )
                        {
                            nextTrialIndex += 1;
                        }
                        nextShape = trialIds [ nextTrialIndex ];
                    } 
                    else
                    {
                        var nextColorIndex = lu.getRandomInt( 0, colorIds.length - 1, false, true );

                        if ( nextColorIndex >= colorIndex )
                        {
                            nextColorIndex += 1;
                        }
                        nextColor = colorIds [ nextColorIndex ];
                    }

                    this.setSymbolId( nextShape, nextColor);
                          
                }
                else if ( showMatch )
                {
                    this.setSymbolId( this.currentSymbolId, this.currentColorId);
                }
                else
                {
                    var currentIndex = trialIds.indexOf( this.currentSymbolId );
                    var colorIndex = colorIds.indexOf( this.currentColorId );
                    if ( currentIndex === -1 )
                    {
                        cc.log("Could not find current symbol id in SpeedMatch::getNextTrialSymbol(), returning");
                        return "";
                    }
    
                    var nextTrialIndex = lu.getRandomInt( 0, trialIds.length - 1, false, true );
    
                    if ( nextTrialIndex >= currentIndex )
                    {
                        nextTrialIndex += 1;
                    }

                    var nextColorIndex = lu.getRandomInt( 0, colorIds.length - 1, false, true );

                    if ( nextColorIndex >= colorIndex )
                    {
                        nextColorIndex += 1;
                    }
    
                    this.setSymbolId( trialIds[ nextTrialIndex ], colorIds[ nextColorIndex ] );
                }
            }
        },

        markTrialStartTime: function()
        {
            this.trialStartTime = lumosity.game.frametime;
        },

        markTrialFinishTime: function()
        {
            this.reactionTimes.push( lumosity.game.frametime - this.trialStartTime );
        },

        markResponse: function( correct )
        {
            this.totalResponses += 1;
            
            var showingHint = this.forcePartialTrialCount > 0;
            var responseTime = lumosity.game.frametime - this.trialStartTime;
            
            if ( !showingHint ) 
            {
                this.metadata.num_no_hint++;
            }


            if ( this.lastTrialWasColorMatch && this.lastTrialWasShapeMatch)
            {
                this.metadata.num_full_match++;
                this.responseTimesMatch.push( responseTime );
                this.metadata.response_time_full_match =  Math.round( lu.convertSecondsToMilliseconds( this.findAverageReactionTimeMatch() ) );
            }else if ( this.lastTrialWasColorMatch )
            {
                this.metadata.num_color_match++;
                this.responseTimesColorMatch.push( responseTime );
                this.metadata.response_time_color_match = Math.round( lu.convertSecondsToMilliseconds( this._findAverageReactionTime(this.responseTimesColorMatch) ) );
            }else if ( this.lastTrialWasShapeMatch )
            {
                this.metadata.num_shape_match++;
                this.responseTimesShapeMatch.push( responseTime );
                this.metadata.response_time_shape_match = Math.round( lu.convertSecondsToMilliseconds( this._findAverageReactionTime(this.responseTimesShapeMatch) ) ); 
            }
            else
            {
                this.metadata.num_no_match++;
                this.responseTimesNoMatch.push( responseTime );
                this.metadata.response_time_no_match = Math.round( lu.convertSecondsToMilliseconds( this.findAverageReactionTimeNoMatch() ) );
            }

            if ( correct )
            {
                this.totalResponsesCorrect += 1;
                this.consecutiveResponsesCorrect += 1;

                if( !showingHint )
                {
                     this.metadata.num_no_hint_correct++;
                     this.responseTimesNoHintCorrect.push( responseTime );
                     this.metadata.response_time_no_hint_correct = Math.round( lu.convertSecondsToMilliseconds( this._findAverageReactionTime(this.responseTimesNoHintCorrect) ) );
                }
                
                if ( this.lastTrialWasColorMatch && this.lastTrialWasShapeMatch)
                {
                    this.metadata.num_correct_full_match++;
                    this.responseTimesMatchCorrect.push( responseTime );
                    this.metadata.response_time_correct_full_match = Math.round( lu.convertSecondsToMilliseconds( this._findAverageReactionTime(this.responseTimesMatchCorrect) ) ); 
                }else if ( this.lastTrialWasColorMatch )
                {
                    this.updatePartialProbability(true);

                    this.metadata.num_correct_color_match++;
                    this.responseTimesColorMatchCorrect.push( responseTime );
                    this.metadata.response_time_correct_color_match = Math.round( lu.convertSecondsToMilliseconds( this._findAverageReactionTime(this.responseTimesColorMatchCorrect) ) );
                }else if ( this.lastTrialWasShapeMatch )
                {
                    this.updatePartialProbability(true);

                    this.metadata.num_correct_shape_match++;
                    this.responseTimesShapeMatchCorrect.push( responseTime );
                    this.metadata.response_time_correct_shape_match = Math.round( lu.convertSecondsToMilliseconds( this._findAverageReactionTime(this.responseTimesShapeMatchCorrect) ) ); 
                }
                else
                {
                    this.metadata.num_correct_no_match++;
                    this.responseTimesNoMatchCorrect.push(responseTime );
                    this.metadata.response_time_correct_no_match = Math.round( lu.convertSecondsToMilliseconds( this._findAverageReactionTime(this.responseTimesNoMatchCorrect) ) );
                }
            }
            else
            {
                this.consecutiveResponsesCorrect = 0;
                
                if( !(this.lastTrialWasColorMatch && this.lastTrialWasShapeMatch) && (this.lastTrialWasColorMatch || this.lastTrialWasShapeMatch) )
                {
                    this.updatePartialProbability(false);
                }
            }
            
            if ( this.persistentData.isPracticeMode && !this.hasFinishedPractice() )
            {
                if ( correct )
                {
                    this.practiceProgress++;
                }
            }
        },
        updatePartialProbability:function( correct )
        {
            this.recentPartialResults.push(correct);
            if( this.recentPartialResults.length > 4 )
            {
                if( !this.recentPartialResults.unshift() ) 
                {
                    this.recentWrongPartialAnswers--;
                }
            }
            if( !correct )
            {
                this.recentWrongPartialAnswers++;
                if (this.recentWrongPartialAnswers >= 3)
                {
                    this.forcePartialTrialCount = 2;
                }
            }
        },
        getLastColorIndex:function()
        {
            return this.currentColorSet.getSymbolIds().indexOf(this.lastColorId);
        },
        getLastShapeIndex:function()
        {
            return this.currentTrialSet.getSymbolIds().indexOf(this.lastSymbolId);
        },
        getCurrentShapeIndex:function()
        {
            return this.currentTrialSet.getSymbolIds().indexOf(this.currentSymbolId);
        },
        getCurrentColorIndex:function()
        {
            return this.currentColorSet.getSymbolIds().indexOf(this.currentColorId);
        },
        
        hasFinishedPractice:function()
        {
            return this.persistentData.isPracticeMode && ( !this.scriptedTrials || this.practiceProgress >= this.scriptedTrials.length );
        },
        
        findAverageReactionTimeMatch: function()
        {
            return this._findAverageReactionTime( this.responseTimesMatch );
        },
        
        findAverageReactionTimeNoMatch: function()
        {
            return this._findAverageReactionTime( this.responseTimesNoMatch );
        },
        
        _findAverageReactionTime: function( array )
        {
            var averageReactionTime = 0;
            
            if ( array.length > 0 )
            {
                var reactionTimeIndex;
                for ( reactionTimeIndex = 0; reactionTimeIndex < array.length; reactionTimeIndex++ )
                {
                    averageReactionTime += array[ reactionTimeIndex ];
                }
                averageReactionTime /= array.length;
            }
            return averageReactionTime;
        }

    } );
}() );
