/*global cc*/
/*global require*/
/*global LumosityUtils*/
/*global lg*/
/*global InteractionType*/
/*global InteractionSource*/
/*global CardinalDirections*/
/*global GameComponentScoreMultiplier*/
/*global metadataAggregateKeys*/
/*global metadataTrialKeys*/
/*global lumosity*/

require( "src/Game/Base/Game.js" );
/*global Game*/

require( "src/Game/SpeedMatchSettings.js" );
/*global SpeedMatchSettings*/

require( "src/Game/SpeedMatchModel.js" );
/*global SpeedMatchModel*/

require( "src/Game/SpeedMatchView.js" );
/*global SpeedMatchView*/

var SpeedMatch;
(function()
{
    "use strict";

    SpeedMatch = Game.extend(
    {
        ctor: function()
        {
            this._super();
        },
        
        _createSettings: function()
        {
            return new SpeedMatchSettings();
        },

        _createModel: function()
        {
            return new SpeedMatchModel( this.settings );
        },

        _createView: function( screenSize )
        {
            return new SpeedMatchView( screenSize, this.settings, this.model, this );
        },
        
        onBegin: function()
        {
            this._super();
            this.feedback = this.view.getFeedbackWidget();
            this.scoring = this.getScoringComponent();
            this.allowButtonInput = false;
            this.view.enableButtons( false );
            this.model.chooseNextSymbol();
            
            this.view.updateSymbols();
            this.view.updateText( true );
        },
        
        onStartGameplay: function()
        {
            this._super();
            this.view.enableButtons( true );
            this.model.markTrialStartTime();
            
            this._showNewStimulus();
        },
        
        onEndGameplay: function()
        {
            //we need to make sure button mashing etc won't carry on after the timer runs out
            this.view.enableButtons( false );
            this.view.stopAnimation( "problem" );
            
            this.getMetadataComponent()._trialCount = this.model.totalResponses;
            return this._super();
        },
        
        handleResponse: function( correct, userAnsweredYes )
        {            
            if ( LumosityUtils.isDebugModeOn() )
            {
                correct = true;
            }
            
            //this must be called before the multiplier etc are increased
            this._recordTrialMetadata( correct, userAnsweredYes );
            
            this.model.markResponse( correct );
            
            if ( correct )
            {
                this.scoring.increaseMultiplierProgress();
                
                //update the text in practice mode, but don't do it on the last correct since the text will fade out
                if ( this.model.persistentData.isPracticeMode && !this.model.hasFinishedPractice() )
                {
                    this.view.updateText( false, false, userAnsweredYes );
                }
            }
            else
            {
                this.scoring.decreaseMultiplierProgress();
                
                //if in practice mode, we want to show feedback about what they did wrong
                if ( this.model.persistentData.isPracticeMode )
                {
                    this.view.updateText( false, true, !userAnsweredYes );
                    this.model.maintainString = true;
                    
                    //don't advance to the new card until they get it right
                    this.feedback.play( correct);
                    return;
                }
            }
            
            this.feedback.play( correct );
            
            this.allowButtonInput = false;
            
            //don't show anything else if we have finished gameplay in practice mode
            if ( this.model.hasFinishedPractice() )
            {
                return;
            }
            
            this.model.chooseNextSymbol();
            this.view.updateSymbols();
            this.model.markTrialStartTime();
            
            //we can't use stopAnimation because that stops all animations on all the cards, so we do this instead
            this.view.card1.stopAllActions();
            this.view.card2.stopAllActions();
            
            this._showNewStimulus();
        },
        
        cue_allowInput: function()
        {
            if ( this.model.hasFinishedPractice() )
            {
                return;
            }
            
            this.model.hasAllowedInput = true;
            
            //bleh kinda hacky
            if ( !this.model.maintainString )
            {
                this.view.updateText( false );
            }
            this.model.maintainString = false;
            
            this.allowButtonInput = true;
            this.model.trialStartTime = lumosity.game.frametime;
            this.beginTrial();
        },
        
        _showNewStimulus: function()
        {
            this.view.playAnimation( "problem" );
        },
        
        _recordTrialMetadata: function( correct, userAnsweredYes )
        {
            //trial metadata
            var trial =
            {
                stimulus : this.settings.stimulusMetadataIndices[ this.model.currentSymbolId ],
                response : userAnsweredYes ? "T" : "F",
                correct : correct ? "T" : "F",
                multiplier : this._components[ GameComponentScoreMultiplier.prototype.getType() ]._multiplier,
                progress : this._components[ GameComponentScoreMultiplier.prototype.getType() ].getProgressMetadata(),
                back : this.settings.stimulusMetadataIndices[ this.model.lastSymbolId ],
                match : this.model.doSymbolsMatch() ? "T" : "F" 
            };
            
            // NAF: Removing for now because simulating a button press rather than going through the interaction reciever doesn't
            //      allow us to pass through the sourceInteraction to determine whether it was a button or keyboard input.
            // var inputType = "T";
            // if ( sourceInteraction && sourceInteraction.getSource() === InteractionSource.Keyboard )
            // {
            //     if ( sourceInteraction.isKeyboardButton( cc.KEY.w, true ) || sourceInteraction.isKeyboardButton( cc.KEY.a, true ) ||
            //          sourceInteraction.isKeyboardButton( cc.KEY.s, true ) || sourceInteraction.isKeyboardButton( cc.KEY.d, true ) )
            //     {
            //         inputType = "L";
            //     }
            //     else
            //     {
            //         inputType = "R";
            //     }
            // }
            // trial.input_type = inputType;
            
            this.endTrial( trial, correct, false, false, false, false, true );
        },
        
        feedback_isFinished: function()
        {
            //end the gameplay if we finished practice
            if ( this.model.hasFinishedPractice() )
            {
                this.onEndGameplay();
            }
        },
        
        interaction_alias:function( alias, data )
        {
            this._super( alias, data );
            
            if ( alias === "correct" )
            {
                this.handleResponse( true, this.model.doSymbolsMatch(), null );
            }
            else if ( alias === "incorrect" )
            {
                this.handleResponse( false, !this.model.doSymbolsMatch(), null );
            }
        }
    } );
}() );
