/*global require*/
/*global cc*/
/*global GameModel*/
/*global SpeedPackDoorModel*/

require( "src/Game/Base/GameModel.js" );
require( "src/Game/SpeedPackDoorModel.js");

var SpeedPackTrialModel;
( function()
{
    "use strict";

    SpeedPackTrialModel = GameModel.extend(
    {
        /** @lends SpeedPackTrialModel.prototype */

        totalnum_items: 0,
        id: 0,
        net: "null_net",
        doors: [],

        ctor: function()
        {
            this._super();
        },

        /**
         * returns true/false if there was a collision during folding
         * @returns {*}
         */
        checkDoorOverlap: function()
        {
            var folded_doors = [];
            var combined_door;
            //if there are doors to fold
            if ( this.doors.length >= 2 )
            {
                //go through all doors
                var door;
                for ( door = 0; door < this.doors.length; ++door )
                {
                    //push the main door (doesn't get folded) onto the array first
                    if ( this.doors[door].direction === 0 )
                    {
                        folded_doors.push( this.doors[door] );
                    }
                }

                //if there was a main door
                if ( folded_doors.length > 0 )
                {
                    //go through all doors...
                    var i;
                    for ( i = 0; i < this.doors.length; ++i )
                    {
                        //...that aren't the main door
                        if ( this.doors[i].direction !== 0 )
                        {
                            //combine it with the main door
                            combined_door = this.createCombinedDoor( folded_doors[folded_doors.length - 1], this.doors[i] );
                            folded_doors.push( combined_door );
                        }
                    }
                }
            }

           if ( folded_doors.length > 0 )
           {
               return ( this.isSlotOverlap( folded_doors[folded_doors.length - 1] ) );
           }

            return false;
        },

        /**
         * folds doors together, returning combined door
         * @param firstDoor
         * @param secondDoor
         * @returns {*}
         */
        createCombinedDoor: function( firstDoor, secondDoor )
        {
            var base_door = null;
            var moved_door = null;
            //set the main door as the base door
            if ( firstDoor.direction === 0 )
            {
                moved_door = secondDoor;
                base_door = firstDoor;
            }
            else if ( secondDoor.direction === 0 )
            {
                moved_door = firstDoor;
                base_door = secondDoor;
            }

            //if its a special case mismatched door (folding left or up)
            if( firstDoor.slots.length !== secondDoor.slots.length )
            {
                //handle the special case mismatch so indexes match the main door
                moved_door = this.handleMismatch( moved_door );
            }
            //otherwise flip the door's indexes to match the main door
            else
            {
                moved_door = this.flipDoorIndexByDirection( moved_door );
            }

            //add the filled slots from flipped door to base door, creating folded door
            var i;
            var j;
            for ( i = 0; i < moved_door.slots.length; ++i )
            {
                for ( j = 0; j < base_door.slots.length; ++j )
                {
                    if ( moved_door.slots[i].index === base_door.slots[j].index )
                    {
                        base_door.slots[j].filled += moved_door.slots[i].filled;
                    }
                }
            }

            return base_door;
        },

        /**
         * handles the special cases when 2x3 doors are folded onto 3x3 doors
         * @param door
         * @returns {*}
         */
        handleMismatch: function ( door )
        {
            var i;
            //alter door folding up
            if( door.direction === 2 )
            {
                for ( i = 0; i < door.slots.length; i++ )
                {
                    if ( door.slots[i].index % 5 === 0 )
                    {
                        door.slots[i].index += 2;

                    }
                }
            }
            //alter door folding left
            else if ( door.direction === 8 )
            {
                for ( i = 0; i < door.slots.length; i++ )
                {
                    if ( door.slots[i].index <= 2 )
                    {
                        door.slots[i].index += 10;

                    }
                }
            }
            //when mismatch door direction is 4 or 6 we can treat it as a normal flip (non mismatched)
            else
            {
                door = this.flipDoorIndexByDirection( door );
            }

            return door;
        },

        /**
         * checks a door for collisions (i.e. a slot with filled > 1)
         * @param door
         * @returns {boolean}
         */
        isSlotOverlap: function( door )
        {
            var doorSlots = door.slots;

            var i;
            for ( i = 0; i < doorSlots.length; ++i )
            {
                if ( doorSlots[i].filled > 1 )
                {
                    return true;
                }
            }

            return false;
        },

        /**
         * flips a door based on its intended direction
         * @param door
         * @returns {*}
         */
        flipDoorIndexByDirection: function( door )
        {
            var slots = door.slots;

            // Vertical flip
            if ( door.direction === 2 || door.direction === 6 )
            {
                slots = this.flipVertical( slots );
            }
            // Horizontol flip
            else if ( door.direction === 4 || door.direction === 8 )
            {
                slots = this.flipHorizontal( slots );
            }
            // Diagonal flip

            else
            {
                if ( door.direction === 1 || door.direction === 5 )
                {
                    slots = this.flipHorizontal( slots );
                    slots = this.rotateRightAngle( slots, door.width );
                }
                else if ( door.direction === 3 || door.direction === 7 )
                {
                    slots = this.flipVertical( slots );
                    slots = this.rotateRightAngle( slots, door.width );
                }
            }

            door.slots = slots;

            return door;
        },

        flipVertical: function ( slots )
        {
            var currentVerSlotIndex = 0;
            var prevVerArrayIndex = 0;
            var i;
            for ( i = 0; i < slots.length; ++i )
            {
                currentVerSlotIndex = slots[ i ].index;
                if ( currentVerSlotIndex === 5 ||
                    currentVerSlotIndex === 10 ||
                    currentVerSlotIndex === 15 ||
                    currentVerSlotIndex === 20 )
                {
                    slots = this.flipArray( slots, prevVerArrayIndex, i - 1 );
                    prevVerArrayIndex = i;
                }
            }
            return this.flipArray( slots, prevVerArrayIndex, slots.length - 1 );
        },

        flipHorizontal: function ( slots )
        {
            var currentHorSlotIndex = 0;
            var prevHorArrayIndex = 0;
            var j;
            for ( j = 0; j < slots.length; ++j )
            {
                currentHorSlotIndex = slots[ j ].index;
                if ( currentHorSlotIndex === 5 ||
                    currentHorSlotIndex === 10 ||
                    currentHorSlotIndex === 15 ||
                    currentHorSlotIndex === 20 )
                {
                    slots = this.flipArray( slots, prevHorArrayIndex, j - 1 );
                    prevHorArrayIndex = j;
                }
            }
            slots = this.flipArray( slots, prevHorArrayIndex, slots.length - 1 );
            return this.flipArray( slots, 0, slots.length - 1 );
        },

        /**
         * rotates the slots using the slot indexes only. slots.filled stay in place
         * @param slots
         * @param width
         * @returns {*}
         */
        rotateRightAngle: function ( slots, width )
        {
            var indexes = [];

            var row;
            var col;
            for ( col = width - 1; col >= 0; --col )
            {
                for ( row = 0; row < width; ++row )
                {
                    indexes.push( slots[ row * width + col ].index );
                }
            }

            var index;
            for ( index = 0; index < slots.length; ++index )
            {
                slots[ index ].index = indexes[ index ];
            }

            return slots;
        },

        /**
         * reverses an array's order between 2 indexes
         * @param array
         * @param startIndex
         * @param endIndex
         * @returns {*}
         */
        flipArray: function ( array, startIndex, endIndex )
        {
            var new_array = [];
            var temp = null;

            var k;
            for ( k = startIndex; k <= endIndex; ++k )
            {
                new_array.push( array[ k ] );
            }

            var i;
            for ( i = 0; i < new_array.length / 2; ++i )
            {
                temp = new_array[ i ].index;
                new_array[ i ].index = new_array[ new_array.length - 1 - i ].index;
                new_array[ new_array.length - 1 - i ].index = temp;
            }

            var j;
            for ( j = startIndex; j <= endIndex; ++j )
            {
                array[ j ] = new_array[ j - startIndex ];
            }

            return array;
        }
    } );
}() );
