/*global cc*/
/*global require*/
/*global Game*/
/*global StarSearchSettings*/
/*global StarSearchModel*/
/*global StarSearchView*/
/*global LumosityUtils*/
/*global StageSelect*/
/*global audio*/
/*global LumosityButton*/
/*global LumosityButtonState*/
/*global LevelKey*/

var StarSearch;
( function()
{
    "use strict";
    require( "src/Game/Base/Game.js" );
    require( "src/Game/StarSearchSettings.js" );
    require( "src/Game/StarSearchModel.js" );
    require( "src/Game/StarSearchView.js" );
    require( "src/Utility/LumosityUtils.js" );
    require( "src/View/LumosityButton.js" );
    require( "src/Game/LevelKey.js" );

    StarSearch = Game.extend(
    {
        /** @lends StarSearch.prototype */
        
        /**
         *  StarSearch is the main class for the game, Star Search!
         *
         *  @constructs
         */
        _debugMode: null,
        _debugLabel1: null,
        _debugLabel2: null,
        _debugLabel3: null,
        _debugLabel4: null,
        _debugLabel5: null,
        _cheatMode: false,

        ctor:function()
        {
            this._super();

            // setup for debug mode
            if ( LumosityUtils.isDebugModeOn() )
            {
                var font = "fonts/MuseoSans-700.ttf";
                var fontSize = 25;
                var cheatText;
                if ( this._cheatMode )
                {
                    cheatText = "On";
                }
                else
                {
                    cheatText = "Off";
                }
                //add labels
                this._debugLabel1 = LumosityLabel.create( "Level: " + this.model.current_level.toString(), font, fontSize );
                this._debugLabel2 = LumosityLabel.create( "Screen Size: " + this.view.getContentSize().width + ", " + this.view.getContentSize().height, font, fontSize );
                this._debugLabel3 = LumosityLabel.create( "Star Size: " + "xxx", font, fontSize );
                this._debugLabel5 = LumosityLabel.create( "Cheat Mode: " + cheatText.toString(), font, fontSize );
                this._debugLabel1.setPosition( 70, 105 );
                this._debugLabel2.setPosition( 180, 80 );
                this._debugLabel3.setPosition( 130, 55 );
                this._debugLabel5.setPosition( 130, 130 );
                this.view._gameStage.addChild( this._debugLabel1, 200 );
                this.view._gameStage.addChild( this._debugLabel2, 200 );
                this.view._gameStage.addChild( this._debugLabel3, 200 );
                this.view._gameStage.addChild( this._debugLabel5, 200 );

                //add objects for the Debug Menu
                var debugMenuObject = {
                    "labelText": "Cheat Mode",
                    "callback": this.toggleCheatMode,
                    "target": this,
                    "closeMenuOnCallback": false };

                var debugMenuObject2 = {
                    "labelText": "Unlock Planets",
                    "callback": this.toggleUnlock,
                    "target": this,
                    "closeMenuOnCallback": true };

                this.view._debugMenu.addDebugButton( debugMenuObject );
                this.view._debugMenu.addDebugButton( debugMenuObject2 );

            }
        },

        //when the game loads
        onBegin: function()
        {
            //get screen data
            this.model.width =  this.view.getContentSize().width;
            this.model.height =  this.view.getContentSize().height;

            //play nebula intro animation
            this.view.nebulaParentNode.setupChildren();
            this.view.nebulaParentNode.playNebulaAnimation();

            //setup current level based on past metadata level
            this.calculateStartingDifficulty();

            //set up level select prompt from localized string
            this.view.levelSelectLabelTop.setString( localize ("levelSelectLabel1"));
            this.view.levelSelectLabelBottom.setString( localize( "levelSelectLabel2" ).replace( "%d", this.model.unlockedLevel ).replace( "%d", this.settings.maxLevel ));

            //set up level select buttons
            this.levelButtonSet();

            //bring level select button onscreen
            this.view.playAnimation( "menuIn" , this.drawLines );
            LumosityUtils.callFunctionOnCollection( this._components, "onBegin" );
        },

        //controls effect of enabling Debug Mode for QA (set in App during runtime)
        debugModeText: function ()
        {
            var cheatText;
            if ( this._cheatMode )
            {
                cheatText = "On";
            }
            else
            {
                cheatText = "Off";
            }

            //display values on screen
            this._debugLabel1.setString( "Level: " + this.model.current_level.toString() );
            this._debugLabel5.setString( "Cheat Mode: " + cheatText.toString() );
        },

        //cheat mode toggling for debug mode
        toggleCheatMode: function ()
        {
            this._cheatMode = !this._cheatMode;
        },

        //switches the flag to unlock all planet levels
        toggleUnlock: function ()
        {
            this._allUnlock = !this._allUnlock;
            this.onRestart();
        },

        //draw level select connecting lines
        drawLines: function ()
        {
            this.view.drawLines( this.model.unlockedLevel );
        },

        //start game and countdown
        playStartingAnimation: function ()
        {
            this._super();
        },

        onStartGameplay:function()
        {
            this._super();
            this.model.gameState = "game";
            this.model.gameTimer = 0;
            this._resetTrial();
        },

        //calculates users current level based on their previous final level         *
        calculateStartingDifficulty: function()
        {
            var startingLevel = 0;
            // look for app params
            if( lumosity.game.params && lumosity.game.params.last_results && lumosity.game.params.last_results.gameStats )
            {
                //get user's last level
                var prevLastLevel = lumosity.game.params.last_results.gameStats[ this.model.previousGameStatKey ];
                if( LumosityUtils.isNumeric( prevLastLevel ) )
                {
                    // previous max level exists, so pull it in
                    startingLevel = prevLastLevel;

                    // Make sure starting level is not too low
                    if ( startingLevel < this.settings.startingLevel )
                    {
                        startingLevel = this.settings.startingLevel;
                    }
                    // Make sure it is not too high (in case we have users who cheated or achieved higher max tile count on another device or older version)
                    if ( startingLevel > this.settings.maxLevel )
                    {
                        startingLevel = this.settings.maxLevel;
                    }
                    // Set session level in meta data
                    var metadata = this.model.metadata;
                    if ( metadata )
                    {
                        metadata[ metadataAggregateKeys.userLevel ] = this.model.current_level;
                    }
                }
            }

            // No previous results found
            if ( startingLevel === 0 )
            {
                startingLevel = this.settings.startingLevel;
            }
            if (this._allUnlock)
            {
                startingLevel = 200;
            }
            this.model.unlockedLevel = startingLevel;
        },

        onEndGameplay:function()
        {
            this.model.gamestate = "end";
            this.getRemainingStars();
            //update the score HUD
            this.getScoringComponent().setScore(this.model.score);

            // if the user started on their maximum level but they ended lower, lower their maximum level :(
            if ( this.model.selected_level == this.model.unlockedLevel && this.model.current_level < this.model.selected_level )
            {
                this.model.unlockedLevel = this.model.current_level;
            }
            //animate the stars out so they dont show up behind the score popup
            if(this.model.animating == false)
            {
                //if it's the first level point it to level 1
                if (this.model.arrayOfLevels[ this.model.current_level - 1 ] == undefined)
                {
                    //if you're on level one make sure the model is looking at the right info
                    this.view.animateStarsOut(
                    (this.model.arrayOfLevels[ this.model.current_level ][ LevelKey.ColumnFill ] * this.model.arrayOfLevels[ this.model.current_level ][ LevelKey.RowFill ] ),
                    this.settings.starAnimationLength,
                    this._super(),
                    this );
                }
                else
                {
                    this.view.animateStarsOut(
                        (this.model.arrayOfLevels[ this.model.current_level - 1 ][ LevelKey.ColumnFill ] * this.model.arrayOfLevels[ this.model.current_level - 1 ][ LevelKey.RowFill ] ),
                        this.settings.starAnimationLength,
                        this._super(),
                        this );
                }
            }
            else
            {
                this._super();
            }
        },

        onRestart: function()
        {
            this.getRemainingStars();
            this._super();
        },

        onUpdate: function( delta )
        {
            this._super( delta );
            this.model.onUpdate( delta );
            //check for hint time
            if (this.model.isHintingNow)
            {
                this.view.giveHint();
            }
            //keep the shooting star background anim going
            this.view.playStarShow( this.model.gameTimer );
            this.updateHud();
            if ( lu.isDebugModeOn() )
            {
                this.debugModeText();
            }
        },

        updateHud: function ()
        {
            this.view.setScore( this.model.score );
        },

        //setup level select buttons and activate functionality
        levelButtonSet:function()
        {
            for( var i = 0; i < this.view.levelSelectButtons.length; i++ )
            {
                //if the level button is lower than the user's highest record, enable it
                var levelNum = i * 20;
                if ( levelNum <= this.model.unlockedLevel )
                {
                    var color = this.settings.buttonColors[0];
                    var tintAction = cc.TintTo.create( 0.0, color.r, color.g, color.b );

                    //when it's the max available level
                    if ( levelNum == ( this.model.unlockedLevel - (this.model.unlockedLevel % 20)) )
                    {
                        this.view.levelSelectButtons[i].getParent().playActiveInAnimation();
                        this.view.activePlanet = this.view.levelSelectButtons[i].getParent();
                        color = this.settings.buttonColors[i];
                        tintAction = cc.TintTo.create( 0.0, color.r, color.g, color.b );
                        var rotateAction = cc.RotateTo.create( 50000, 1000000 );
                        this.view.levelSelectButtons[i].getParent().orbitTriangles.runAction( rotateAction );
                    }
                    else
                    {
                        this.view.levelSelectButtons[i].getParent().playUnlockedAnimation();
                    }
                    this.view.levelSelectButtons[i].getParent().circle.runAction( tintAction );
                    //if it's the lowest level, make it say 1 instead of 0 because people hate zeros
                    if ( levelNum === 0 )
                    {
                        this.view.levelSelectButtons[i].getParent().number.setString( "1" );
                    }
                    else
                    {
                        this.view.levelSelectButtons[i].getParent().number.setString( levelNum.toString());
                    }
                    this.view.levelSelectButtons[i].setReleasedCallback( this.levelSelectButtonReleased.bind( this, this.view.levelSelectButtons[i].getTag() ) );
                    this.view.levelSelectButtons[i].setEnabled(false);
                }
            }
        },

        //what each level select button actually does when released
        levelSelectButtonReleased: function ( buttonIndex )
        {
            //disable level select buttons
            this.view.disableLevelSelectButtons();

            this.model.selected_level = buttonIndex * 20;

            //if they have progressed towards next level, start them at their highest progress
            if (this.model.selected_level == this.model.unlockedLevel - (this.model.unlockedLevel % 20))
            {
                this.model.selected_level = this.model.unlockedLevel;
            }
            //if they selected the first planet, make it the first level
            if( this.model.selected_level == 0)
            {
                this.model.selected_level = 1;
            }

            // Set the current stage from the selected button.
            this.model.setStage( this.model.selected_level );

            //animate out current active button
            this.view.activePlanet.playActiveOutAnimation();
            var color = this.settings.buttonColors[0];
            var tintAction = cc.TintTo.create( 0.0, color.r, color.g, color.b );
            this.view.activePlanet.circle.runAction ( tintAction );

            //animate in the clicked button
            this.view.activePlanet =  this.view["button" + (buttonIndex + 1)];
            this.view.activePlanet.playSelectAnimation();
            var rotateAction = cc.RotateTo.create( 2, 30 );
            this.view.activePlanet.orbitTriangles.runAction( rotateAction );
            rotateAction = cc.RotateTo.create( 2, -30 );
            this.view.activePlanet.orbitInner.runAction( rotateAction );
            rotateAction = cc.RotateTo.create( 2, 60 );
            this.view.activePlanet.orbitOutter.runAction( rotateAction );
            color = this.settings.buttonColors[ buttonIndex ];
            tintAction = cc.TintTo.create( 0.0, color.r, color.g, color.b );
            this.view.activePlanet.circle.runAction ( tintAction );

            var delayForSelectAnim = cc.DelayTime.create (.65 );
            var removeLines = cc.CallFunc.create ( this._fadeOutLines, this );
            var removeMenu = cc.CallFunc.create ( this._selectedLevel, this );
            var delayForRemoveMenu = cc.DelayTime.create (.75 );

            var startGame = cc.CallFunc.create ( this.playStartingAnimation, this );
            var sequence = cc.Sequence.create( removeLines, delayForSelectAnim, removeMenu, delayForRemoveMenu, startGame );

            this.view.runAction( sequence );
        },

        _selectedLevel: function ()
        {
            this.view.playAnimation( "menuOut" );
        },

        _fadeOutLines: function ()
        {
            this.view.fadeOutLines();
        },

        //Deals with player input to star buttons
        factorResults: function( success, star )
        {
            //disable buttons so the user can't double press
            this.getRemainingStars();
            this.view.disableStars();
            star.selected = true;
            this.model.animating = true;
            //put remaining stars into an array

            //calculate score
            if( this._cheatMode )
            {
                success = true;
            }
            var pointsEarned = this.model.factorResults(success, this.settings);
            //play sound based on answer speed
            if( success )
            {
                audio.playEffect("sounds/levelUp.mp3");
                if( this.model.trialTimer >= this.settings.speedIndex1 )
                {
                    audio.playEffect( "sounds/answerCorrect.mp3" );
                }
                else if ( this.model.trialTimer >= this.settings.speedIndex2 )
                {
                    audio.playEffect( "sounds/answerCorrectX2.mp3" );
                }
                else if ( this.model.trialTimer >= this.settings.speedIndex2 )
                {
                    audio.playEffect( "sounds/answerCorrectX3.mp3" );
                }
                else
                {
                    audio.playEffect( "sounds/answerCorrectX4.mp3" );
                }
            }
            else
            {
                audio.playEffect( "sounds/answerWrong.mp3" );
            }


            //store trial info
            this._recordTrialMetadata( star, success, pointsEarned );

            this.view.showFeedbackOverStar( success, star, this.model.gameState );

            var delayForFeedback;
            if ( success )
            {
                delayForFeedback = cc.DelayTime.create ( .5 );
            }
            else
            {
                delayForFeedback = cc.DelayTime.create ( 1.1 );
            }
            var animateStarsOut = cc.CallFunc.create( this._animateStarsOut, this);
            //reset trial
            this.view.runAction( cc.Sequence.create(  delayForFeedback, animateStarsOut ));  //, delayForStarsOut, resetTrial ) );
        },

        getRemainingStars: function()
        {
            this.view.remainingStars = [];
            var cols;
            var rows;
            if( this.view.starArray != undefined)
            {
                cols = this.view.starArray.length;
            }


            for ( var i = 0; i < cols; i++ )
            {
                if( this.view.starArray[i] != undefined)
                {
                    rows = this.view.starArray[i].length;
                }
                for ( var j = 0; j < rows; j++ )
                {
                    if ( this.view.starArray[i][j] && !this.view.starArray[i][j].hidden )
                    {
                        this.view.remainingStars.push(this.view.starArray[i][j]);
                    }
                }
            }
        },

        _animateStarsOut: function ()
        {
            this.model.delayStarsAnimation = this.view.animateStarsOut(

                (this.model.arrayOfLevels[ this.model.current_level ][ LevelKey.ColumnFill ] *
                this.model.arrayOfLevels[ this.model.current_level ][ LevelKey.RowFill ] ),
                this.settings.starAnimationLength, this._resetTrial, this );
        },



        _recordTrialMetadata: function ( star, success, pointsEarned )
        {
            var level_details = this.model.arrayOfLevels[this.model.current_level];
            //trial metadata
            if (star)
            {
                this.model.diagram = this.model.createTrialDiagram(star);
            }
            var trial =
            {
                //Star Search csv
                color : level_details[ LevelKey.Color ],
                level : this.model.current_level,
                points : pointsEarned,
                color_component0 : level_details[ LevelKey.Texture ],
                columnCount : level_details[ LevelKey.ColumnCount ],
                columnFill : level_details[ LevelKey.ColumnFill ],
                diagram : this.model.diagram,
                distractorVariety : level_details[ LevelKey.DistractorVariety ],
                motion : level_details[ LevelKey.ConstantRotation ],
                palette : level_details[ LevelKey.Palette ],
                rotate : level_details[ LevelKey.SingleRotation ],
                rowCount : level_details[ LevelKey.RowCount ],
                rowFill : level_details[ LevelKey.RowFill ],
                selection: (success) ? "a" : star.starIndex.toString(),
                shape : level_details[ LevelKey.Shape ],
                up : (pointsEarned > 0) ? "T" : "F",
                types: this.model.starLevelInfo.possibleStars.length,
                counts: this.getCounts(),
                colors: this.model.distractor_colors_string.toString(),
                textures: this.model.distractor_textures_string.toString(),
                shapes: this.model.distractor_shapes_string.toString(),
                motions: this.model.distractor_motions_string.toString(),
                rotations: this.model.distractor_rotations_string.toString(),
                //common csv
                correct : success ? "T" : "F",
                response_time : this.model.response_time,
                time_offset : this.model.time_offset + ""

            };
            this.endTrial( trial, success, false, false, false, false );
        },

        //metadata - counts distractor types
        getCounts: function()
        {
            var counts = [0,0,0,0,0,0,0,0,0,0];
            for( var i = 0; i < this.model.starLevelInfo.filled2dArray.length; i++ )
            {
                for ( var j = 0; j < this.model.starLevelInfo.filled2dArray[i].length; j++)
                {
                    if(this.model.starLevelInfo.filled2dArray[i][j])
                    {
                        counts[this.model.starLevelInfo.filled2dArray[i][j].starIndex]++;
                    }
                }
            }
            counts = counts.toString();
            var endpoint = 20;
            for (var j = 0; j < counts.length; j++)
            {
                if( counts[j] == "0" && endpoint == 20 )
                {
                    endpoint = j-1;
                }
            }
            counts = counts.substring(0, endpoint);
            counts = counts.split("," ).join(";");
            return counts;

        },

        _resetTrial: function()
        {
            if( this.model.isGameRunning && this.model.gamestate != "end")
            {
                this.model.time_offset = this.model.gameTimer;
                //Initialize Metadata
                this.getMetadataComponent().beginMetadataTrial();
                this.model.trialTimer = 0;
                this.model.distractorDelayIndex = 0;
                this.model.hintGiven = false;

                this._fillStarArray();
                this.model.animating = false;

                this.model.setHintDelay(this.settings.hintDelay);
            }
        },

        _fillStarArray: function()
        {
            this.model.starLevelInfo = this.model.starLevelCreator.createAndReturnStarArray(
                this.model.arrayOfLevels[this.model.current_level],
                this.settings.palettes,
                this.view.centeredContent.getContentSize().width,
                this.view.centeredContent.getContentSize().height,
                this.settings.starAnimationLength );

            //set some metadata stuff for distractors
            this.saveDistractorData();

             //Pass it into the view to be created
            this.view.fillStarArray( this.model.starLevelInfo.filled2dArray, this.factorResults, this );

        },

        //metadata
        saveDistractorData: function()
        {
            //clear old data
            this.model.resetTrialData();
            var distractorData = this.model.starLevelInfo.possibleStars;
            for (var i = 0; i < distractorData.length; i++)
            {
                this.addColors(distractorData[i].arrayOfStarElements[1]);
                this.addTextures(distractorData[i].arrayOfStarElements[4]);
                this.addShapes(distractorData[i].arrayOfStarElements[0]);
                this.addMotions(distractorData[i].arrayOfStarElements[3]);
                this.addRotations(distractorData[i].arrayOfStarElements[2]);
            }

            var find = ',';
            var re = new RegExp(find, 'g');

            this.model.distractor_colors_string = this.model.distractor_colors.join(";");
            this.model.distractor_colors_string = this.model.distractor_colors_string.replace(re, ';' );

            this.model.distractor_shapes_string = this.model.distractor_shapes.join(";");
            this.model.distractor_shapes_string = this.model.distractor_shapes_string.replace(re, ';');

            this.model.distractor_motions_string = this.model.distractor_motions.join(";");
            this.model.distractor_motions_string = this.model.distractor_motions_string.replace(re, ';');

            this.model.distractor_rotations_string = this.model.distractor_rotations.join(";");
            this.model.distractor_rotations_string = this.model.distractor_rotations_string.replace(re, ';');

            this.model.distractor_textures_string = this.model.distractor_textures.join(";");
            this.model.distractor_textures_string = this.model.distractor_textures_string.replace(re, ';');
        },

        //metadata
        addColors: function ( color )
        {
            var hex = (this.componentToHex(color.r) + this.componentToHex(color.g) + this.componentToHex(color.b )).toUpperCase();
            this.model.distractor_colors.push(hex);

        },
        //metadata
        componentToHex: function (color)
        {
            var hex = color.toString(16);
            return hex.length == 1 ? "0" + hex : hex;
        },
        //metadata
        addShapes: function ( shape )
        {
            this.model.distractor_shapes.push(shape);
        },
        //metadata
        addMotions: function ( motion )
        {
            this.model.distractor_motions.push(motion);
        },
        //metadata
        addRotations: function ( rotation )
        {
            this.model.distractor_rotations.push(rotation);
        },
        //metadata
        addTextures: function ( texture )
        {
            this.model.distractor_textures.push(texture);
        },

        //set functionality for star buttons
        starButtonSet:function()
        {
            var colsIndex;
            var rowsIndex;
            var cols = this.view.starArray.length;
            var rows = this.view.starArray[0].length;

            for( colsIndex = 0; colsIndex< cols; colsIndex++ )
            {
                for( rowsIndex = 0; rowsIndex < rows; rowsIndex++ )
                {
                    if( this.view.starArray[colsIndex][rowsIndex] )
                    {
                        var star = this.view.starArray[colsIndex][rowsIndex];
                        //if this is the unique star, the player will get it right
                        if( this.view.starArray[colsIndex][rowsIndex].amIUnique )
                        {
//                            star.setReleasedCallback( this.factorResults.bind( this, true, star ) );
                        }
                        //if it is non unique, the player will get it wrong
                        else
                        {
//                            star.setReleasedCallback( this.factorResults.bind( this, false, star ) );
                        }
                    }
                }
            }
        },

        onFinish: function ()
        {
            this._super();
        },

        feedback_isLargest:function( feedbackWidget )
        {
            this.view.playAnimation( "problem" );
        },
        
        cue_showCountdown: function()
        {
            this.view.showCountdown();
        },

        _createSettings:function()
        {
            return new StarSearchSettings( "settings.json" );
        },

        _createModel:function()
        {
            return new StarSearchModel();
        },

        _createView:function( screenSize )
        {
            return new StarSearchView( screenSize, this.settings, this.model, this, true );
        }
    } );
}() );
