"use strict";
var Blender = cc.Class.extend(
{
    activateTime: -1,
    active: undefined,
    addIngredientEnabled: undefined,
    justTrashed: undefined,
    isTimerVisible: undefined,
    isTimerVisibleMilliseconds: undefined,
    metadata: undefined,
    serveEnabled: undefined,
    ingredients: undefined,
    startEnabled: undefined,
    trashEnabled: undefined,
    _now: undefined,
    previousTrashAlert: undefined,
    startTime: -1,
    duration: 0,

    /**
     * Example @see TestBlender
     */
    ctor: function(metadata)
    {
        this.activateTime = -1;
        this.startTime = -1;
        this.duration = 0;
        this.isTimerVisible = true;
        this.isTimerVisibleMilliseconds = -1;
        this.metadata = metadata;
        this.clear();
    },

    isSingle: function()
    {
        return this.ingredients.length < Blender.ingredientsMax;
    },

    /**
     * Sort ingredients.
     * Example @see TestBlender
     */
    addIngredient: function(ingredient, now)
    {
        if (undefined === now) {
            now=-1;
        }
        if (this.addIngredientEnabled && this.ingredients.length < Blender.ingredientsMax) {
            this.isTimerVisible = true;
            this.ingredients.push(ingredient);
            this.ingredients.sort(Array.NUMERIC);
            if (Blender.ingredientsMax <= this.ingredients.length) {
                this.addIngredientEnabled = false;
            }
            this.activateTime = now;
            this.startEnabled = true;
            this.trashEnabled = true;
            this.metadata.addIngredient(this, now);
            if (null != Blender.onAddIngredientSelected) {
                Blender.onAddIngredientSelected();
            }
            return true;
        }
        else {
            return false;
        }
    },

    /**
     * Explicitly set, or multiple of ingredients.
     * Example @see TestBlender
     * https://lumoslabs.atlassian.net/browse/WG-364
     */
    idealPeriod: function()
    {
        return this.ingredients.length * Blender.period;
    },

    /**
     * @param   disableJustTrashed  If true, was not just trashed and do not record metadata.  Test case:  2015-10-20 Serve wrong.  Expect 2 trials.  Got 3 trials.
     */
    trash: function(now, disableJustTrashed)
    {
        this._now = now;
        if (this.trashEnabled) {
            this.justTrashed = !disableJustTrashed;
            if (!disableJustTrashed) {
                this.metadata.endTrial(this, now, -1, 
                    this.ingredients[0],
                    false, 
                    0,
                    this.trashAlerting() ? "O" : "E");
            }
            this.clear();
            var trashed = this.ingredients.slice();
            this.ingredients = [];
            this.startEnabled = false;
            this.serveEnabled = false;
            this.addIngredientEnabled = true;
            this.activateTime = now;
            if (null != Blender.onAction) {
                Blender.onAction();
            }
            return trashed;
        }
        else {
            return null;
        }
    },

    trashAlerting: function()
    {
        return this.trashEnabled && this.active && this.idealPeriod() < this.duration && !this.serveEnabled;
    },

    /**
     * Expects duration was already set, by calling fraction.
     * @return  Label to play on trashAlert:  "none" or "active".  If not changed, return null.
     * Example @see TestBlender
     * 14/2/11 Overflow.  Patty expects to see how to trash  (14/3/3 +Joe Palmieri x6) 
     */
    trashAlert: function()
    {
        var alert = this.trashAlerting();
        var label = alert ? "active" : "none";
        if (this.previousTrashAlert != label) {
            this.previousTrashAlert = label;
        }
        else {
            label = null;
        }
        return label;
    },

    /**
     * Clicking cup does not serve.  It always trashes.  Would confuse subscribers.
     * Test case:  2014-04-04 Ben expects clicking cup may serve or trash.  But not if overflowing.
     */
    trashOrSelect: function(now)
    {
        if (this.trashEnabled) {
            return null != this.trash(now);
        }
        else {
            var selected = this.select(now);
            if (selected) {
                if (null != Blender.onSelect) {
                    Blender.onSelect();
                }
            }
            return selected;
        }
    },

    /**
     * Disable select.
     * @return if starting.
     * Example @see TestBlender
     */
    start: function(now)
    {
        this._now = now;
        if (this.startEnabled && !this.active) {
            this.startTime = now;
            this.activateTime = -1;
            this.duration = 0;
            this.startEnabled = false;
            this.serveEnabled = true;
            this.addIngredientEnabled = false;
            this.active = true;
            this.metadata.startTrial(this, now);
            if (null != Blender.onStart) {
                Blender.onStart();
            }
            if (null != Blender.onAction) {
                Blender.onAction();
            }
            return true;
        }
        else {
            return false;
        }
    },

    /**
     * If an ingredient inside, then trash.
     * Test case:  2014-11-06 Tap cup with ingredient just outside of trash button to select.  Cup becomes invisible.  Mark Scoptur expects cup never becomes invisible.  (2014-11-11 +Pranjal)
     */
    select: function(now)
    {
        if (this.addIngredientEnabled) {
            this.activateTime = now;
            return true;
        }
        else {
            return false;
        }
    },

    /**
     * Also select this blender.
     * On serve, hide the timer and schedule showing the timer when the animation would be finished.  2015-01-05 Feedback.  [~wouter] expects to hide timer.
     * Example @see TestBlender
     */
    serve: function(now)
    {
        this._now = now;
        if (this.serveEnabled && this.active) {
            this.clear();
            this.activateTime = -1;
            this.isTimerVisible = false;
            this.isTimerVisibleMilliseconds = now + Blender.hideTimerMilliseconds;
            if (null != Blender.onAction) {
                Blender.onAction();
            }
            return true;
        }
        else {
            return false;
        }
    },

    update: function(now)
    {
        this._now = now;
        if (!this.isTimerVisible) {
            if (this.isTimerVisibleMilliseconds <= now) {
                this.isTimerVisible = true;
            }
        }
        if (this.active) {
            this.serveEnabled = Blender.accuracyMinServeEnabled <= this.accuracy(this._now);
        }
    },

    /**
     * @param   now     Milliseconds.
     * @return If no ingredients, return 0.0.  fraction: 0.0 too early 0.5 full 1.0 too late. FractionClip [0.0 ... 1.0].
     * Example @see TestBlender
     */
    fraction: function(now)
    {
        if (undefined === now) {
            now=-1;
        }
        if (-1 == now) {
            now = this._now;
        }
        if (this.active) {
            this.duration = now - this.startTime;
        }
        var ideal = this.idealPeriod();
        if (ideal <= 0) {
            return 0.0;
        }
        else {
            var fraction = Math.min(1.0, 0.5 * this.duration / ideal);
            return fraction;
        }
    },

    /**
     * Example @see TestBlender
     */
    accuracy: function(now)
    {
        return this._accuracy(this.fraction(now));
    },

    /**
     * +2% multiplier to accuracy.  14/2/26 Ben expects more than 60 ms to receive 100% correct.  Expects when line is touching to get 100%.
     * https://lumoslabs.atlassian.net/browse/WG-465
     */
    _accuracy: function(fraction)
    {
        var accuracy = (Blender.toleranceMultiplier * Blender.accuracyMax) 
            * (0.5 - Math.abs(0.5 - fraction)) / 0.5;
        accuracy = Blender.accuracyServeScore(accuracy, fraction);
        return Math.min(Blender.accuracyMax, Math.ceil(accuracy));
    },

    clear: function()
    {
        this.duration = 0;
        this.ingredients = [];
        this.startEnabled = Blender.ingredientsMax <= 0;
        this.serveEnabled = false;
        this.trashEnabled = false;
        this.addIngredientEnabled = 1 <= Blender.ingredientsMax;
        this.active = false;
    }
});

/**
 * Shorten serving time window from 75% or higher to 85% or higher, to reward more patient attention than rapid clicking.  14/2/26 Joe expects to think between clicks.  14/2/26 Dave Wu expects to batch operations.  14/2/26 Eli expects not a click fest.
 * https://lumoslabs.atlassian.net/browse/WG-465
 */
Blender.accuracyMinServeEnabled = 90;
Blender.accuracyMinGood = 94;
Blender.accuracyMinExcellent = 98;
/**
 * 0: May start any blender immediately.
 * 1 or higher: Must add ingredient before starting.
 * Example @see TestBlender
 */
Blender.ingredientsMax = 2;
Blender.hideTimerMilliseconds = Math.round(22 / 30 * 1000);
/**
 * Milliseconds
 * 14/2/25 Eli may expect to try 10 seconds, 20 seconds
 */
Blender.period = 10000;
Blender.onAction;
Blender.onAddIngredientSelected;
Blender.onSelect;
Blender.onStart;
Blender.toleranceMultiplier = 1.02;
Blender.accuracyMax = 100.0;

Blender.correctLabel = function(correct, accuracy)
{
    var label;
    if (!correct) {
        label = "answerWrong";
    }
    else if (Blender.accuracyMinExcellent <= accuracy) {
        label = "answerCorrectExcellent";
    }
    else if (Blender.accuracyMinGood <= accuracy) {
        label = "answerCorrectGood";
    }
    else {
        label = "answerCorrect";
    }
    // ll.log("Blender.correctLabel: " + label);
    return label;
}

/**
 * WG-1018 If in group "early", skew accuracy to peak at minimal fraction within enabled serving duration, retaining same integral of accuracies as "control" which is a truncated triangle whose peak is in the center of the serving window.  14/4/22 [~bahroni] expects to split test serving score function peak in time window center (control) or beginning (early), hoping for higher DAU or smoother median nth session level.  Ben noted that some coworkers and subscribers do not realize the highest score per serve is in the middle of the serve window and they behave as if the highest serving score is at the beginning of the serve time window.
 * Example @see TestModel
 */
Blender.accuracyServeScore = function(accuracy, fraction)
{
    if (Blender.accuracyMinServeEnabled <= Math.ceil(accuracy)) {
        var fractionCenter = 0.5;
        var fractionRadius = fractionCenter - fractionCenter 
            * Blender.accuracyMinServeEnabled / Blender.toleranceMultiplier / Blender.accuracyMax;
        var fractionMin = fractionCenter - fractionRadius;
        accuracy = (Blender.accuracyMax * (fractionMin - fraction) + Blender.accuracyMax) 
            * (Blender.toleranceMultiplier * 2.0 - 1.0);
    }
    return accuracy;
}

Blender.isOverflow = function(fraction)
{
    var isOver = false;
    if (0.5 <= fraction) {
        isOver = Blender.prototype._accuracy(fraction) < Blender.accuracyMinServeEnabled;
    }
    return isOver;
}
