/*jslint node: true */
"use strict";

require("src/Game/BlenderView.js");
require("src/Game/CafeModel.js");
require("src/View/CCBReaderHelper.js");

var BlenderController = cc.Class.extend(
{
    blenders: undefined,
    model: undefined,

    ctor: function()
    {
    },

    /**
     * Clicking on timer tries to select the cup, which may throw away its ingredients.
     * Test case:  2014-11-06 Tap cup with ingredient just outside of trash button to select.  Expect cup visible.  Mark Scoptur expects cup never becomes invisible.  (2014-11-11 +Pranjal)
     * @param   blenderButtonParents    View of each blender to be listened to.
     * @param   ingredientButtons   Expects each button already constructed.
     * Example @see TestBlenderController
     */
    populate: function(model, blenderButtonParents, ingredientButtons, scrollButtons)
    {
        this.model = model;
        this.listenToBlenderButtons(blenderButtonParents);
        this.listenToButtons(ingredientButtons, this.addIngredient);
        this.listenToButtons(scrollButtons, this.scroll);
    },

    /**
     * If not playing, disable trash or select.  Test case:  2015-01-20 During 3-2-1 countdown.  Tap cup.  Anthony Chaseendran expects to not hear a sound.  Got a cup sound.
     */
    select: function(cup)
    {
        var mc = cup.getParent().getParent();
        if (this.model.playing) {
            return mc.model.trashOrSelect(this.model.now);
        }
    },

    start: function(button)
    {
        var mc = button.getParent();
        mc.model.start(this.model.now);
    },

    /**
     * Would be more MVC if model stored the arguments to apply to view next update.
     * Would also delay feedback until next frame.
     */
    serve: function(button)
    {
        var mc = button.getParent();
        var args = this.model.serveArguments(mc.model);
        if (args) {
            args = [mc].concat(args);
            BlenderView.serveFeedback.apply(this, args);
        }
    },

    /**
     * Clicking cup does not serve.  It always trashes.  Would confuse subscribers.
     * 14/4/4 Ben expects clicking cup may serve or trash.  But not if overflowing.
     */
    trash: function(button)
    {
        var mc = button.getParent();
        mc.model.trash(this.model.now);
    },

    /**
     * @param   target          Expects target has name like "ingredient_2".
     * @param   localMethod     Expects in the BlenderController instance's context.
     *
     * Bug reported here:
     * https://github.com/cocos2d/cocos2d-x/issues/9480
     */
    listenToSelected: function(target, localMethod)
    {
        // ll.log("BlenderController.listenToSelected: " + target.getName());
        cc.BuilderReaderHelper.addTargetWithActionForControlEvents(target, 
            this, localMethod, 
            cc.CONTROL_EVENT_TOUCH_DOWN);
    },

    listenToButtons: function(buttons, localMethod)
    {
        for (var b = 0; b < buttons.length; b++) {
            this.listenToSelected(buttons[b], localMethod);
        }
    },

    listenToBlenderButtons: function(blenders)
    {
        for (var b = 0; b < blenders.length; b++) {
            var blender = blenders[b];
            this.listenToSelected(blender.start, this.start);
            this.listenToSelected(blender.serve, this.serve);
            this.listenToSelected(blender.trashAlert, this.trash);
            this.listenToSelected(blender.timer_1.select, this.select);
            this.listenToSelected(blender.timer_2.select, this.select);
        }
    },

    /**
     * Auto-place in next available blender.
     * Replaces current selection.
     * If already selected, clear.
     * @param   e   Expects target has name like "ingredient_2".
     */
    addIngredient: function(target, controlEventIgnored)
    {
        // ll.log("BlenderController.addIngredient: " + target.getName() + " event " + controlEventIgnored);
        var name = target.getName();
        var selectedIndex = parseInt(name.split("_")[1]);
        this.model.addIngredientToBlenderAvailable(selectedIndex, this.model.now);
    },

    /**
     * Scroll time 0.1.  2014-03-21 Scroll 0.25 seconds.  Ben expects faster scroll.
     * Scrolls parent of button.  Disables button during transition, then reenables.
     */
    scroll: function(button, controlEventIgnored)
    {
        this.model.scroll(button.getName());
    }
});
