/*jslint node: true */
"use strict";

require("src/View/CCBReaderAdapter.js" );

require("src/Game/Blender.js");
require("src/Game/CafeModel.js");

var BlenderView = cc.Class.extend(
{
    blenders: undefined,
    model: undefined,

    ctor: function()
    {
    },

    /**
     * Clicking on timer tries to select the cup, which may throw away its ingredients.
     * Test case:  2014-11-06 Tap cup with ingredient just outside of trash button to select.  Expect cup visible.  Mark Scoptur expects cup never becomes invisible.  (2014-11-11 +Pranjal)
     * @param   modelBlenders   Blenders in the model to bind each view to a model.
     * @param   blender     Assumes at least one blender already placed at "blender".  
     *                      Place each blender starting from left of "blender".
     * @param   ingredientButtons   Expects each button already constructed.
     * Example @see TestBlenderController
     */
    populate: function(model, layouts, ingredientButtons)
    {
        this.model = model;
        this.blenders = [];
        this.blenders = this.populateBlenders(model, layouts);
    },

    update: function()
    {
        if (this.blenders) {
            for (var b = 0; b < this.blenders.length; b++) {
                var mc = this.blenders[b];
                try {
                    BlenderView.view(mc, this.blenders, this.model.blenderAvailable());
                }
                catch (err) {
                    var audit = mc.getName();
                    ll.error(audit + ": " + err);
                }
            }
        }
    },

    /**
     * GAME-2403 If loaded, then reset loaded view of level.
     */
    isLoaded: function()
    {
       return undefined != this.loadedLevels && false != this.loadedLevels[this.model.level];
    },

    /**
     * Expects first blender was populated.
     * Expects layouts were populated.
     *
     * XXX Do not reparent.  Test case:  2015-01-13 Cocos2d-x.  How to Play mode.  See coffee machine.  Skip tutorial.  Select level 1.  See cafe.  Game stops responding.  CPU 100%.  Appears to be adding nodes.  2015-01-11 Cocos2d-x.  Play.  How to Play.  cocos2d-html5 ok.  Test case:  2015-01-13 Level 3.  Restart.  Level 2.  Cafe slides into view.  Expect 3-2-1 countdown.  Got no response.  CPU 100%.  
     */
    populateBlenders: function(model, layouts)
    {
        if (null == this.blenders) {
            this.blenders = [];
        }
        var modelBlenders = model.blenders;
        for (var b = 0; b < modelBlenders.length; b++) {
            try {
                var roomIndex = model.roomAt(b);
                var remainder = model.roomRemainder(b);
                var room = layouts[roomIndex];
                var name = "blender_" + remainder;
                var next = room[name];
                if (!next.timer_1) {
                    next.removeAllChildren();
                    this.loadedLevels = this.loadedLevels || {};
                    this.loadedLevels[model.level] = true;
                    var cacheName = "blender_" + b + "_level_" + model.level;  // XXX I feel confused with stub key.
                    next = CCBReaderAdapter.adoptOrCreate(cacheName, "BlenderLoaded", room[name]);
                    cc.BuilderReaderHelper.replace(room, name, next);
                }
            }
            catch (err) {
                throw new TypeError("Expected to find layout " + roomIndex
                    + " remainder " + remainder + " of blender " + b 
                    + " model this.blenders " + modelBlenders.length);
            }
            this.construct(next, modelBlenders[b], this.blenders);
            this.blenders.push(next);
            BlenderView.view(next, this.blenders);
        }
        return this.blenders;
    },

    /**
     * Stop available on any label and disable mouse.
     * Disable mouse on feedback text. 14/1/10 Serve. Refill. Expect to fix can click start.
     * https://trello.com/c/dD7xXA96/947-13-multitasking-aaron-and-dave-wu-expect-to-play-a-simple-game-with-a-format-like-papa-s-burgeria
     */
    construct: function(mc, modelBlender, blenders)
    {
        mc.model = modelBlender;
        IngredientsView.init(mc, IngredientsView.getIngredientClasses());
        cc.BuilderReaderHelper.runAnimations(mc.feedback_2, "none", true);
        cc.BuilderReaderHelper.runAnimations(mc.feedback_1, "none", true);
        this.populateFraction(mc.timer_2);
        this.populateFraction(mc.timer_1);
        mc.timers = [mc.timer_1, mc.timer_2];
        mc.timer = mc.timers[0];
    },

    populateFraction: function(timer)
    {
        if (null == timer.fraction) {
            throw new Error("Expected a child of timer to be named 'fraction'.");
        }
    }
});

BlenderView.periodClasses = ["CupSingle", "CupSingle", "CupDouble"];

/**
 * Feedback percent.
 *
 * If not correct, also select feedback to display.
 * Test case: 2014-12-18 Level 9. Serve wrong order?  Matt Leung expects feedback.  Got crash.
 * Reuse feedback.  Serve.  At blender, animate with text like 95%.  Test case:  Serve.  2014-01-09 Stephanie expects to feel aware of success.
 * https://trello.com/c/dD7xXA96/947-13-multitasking-aaron-and-dave-wu-expect-to-play-a-simple-game-with-a-format-like-papa-s-burgeria
 * Test case:  2014-03-24 Serve correctly. Wouter expects program to play custom animation by cup size.
 * Measure cup before serving.  2014-03-31 Wouter expects to fix select double accuracy feedback for 2 ingredients.
 * GAME-588 Wrong.  animate feedback.  2014-04-03 Wrong. Wouter expects to fix to see wrong feedback animation.
 */
BlenderView.serveFeedback = function(mc, single, correct, label, points)
{
    var string = correct ? points.toString() : "";
    var feedback = single ? mc.feedback_1 : mc.feedback_2;
    feedback.txt.setString(string);
    cc.BuilderReaderHelper.runAnimations(feedback, label);
    Sounds.answer(label);
}

BlenderView.show = function(obj)
{
    obj.setVisible(true);
}

/**
 * Only view start xor serve button, because start and serve buttons overlap.
 * https://lumoslabs.atlassian.net/browse/WG-368
 */
BlenderView.view = function(mc, blenders, availableIndex)
{
    if (undefined === availableIndex) {
        availableIndex=-1;
    }
    var blender = mc.model;
    mc.start.setVisible(true);
    mc.start.setOpacity(255);
    mc.start.setEnabled(blender.startEnabled);
    if (blender.serveEnabled) 
    {
        var breakHere = true;
    }
    mc.serve.setEnabled(blender.serveEnabled);
    mc.serve.setVisible(blender.serveEnabled);
    IngredientsView.setIngredients(mc.ingredients, blender.ingredients, 
        IngredientsView.getIngredientClasses());
    BlenderView.viewFraction(mc);
    BlenderView.viewAvailable(mc, blenders, availableIndex);
    BlenderView.viewTrash(mc);
}

/**
 * If just trashed, then play animation.
 * Program expects each Blender has child named "trash" and "trashAlert".
 * 14/3/17 Wouter expects available and trash inside timer.
 * WG-444
 * 14/3/31 Level 3, 5.  Overflow.  Guest Tori may expect to see button to trash.
 * WG-587
 */ 
BlenderView.viewTrash = function(mc)
{
    if (mc.model.justTrashed) {
        mc.model.justTrashed = false;
        cc.BuilderReaderHelper.runAnimations(mc.trash, "selected");
    }
    mc.trashAlert.setEnabled(mc.model.trashAlerting());
    mc.trashAlert.setVisible(mc.model.trashAlerting());
}

/**
 * Goto and play if movieclip mc.available is not already "none" or "active".
 * @return  new label or null if not changing labels.
 * available:  Open and emphasize next available blender on screen. Preview which blender ingredient will go to. 14/1/19 Level 5. Play frenetically. Hover over ingredient. Stephanie expects to see the ingredient will go to 3rd blender. Got confused and thought ingredient would go to 4th blender that had one ingredient already.
 * https://lumoslabs.atlassian.net/browse/WG-233
 * Example @see TestBlenderView
 * When this blender is available to add ingredients into, this timer plays "enable" label or "enabled" label.  When not available, timer plays "disabled" label.  14/3/17 Wouter expects available pops forward.
 * https://lumoslabs.atlassian.net/browse/WG-444
 */
BlenderView.viewAvailable = function(mc, blenders, availableIndex)
{
    var index = blenders.indexOf(mc);
    if (index <= -1) {
        throw new Error("Expected " + mc + " in blenders " + blenders);
    }
    var isEnabled = index != availableIndex || mc.model.trashEnabled;
    mc.timer_1.select.setEnabled(isEnabled);
    mc.timer_2.select.setEnabled(isEnabled);
}

/**
 * May show a different timer.
 * https://lumoslabs.atlassian.net/browse/WG-364
 *
 * To animate tip forward, animate cup transition and fraction not visible until last frame of "disabled".  And fraction not visible in "enabled" label.
 */
BlenderView.viewFraction = function(mc)
{
    var periodClassIndex = Math.max(0, mc.model.ingredients.length - 1);
    var single = 0 == periodClassIndex;
    var fraction = mc.model.fraction();
    var label = 0 < fraction ? "fraction" : "none";
    label = CafeTutor.fractionState(mc.model, label);
    var timer = single ? mc.timer_1 : mc.timer_2;
    cc.BuilderReaderHelper.setState(timer, label);
    cc.BuilderReaderHelper.setState(timer.fraction.content.clippingNode.root, label);
    var isVisible = mc.model.isTimerVisible;
    try {
        mc.timer_1.setVisible(isVisible && single);
        mc.timer_2.setVisible(isVisible && !single);
    }
    catch (err) {
        ll.error(mc.getName() + ": " + err);
    }
}
