/*jslint node: true */
"use strict";

require("src/Utility/Count.js");
require("src/Utility/LumosityUtils.js");
require("src/Game/Blender.js");
require("src/Game/CafeTutor.js");
require("src/Game/Metadata.js");


/**
 * Abstract algorithms and data to evaluate rules of managing timers.
 */
var CafeModel = cc.Class.extend(
{
    accuracies: undefined,
    accuracy: undefined,
    /**
     * @return  If any blender is active.  2014-09-03 Sarah expects coffee brew sound is only heard when one or more coffee machines are flowing.
     */
    active: false,
    blenders: [],
    blendersPerRooms: undefined,
    customerRoom: 0,
    /**
     * What each customer is ordering right now, which visually may correspond to blenders.
     */
    customers: undefined,
    enabled: undefined,
    level: undefined,
    ingredientsVisible: true,
    metadata: undefined,
    nextTrialTime: undefined,
    now: 0,
    num_blenders: undefined,
    num_ingredients: undefined,
    num_ingredients_visible: undefined,
    num_correct: 0,
    num_trials: undefined,
    numRooms: undefined,
    occupied: undefined,
    /**
     * Expects an even number.
     */
    orderLengthConsecutiveMax: 6,
    orderLengthMax: Blender.ingredientsMax,
    orderLengthMin: 1,
    /**
     * Array of indexes.
     */
    orders: undefined,
    ordersQueue: undefined,
    ordersStarted: undefined,
    /**
     * Most recently added to score.
     */
    points: 0,
    score: 0,
    roomIndex: 0,
    /**
     * Ingredient.
     */
    selected: -1,
    playing: false,
    secondsRemaining: 0.0,
    isScrollStarting: false,
    scrollEnabled: false,
    scrollEnableTime: undefined,
    scrollX: 0,
    scrollXPrevious: 0,
    time: 120,
    timeLimit: 120,
    trial: 0,
    trialPeriod: 4000,
    trialPeriodMin: 250,
    tutor: undefined,
    user_level: undefined,
    consecutive: 0,
    consecutiveGap: 0,
    consecutiveLengthMin: undefined,
    customerStates: [],
    customersServed: undefined,
    /**
     * Aaron may want to log which orders were displayed.
     */
    orderHistory: undefined,
    orderLengths: [],
    previousOrder: undefined,
    remindRoomTime: undefined,

    ctor: function()
    {
        this.active = false;
        this.blenders = [];
        this.customerRoom = 0;
        this.ingredientsVisible = true;
        this.now = 0;
        this.num_correct = 0;
        this.orderLengthConsecutiveMax = 6;
        this.orderLengthMax = Blender.ingredientsMax;
        this.orderLengthMin = 1;
        this.points = 0;
        this.score = 0;
        this.roomIndex = 0;
        this.selected = -1;
        this.playing = false;
        this.secondsRemaining = 0.0;
        this.isScrollStarting = false;
        this.scrollEnabled = false;
        this.scrollEnableTime = CafeModel.FOREVER;
        this.scrollX = 0;
        this.scrollXPrevious = 0;
        this.time = 120;
        this.timeLimit = 120;
        this.trial = 0;
        this.trialPeriod = 4000;
        this.trialPeriodMin = 250;
        this.consecutive = 0;
        this.consecutiveGap = 0;
        this.customerStates = [];
        this.orderLengths = [];
    },

    /**
     * If level 0 or less, then 
     *      as much time as you like.
     *      ingredients start invisible.
     * 14/2/5 Wouter expects to see total time.
     * 14/2/20 Stephanie got confused.
     * https://lumoslabs.atlassian.net/browse/WG-364
     */
    populate: function(params)
    {
        this.now = 0;
        this.num_correct = 0;
        this.accuracy = 0;
        this.score = 0;
        this.orders = [];
        this.orderHistory = [];
        this.playing = false;
        this.roomIndex = 0;
        this.occupied = -1;
        this.trial = 0;
        this.ordersStarted = 0;
        this.scrollX = 0;
        this.selected = -1;
        this.scrollReset();
        this.accuracies = [];
        this.previousOrder = [];
        this.scrollEnabled = false;
        this.consecutiveLengthMin = false;
        for (var param in params) {
            this[param] = params[param];
        }
        this.numRooms = CafeModel.roomCount(this.num_blenders) + this.customerRoom;
        this.user_level = Math.max(CafeModel.session_level, this.level);
        this.ingredientsVisible = 1 <= this.level;
        this.setTimeLimit(this.timeLimit);
        if (null == this.blendersPerRooms) {
            this.blendersPerRooms = CafeModel.distribute(this.num_blenders);
        }
        this.metadata = new Metadata(this);
        this.blenders = CafeModel.addBlenders(this.blenders, this.num_blenders, this.metadata);
        if (1 <= this.customerRoom) {
            for (var c = 0; c < this.customerRoom; c++) {
                if (0 != this.blendersPerRooms[c]) {
                    this.blendersPerRooms.unshift(0);
                }
            }
        }
        this.updateShowIngredients(this.roomIndex, this.customerRoom);
        this.customers = [];
        this.customerStates = [];
        var b;
        for (b = 0; b < this.num_blenders; b++) {
            this.blenders[b].metadata = this.metadata;
            this.customers.push(null);
            this.customerStates.push("none");
        }
        this.orderLengths = [];
        this.customersServed = [];
        this.trialPeriod = this.calculateTrialPeriod();
        this.nextTrialTime = this.now + this.trialPeriod / 4;
        this.remindRoomTime = undefined;
        this.tutor = new CafeTutor(this);
        // ll.log("CafeModel.populate: this.now " + this.now + " this.trialPeriod " + this.trialPeriod + " this.nextTrialTime " + this.nextTrialTime);
    },

    setTimeLimit: function(timeLimit)
    {
        this.timeLimit = timeLimit;
        this.time = this.level <= 0 ? CafeModel.FOREVER : this.timeLimit;
        this.secondsRemaining = this.time; 
    },

    /**
     * @return Layouts {category, count, x, y}.  Category 0: customers and blenders or blenders only, 1: customers only.  count: number of blenders on that screen or all screens in some rooms.
     * 14/3/28 Wouter expects to customize tutorial layout.
     * WG-585
     * WG-1021 Index 8 or 9, for layout room 2. 14/6/3 Level 9 and 10:  Julie expects custom layout of 3 or 4 coffee machines when on the lefthand side.
     * GAME-2034 Layout customers at index equal to number of blenders.  2014-12-02 Level 5.  Vishu expects to fix layout.
     */
    populateLayouts: function(layout_0_x, layout_0_y, width, customOffsets)
    {
        if (undefined === customOffsets) {
            customOffsets=null;
        }
        var layouts = [];
        for (var roomIndex = 0; roomIndex < this.numRooms; roomIndex++) {
            var layoutIndex = this.blendersPerRooms[roomIndex];
            var category = 0;
            if (0 == roomIndex && (3 == this.numRooms || 5 == this.num_blenders || 6 == this.num_blenders)) {
                category = 1;
                layoutIndex = this.num_blenders;
            }
            else if ((3 == this.numRooms && 2 == roomIndex) 
                  || (2 == this.numRooms && 0 == roomIndex)) {
                layoutIndex += 6;
            }
            var layout = {};
            layout.category = category;
            layout.count = layoutIndex;
            var x = CafeModel.roomOffsets[roomIndex];
            var offset = 0;
            if (customOffsets) {
                offset = customOffsets[roomIndex];
            }
            layout.x = x * width + layout_0_x + offset;
            layout.y = layout_0_y;
            layouts.push(layout);
        }
        return layouts;
    },

    /**
     * Faster with more blenders.  Slower with more average ingredients.
     * Minimum period between trials.
     * https://lumoslabs.atlassian.net/browse/WG-232
     * Sample @see TestModel
     * WG-444 Double rate of customer arrivals.  14/3/13 Level 7.  Ben expects to have customers to serve at all times.
     */
    calculateTrialPeriod: function()
    {
        var factor = 
                            // 0.75;
                            1.0;
                            // 1.5;
        var power = // 1.0;
                           1.5;
                           // 2.0;
        var trialPeriod = factor * (this.orderLengthMin + this.orderLengthMax) * 0.5 * Blender.period 
            / Math.pow(this.num_blenders, power);
        trialPeriod = Math.max(this.trialPeriodMin, trialPeriod);
        return trialPeriod;
    },

    /**
     * Minimum period between trials if any vacant customer.
     * If next trial now, call the next trial.
     * @return  if all complete
     * GAME-235 During model update, update current time of each blender.  Controller and view refer to model's current time.  Test case:  2014-12-04 Level 9. Fill cup.  Melissa Malski expects to fix serve button appears.
     * GAME-691 Orders are removed when started.  Test case:  2014-04-02 Subscriber Tereza expects to see future orders.  2014-04-08 Abhishek expects started order may be overwritten.
     * GAME-2456 If game is complete, set each timer to be visible.  Test case:  2015-01-05 [~wouter] expects cup to disappear during feedback.  2015-01-13 [~sclark] expects an exception to the rule:  that during the last cup of a complete order, the cup is visible.
     */
    update: function(now, secondsRemaining)
    {
        this.now = now;
        this.secondsRemaining = secondsRemaining;
        if (this.playing && this.nextTrialTime <= now && CafeModel.nextTrialMinSecondsRemaining <= secondsRemaining) {
            if (this.moreTrials()) {
                if (0 <= CafeModel.vacant(this.customers)) {
                    this.nextTrialTime += this.trialPeriod;
                    this.nextTrialTime = Math.max(now + this.trialPeriodMin, this.nextTrialTime);
                    this.nextTrial();
                }
            }
        }
        this.updateScroll();
        if (this.playing) {
            for (var b = 0; b < this.blenders.length; b++) {
                this.blenders[b].update(now);
            }
            this.updateShowIngredients(this.roomIndex, this.customerRoom);
            this.updateCustomers();
        }
        this.active = this.playing && 1 <= this.countBlendersActive();
        var complete = this.isComplete();
        if (complete) {
            for (var b = 0; b < this.blenders.length; b++) {
                this.blenders[b].isTimerVisible = true;
            }
        }
        return complete;
    },

    endGame: function()
    {
        this.tutor.isUpdating = false;
        this.playing = false;
        this.user_level = this.nextSessionLevel();
        this.metadata.endGame(this.now);
    },

    /**
     * @param   previousOrder   Next order different from previous, unless length 0.  If no previous, then minimum number of newest ingredient.
     * @return  Random array of ingredients.  Each ingredient is the same.
     * Example @see TestModel
     */
    randomNonconsecutiveOrder: function(length, previousOrder)
    {
        var order = [];
        var ingredient;
        if (null == previousOrder || previousOrder.length <= 0) {
            ingredient = this.num_ingredients - 1;
        }
        else {
            ingredient = this.randomIndex(this.num_ingredients);
        }
        for (var i = 0; i < length; i++) {
            order.push(ingredient);
        }
        if (CafeModel.arraysEqual(order, previousOrder)) {
            var index = this.randomIndex(length);
            var exclude = previousOrder[index];
            ingredient = this.randomIndex(this.num_ingredients - 1);
            if (ingredient == exclude) {
                ingredient++;
            }
            order = [];
            for (i = 0; i < length; i++) {
                order.push(ingredient);
            }
        }
        return order;
    },

    randomIndex: function(length)
    {
        var index = Math.floor(Math.random() * length);
        return index;
    },

    /**
     * Shuffle deck to order ingredient counts:  1, 1, 1, 2, 2, 2.
     * To tutor, first is always 1-ingredient, next is always 2-ingredients.
     * Example @see TestModel
     * https://lumoslabs.atlassian.net/browse/WG-232
     * 14/3/4 Customer asks for 2 ingredients. Mason expects to realize need two ingredients.  14/3/11 Patrick Lutz expects to tutor 2 ingredients.
     * https://lumoslabs.atlassian.net/browse/WG-21
     */
    shiftOrderLength: function(first)
    {
        if (undefined === first) {
            first=false;
        }
        if (this.orderLengths.length <= 0) {
            var duplicates = this.orderLengthConsecutiveMax / 2;
            for (var orderLength = this.orderLengthMin; orderLength <= this.orderLengthMax; orderLength++) {
                for (var d = 0; d < duplicates - 1; d++) {
                    this.orderLengths.push(orderLength);
                }
            }
            if (first) {
                var min = this.orderLengths.shift();
                var max = this.orderLengths.pop();
            }
            LumosityUtils.shuffle(this.orderLengths);
            if (first) {
                this.orderLengths.unshift(max);
                this.orderLengths.unshift(min);
            }
        }
        var length = 
                        this.orderLengths.shift();
        // ll.log("CafeModel.shiftOrderLength: " + length + " from " + this.orderLengths);
        return length;
    },

    /**
     * Level 9 to 11 has some consecutive repetitions of double orders and some single orders that do not repeat.  Level 9 has triplets between every 4 nonconsecutive orders.  Level 10 has doubles every 4 nonconsecutive orders.  Level 11 has triplets every 8 nonconsecutive orders.
     * 14/3/21 Level 9.  Ben expects to feel somewhat more challenged more than level 8.  Got overwhelmed.
     * https://lumoslabs.atlassian.net/browse/WG-577
     */
    pushConsecutive: function(ordersQueue, t)
    {
        if (1 <= this.consecutive && (this.consecutiveGap == 0 
                              || t % this.consecutiveGap == (this.consecutiveGap - 1))) {
            var length;
            var order;
            var c;
            if (this.consecutiveLengthMin) {
                length = this.orderLengthMin;
                for (c = 0; c < this.consecutive; c++) {
                    order = this.randomNonconsecutiveOrder(length, this.previousOrder);
                    ordersQueue.push(order.slice());
                }
            }
            else {
                length = this.orderLengthMax;
                order = this.randomNonconsecutiveOrder(length, this.previousOrder);
                for (c = 0; c < this.consecutive; c++) {
                    ordersQueue.push(order.slice());
                }
            }
            this.consecutiveLengthMin = !this.consecutiveLengthMin;
        }
    },

    nextOrder: function(num_trials)
    {
        if (null == this.ordersQueue || this.ordersQueue.length <= 0) {
            this.ordersQueue = [];
            var order;
            var length;
            for (var t = 0; t < num_trials; t++) {
                length = this.shiftOrderLength(t + this.trial <= 0);
                order = this.randomNonconsecutiveOrder(length, this.previousOrder);
                this.ordersQueue.push(order);
                this.pushConsecutive(this.ordersQueue, t);
                this.previousOrder = this.ordersQueue[this.ordersQueue.length - 1];
            }
        }
        return this.ordersQueue.shift();
    },

    /**
     * On select tutorial or end tutorial, stop playing.
     */
    tutorial: function()
    {
        this.playing = false;
    },

    /**
     * GAME-2016 Serve 0.  2015-01-05  Aaron expects num_correct is 0.  Got 1.
     */
    onStartGameplay: function()
    {
        this.playing = true;
        this.num_correct = 0;
        this.accuracy = 0;
        this.score = 0;
        this.tutor.isUpdating = true;
        this.scrollEnable();
    },

    /**
     * @param   order   Explicit ingredients.  If null, random.
     * @return  Array of order.
     * Example @see TestModel
     */
    nextTrial: function(order)
    {
        if (undefined === order) {
            order=null;
        }
        if (null == order) {
            order = this.nextOrder(this.num_trials);
        }
        // ll.log("CafeModel.nextTrial: " + order);
        this.metadata.nextOrder(this.orderHistory.length, this.now);
        this.orderHistory.push(order.slice());
        this.orders.push(order);
        this.previousOrder = order;
        this.trial++;
        return order;
    },

    /**
     * @return  Iff all trials correct.
     */
    allCorrect: function()
    {
        return this.num_trials <= this.num_correct;
    },

    /**
     * If less than max.
     */
    mayDistributeIngredients: function()
    {
        return this.num_ingredients < CafeModel.levelParams[CafeModel.levelParams.length - 1].num_ingredients;
    },

    /**
     * Of highest level.
     */
    maxIngredientIndex: function()
    {
        return CafeModel.levelParams[CafeModel.levelParams.length - 1].num_ingredients - 1;
    },

    /**
     * @return  Partial arguments for serveFeedback: [single, correct, label, points].  If no ingredients or already served, return null.
     */
    serveArguments: function(blender)
    {
        var single = blender.isSingle();
        var accuracy = blender.accuracy(this.now);
        var index = this.serve(blender, this.now);
        if (-2 == index) {
            return null;
        }
        var correct = 0 <= index;
        var label = Blender.correctLabel(correct, accuracy);
        return [single, correct, label, this.points];
    },

    /**
     * Guards against spam.
     * @return index of order that was first served, or -1.  If spam, return -2.
     * Example @see TestModel
     */
    serve: function(blender, now)
    {
        var ingredients = blender.ingredients;
        if (ingredients.length <= 0) {
            return -2;
        }
        var orderIndex = CafeModel.indexOf(this.orders, ingredients);
        var order_id = -1;
        var ingredient = ingredients[0];
        var correct = 0 <= orderIndex;
        var isRecording = false;
        if (correct) {
            var serveAccuracy = blender.accuracy(now);
            var serveFraction = blender.fraction(now);
            if (blender.serve(now)) {
                // ll.log("CafeModel._serve: this.accuracy " + serveAccuracy + " fraction " + serveFraction);
                this.points = CafeModel.score(1, serveAccuracy, 0);
                this.score += this.points;
                this.num_correct++;
                this.accuracies.push(serveAccuracy);
                this.accuracy = CafeModel.averageNumber(this.accuracies);
                this.orders.splice(orderIndex, 1);
                order_id = this.metadata.order_ids.splice(orderIndex, 1);
                isRecording = true;
            }
            else {
                orderIndex = -2;
            }
        }
        else {
            blender.trash(now, true);
            this.metadata.num_trash++;
            isRecording = true;

        }
        if (isRecording) {
            this.metadata.endTrial(blender, now, order_id, 
                ingredient,
                correct, 
                correct ? this.points : 0,
                correct ? "T" : "I");
        }
        return orderIndex;
    },

    /**
     * @return If may end game.
     */
    isComplete: function()
    {
        if (!this.playing) {
            return false;
        }
        return this.trialCompleted();
    },

    /**
     * All correct, after animation.
     */
    trialCompleted: function()
    {
        return this.num_trials <= this.num_correct;
    },

    /**
     * If more trials could be spawned.  Not necessarily completed.
     */
    moreTrials: function()
    {
        return this.orderHistory.length < this.num_trials;
    },

    updateCustomers: function()
    {
        this.customers = CafeModel.demand(this.orders, CafeModel.ingredientsStarted(this.blenders), this.customers);
    },

    /**
     * @param   currentLabel    Animation state, so as to detect transition from "appear" to "active".
     * @return  State of customer now.
     *      "none"      Vacant.
     *      "appear"    New order appearing.
     *      "active"    Waiting to start.
     *      "serve"     Serving.
     *
     * WG-587 14/4/2 Kristi expects to not see dim until click start.
     * WG-691 14/4/2 Subscriber Tereza expects to see future orders.  14/4/8 Abhishek expects started order may be overwritten.
     * WG-699 If new order immediately replaces old order, replay "appear".  14/4/16 New order replaces old order. Wouter expects to replay order "active" animation.
     * GAME-1872 When a new order appears, play "appear" animation.  2014-12-15 Next order.  Wouter expects to see "appear" animation.  Got static.
     */
    customerState: function(c, currentLabel)
    {
        if (undefined === currentLabel) {
            currentLabel="none";
        }
        var label;
        var previous = this.customerStates[c];
        if (null == this.customers[c]) {
            label = (previous == "appear" 
                  || previous == "serve") ? "serve" 
                : "none";
        }
        else {
            if (previous != "active" && (previous != "appear" || currentLabel == "appear")) {
                label = "appear";
            }
            else {
                label = "appear";
            }
        }
        if ("serve" == label && "serve" != this.customerStates[c]) {
            this.ordersStarted++;
        }
        this.customerStates[c] = label;
        return label;
    },

    maySplice: function(customerIndex)
    {
        if (undefined === customerIndex) {
            customerIndex=-1;
        }
        if (0 <= customerIndex && this.customersServed.indexOf(customerIndex) <= -1) {
            this.customersServed.push(customerIndex);
        }
        if (this.roomIndex == 0 || this.roomIndex < this.customerRoom) {
            CafeModel.splice(this.customers, this.customersServed);
        }
    },

    /**
     * @return  index of the room a blender index is in.  Else -1.
     * Example @see TestModel
     */
    roomAt: function(blenderIndex)
    {
        var roomMin = 0;
        for (var room = 0; room < this.numRooms; room++) {
            var minMax = this.roomMinMax(room);
            if (blenderIndex < minMax[1]) {
                return room;
            }
        }
        return -1;
    },

    /**
     * @param   name    If "customer" and 1 customerRoom, then remainder equals index.
     * @return  index from this room, given a blender index.  Else -1.
     * Example @see TestModel
     */
    roomRemainder: function(blenderIndex)
    {
        var roomMin = 0;
        for (var room = 0; room < this.numRooms; room++) {
            var minMax = this.roomMinMax(room);
            if (blenderIndex < minMax[1]) {
                return blenderIndex - minMax[0];
            }
        }
        return -1;
    },

    updateShowIngredients: function(roomIndex, customerRoom)
    {
        this.num_ingredients_visible = 1 <= roomIndex || customerRoom == 0 
            ? (this.ingredientsVisible ? this.num_ingredients : 0)
            : 0;
    },

    /**
     * Lowest number that is not blending and has space for an ingredient in the current room.
     * Trashing selects that blender.
     * Example @see TestModel
     *
     * Click blender ingredients to trash.  Also selects that blender. 14/1/28 Annie expects to see how to trash ingredients. (+Aaron)  14/1/29 Kelsey expects to see how to move selected blender.
     * Keep blender selected if also is available. 14/1/29 Aaron expects to blender selected not to move. (+Eli)
     * https://lumoslabs.atlassian.net/browse/WG-233
     * Mouse down on ingredient adds to first on screen.  14/1/24 Kelsey expects to fix ingredient added to blender on screen.
     * https://lumoslabs.atlassian.net/browse/WG-174
     * GAME-2403 When not playing, set no blender available.  Test case:  2015-10-20 Before 3-2-1 countdown.  [~wouter] expects cup pops forward once.  Got first cup pops forward, pops backward and pops forward again.
     */
    blenderAvailable: function()
    {
        var index = -1;
        if (this.playing) {
            index = this.blenderActivated();
            if (index <= -1 || !(this.blenders[index].addIngredientEnabled)) {
                index = this.blenderEmpty();
            }
        }
        return index;
    },

    /** 
     * @return  May not be enabled to add an ingredient.
     */
    blenderActivated: function()
    {
        var index = -1;
        var minMax = this.roomMinMax(this.roomIndex);
        var activateTimeMax = 0;
        for (var b = minMax[0]; b < minMax[1]; b++) {
            var activateTime = this.blenders[b].activateTime;
            if (activateTimeMax < activateTime) {
                activateTimeMax = activateTime;
                index = b;
            }
            else if (-1 == index) {
                index = b;
            }
        }
        return index;
    },

    /**
     * Order from left-to-right, top-to-bottom.  2014-11-05 Emma expects to predict which cup will be selected.  (2014-11-11 +Pranjal)
     */
    blenderEmpty: function()
    {
        var index = -1;
        var minMax = this.roomMinMax(this.roomIndex);
        var activateTimeMax = 0;
        for (var b = minMax[1] - 1; minMax[0] <= b; b--) {
            if (this.blenders[b].addIngredientEnabled && this.blenders[b].ingredients.length == 0) {
                index = b;
            }
        }
        return index;
    },

    /**
     * Lowest and highest index of blenders that are currently on screen.
     */
    roomMinMax: function(roomIndex)
    {
        var roomMin = 0;
        for (var room = 0; room < roomIndex; room++) {
            roomMin += this.blendersPerRooms[room];
        }
        var roomMax = roomMin + this.blendersPerRooms[room];
        return [roomMin, roomMax];
    },

    scrollReset: function()
    {
        this.scrollX = 0;
        this.scrollXPrevious = 0;
        this.isScrollStarting = false;
        this.scrollEnable(false);
    },

    /**
     * @return  Booleans of visibility: [lower, higher], [0..2]: index of which button to show.  -1: invisible.  -2 no change since last query.
     * Example @see TestModel
     * 14/2/5 Wouter expects to see a page with a layoutTime on it.
     * WG-364
     * 1 screen of processors.
     * Only higher button visible.
     * 2 screens of processors.
     * Both higher and lower button visible.
     * Screen buttons that update to show target screen not available. 14/3/3 Wouter expects a button unique to that screen and camera pan across rooms or customer screen in center of 2 processors screens. 14/3/3 Wouter expects at most 6 processors per screen.
     * https://lumoslabs.atlassian.net/browse/WG-465
     */
    scrollVisibles: function()
    {
        var lower = false;
        var higher = false;
        if (3 <= this.numRooms) {
            lower = 0 <= CafeModel.roomOffsets[this.roomIndex];
            higher = CafeModel.roomOffsets[this.roomIndex] <= 0;
        }
        else if (2 <= this.numRooms) {
            lower = 1 <= CafeModel.roomOffsets[this.roomIndex];
            higher = CafeModel.roomOffsets[this.roomIndex] <= 0;
        }
        return [lower, higher];
    },

    /**
     * @return  Only in rooms 1 and higher and only if 3 rooms.
     */
    scrollFractionMaxes: function(now)
    {
        if (undefined === now) {
            now=-1;
        }
        if (-1 == now) {
            now = this.now;
        }
        var fractionMaxes = [0.0, 0.0];
        if (3 <= this.numRooms) {
            for (var r = 1; r < this.numRooms; r++) {
                var minMax = this.roomMinMax(r);
                for (var b = minMax[1] - 1; minMax[0] <= b; b--) {
                    var fraction = this.blenders[b].fraction(now);
                    if (fractionMaxes[this.numRooms - 1 - r] < fraction) {
                        fractionMaxes[this.numRooms - 1 - r] = fraction;
                    }
                }
            }
        }
        return fractionMaxes;
    },

    /**
     * @param   direction   "lower" or "higher".
     * @return  (1: higher, 0: middle, or -1: lower).  Also sets this on member variable "scrollX".
     * Scroll time 0.1.  2014-03-21 Scroll 0.25 seconds.  Ben expects faster scroll.
     * Scrolls parent of button.  Disables button during transition, then reenables.
     * Example @see TestModel
     *
     * Change log:
     * 14/2/12 Joe Palmeri expects to fix 5th blender usable.
     * WG-364
     * 1 screen of processors.
     * Only higher button visible.
     * Clicking switches screens.
     * 2 screens of processors.
     * Both higher and lower button visible.
     * Clicking goes up to next or cycles back.
     * Scroll through a static layout (not rotating indefinitely).
     * Change appearance of button to represent which target room it will go to:  Customer, Blender1, Blender2.
     * Screen buttons that update to show target screen not available. 14/3/3 Wouter expects a button unique to that screen and camera pan across rooms or customer screen in center of 2 processors screens. 14/3/3 Wouter expects at most 6 processors per screen.
     * https://lumoslabs.atlassian.net/browse/WG-465
     */
    scroll: function(direction)
    {
        var distance = -1;
        if (!this.scrollEnabled) {
            return 0;
        }
        this.scrollEnable(false);
        var roomDistance = 1;
        if ("lower" == direction) {
            roomDistance *= -1;
        }
        else if ("higher" == direction) {
        }
        else {
            throw new Error("CafeModel.scroll: Expected <lower> or <higher>.  Got <" + direction + ">");
        }
        var previousRoomIndex = this.roomIndex;
        this.roomIndex += roomDistance;
        if (this.roomIndex < 0 || this.numRooms <= this.roomIndex) {
            this.roomIndex = (this.numRooms + this.roomIndex) % this.numRooms;
        }
        this.maySplice();
        this.updateShowIngredients(this.roomIndex, this.customerRoom);
        if (this.playing) {
            this.metadata.moveRoom(this.now, this.roomIndex, this.countBlendersActive(), this.orders.length);
        }
        this.scrollX = distance * CafeModel.roomOffsets[this.roomIndex];
        return this.scrollX;
    },

    /**
     * Set isScrollStarting by previous scroll target position.
     */
    updateScroll: function()
    {
        this.isScrollStarting = this.scrollX != this.scrollXPrevious;
        if (this.isScrollStarting) {
            this.scrollEnableTime = this.now + 1000 * CafeModel.scrollTime;
        }
        if (!this.scrollEnabled && this.scrollEnableTime <= this.now) {
            this.scrollEnable();
        }
        this.scrollXPrevious = this.scrollX;
    },

    tutorLevel: function(enabled)
    {
        if (undefined === enabled) {
            enabled=true;
        }
        this.ingredientsVisible = !enabled;
        this.scrollEnable(enabled);
    },

    /**
     * Callback to re-enable button after a room transition or tutor.
     */
    scrollEnable: function(enabled)
    {
        if (undefined === enabled) {
            enabled=true;
        }
        // ll.log("CafeModel.scrollEnable: " + enabled);
        this.scrollEnabled = enabled;
    },

    /**
     * @return  "lower" off screen lower index, "equal" on screen, "higher" off screen higher index.
     * Example @see TestModel
     */
    nextCustomerScreenLabel: function()
    {
        var minMax;
        if (1 <= this.customerRoom) {
            minMax = [this.num_blenders * this.roomIndex, (this.roomIndex + 1) * this.num_blenders];
        }
        else {
            minMax = this.roomMinMax(this.roomIndex);
        }
        if (this.occupied < minMax[0]) {
            return "lower";
        }
        else if (minMax[1] <= this.occupied) {
            return "higher";
        }
        else {
            return "equal";
        }
    },

    countBlendersActive: function()
    {
        var active = 0;
        for (var b = 0; b < this.blenders.length; b++) {
            if (this.blenders[b].active) {
                active++;
            }
        }
        return active;
    },

    /**
     * 14/2/27 Click different ingredient. Wouter expects to add to next blender.
     * Example @see TestModel
     * If all blenders in a room have one ingredient and the next ingredient does not match, then ignore.  Test case:  2015-01-05 Boiler.  Level 9.  Room 1.  All 6 coffee machines have at least one ingredient.  Select a different ingredient than previous coffee machine.  Expect nothing happens.  Got crash.
     */
    addIngredientToBlenderAvailable: function(ingredient, now)
    {
        var index = this.blenderAvailable();
        if (0 <= index) {
            var ingredients = this.blenders[index].ingredients;
            if (1 <= ingredients.length 
                    && ingredient != ingredients[0]) {
                index = this.blenderEmpty();
            }
            if (0 <= index) {
                this.blenders[index].addIngredient(ingredient, now);
            }
        }
        return index;
    },

    /**
     * To next level.
     */
    progress: function()
    {
        return this.num_correct / this.num_trials;
    },

    scoreEnd: function()
    {
        this.metadata.time_bonus = CafeModel.score(0, 0.0, this.secondsRemaining);
        this.score += this.metadata.time_bonus;
        return this.score;
    },

    blendersSumIngredients: function()
    {
        var sum = 0;
        for (var b = 0; b < this.blenders.length; b++) {
            sum += this.blenders[b].ingredients.length;
        }
        return sum;
    },

    /**
     * @return  0 if no orders. 2 or 3 depending on outstanding order.  Measure order length by ingredients in blender if no customer order appears on screen.  So instructions about double order work when no customer appears.  
     * WG-691 Tutor. Fix instructions of double order.
     */
    firstOrderLength: function()
    {
        return this.orders.length <= 0 ? 0 : 
            this.customers[0] == null ? this.blenders[0].ingredients.length : this.customers[0].length;
    },

    /**
     * Orders loaded for customers, in customer sorting.
     * Ignore trash alerting blender.
     * If this were a runtime performance hotspot, could cache.
     * Example @see TestModel
     * 14/3/25 Tutorial. Overflow. Expect checkmark and tutorial step no longer shows order has started.
     * https://lumoslabs.atlassian.net/browse/WG-564
     */
    started: function(customerIndex, onlyNeedIngredients)
    {
        if (undefined === onlyNeedIngredients) {
            onlyNeedIngredients=true;
        }
        var start = false;
        for (var i = 0; i < this.orders.length; i++) {
            var order = this.orders[i];
            if (null != order) {
                var supplies = 0;
                for (var b = 0; b < this.blenders.length; b++) {
                    if ((onlyNeedIngredients || this.blenders[b].active) 
                            && CafeModel.arraysEqual(this.blenders[b].ingredients, order) 
                            && !this.blenders[b].trashAlerting()) {
                        start = true;
                    }
                }
            }
        }
        return start;
    },

    /**
     * 13/10/9 Next session. If all correct, Dave Alleca expects to unlock next level.
     * https://trello.com/c/dmmI2iTh/811-8-taxi-planner-dave-alleca-expects-to-adjust-next-session-level
     * If end at level 3.  Reset level to 0 to prompt tutorial again next session.  14/3/31 Marylou expects to see tutorial again.
     * WG-1019 At level 13 or higher, failing to finish 5 or more orders lowers next session level.  At level 12 or higher, completing all orders increases next session number of orders up to a maximum of 72 orders at level 22.  14/6/2 Level 12. Complete.  [~aaron] expects number of trials to increase up hypothetical maximum number of trials in worst case scenario.
     * WG-700 14/6/19 Aaron expects no tutor if close.
     */
    nextSessionLevel: function()
    {
        if (this.trialCompleted()) {
            return Math.min(Math.max(CafeModel.session_level, this.level + 1), CafeModel.getLevelMax());
        }
        else {
            var tolerance = 4;
            var tutorialPromptMax = 11;
            if (CafeModel.selected_level <= CafeModel.visibleLevelMax || this.num_trials - tolerance <= this.num_correct) {
                return (CafeModel.session_level <= CafeModel.firstLevel 
                    && this.num_correct <= tutorialPromptMax) ? 0 : CafeModel.session_level;
            }
            else {
                return CafeModel.session_level - 1;
            }
        }
    },

    isLevelUp: function()
    {
        return CafeModel.session_level < this.user_level;
    },

    /**
     * @return  If would Remind blender.
     */
    wouldRemindRoom: function(msRemaining, alwaysRemind)
    {
        if (undefined === alwaysRemind) {
            alwaysRemind=false;
        }
        var remindInterval = alwaysRemind ? 3000 : 6000;
        var reminding = false;
        if (this.roomIndex <= 0 && (alwaysRemind || 1 <= this.orders.length)
                && 1 <= this.customerRoom) {
            if (isNaN(this.remindRoomTime)) {
                this.remindRoomTime = msRemaining - remindInterval;
                // ll.log("CafeModel.wouldRemindRoom: reminding at " + this.remindRoomTime + " this.now " + msRemaining);
            }
        }
        else {
            this.remindRoomTime = undefined;
        }
        if (msRemaining <= this.remindRoomTime) {
            this.remindRoomTime = msRemaining - remindInterval;
            reminding = true;
        }
        return reminding;
    },

    wouldShowHint: function()
    {
        return this.tutor.wouldShowHint(this.level, this.orders, this.blenders);
    }
});

CafeModel.FOREVER = 99999999;
CafeModel.nextTrialMinSecondsRemaining = 6.0;
/**
 * Max.  For multipanels, set to 4.
 * 14/3/3 Wouter expects max of 6 processors per room.
 * https://lumoslabs.atlassian.net/browse/WG-465
 */
CafeModel.blendersPerRoom = 6;
CafeModel.customersMax = 12;
CafeModel.invisible = -1;
CafeModel.noChange = -2;
CafeModel.roomOffsets = [0, 1, -1];
CafeModel.scrollTime = 0.1;
CafeModel.selected_level;
/**
 * Start at 0.  Test case:  2014-12-13 Cocos2d-html5.  Max of session_level or level.  If level is NaN the result is NaN.
 */
CafeModel.session_level = 0;
CafeModel.visibleLevelMax = 12;
/**
 * 13/8/1 Star Search: Ben expects most scores between about 10,000 and 90,000.
 */
CafeModel.scorePerAccuracy = 10;
/**
 * 13/8/1 Star Search: Ben expects most scores between about 10,000 and 90,000.
 * 14/3/13 Ben expects to emphasize accuracy over speed.
 * WG-689 Double time bonus from x10 to x20.  Expect to discriminate low score percentiles.
 */
CafeModel.scorePerTime = 20;
/**
 * Level at first session.
 * 14/2/27 Eric Joe expects to start with 3 or 4 blenders.
 * 14/3/13 Wouter expects to start with second screen sooner.
 * WG-577 14/3/31 Median subscribers takes 4 sessions to pass first level.  Expected 1 session.
 * WG-1019 Set first level according to group "control" (2), "3", or "4".  14/6/2 [~aaron] expects to split test starting at a higher level.  
 * WG-700 [~bahroni] expects early serve score wins.  [~aaron] expects start level 3 wins.
 */
CafeModel.firstLevel = 3;
/**
 * Only require serving 3 per blender.  14/1/9 Level 1. Stephanie serves 4 of 4 with 0:01 seconds remaining.
 * Only require serving 4 per blender.  14/1/9 Level 1. Stephanie serves 4 of 4 with 0:01 seconds remaining.
 * Level 2+, trials = 5 trials per blenders.  14/1/14 Aaron expects to feel challenged.  Got 4 trials per blender.
 * Level 5+ 4 per blender.  14/2/14. 5 per blender.  5 blenders.  Miguel feels overwhelmed.
 * Only 2 or 3 ingredients.  Prepare to remember ingredients.  14/2/18 Aaron expects multiple screens.
 * Level 6 or higher, +2 trials per blender.  14/2/24 Hard.
 * https://lumoslabs.atlassian.net/browse/WG-173
 * 14/2/28 Wouter expects to not pass second to highest level (11 blenders with 30 trials).
 * https://lumoslabs.atlassian.net/browse/WG-465
 * 14/3/7 Level 7. Erin expects to feel encouraged.  Got overwhelmed.
 * 14/3/19 Level 7 and 8, experiment with some blenders on customer screen.
 * WG-444 14/3/20 Level 9 and 10, Wouter expects to keep one screen with 6 blenders.
 * WG-689:
 *      Level 5.  Add consecutives and increase trials.  14/4/15 Expect score up with intro of 2 screens.
 *      Level 10, 11. Remove consecutives.  14/4/15 Expect few users arriving to level 12.  
 *      Level 6, 7, 8, 11, 12.  Increase trials.  14/4/15 Expect smooth correctness rate curve.
 * WG-698:
 *      Level 5.  Double frequency of consecutive duplicate orders (from every 4 to every 2).  14/4/29 Level 5 median subscriber expects score higher than level 4.
 *      Level 12.  Increase trials from 48 to 50.  14/4/29 Level 12 median subscriber expects completion rate lower than level 11.  14/4/29 Michael, a level 12 top performer expects score higher for excellent performance.  https://lumosity.zendesk.com/agent/#/tickets/1398105
 * WG-820:
 *      Level 5.  Double frequency of consecutive duplicate orders (gap from 2 to 1).  Reduce average order length from about 1.8 to about 1.5.  14/5/13 Level 5 median subscriber expects score higher than level 4.
 *      Level 9.  Increase frequency of consecutive duplicate orders (gap from 4 to 3).  Reduce average order length from about 1.6 to about 1.5.  14/5/13 Level 9 median subscriber expects score interpolates level 8 and 10.
 *      Level 12.  Increase trials from 50 to 52.  14/4/29 Level 12 median subscriber expects completion rate about same as level 11.
 *      [~aaron] expects up to theoretical maximum in worst case scenario:  72 orders with 12 processors at 20 seconds per process.  
 * WG-970:
 *      Level 5:  Increase frequency of consecutive duplicate orders (count from 3 to 4, gap from 1 to 0).  Increase number of orders from 22 to 24.  14/6/3 Level 5 median subscriber expects score higher than level 4.
 *      Level 2, 3, 4, 6, 7, 11:  Increase number of orders by 2.  Level 8, 9, 10:  Increase number of orders by 4.  14/5/27 [~aaron] expects growth of median nth sessions at level has no outliers.
 * WG-1020:
 *      Level 9 to 11: Increase orders by 2.  Level 9:  Increase orders from 36 to 38.  Level 10:  Increase number of orders from 40 to 42.  Level 11 from 44 to 46.  14/6/15 Median subscribers may expect more sessions at level 9 than at level 8, yet that level 9 score is higher than level 8.
 * WG-1140:
 *      Level 9 to 11: Increase orders by 2.  14/7/8 Median subscriber may expect more sessions at level 9 than at level 8.
 */
CafeModel.levelParams = [
            {num_blenders:  1, num_ingredients: 2, num_trials:   3},  // tutorial, level 0
            {num_blenders:  2, num_ingredients: 2, num_trials:   6},
            {num_blenders:  3, num_ingredients: 2, num_trials:  12},
            {num_blenders:  4, num_ingredients: 2, num_trials:  16},
            {num_blenders:  4, num_ingredients: 3, num_trials:  22},
            {num_blenders:  5, num_ingredients: 3, num_trials:  24, customerRoom: 1, consecutive: 4, consecutiveGap: 0},
            {num_blenders:  6, num_ingredients: 4, num_trials:  28, customerRoom: 1},
            {num_blenders:  7, num_ingredients: 4, num_trials:  30, customerRoom: 0, blendersPerRooms: [1, 6]},
            {num_blenders:  8, num_ingredients: 5, num_trials:  34, customerRoom: 0, blendersPerRooms: [2, 6]},
            {num_blenders:  9, num_ingredients: 5, num_trials:  40, customerRoom: 1, blendersPerRooms: [6, 3], consecutive: 3, consecutiveGap: 4},
            {num_blenders: 10, num_ingredients: 5, num_trials:  44, customerRoom: 1, blendersPerRooms: [6, 4]},
            {num_blenders: 11, num_ingredients: 5, num_trials:  48, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  52, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  54, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  56, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  58, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  60, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  62, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  64, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  66, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  68, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  70, customerRoom: 1},
            {num_blenders: 12, num_ingredients: 5, num_trials:  72, customerRoom: 1}
        ];

/**
 * Copied frome ArrayHelper to avoid porting several ArrayHelper functions to JavaScript.
 * Returns the average of the given Array of Numbers, returns NaN if the Array is empty
 * Assert that the element is a number.
 * Example @see TestArrayHelper.as
 */
CafeModel.averageNumber = function(numberArray)
{
    var averageNumber = 0;
    for (var n=0; n < numberArray.length; n++) {
        if (isNaN(numberArray[n])) {
            throw new Error("Expected a number, got " + numberArray[n]);
        }
        averageNumber += numberArray[n];
    }
    averageNumber /= numberArray.length;

    return averageNumber;
}

CafeModel.getLevelMax = function()
{
    return CafeModel.levelParams.length - 1;
}

/**
 * Select 12th button.  [~aaron] expects to actually select maxim
 */
CafeModel.selectLevel = function(value)
{
    if (value < CafeModel.visibleLevelMax) {
        CafeModel.selected_level = value;
    }
    else {
        CafeModel.selected_level = Math.max(value, CafeModel.session_level);
    }
}

CafeModel.milestone = function()
{
    return Math.min(CafeModel.visibleLevelMax, CafeModel.session_level);
}

CafeModel.setLevel = function(newLevel)
{
    newLevel = CafeModel.clampLevel(newLevel);
    CafeModel.selectLevel(newLevel);
    return newLevel;
}

/**
 * @return Each blender in an array.
 */
CafeModel.addBlenders = function(blenders, num_blenders, metadata)
{
    blenders = [];
    for (var b = blenders.length; b < num_blenders; b++) {
        blenders.push(new Blender(metadata));
    }
    return blenders;
}

/**
 * @param   seconds     Ceiling of seconds remaining.  Tyler expects time on clock matches time bonus.
 */
CafeModel.score = function(num_correct, accuracy, seconds)
{
    return Math.floor(num_correct * accuracy) * CafeModel.scorePerAccuracy 
        + Math.ceil(seconds) * CafeModel.scorePerTime;
}

/**
 * Compare existing session level. 13/11/28 Select lower level. Tutorial. Skip. Expect to fix level available at maximum and maximum selected. Got higher level disabled.
 * https://trello.com/c/7Ovjhrf0/893-13-taxi-planner-abhishek-expects-to-polish-hud-and-tutorial
 */
CafeModel.initSessionLevel = function(level)
{
    CafeModel.session_level = Math.max(CafeModel.firstLevel, CafeModel.session_level, Math.min(level, CafeModel.getLevelMax()));
    return CafeModel.session_level;
}

/**
 * 0: tutorial.
 */
CafeModel.clampLevel = function(level)
{
    return Math.max(0, Math.min(CafeModel.levelParams.length - 1, level));
}

/**
 * Not accurate if model dynamically modifies number of trials later.
 * How can be tested if dynamically setting number of trials?
 * Example @see TestModel
 */
CafeModel.num_trials = function(level)
{
    return CafeModel.levelParams[level].num_trials;
}

CafeModel.loadLevel = function(level, overrideParams)
{
    if (undefined === overrideParams) {
        overrideParams=null;
    }
    var model = new CafeModel();
    var params = CafeModel.levelParams[CafeModel.clampLevel(level)];
    params.level = level;
    if (null != overrideParams) {
        for (var param in overrideParams) {
            params[param] = overrideParams[param];
        }
    }
    model.populate(params);
    return model;
}

/**
 * @param   excludes    Indexes that would not be returned.
 * @return Index where one ingredients' array of complete values appear inside another array of orders.
 * Example @see TestModel
 */
CafeModel.indexOf = function(orders, ingredients, excludes)
{
    if (undefined === excludes) {
        excludes=null;
    }
    var index = -1;
    for (var o = 0; o < orders.length; o++) {
        var order = orders[o];
        if (null != order && order.length == ingredients.length) {
            index = o;
            for (var i = 0; i < order.length; i++) {
                if (order[i] != ingredients[i]) {
                    index = -1;
                    break;
                }
            }
            if (null != excludes && 0 <= excludes.indexOf(index)) {
                index = -1;
            }
            if (0 <= index) {
                break;
            }
        }
    }
    return index;
}

CafeModel.roomCount = function(num_blenders)
{
    return Math.ceil(num_blenders / CafeModel.blendersPerRoom);
}

/**
 * Evenly distribute excess blenders between screens.  
 * @return Counts of blenders in that room.  Higher numbers first.
 * Example @see TestModel
 */
CafeModel.distribute = function(num_blenders)
{
    var count = CafeModel.roomCount(num_blenders);
    var counts = [];
    var remainder = num_blenders;
    for (var c = 0; c < count; c++) {
        counts[c] = Math.floor(num_blenders / count);
        remainder -= counts[c];
    }
    for (c = 0; 1 <= remainder && c < count; c++) {
        counts[c]++;
        remainder--;
    }
    return counts;
}

/**
 * Evenly distribute excess blenders between screens.
 */
CafeModel.offset = function(num_blenders)
{
    if (num_blenders <= CafeModel.blendersPerRoom) {
        return 0;
    }
    var excess = num_blenders % CafeModel.blendersPerRoom;
    if (1 <= excess && excess <= 2) {
        return 1;
    }
    else {
        return 0;
    }
}

/**
 * First vacant seat.
 * Example @see TestModel
 * https://lumoslabs.atlassian.net/browse/WG-234
 */
CafeModel.vacant = function(customers)
{
    for (var c = 0; c < customers.length; c++) {
        if (null == customers[c]) {
            return c;
        }
    }
    return -1;
}

/**
 * Customer takes first seat available, on any screen.  
 * If full, do nothing.
 * Example @see TestModel
 * @param   customers   Modified in-place.
 * @return  customers.
 */
CafeModel.occupy = function(customers, order)
{
    var index = CafeModel.vacant(customers);
    if (0 <= index) {
        customers[index] = order;
    }
    return customers;
}

/**
 * @return  Lists of ingredients of blenders in progress that could be served between now and later.
 */
CafeModel.ingredientsStarted = function(blenders)
{
    var starts = [];
    for (var b = 0; b < blenders.length; b++) {
        if (blenders[b].active && !blenders[b].trashAlerting()) {
            starts.push(blenders[b].ingredients);
        }
    }
    return starts;
}

/**
 * @param   orders              Unserved ingredient requests.
 * @param   ingredientsStarted  Already filtered trash alerts and whether started.
 * @param   customers           Currently unstarted orders.
 * @param   model               Only needs to update "occupied" index of last vacant.
 * Example @see TestModel
 * @return  customers   Evolution of unstarted orders, adding or deleting from customers.  add customer at first vacant.  remove customer first matching customer.
 */
CafeModel.demand = function(orders, ingredientsStarted, customers)
{
    var customersNow = customers.concat();
    var orderCounts = Count.each(orders);
    var ingredientsStartedCounts = Count.each(ingredientsStarted);
    var customersCounts = Count.each(customers);
    var customersUnstarted = Count.subtract(orderCounts, ingredientsStartedCounts);
    var customersDifference = Count.subtract(customersUnstarted, customersCounts);
    for (var orderText in customersDifference) {
        var order;
        var i;
        var difference = customersDifference[orderText];
        if (0 < difference) {
            order = CafeModel.parseOrderText(orderText);
            for (i = 0; i < difference; i++) {
                CafeModel.occupy(customersNow, order);
            }
        }
        else if (difference < 0) {
            order = CafeModel.parseOrderText(orderText);
            i = 0;
            for (var c = 0; difference < i && c < customersNow.length; c++) {
                if (CafeModel.arraysEqual(customersNow[c], order)) {
                    i--;
                    customersNow[c] = null;
                }
            }
        }
    }
    return customersNow;
}

/**
 * Average order length percent, Integer is easier to store and process than floating point.
 * Example @see TestModel
 */
CafeModel.order_length_percent = function(trial_orders)
{
    var average = 0.0;
    if (1 <= trial_orders.length) {
        for (var trial = 0; trial < trial_orders.length; trial++) {
            average += trial_orders[trial].length;
        }
        average /= trial_orders.length;
        average = Math.round(100.0 * average);
    }
    return average;
}

CafeModel.parseOrderText = function(orderText)
{
    var order = orderText.split(",");
    for (var i = 0; i < order.length; i++) {
        order[i] = parseInt(order[i]);
    }
    return order;
}

/**
 * @return  true if every element is strictly equal.  Does not inspect nested arrays.
 * Example @see TestModel
 */
CafeModel.arraysEqual = function(a, b)
{
    if (a == null || b == null) {
        return a == b;
    }
    if (a.length != b.length) {
        return false;
    }
    for (var i = a.length - 1; 0 <= i; --i) {
        if (a[i] !== b[i]) {
            return false;
        }
    }
    return true;
}

/**
 * @param   customers   Vacant customers sorted descending, in-place.
 * @param   customerIndexes   Consumed in-place.
 * Example @see TestModel
 * WG-691 14/4/8 Abhishek expects orders do not collapse.
 */
CafeModel.splice = function(customers, customerIndexes)
{
    while (1 <= customerIndexes.length) {
        var customerIndex = customerIndexes.pop();
        customers[customerIndex] = null;
    } 
    return customers;
}
