/*jslint node: true */
"use strict";

require("src/Game/CafeModel.js");
require("src/Utility/Count.js");



/**
 * Rules of when and what to tutor.
 */
var CafeTutor = cc.Class.extend(
{
    /**
     * What the animation is currently playing.  When the animation has exited that label after its transition, this function expects the currentLabel to change.
     */
    currentLabel: undefined,
    dim: undefined,
    /**
     * read-only
     */
    next: undefined,
    isUpdating: false,
    model: undefined,
    state: undefined,
    isBeginning: true,
    /**
     * read-only
     */
    levelStep: 0,
    isLevelEnding: false,
    levelIndex: -1,

    ctor: function(model)
    {
        this.isUpdating = false;
        this.isBeginning = true;
        this.levelStep = 0;
        this.isLevelEnding = false;
        this.levelIndex = -1;
        this.model = model;
        this.dim = "none";
        this.state = "";
        this.levelStep = 0;
    },

    /**
     * GAME-582 Start ingredients invisible.  14/5/2 Abhishek expects to show ingredients at arbitrary time during tutorial.
     */
    begin: function()
    {
        ll.log("CafeTutor.begin");
        this.model.playing = true;
        this.model.num_trials = 3;
        this.state = "";
        this.dim = "none";
        this.model.ingredientsVisible = false;
        this.overflow();
        this.isBeginning = false;
    },

    /**
     * Setup cup already overflowing.
     * 14/3/28 Miguel expects to tutor trash (14/3/31 +Guest Tori).
     * GAME-564
     * Cake Mania series tutors how to trash first.
     */
    overflow: function()
    {
        this.model.customers = [null];
        this.model.orders = [];
        this.model.trial = 0;
        this.model.blenders[0].addIngredient(1);
        this.model.blenders[0].start(-CafeModel.FOREVER);
        this.model.blenders[0].update(0);
        this.model.nextTrialTime = CafeModel.FOREVER;
        this.state = "trialStartOverflow";
    },

    /**
     * GAME-582 2014-05-02 Abhishek expects order appears before text.
     * GAME-582 2014-05-02 Abhishek expects to tutor 1-sugar, 2-sugar.  Then I expect to tutor 1-coffee.
     * GAME-1874 Now I tutor 1 coffee before 2 sugar.  Test case:  2014-11-05 Tutor 2 orders.  Filling.  Beverly expects to feel interested.  Got impatient.  (2014-12-01 +Emma, 2014-12-02 +Vishu, 2014-12-18 +Guest Harold, 2015-01-06 +Sonia).
     */
    nextTrial: function()
    {
        var nextTrials = [[1], [0], [1, 1]];
        var order = nextTrials[this.model.trial];
        this.model.nextTrial(order);
        ll.log("CafeTutor.nextTrial: " + order + " this.model.trial " + this.model.trial);
    },

    /**
     * GAME-1874 Show ingredients at start of timeline "trialIngredients".  2014-01-06 [~wouter] expects callout earlier.
     * Tutor.  Expect not to see message to trash.  Expect not to see next order until first dealt with.
     * @return  Next label of step.  Return "" (empty) when finished or nothing to change.  When trialEnd has been advanced to, then no more updates are performed.
     */
    update: function(tutorialMode, tutorCurrentLabel)
    {
        this.next = "";
        if (tutorialMode) {
            this.currentLabel = tutorCurrentLabel;
        }
        else {
            this.isUpdating = false;
        }
        this.updateDim();
        if (1 <= this.model.level) {
            this.levelUpdate();
            return this.next;
        }
        if (!this.isUpdating) {
            return this.next;
        }
        if (this.isBeginning) {
            this.begin();
        }
        this.next = CafeTutor.updateLabel(
            this.model.firstOrderLength(),
            this.model.started(0, true), 
            this.wouldShowHint(1, this.model.orders, this.model.blenders),
            this.model.started(0, false),
            this.model.blenders[0].trashAlerting(),
            this.model.num_correct, 
            this.state,
            this.currentLabel);
        if ("trialEnd" == this.state) {
            this.state = "";
            this.next = "";
        }
        if (this.next == this.state) {
            this.next = "";
        }
        if ("" != this.next && this.state != this.next) {
            ll.log("CafeTutor.update: from this.state <" + this.state 
                + "> to <" + this.next + "> when animation was at <" + tutorCurrentLabel + ">"
                + " at " + this.model.now + " milliseconds.");
            if (0 == this.next.indexOf("trialOrdersEnd")) {
                this.nextTrial();
            }
            else if ("trialIngredients" == this.next) {
                this.model.ingredientsVisible = true;
            }
            this.state = this.next;
        }
        return this.next;
    },

    /**
     * GAME-1885 At end of any "trialOrders" tutor sets dim to "begin", and at end of any "trialOrdersEnd" sets dim to "end".  Every other frame dim is null.  Test case:  2015-01-07 Dim.  Instantly add ingredient.  [~wouter] expects dim disappears.  Got remains.  Test case:  2015-01-10 How to Play.  Dim appears.  Skip tutorial.  Pause.  How to Play.  Read intro.  I expect screen stays bright.  Got dims.
     */
    updateDim: function()
    {
        if (0 == this.state.indexOf("trialOrdersEnd")) {
            this.dim = "end";
        }
        else if (0 == this.state.indexOf("trialOrders")) {
            this.dim = "begin";
        }
    },

    /**
     * @return  If currently tutoring a level.  
     * GAME-1874 At a level to tutor and a step is active.  2014-12-18 Level 9.  Matt Leung expects to fix serve single order at 10 seconds.  Got serve at 5 seconds.
     */
    isLevel: function()
    {
        return 0 <= this.levelIndex && 1 <= this.levelStep;
    },

    /**
     * Optionally tutor level 5 and 9.
     * @return  -1 if not starting now or already started.  
     *          0 if starting level 5.  
     *          1 if starting level 9.
     *
     * If session level 5 and select level 5: Before countdown, a sign reads "Now the coffee shop has TWO rooms! Click the right room button to see the coffee machines." The right room button flashes every 6 seconds. The ingredient buttons are invisible. After clicking, and seeing the coffee machines, a sign reads "Click the LEFT room button to see customer orders." The left room button flashes every 6 seconds. After clicking, return to customers and countdown 3-2-1.
     * 14/3/21 Level 5.  Ben expects to tutor switching rooms.  (14/3/25 +Joe Hardy, 14/3/26 +Mark Palange, Dave Wu)
     *
     * If session level 9 and select level 9: Before countdown, a sign reads "Now the coffee shop has THREE rooms! Click the left room button to see the left room of coffee machines." The left room button flashes every 6 seconds. After clicking, and seeing the coffee machines, a sign reads "Click the right room button to see customer orders. On level 9 or higher, the room button will also show how full a coffee machine in that room is." The right room button flashes every 6 seconds. After clicking, return to customers and countdown 3-2-1.
     *
     * GAME-583
     * GAME-582 At level 9, disable right room button until going to the left room.
     */
    levelBegin: function(session_level)
    {
        this.model = this.model;
        if (1 <= this.levelStep) {
            return -1;
        }
        this.levelIndex = -1;
        if (this.model.level == 5 && session_level == 5) {
            this.levelIndex = 0;
            this.levelStep = 1;
            this.model.tutorLevel();
        }
        else if (this.model.level == 9 && session_level == 9) {
            this.levelIndex = 1;
            this.levelStep = 1;
            this.model.tutorLevel();
            this.model.ingredientsVisible = false;
        }
        return this.levelIndex;
    },

    /**
     * Null check view and model.  2014-09-19 Level 5. Tutor second room.  Click (?), Yes.  Expect to fix tutorial begins.  Got crash.
     * @return  0:  ending.  1: started.  2: in target room.
     */
    levelUpdate: function()
    {
        var next = this.levelStep;
        this.isLevelEnding = 3 == next;
        if (!this.isLevel()) {
            return next;
        }
        if (3 <= this.levelStep) {
            next = 0;
        }
        else if (1 <= this.levelStep && null != this.model) {
            var target = this.levelIndex + 1;
            if (this.model.roomIndex == target) {
                next = 2;
                this.model.scrollEnable();
            }
            else if (this.model.roomIndex == 0 && 2 <= this.levelStep) {
                next = 3;
                this.model.tutorLevel(false);
            }
        }
        if (this.levelStep != next) {
            ll.log("CafeTutor.levelUpdate: from <" + this.levelStep + "> to <" + next + ">");
            this.levelStep = next;
        }
        this.isLevelEnding = 3 == next;
        return next;
    },

    /**
     * GAME-699 14/4/22 Level 1, 2 or 3: Start wrong. Tutor order area.
     */
    wouldShowHint: function(level, orders, blenders)
    {
        var show = false;
        if (1 <= level && level <= 3) {
            show = CafeTutor.wrongOrderAny(orders, blenders);
        }
        return show;
    }
});



/**
 * Evolve finite state machine of the tutorial steps:
 *      0. Count orders.
 *      1. Add ingredient.
 *      2. Start blender.
 *      3. Serve customers.
 *      4. If ingredients don't match an order, go back to step 1.
 *      5. If ingredients are wrong, prompt to trash ingredients.
 *
 * @param   orderLength     Number of orders.
 * @param   trash           If should trash and happened to play to end of trash, reset step to trash.
 * @param   currentLabel    What the animation is currently playing.  When the animation has exited that label after its transition, this function expects the currentLabel to change.
 *
 * 14/3/18 Sushmita or Pam suggested revised steps for tutorial:
 * 1. Look at exactly one order. Only one coffee machine. 
 * 3. Instruct to click ingredient button. On first time, callout ingredient button.
 * 4. Only after selecting correct ingredients, instruct to start. Callout start button.
 * 5. While waiting to serve for 10 seconds, do not discuss score. Mention overflow.
 * 6. After serve, second order of two ingredients appears. Instruct to click 2 ingredients.
 * 7. Only after selecting correct ingredients, instruct to start.
 * 8. While waiting to serve for 20 seconds, may mention to score higher by serving when full.
 * 9. Serve second order. End tutorial.
 * 10. If wrong ingredient or overflow, instruct and callout to trash by clicking cup.
 *
 * Playtests:
 * 14/3/14 Tutor. Click ingredient. Nicole expects to recognize button to start coffee machine.
 * 14/3/14 Tutor. Start a coffee machine. Nicole expects to recognize button to serve coffee machine.
 * 14/3/14 Tutor. Start a coffee machine. Nicole expects to not click ingredients to trash contents.
 * 14/3/14 Overflow. Nicole expects to realize can no longer serve.
 * 14/3/24 Miguel expects to tutor to trash.
 * 14/3/25 Tutorial. Before an order. Jessica may expect to tutor number on card represents number of customers waiting.
 * https://lumoslabs.atlassian.net/browse/GAME-564
 *
 * If should trash and happened to play to end of trash, reset step to trash. 14/3/31 Tutor 1 ingredient. Unknown steps of adding wrong ingredients and trashing.  Maybe timed to refill the wrong ingredient while the animation of the trash instruction is fading away.  Breanna expects to fix that trash message does not loop.
 * https://lumoslabs.atlassian.net/browse/GAME-656
 *
 * GAME-582 14/5/2 Tutor.  Overflow.  Pam expects different message than if wrong ingredient.
 * GAME-582 14/5/23 Abhishek expects to fix to see message for 2-sugar.
 * GAME-582 If skipping, do not alter tutorial step.  14/5/23 Skip tutorial.  Still see tutorial.  Wouter expects to fix can still skip.
 *
 * The code for this might be simplified by handling labels that end with "End" as transitions to delay further state transition.
 *
 * MOBGAMES-1837 Third order has text like "You got it!  Now make the order shown above."  Test case:  2014-03-25 Abhi may expect tutor third trial with some text.  (2014-11-05 +Beverly, +Gino).  2014-11-11 End tutorial.  Brief 2 lines of text, as in mobile.  Got placeholder (Kelsey, +Derek)  * First order text says "above" instead of "here".  2014-04 Subscribers expect to realize where orders are.
 *
 * GAME-1885 If tutoring third trial and wrong order or overflow, then instruct to trash.  Test case:  2015-01-09 Third trial.  Wrong order.  Overflow.  I expect instruction to trash.  Got no instruction.
 */
CafeTutor.updateLabel = function(orderLength, ingredients, trash, started, overflow, num_correct, state, currentLabel)
{
    if (null == currentLabel) {
        currentLabel = "";
    }
    state = state.split("2")[0];
    currentLabel = currentLabel.split("2")[0];
    if ("" == state) {
    }
    else if (3 <= num_correct) {
        state = "trialEnd";
    }
    else if (null == state) {
    }
    else if ("skip" == currentLabel || "skip" == state) {
        state = "skip";
    }
    else if ("trialStartOverflow" == state) {
        if (!overflow) {
            state = "trialStartOverflowEnd";
        }
    }
    else if ("trialOverflow" == state) {
        if (!overflow) {
            state = "trialOverflowEnd";
        }
    }
    else if (0 <= ["trialOrders", "trialStartOverflowEnd", "trialOverflowEnd"].indexOf(state)) {
        if (trash) {
            state = "trialBlenderTrash";
        }
        else if (overflow) {
            state = "trialOverflow";
        }
        else if ("trialOrdersEnd" == currentLabel) {
            state = "trialOrdersEnd";
        }
        else if (ingredients) {
            state = "trialIngredientsEnd";
        }
        else if (0 <= ["trialOrders", "trialStartOverflowEnd", "trialOverflowEnd"].indexOf(currentLabel)) {
            if ("trialOrders" == currentLabel) {
                state = "trialOrders";
            }
        }
        else {
            if (num_correct <= 1) {
                state = "trialIngredients";
            }
            else {
                state = "trial";
            }
        }
    }
    else if (0 <= ["trialOrdersEnd", "trialIngredients"].indexOf(state)) {
        if (trash) {
            state = "trialBlenderTrash";
        }
        else if (overflow) {
            state = "trialOverflow";
        }
        else if (orderLength <= 0) {
            state = "trialOrders";
        }
        else if (ingredients) {
            state = "trialIngredientsEnd";
        }
        else if (0 <= ["trialOrdersEnd", "trialIngredients"].indexOf(currentLabel)) {
        }
        else {
            if (num_correct <= 1) {
                state = "trialIngredients";
            }
            else {
                state = "trial";
            }
        }
    }
    else if (0 <= ["trialIngredientsEnd", "trialBlenderStart"].indexOf(state)) {
        if (started) {
            state = "trialBlenderStartEnd";
        }
        else if (trash) {
            state = "trialBlenderTrash";
        }
        else if (overflow) {
            state = "trialOverflow";
        }
        else if (orderLength <= 0) {
            state = "trialOrders";
        }
        else if (!ingredients) {
            state = "trialIngredients";
        }
        else if ("trialBlenderStart" == currentLabel) {
            state = "trialBlenderStart";
        }
    }
    else if (0 <= ["trialBlenderStartEnd", "trialBlenderServe"].indexOf(state)) {
        if (trash) {
            state = "trialBlenderTrash";
        }
        else if (overflow) {
            state = "trialOverflow";
        }
        else if (orderLength <= 0) {
            state = "trialOrders";
        }
        else if (!ingredients) {
            state = "trialIngredients";
        }
        else if (!started) {
            state = "trialBlenderStart";
        }
        else if ("trialBlenderServe" == currentLabel) {
            state = "trialBlenderServe";
        }
    }
    else if (0 <= ["trial", "trialBlenderServeEnd"].indexOf(state)) {
        if (3 <= num_correct) {
            state = "trialEnd";
        }
        else if (trash) {
            state = "trialBlenderTrash";
        }
        else if (overflow) {
            state = "trialOverflow";
        }
        else if ("trialBlenderServeEnd" != currentLabel) {
            state = "trial";
        }
    }
    else if ("trialEnd" == state) {
    }
    else if (0 <= ["trialBlenderTrash", "trialBlenderTrashEnd"].indexOf(state)) {
        if (trash) {
            state = "trialBlenderTrash";
        }
        else if (orderLength <= 0) {
            state = "trialOrders";
        }
        else if (0 <= ["trialBlenderTrash", "trialBlenderTrashEnd"].indexOf(currentLabel)) {
            state = "trialBlenderTrashEnd";
        }
        else if (!ingredients) {
            if (2 <= num_correct) {
                state = "trial";
            }
            else {
                state = "trialIngredients";
            }
        }
        else if (!started) {
            state = "trialBlenderStart";
        }
        else {
            state = "trialBlenderStartEnd";
        }
    }
    else {
        throw new Error("Expected to find tutorial step " + state);
    }
    if (state &&
        (1 <= num_correct || 3 <= orderLength) 
        && state.indexOf("trialBlenderTrash") <= -1
        && state.indexOf("trialOverflow") <= -1
        && ["trial", "trialEnd", "skip"].indexOf(state) <= -1
    ) {
        state += "2";
    }
    return state;
}

/**
 * @return Any non-zero list of ingredients in any blender that is not overflowing.
 */
CafeTutor.ingredientsLoaded = function(blenders, isActiveOnly)
{
    var loads = [];
    for (var b = 0; b < blenders.length; b++) {
        var blender = blenders[b];
        if (1 <= blender.ingredients.length && !blender.trashAlerting() 
        && (!isActiveOnly || blender.active)) {
            loads.push(blender.ingredients);
        }
    }
    return loads;
}

/**
 * Ignore if first ingredient of a double ingredient order.  Test case: 2014-03-26 Tutor.  Add 2 ingredients when only 1 ingredient on order.  Ben expects to fix to see to trash.  Got instruct to click ingredient.
 * GAME-1885 If start with first ingredient of a double ingredient order, then call recursively to return wrong.  Test case:  2015-01-12 Debug option overflow early.  Tutor 2 sugar.  Start 1 sugar.  I expect to see message was wrong.
 * GAME-564 Second order.  Start with only one ingredient.  Expect trash.
 * GAME-582 If next trial time is disabled, and ingredient, then should trash.  Test case:  2014-05-02 Tutor.  Overflow.  Pam expects different message than if wrong ingredient.
 * If started 1-ingredient for 2-ingredient order, wrong order.
 * @param   blenders        Array blenders, not necessarily started.
 * @param   isActiveOnly    Filter to only blenders that have started.
 * @return  Any blender has an ingredient that was not ordered and would not be ordered if another copy of the ingredient were added.
 */
CafeTutor.wrongOrderAny = function(orders, blenders, isActiveOnly)
{
    var loads = CafeTutor.ingredientsLoaded(blenders, isActiveOnly);
    var loaded = Count.each(loads);
    var ordered = Count.each(orders);
    var excess = Count.subtract(loaded, ordered);
    var wrong = false;
    for (var ingredients in excess) {
        if (1 <= excess[ingredients]) {
            wrong = true;
            var elements = ingredients.split(",");
            var one = 1 <= elements.length && elements.length <= 2;
            if (one) {
                var two = ingredients + "," + elements[0];
                if (two in ordered) {
                    if (!isActiveOnly) {
                        wrong = CafeTutor.wrongOrderAny(orders, blenders, !isActiveOnly);
                    }
                }
            }
        }
    }
    return wrong;
}

CafeTutor.fractionState = function(blender, state)
{
    if ("fraction" == state && blender.startTime <= -9999) {
        state = "tutor_overflow";
    }
    return state;
}
