/*jslint node: true */
"use strict";

require( "src/Game/TroubleBrewing.js" );

var DebugMenu = cc.Class.extend(
{
    timeLimit: undefined,

    /**
     * @param   controller  Game that has model and view property.
     */
    ctor: function(controller)
    {
        this.controller = controller;
        // XXX Not for production:
        // lumosity.game.params.short_game = true;  // XXX Is there a way to set short game only in my LumosityMobile?
        // lumosity.game.params.debug_mode = true;  // XXX Is there a way to set debug only in my LumosityMobile?
        this._shortGamesSetup();
        this._addDebugMenuButtons();
        DebugMenu.html5Accessor(controller);
    },

    isOpen: function()
    {
        return lu.isDebugModeOn() 
            && this.controller.view._debugMenu 
            && !this.controller.view._debugMenu.openButton.isVisible();
    },

    /**
     * @param   root    If no root, then ignore.  Test case:  2015-02-10 Android.  Select Trouble Brewing.  John Tighe expects to begin.  Got crash.  Log message: "JavaScript error: DebugMenu.js line 39.  TypeError: root is undefined"  (+Dom).  Xcode Locomotive iOS simulator:  No problem.
     * GAME-2177 Hide scene if debug menu is open.  Test case:  2014-12-16 Touch D.  Touch button.  Expect debug button activates.  Got level select button behind it activates.  
     *
     * GAME-2177 Unfortunately, only a few pixels on the outside edge of the debug button respond to a mouse click in Boiler, and probably none respond to a tap on a mobile device.  [~nick], On a mobile device, do you know how to select a debug button that is in front of a game button?  As [~shalin] mentioned in his Train of Thought post-mortem, unfortunately the level select LumosityButtons from LayoutLoader, although behind the debug buttons, block the click on the debug buttons.  Likewise the ingredient ControlButtons while visually behind the "D" debug button, block the debug button menu altogether.  When the debug buttons are drawn in front game content, I hope they could process a touch or a click.
     *
     * When not in level select mode, do not auto-show scene.  Test case:  2015-02-13 At level select.  Pause.  How to Play.  Mark Torii expects level select no longer visible.  Got appears in background.  Quit.  Select Debug mode.  Repeat same steps.
     */
    hideIfOpen: function(root, isEnabled)
    {
        if (!lu.isDebugModeOn()) {
            return;
        }
        if (!root) {
            return;
        }
        var isBlocked = !isEnabled || this.isOpen();
        if (root.isVisible() == isBlocked) {
            ll.log("DebugMenu.hideIfDebugMenuOpen: " + isBlocked);
            root.setVisible(!isBlocked);
        }
    },

    /**
     * short game:  5 seconds, may overflow early, and only one order.
     */
    _shortGamesSetup: function()
    {
        if (lu.isShortGamesModeOn()) {
            this._overflowEarly();
            this._orders(1);
            CafeModel.nextTrialMinSecondsRemaining = 0.0;
        }
    },

    _overflowEarly: function()
    {
        Blender.accuracyMinServeEnabled = 5;
        Blender.period = 1000;
    },

    _addDebugMenuButtons: function()
    {
        this.timeLimit = null;
        var view = this.controller.view;
        view.addDebugMenuButton( { labelText: "Level up", 
            callback: this._levelUp, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "2 rooms at Lv 5", 
            callback: function(){this._levelUp(5)}, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "3 rooms at Lv 9", 
            callback: function(){this._levelUp(9)}, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "0 orders", 
            callback: this._orders, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "1 order", 
            callback: function(){this._orders(1)}, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "Overflow early", 
            callback: this._overflowEarly, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "12 minutes", 
            callback: this._minutes12, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "Validate layouts", 
            callback: this._validateLayouts, 
            target: this, closeMenuOnCallback: true } );
        view.addDebugMenuButton( { labelText: "Test layout", 
            callback: this._testLayout, 
            target: this, closeMenuOnCallback: true } );
    },

    _levelUp: function(level)
    {
        this.controller.model._levelUp(level);
        this.controller.onRestart();
    },

    _orders: function(remaining)
    {
        if (undefined === remaining) {
            remaining = 0;        
        }
        var params = CafeModel.levelParams;
        for (var p = 0; p < params.length; p++) {
            params[p].num_trials = remaining;
        }
        this.controller.controller.model.num_trials = remaining;
        this.controller.levelSelectSetup();
    },

    _minutes12: function()
    {
        this.timeLimit = 720;
        this.controller._components.GameComponentTimer.reset(this.timeLimit);
    },

    _testLayout: function()
    {
        var stage = this.controller.view._gameStage;
        CCBReaderAdapter.testLayout(stage, "TestLayout");
    },

    _validateLayouts: function()
    {
        for (var level = 0; level <= CafeModel.visibleLevelMax; level++) {
            ll.log("DebugMenu._validateLayouts: level " + level);
            this.controller.loadLevel(level, this.controller.scene.root.board);
        }
    }
});

/**
 * @param   controller  Assign to "lumosity.game.instance" for convenient debugging.
 */
DebugMenu.html5Accessor = function(controller)
{
    if ( lumosity.device.id === "html5" ) 
    {
        if (!lu.isDebugModeOn())
        {
            ll.CurrentLoggingLevel = LoggingLevel.Normal;
        }
        try
        {
            lumosity.game.instance = controller;
        }
        catch (err)
        {
        }
    }
};

