/*jslint node: true */
"use strict";

require("src/Utility/Count.js");
require("src/Utility/LumosityUtils.js");
require("src/Utility/Variance.js");
require("src/Game/CafeModel.js");
var lu = LumosityUtils;


/**
 * GAME-1628, GAME-2016
 * Abstract algorithms and data to measure player performance.
 */
var Metadata = cc.Class.extend(
{
    blender_num_serve: [],
    num_invalid: 0,
    num_late: 0,
    num_serve: 0,
    num_start: 0,
    num_trash: 0,
    order_id_attempts: [],
    order_ids: [],
    order_times: [],
    rooms: [],
    start_times: [],
    time_bonus: 0,
    trials: [],
    ingredient_times: [],
    model: undefined,
    time_in_room: [],
    moveTime: 0,
    currentRoom: 0,

    ctor: function(model)
    {
        this.num_invalid = 0;
        this.num_late = 0;
        this.num_serve = 0;
        this.num_start = 0;
        this.num_trash = 0;
        this.order_id_attempts = [];
        this.order_ids = [];
        this.order_times = [];
        this.rooms = [];
        this.start_times = [];
        this.time_bonus = 0;
        this.trials = [];
        this.ingredient_times = [];
        this.time_in_room = [];
        this.moveTime= 0;
        this.currentRoom= 0;
        this.model = model;
        this.blender_num_serve = [];
        for (var b = 0; b < model.num_blenders; b++) {
            this.start_times[b] = 0;
            this.ingredient_times[b] = [];
            this.blender_num_serve[b] = 0;
        }
        for (var r = 0; r < model.numRooms; r++) {
            this.time_in_room[r] = 0;
        }
    },

    /**
     * 2014-11-24 [~aaron] expects to add:
     * 
     * num_room_moves: [integer] Number of switches between screens
     * num_late: [integer] Number of overflowed orders
     * num_invalid: [integer] Number of invalid orders
     * time_in_room: [integer list] Comma separated list of milliseconds spent on each screen
     * time_bonus: [integer] time bonus
     * num_blenders: [integer] number of machines
     *
     * 2015-01-05 [~aaron] expects to add:
     *
     * ingredient_counts:  a comma-separated list of total ingredient counts as a way to help detect any potential bias or confusion of ingredient art.
     *
     * 2015-01-21 [~aaron] expects to add:
     *
     * GAME-2016 In "blender_num_serve", Count correct serve on blender index into semi-colon separated values.  Test case:  2015-01-21 Level with 4 coffee machines.  Serve an order on left machine. Serve an order on right machine.  Serve something no one asked for on the right machine.  [~aaron] expects to record "1;0;0;1".
     */
    hashData: function(dataHash)
    {
        dataHash.accuracy = Math.round(this.model.accuracy);
        var ingredient_counts = Metadata.ingredient_counts( 
            this.trials, this.model.num_ingredients);
        dataHash.ingredient_counts = ingredient_counts.join(";");
        dataHash.ingredient_count_variance_percent = Metadata.ingredient_count_variance_percent( 
            ingredient_counts);
        dataHash.num_blenders = this.model.num_blenders;
        dataHash.blender_num_serve = this.blender_num_serve.join(";");
        dataHash.num_correct = this.model.num_correct;
        dataHash.num_invalid = this.num_invalid;
        dataHash.num_late = this.num_late;
        dataHash.num_rooms = this.model.numRooms;
        dataHash.num_room_moves = this.rooms.length;
        dataHash.num_start = this.num_start;
        dataHash.num_trash = this.num_trash;
        dataHash.num_trials = this.model.num_trials;
        dataHash.num_tries = this.trials.length;
        dataHash.order_length_percent = Metadata.order_length_percent(this.trials);
        dataHash.score = this.model.score;
        dataHash.selected_level = CafeModel.selected_level;
        dataHash.session_level = CafeModel.session_level;
        dataHash.time = this.model.now;
        dataHash.time_bonus = this.time_bonus;
        dataHash.time_in_room = this.time_in_room.join(";");
        dataHash.room_csv = lu.jsToCsv(this.rooms);
        dataHash.trial_csv = lu.jsToCsv(this.trials);
        dataHash.user_level = this.model.user_level;
    },

    nextOrder: function(id, now)
    {
        this.order_ids.push(id);
        this.order_id_attempts.push(0);
        this.order_times.push(now);
    },

    startTrial: function(blender, now)
    {
        var b = this.model.blenders.indexOf(blender);
        this.start_times[b] = now;
        this.num_start++;
    },

    addIngredient: function(blender, now)
    {
        var b = this.model.blenders.indexOf(blender);
        this.ingredient_times[b].push(now);
    },

    /**
     * 2014-11-24 Aaron expects to analyze these order correlations:
     *
     * trial_csv: (Tracks orders, as well as drinks not asked for. Each time an order reappears due to a mistake, this creates a new entry with the same order id number)
     * order_id: [integer] order number, "" for invalid orders or trashed orders.
     * response: [character] T (correct) / F (missed order) / O (order overflowed) / E (trashed order early) / I (invalid order)
     * response_time: [integer] time_offset - initial order_time.  current time for invalid orders.
     * correct: [boolean] T/F (F is used any time response is not T)
     * points: [integer] Points received
     * time_offset: [integer] Time order completed in milliseconds since beginning of game
     * order_time: [integer] Time order appeared in milliseconds since beginning of game, "" for invalid orders
     * start_time: [integer] Time order started in milliseconds since beginning of game (might be before the order_time).  Would be 0 if never started.
     * ingredient_time: [integer list] Semicolon separated list of time offsets of each ingredient addition in milliseconds since beginning of game
     * size: [integer] Number of ingredients
     * ingredient: [integer] Index of ingredient.
     * attempt: [integer] Attempt number on this order, empty for invalid orders
     * blender: [integer] Index of machine order made on
     *
     * GAME-2016 On serve, record active, max fill, orders, count serve.  Test case:  2015-01-05 Serve.  Observe number of machines active.  Finish game.  In trial_csv, [~aaron] expects to record number of machines "active" and number of "orders".
     */
    endTrial: function(blender, trialTime, order_id, ingredient, correct, points, response)
    {
        var b = this.model.blenders.indexOf(blender);
        if (0 <= order_id) {
            this.order_id_attempts[order_id]++;
        }
        var trial = {
            active:  this.model.countBlendersActive(),
            attempt: correct ? this.order_id_attempts[order_id] : "",
            blender: b,
            correct: correct ? "T" : "F",
            ingredient: ingredient,
            ingredient_time: this.ingredient_times[b].join(";"),
            max_fill: this.max_fill(),
            orders:  this.model.orders.length,
            order_id: correct ? order_id : "",
            order_time: correct ? this.order_times[order_id] : "",
            points: points,
            response: response,
            response_time: trialTime - this.start_times[b],
            start_time: this.start_times[b],
            size: this.ingredient_times[b].length,
            time_offset: trialTime
        };
        if (correct) {
            this.blender_num_serve[b]++;
        }
        this.trials.push(trial);
        this.ingredient_times[b] = [];
        this.start_times[b] = 0;
        if ("O" == response) {
            this.num_late++;
        }
        else if ("I" == response) {
            this.num_invalid++;
        }
    },

    /**
     * At end of play, add to time in room.  Test case:  2014-11-24 Do not move rooms.  Aaron expects to record time in first room.
     */
    endGame: function(time_offset)
    {
        this.time_in_room[this.currentRoom] += time_offset - this.moveTime;
    },

    /**
     * @return  Maximum fraction in any blender.  Expected by metadata.
     * Now I store max_fill as a percent from (0..100).  A perfect serve is less than 50.  To clarify, the representation of serve at halfway is 0.5 (to represent values from 0 to twice ideal).  Any time after double the fullest cup's duration is 100.  Example:  21 seconds for a 10 second cup is recorded as 100.
     */
    max_fill: function()
    {
        var max = 0.0;
        var blenders = this.model.blenders;
        for (var b = blenders.length - 1; 0 <= b; b--) {
            var fraction = blenders[b].fraction();
            if (max < fraction) {
                max = fraction;
            }
        }
        return Math.round(100.0 * max);
    },

    /**
     * If movign to same room, do not record.
     *
     * room_csv: (track screen switches)
     * time_offset: [integer] Milliseconds since start of game
     * room: [integer] Which screen was switched to
     * active: [integer] Number of active machines at switch time
     * orders: [integer] Number of visible orders at switch time
     *
     * GAME-2014 Now I store max_fill as a percent from (0..100).  A perfect serve is less than 50.  To clarify, the representation of serve at halfway is 0.5 (to represent values from 0 to twice ideal).  Any time after double the fullest cup's duration is 100.  Example:  21 seconds for a 10 second cup is recorded as 100.
     */
    moveRoom: function(time_offset, room, active, orders)
    {
        if (room == this.currentRoom) {
            return;
        }
        var move = {
            active: active,
            max_fill: this.max_fill(),
            orders: orders,
            room: room,
            time_offset: time_offset
        }
        this.rooms.push(move);
        this.time_in_room[this.currentRoom] += time_offset - this.moveTime;
        this.moveTime = time_offset;
        this.currentRoom = room;
    }
});



/**
 * Average order length percent, Integer is easier to store and process than floating point.
 * Example @see TestModel
 */
Metadata.order_length_percent = function(trials)
{
    var average = 0.0;
    if (1 <= trials.length) {
        for (var trial = 0; trial < trials.length; trial++) {
            average += trials[trial].size;
        }
        average /= trials.length;
        average = Math.round(100.0 * average);
    }
    return average;
}

Metadata.ingredient_counts = function(trials, num_ingredients)
{
    var variance = 0.0;
    var ingredientCounts = [];
    var counts = {};
    var ingredient;
    var key;
    for (var trial = 0; trial < trials.length; trial++) {
        ingredient = trials[trial].ingredient;
        key = ingredient.toString();
        Count.increment(counts, key);
    }
    for (ingredient = 0; ingredient < num_ingredients; ingredient++) {
        var count = 0;
        key = ingredient.toString();
        if (key in counts) {
            count = counts[key];
        }
        ingredientCounts.push(count);
    }
    return ingredientCounts;
}

/**
 * Variance in counts of ingredients.  Integer is easier to store and process than floating point.
 * Does not count double-order as two separate instances of the ingredient.
 */
Metadata.ingredient_count_variance_percent = function(ingredientCounts)
{
    var variance = Variance.variance(ingredientCounts);
    return Math.round(100 * variance);
}
