/*jslint node: true */
"use strict";

require( "src/Game/TroubleBrewing.js" );

/**
 * Rollup score and rolldown time.
 * Modifies model score and timeRemaining.
 * Modifies timer component.
 * Expects model and view.
 * Expects view getWidgetByClass.
 */
var TimeRollup = cc.Class.extend(
{
    bonus: undefined,
    isEnabled: undefined,
    timeRemained: undefined,

    /**
     * @param   controller  Game that has model and view property.
     */
    ctor: function(controller)
    {
        this.controller = controller;
    },

    isTransferringTime: function()
    {
        return this.controller.model.timeRemaining > 0 && 
            this.isEnabled;
    },

    /**
     * If in practice mode or no bonus points, do not setup time rollup.  Test case:  2015-02-11 Short game mode.  How to Play mode.  Finish tutorial trials.  Expect popup.  Got blank screen.  Tap Skip Tutorial.  Popup appears.  Test case:  2015-02-11 Short game mode.  Play.  Start three cups.  Get two wrong.  Do not finish final cup.  Let time run out.  Expect popup with score.  Got no response.
     */
    setup: function(scoreBase, points)
    {
        var model = this.controller.model;
        var isSetupNow = !this.isEnabled 
            && !model.persistentData.isPracticeMode
            && 0 < model.timeRemaining
            && 1 <= points;
        if (isSetupNow)
        {
            this.timeRemained = model.timeRemaining;
            model.score = scoreBase;
            this.scoreBase = scoreBase;
            this.points = points;
            this.viewBonus(points);
        }
        return isSetupNow;
    },

    start: function()
    {
        if (this.controller.model.persistentData.isPracticeMode) {
            return;
        }
        // ll.log("TimeRollup.start");
        this.isEnabled = true;
    },

    /**
     * @return {Boolean} if transferring time now.  
     */
    onUpdate: function(dt, animationName)
    {
        var isTransferring = this.isTransferringTime();
        if ( isTransferring )
        {
            if (!this.step(dt))
            {
                // ll.log("TimeRollup.onUpdate: finish");
                this.isEnabled = false;
                this.controller.view.playAnimation( animationName );
                var timer = this.controller._components.GameComponentTimer;
                timer._timeRemaining = 1;
                timer._timeDelta = 0;
            }
        }
        this.viewRollup();
        return isTransferring;
    },

    /**
     * When 1 second or less, set timer to 1 second left and score to base plus bonus.  Test case:  2015-02-11 Time rollup.  Expect Time 0:00.  Got text like -1:59 for a fraction of a second during HUD exit.  Test case:  2015-02-16 Short game.  Serve.  790 points.  Time bonus 40.  Expect score 830.  Got score 810.  The score per second is 20 points.  It appears as if one second were not added.  In regular game mode, no problem observed.  
     */
    isTimeRemaining: function()
    {
        return 1 < this.controller.model.timeRemaining;
    },

    /**
     * Step of rolling down time and rolling up bonus, clamp at end.
     * Not cc.ActionTween, because that expects a cc.Node, and there are two parameters tweened.
     *
     * Expects this instance has set "scoreBase" and "points";
     * Change log:
     * GAME-1399 End. Rolldown remaining time and rollup score in steps of 10.  2014-08-05 Ben expects to rolldown time and rollup score in steps of 10.  Flash version has an end screen to read a time bonus.  Mobile version does not.
     * @return  {boolean} If time is remaining.
     */
    step: function(dt)
    {
        var speed = this.timeRemained;
        var model = this.controller.model;
        model.timeRemaining = Math.max(0.0, model.timeRemaining - speed * dt);
        var isTimeRemaining = this.isTimeRemaining();
        if (isTimeRemaining)
        {
            var fraction = 1.0 - model.timeRemaining / this.timeRemained;
            var bonusLerp = lu.lerp(fraction, 0, this.points);
            bonusLerp = this.roundUp(bonusLerp);
            this.bonus = Math.min(this.points, bonusLerp);
        }
        else 
        {
            this.bonus = this.points;
        }
        model.score = this.scoreBase + this.bonus;
        return isTimeRemaining;
    },

    /**
     * Round up to nearest 10.  So it is always at least 10.
     */
    roundUp: function(points)
    {
        var rounding = 10;
        var rounded = Math.round(points / rounding) * rounding;
        rounded = Math.max(rounded, rounding);
        // ll.verbose("TimeRollup.roundUp: " + rounded + " points " + points);
        return rounded;
    },

    viewRollup: function()
    {
        this.controller.view.setScore( this.controller.model.score );
        var widget = this.controller.view.getWidgetByClass(GameWidgetHud)
        if (widget)
        {
            widget.setTimeRemaining( Math.floor(this.controller.model.timeRemaining) );
        }
    },

    viewBonus: function(bonus)
    {
        var label = this.controller.view["scorePoints"];
        if ( label )
        {
            label.setString( "+" + bonus.toString() );
        }
    },

    viewHighlight: function(highlight)
    {
        var widget = this.controller.view.getWidgetByClass(GameWidgetHud)
        if (widget)
        {
            widget.setScoreHighlighted(highlight);
        }
    }
});
