/*jslint node: true */
"use strict";


require( "src/Utility/LumosityUtils.js" );
require( "src/Utility/LumosityLogger.js" );
require( "src/Game/Base/Game.js" );
require( "src/Game/Base/GameSettings.js" );
require( "src/Game/Component/GameComponentAudioMixer.js" );

require( "src/Game/CafeController.js" );
require( "src/Game/LevelSelect.js" );
require( "src/Game/Scene.js" );
require( "src/Game/Sounds.js" );
require( "src/Game/TutorScene.js" );

require( "src/Game/TroubleBrewingModel.js" );
require( "src/Game/TroubleBrewingView.js" );
require( "src/Game/DebugMenu.js" );
require( "src/Game/TimeRollup.js" );

var lu = LumosityUtils;


var TroubleBrewing = Game.extend(
{
    /** @lends TroubleBrewing.prototype */
    controller: undefined,
    levelSelect: undefined,
    scene: undefined,
    
    /**
     * TroubleBrewing is the main class for the game, TroubleBrewing!
     *
     * @constructs
     */
    ctor: function()
    {
        this._super();
        this.debugMenu = new DebugMenu(this);
        this.timeRollup = new TimeRollup(this);
        this._createSounds();
    },

    /**
     * MOBGAMES-1898 Adapted from Pet Detective, if at level select screen and pressing system back button, then exit game prematurely.  Consistent with GameView, although inconsistent with Pet Detective and Train of Thought, the game quit instead of completed.  2014-11-10 Android.  TroubleBrewing.  Tap system back button.  QA Temp Daniel expects to exit game and return to pregame screen in the app.
     * GAME-2471 On Android back button tapped, only return to app if at level select screen have not yet selected a level.  Test case:  2015-01-21 Android.  Select a level.  But 3-2-1 countdown has not finished.  Press hardware back button.  [~rivey] expects to pause.
     */
    _backButtonTapped: function()
    {
        if (this.model.persistentData.isLevelSelectMode && !this.levelSelected)
        {
            this.onFinish(false);
            return;
        }
        this._super();
    },

    _createSettings: function()
    {
        var settings = new GameSettings( "settings.json" );
        return settings;
    },

    _createModel: function()
    {
        this.model = new TroubleBrewingModel(this.settings);
        return this.model;
    },

    _createView: function( screenSize )
    {
        var view = new TroubleBrewingView( screenSize, this.settings, this.model, this );
        this._createCafe(view._gameStage);
        return view;
    },

    /**
     * Game constructor configures search paths.  
     * CafeController loads spritesheets.  
     * View constructor loads layout.
     * Test case: 2014-12-12 iOS simulator. Start. Error: spritesheet_0.png not found.
     */
    _createCafe: function(parent)
    {
        this.scene = new Scene(this, parent);
        this.levelSelect = new LevelSelect();
        this.tutorScene = new TutorScene(this, parent);
        this.controller = new CafeController();
        this.controller.init(true);
    },

    /**
     * Expects components already created.
     */
    _createSounds: function()
    {
        var sounds = Sounds.createWithMixer(this);
        this.controller.view.wireSounds(sounds);
    },

    cue_hudScoreHighlight: function()
    {
        this.timeRollup.viewHighlight(true);
    },

    cue_hudScoreNormal: function()
    {
        this.timeRollup.viewHighlight(false);
    },

    /**
     * Preload layout of highest level.  Tap level.  Expect no delay.
     * GAME-2565 At level select, block input until after level loaded.  Test case:  2015-01-27 Tap "Play" then spam tapping. Ethan expects level selected.  Got load empty level.
     */
    loadLevel: function(level, board)
    {
        if (this.loadedLevel !== level) {
            this.model.currentLevel = level;
            this.loadedLevel = level;
            CafeModel.setLevel(level);
            this.controller.loadLevel(CafeModel.selected_level);
            this.controller.populate(board, 1 <= level);
        }
        this.model.isBlockingInput = false;
    },

    populateSessionLevel: function()
    {
        this.scene.loadLevel();
        this.controller.populate(this.scene.root.board);
    },

    /**
     * Clear loaded level.
     * In practice mode, hide scene and begin tutor scene.  Test case:  2015-01-07 Level 3.  Pause.  How to Play.  Expect to see 1 coffee machine and 1 order.  Got invisible.
     */
    onBegin: function()
    {
        this.levelSelected = false;
        this.loadedLevel = undefined;
        this._super();
        this.controller.onBegin();
        this.model.onBegin();
        CafeModel.initSessionLevel(this.model.session_level);
        var isLevelSelect = this.model.persistentData.isLevelSelectMode;
        this.model.isBlockingInput = isLevelSelect;
        if (isLevelSelect) {
            this.scene.onBegin();
            this.listenToSelectLevel(this.scene.begin.levelSelect, "milestone_");
            this.levelSelectSetup();
        }
        else if (this.model.persistentData.isPracticeMode) {
            this.tutorScene.onBegin();
        }
        this.view.onBegin();
        this.timeRollup.isEnabled = false;
    },

    /**
     * GAME-2394 Before pause, stop coffee brew.  Test case:  2015-01-06 Hear coffee brew.  Pause.  Resume.  Expect to hear coffee brew.
     * GAME-1885 On pause, pause all animated nodes in board.  Test case:  2015-01-06 Begin filling a cup.  Observe height of liquid.  Pause.  Wait 10 seconds.  Resume.  Expect liquid at same height.  Expect button lights when liquid near top.  Serve.
     * GAME-1885 Disable animation cache.  Test case:  2015-01-08 Play.  Restart.  Play.  Restart.  Play.  Restart.  Beverly expects to see level select.  Got crash.
     */
    onPause: function(instant)
    {
        cc.BuilderReaderHelper.pause(this.controller.view.board);
        this.controller.view.sounds.coffeeBrewLoopStop();
        this._super(instant);
    },

    onResume: function()
    {
        this._super();
        cc.BuilderReaderHelper.resume(this.controller.view.board);
        this.controller.view.updateCoffeeBrewLoop();
    },

    /**
     * Stop animations and set tutor invisible.
     *
     * GAME-1874 Set tutor scene invisible.  Test case:  2014-12-15 Tutor. Skip. Expect see level select only.  Got tutor scene in background.
     * GAME-1405 Stop all animations in view before base class plays end tutorial animations.  Test case:  2014-10-22 Complete first trial and first board.  You cleared board animation begins. Immediately click Skip tutorial.  Expect popup.  Got nothing happens.
     * Stop tutor.  Race precaution:  Program hides the tutor scene if they would still appear when skipping.  There is an animation to make them disappear.
     * GAME-1405 End game now.
     *
     * GAME-1403 If skip tutorial, stop all animations, actions, and sequences.  Example:  Tutor.  Animating between boards.  Quickly click skip tutorial.  Expect next board does not appear.
     * GAME-1405 At end of tutorial trials, remove dim layer if it is still present.  Test case:  2014-10-20 How to Play mode.  Second trial.  Dim.  Immediately click skip tutorial.  Expect dim layer removed.  
     * GAME-1405 When starting next tutorial animation, clear previous animations.  Test case:  2014-10-20 How to play mode.  Answer quickly.  Expect previous dim layer disappeared.
     * GAME-1885 When skipping tutorial, interrupt tutor scene animation.  Test case:  2015-01-10 Skip tutorial.  Wait 5 seconds.  Examine log.  Expect no new messages about tutor.  Got tutor playing animations.
     */
    onCompleteTutorial: function( skipped )
    {
        if (skipped) {
            this._endGameNow();
            this.tutorScene.runAnimations("none");
        }
        this.tutorScene.hide();
        this.view.onCompleteTutorial(skipped);
        this.controller.view.sounds.gameMusicLoopStop();
        this._super(skipped);
    },

    animationFinished_endTutorial: function()
    {
        this.onEndTutorial();
    },

    /**
     * Do not trigger ending when time bonus rolls up.
     */
    onComponentTimerEnd:function()
    {
        if ( !this.timeRollup.isEnabled ) 
        {
            this._super();
        }
    },

    /**
     * Guard until not in level select.
     * In practice mode, do not start cafe yet.  TutorScene will finish its intro animation first, analogous to default "practiceBegin" intro.
     */
    onStartGameplay:function()
    {
        if (this.model.persistentData.isLevelSelectMode) 
        {
            return;
        }
        // ll.verbose( "TroubleBrewing", "onStartGameplay" );

        var timeLimit = this.debugMenu.timeLimit || this.model.onStartGameplay();
        this._components.GameComponentTimer.reset(timeLimit);
        this.controller.model.setTimeLimit(timeLimit);
        if (!this.model.persistentData.isPracticeMode) 
        {
            this.controller.onStartGameplay();
        }
        this._super();
    },

    /**
     * If increasing next session level, disable default finish to show level up.
     * GAME-1403 Practice: Complete tutorial instead of end game.
     * MOBGAMES-1792 If end gameplay already called, ignore second call.  2014-10-30 Practice: End game.  QA Temp1 expects to see end tutorial message once.
     * Hide pause button when bonus begins.
     */
    onEndGameplay: function()
    {
        if (this._endGameNow())
        {
            // ll.log( "TroubleBrewing", "onEndGameplay" );
            if (!this.model.persistentData.isPracticeMode) 
            {
                this.view.playAnimation("hideHud");
                this.view.cue_hidePauseButton();
                this._super();
            }
        }
    },

    /**
     * Override default view to play "gameEnd".  
     * Instead, might see level up.
     * @param   isFinish    If true, play "gameEnd" and finish.
     */
    playEndAnimation: function(isFinish)
    {
        if (isFinish) {
            this.view.playAnimation( "gameEnd", this.onFinish );
        }
        else if ( !this.model.persistentData.isPracticeMode )
        {
            lumosity.game.report( this.model.exportResults() );
            this.levelSelect.levelUp(CafeModel.session_level < CafeModel.getLevelMax(),
                this.controller.model.user_level,
                CafeModel.num_trials(this.controller.model.user_level),
                this.scene.root.levelUp);
            var label = this.controller.model.isLevelUp() 
                ? "gameEndLevelUp" : "gameEnd";
            this.scene.runAnimations(label);
        }
        else
        {
            this.view.playAnimation( "endTutorial", this.onStartFitTest );
        }
    },

    /**
     * Base game expects metadata copied from cafe to base model.
     *
     * @return  If this call ends the game.
     */
    _endGameNow: function()
    {
        var ending = this.model.isGameRunning;
        if (ending)
        {
            this.model.isGameRunning = false;
            this.controller.model.endGame();
            this.model.user_level = this.controller.model.user_level;
            this.model.recordUserLevel();
        }
        return ending;
    },

    /**
     * XXX Save trials to private metadata variables.
     * GAME-2016 Stomp shared metadata faulty num_correct with this game's num_correct.  Test case:  2015-01-05 Short game.  No answer.  [~aaron] expects num_correct 0.  Got 1.
     * In the metadata component "onEndGameplay":  One improvement would be for each key to ignore writing a value if they already contains a defined value.  Test case:  2014-12-13 Trouble Brewing has a clean and completely portable Metadata class that manages all the game-specific values, including "num_correct", "num_trials".  This generic metadata stomps those values with a complicated, rigid expectation to use the beginMetadataTrial and endMetadataTrial functions.  This component's onEndGameplay function is called after "finalizeMetadata" function, so there is no forward-compatible place to overwrite the metadata's bloated "onEndGameplay" values.
     *
     * Unlike Flash version, remove score from game results data.  Score is collected from top level.  Do record score in metadata, so that metadata code can be automatically converted from Flash version, that relies on score in metadata.  Test case:  2015-02-11 Aaron expects score is not duplicated in game result data.
     */
    finalizeMetadata: function()
    {
        this._super();
        var metadata = this.controller.model.metadata;
        metadata.hashData(this.model.metadata);
        if ("score" in this.model.metadata) {
            delete this.model.metadata.score;
        }
        this.model.num_correct = this.controller.model.num_correct;
        var component = this.getMetadataComponent();
        component._trials = metadata.trials;
        component._trialCount = component._trials.length;
        component._trialCorrectTimes = [];
        component._trialCorrectTimes.length = this.model.metadata.num_correct;
    },

    /**
     * GAME-1405 If game complete, and not started and finished transferring time, then pause timer.  Test case:  2014-10-15 Answer in last second.  +10 bonus.  Expect only one popup.  Got 2.  Expect game not end in postgame view.  Got ends.  (+2014-10-20)
     * MOBGAME-1822 Tutor updates with accumulated time between rounds.  Test case:  2014-11-01 End tutor.  Expect text transitions off.
     * GAME-2605 During update, if paused, do not update the cafe, which would cause the coffee pour to restart.  Test case:  2015-01-29 Fill coffee.  Pause.  Quit.  Craig Petzel expects no sound.  Heard coffee filling sound.
     */
    onUpdate: function( dt )
    {
        this._super( dt );
        var consumed = this._components.GameComponentTimer.getTimeConsumed();
        var tutor = this.controller.model.tutor;
        consumed += this.accumulateTutorLevelTime(tutor && tutor.isLevel(), dt);
        var now = Math.round(1000.0 * consumed);
        var time = this.model.timeRemaining;
        var tutorialMode = this.model.isGameRunning && this.model.persistentData.isPracticeMode;
        if (!this.model.isGamePaused) {
            var isComplete = this.controller.update(now, time,
                tutorialMode,
                this.tutorScene.getCurrentState());
            this.scene.update();
            this.tutorScene.update(tutor);
            if (tutor) 
            {
                if (tutor.isLevelEnding)
                {
                    this.playBeginNow();
                }
            }
        }

        var transferring = this.timeRollup.onUpdate(dt, "bonusPointsEnd");
        if ( !transferring )
        {
            if (!this.controller.model.trialCompleted()) {
                this.model.score = this.controller.model.score;
                this.view.setScore( this.model.score );
            }
            if (isComplete) {
                this.onCompleteTrials();
            }
        }
        this._components.GameComponentTimer.setIsComponentPaused(
            transferring || !this.model.isGameRunning );
        this.debugMenu.hideIfOpen(this.scene.begin, this.model.persistentData.isLevelSelectMode);
    },

    accumulateTutorLevelTime: function(isLevel, dt)
    {
        if (isLevel)
        {
            if (undefined === this.tutorLevelTime) 
            {
                this.tutorLevelTime = 0;
            }
            this.tutorLevelTime += dt;
        }
        else
        {
            this.tutorLevelTime = 0;
        }
        return this.tutorLevelTime;
    },

    /**
     * Prepare time bonus points.
     * The fragile order of assignments frustrated me.
     * I suspect there is an opportunity to extract a time bonus view class.
     */
    onCompleteTrials: function()
    {
        // ll.log( "TroubleBrewing", "onCompleteTrials" );
        var scoreBase = this.controller.model.score;
        var timeRollupPoints = CafeModel.score(0, 0.0, this.controller.model.secondsRemaining);
        var isBonus = this.timeRollup.setup(scoreBase, timeRollupPoints);
        this.controller.model.playing = false;
        this.controller.model.scoreEnd();
        this.view.setScore( this.model.score );
        if (isBonus) {
            this.view.playAnimation( "bonusPoints" );
        }
        else {
            this.animationFinished_bonusPointsEnd();
        }
    },

    /**
     * GAME-1885 When quitting Trouble Brewing, remove and stop ambient music.  Test case:  2015-01-21 iOS 8 iPad Mini.  Trouble Brewing.  Play.  Hear ambient music of guitar playing.  Quit.  Start Ebb and Flow.  Expect not to hear guitar playing.  Hear guitar playing.
     */
    onFinish: function(isComplete)
    {
        if (this.controller && this.controller.view && this.controller.view.sounds) {
            this.controller.view.sounds.stopAll();
        }
        this._super();
    },

    animationFinished_bonusPoints: function()
    {
        this.timeRollup.start();
    },

    /**
     * End game or transit to next round.
     * GAME-1403 If continuing practice, animate checkmark.
     */
    animationFinished_bonusPointsEnd: function()
    {
        this.onEndGameplay();
    },

    /**
     * GAME-2019 At level select, set text and position "levelSelect" child named "highlight" at highest unlocked milestone.  Test case:  2015-01-14 Level select screen.  Sarah expects animation at highest unlocked milestone.
     */
    levelSelectSetup: function()
    {
        if (this.scene.begin)
        {
            var parent = this.scene.begin.levelSelect;
            this.levelSelect.localizeText(
                CafeModel.session_level < CafeModel.getLevelMax(),
                CafeModel.num_trials(CafeModel.session_level),
                parent, CafeModel.session_level);
            var name = "milestone_" + CafeModel.milestone().toString();
            var position = parent[name].getPosition();
            parent.highlight.setPosition(position);
        }
    },

    listenToSelectLevel: function(buttonsParent, prefix)
    {
        var unlocked = CafeModel.session_level;
        for (var milestone = 1; milestone <= CafeModel.visibleLevelMax; milestone++) 
        {
            var name = prefix + milestone.toString();
            var button = buttonsParent[name];
            if (button) {
                this.enableLevelButton(button, milestone, unlocked);
                cc.BuilderReaderHelper.addTargetWithActionForControlEvents(button, 
                    this, this.selectLevelFromControlButton, 
                    cc.CONTROL_EVENT_TOUCH_DOWN);
            }
            else {
                ll.error("TroubleBrewing.listenToSelectLevel: Button not found: " + name );
            }
        }
    },

    /**
     * GAME-1885 Highlight lower of unlocked level or max visible level.  Test case:  2015-01-20 Level 13 or higher.  Expect last cup is highlighted.  Got no cup highlighted.
     */
    enableLevelButton: function(button, milestone, unlocked)
    {
        var enabled = milestone <= unlocked;
        button.setEnabled(enabled);
        if (enabled)
        {
            var highlighted = milestone == Math.min(CafeModel.visibleLevelMax, unlocked);
            button.setHighlighted(highlighted);
        }
    },

    /**
     * You can now use 12 level select buttons in SpriteBuilder and we could then get rid of the Lumos xJSFL level select layout.  To identify the level the player wants, the program expects the buttons to be named "milestone_1" to "milestone_12".
     */
    selectLevelFromControlButton: function(button, controlEventIgnored)
    {
        if (this.model.isBlockingInput || this.debugMenu.isOpen())
        {   
            ll.warning("selectLevelFromControlButton: Input blocked or debug menu open.");
            return;
        }
        if (this.model.persistentData.isLevelSelectMode)
        {
            var milestone = parseInt(String.suffix(button.getName()));
            this.selectLevel(milestone);
        }
    },

    /**
     * GAME-1405 Do not disable level buttons now, because their visual appearance would change and distract.
     * MOBGAMES-1828 Upon selecting level, cleanup level select animations, in case the player has rapidly tapped the level before text animates into view.  2014-11-04 Julie expects delay of some seconds.
     */
    selectLevel: function(milestone)
    {
        this.levelSelected = true;
        this.model.isBlockingInput = true;
        this.model.setMilestone(milestone);
        this.view.stopAllAnimations();
        this.loadLevel(this.model.currentLevel, this.scene.root.board);
        this.scene.playBegin();
    },

    /** 
     * Prepare for the first round.
     * Enable starting play after playing "gameplay" animation.
     * GAME-1405 After removing level buttons from screen, disable level buttons.  Test case:  2014-10-24 Boiler.  Play.  Pause.  Click Restart.  Expect to restart.  Got level button intercepts click.
     * GAME-2403 Populate before level tutor.  Test case:  2015-01-13 Level 9.  Expect to tap button.  Got no response.  Test case:  2015-01-13 Level 3.  3-2-1 expect no order appears yet.  Got order appears.
     */
    playBegin: function()
    {
        this.controller.populate();  // XXX Would be nice during 3-2-1 if room buttons could be enabled.
        if (!this.controller.tutorLevelMayBegin()) 
        {
            this.playBeginSoon();
        }
    },

    /** 
     * GAME-2403 Delay 3-2-1 countdown 0.1 seconds after loading.  Test case:  2015-01-13 Level 3 or higher.  3-2-1.  Expect no order yet.  Got an order.
     */
    playBeginSoon: function()
    {
        var act = cc.CallFunc.create(this.playBeginNow, this);
        var delay = cc.DelayTime.create(0.1);
        var sequence = cc.Sequence.create(delay, act);
        this.controller.view.board.runAction(sequence);
    },

    playBeginNow: function()
    {
        // ll.log("TroubleBrewing: playBeginNow");
        this.view.playAnimation( "playBegin", this.onStartGameplayFromLevelSelect );
    },

    /**
     * Test case:  2014-11-06 Fade in.  Rapidly tap level.  Expect spots appear once.  Got double spots.  Tap.  Expect to place.  Got crash.
     */
    onStartGameplayFromLevelSelect: function()
    {
        this.model.persistentData.isLevelSelectMode = false;
        this.onStartGameplay();
    }
});

String.suffix = function(name)
{
    var names = name.split("_");
    var lastName = names[names.length - 1];
    return lastName;
};
