/*jslint node: true */
"use strict";

var require = require;

require( "src/Utility/LumosityUtils.js" );
require( "src/Game/Base/GameModel.js" );

require( "src/Game/CafeModel.js" );


var GameModel = GameModel;
var lu = LumosityUtils;

var TroubleBrewingModel = GameModel.extend(
{
    /** @lends TroubleBrewingModel.prototype */

    MILLISECOND_MAX_VALUE: Math.pow(2, 53 - 10),

    bonus: 0,
    firstLevel: CafeModel.firstLevel,
    maxLevel: CafeModel.getLevelMax(),
    score: 0,
    /**
     * Maximum float that resolves to 4 significant digits is  2 ** (53 - 10), since 2 ** 10 = 1024.
     * 2 ** 43 seconds is over 200,000 years.
     * http://stackoverflow.com/questions/307179/what-is-javascripts-max-int-whats-the-highest-integer-value-a-number-can-go-t
     */
    timeRemaining: 120,

    currentLevel: 1,
    session_level: 1,
    user_level: 1,

    secondsPerMinute: 60,
    seconds: 120,

    /**
     * GAME-1405 Latest results of "{}".  Test case:  2014-10-28 Boiler.  How to Play.  Quit.  Boiler expects JSON string to parse.  Got "undefined".
     * @constructs
     */
    ctor:function( settings )
    {
        this._super();
        
        this.seconds = settings.getGameLengthInSeconds();
        this.secondsOriginal = settings.gameLengthInSeconds;
        this.latestResults = undefined === this.latestResults ? "{}" : this.latestResults;
        if (undefined === this.persistentData.session_level) {
            this.persistentData.session_level = this.loadFromLastResults(this.firstLevel);
        }
        this.initSessionLevel(this.persistentData.session_level);
    },

    _levelUp: function(level)
    {
        if (undefined === level) {
            level = (this.persistentData.session_level + 1) 
                % CafeModel.getLevelMax();
        }
        this.persistentData.session_level = level;
    },

    /**
     * GAME-1403 At least first milestone.  2014-10-16 Finish session.  Do not level up.  Start next session.  Expect to fix first milestone available.  Got no milestone available.
     */
    initSessionLevel: function( newSessionLevel )
    {
        this.session_level = Math.max(this.firstLevel, this.session_level, 
            Math.min(newSessionLevel, this.maxLevel));
        this.currentLevel = this.session_level;
        this.user_level = this.session_level;
    },

    /**
     * Creates a results JSON string to send back to the application layer.
     * Save next session level to gameStats.user_level.  
     * Analogous to Penguin Pursuit (endingLevelIndex) and Train of Thought (sessionLevel), yet their keys differ.
     * A standard key would be more clear.
     *
     * Following Memory Matrix, TroubleBrewing also exports total orders cleared into metadata.  Test case: 2014-12-11 Lumosity Mobile.  Play TroubleBrewing.  Fill 2 orders.  End.  Results screen.  Expect to read "2 Orders".
     *
     * @param  {Object} stats An object that contains keys describing particular stats.
     * @return {String}       String version of results object.
     */
    exportResults: function( stats, bestStatKey )
    {
        if ( !stats )
        {
            stats = {};
        }
        stats[ "Orders" ] = this.num_correct;
        return this._super( stats, "Orders" );
    },

    recordUserLevel: function()
    {
        this.metadata[ metadataAggregateKeys.userLevel ] = this.user_level;
    },
   
    /**
     * Set level select by not being in practice mode so as not to override start play at end of gameBegin animation.
     * GAME-1403 May show tutor message.
     * GAME-1885 On begin, disable updating controller.  2015-01-02 Level 9.  Countdown.  Restart.  Expect no crash.
     */
    onBegin: function()
    {
        this.firstLevel = CafeModel.firstLevel;
        this.initSessionLevel(this.session_level);
        this.persistentData.isLevelSelectMode = !this.persistentData.isPracticeMode;
        if (this.persistentData.isLevelSelectMode) {
            this.isBlockingInput = false;
        }
    },

    /**
     * @return  Time remaining.  If practice mode, return max value.
     *
     * Change log:
     * GAME-1403 If practice mode, return max value of time remaining.  
     * Test cases:  
     * 2014-10-07 Practice:  Wait 2 minutes. Expect no tick or time out.  
     * 2014-10-07 Practice:  Set blocked tile.  Clear all tiles.  Wait 5 seconds.  
     * Expect blocked tile has removed itself and board has cleared.
     */
    onStartGameplay: function()
    {
        var timeRemaining;
        if (this.persistentData.isPracticeMode) {
            timeRemaining = this.MILLISECOND_MAX_VALUE;
        }
        else {
            timeRemaining = this.seconds;
        }
        return timeRemaining;
    },

    setMilestone: function(milestone)
    {
        this.currentLevel = this.milestoneToLevel(milestone);
        // ll.verbose("TroubleBrewingModel.setMilestone: " + this.currentLevel);
    },

    getMilestone: function()
    {
        return this.milestone(this.currentLevel);
    },

    /**
     * @return  Level corresponding to milestone or max achieved, whichever is greater.  
     */
    milestoneToLevel: function(milestone)
    {
        var level;
        if (milestone != this.milestoneUnlocked()) {
            level = milestone;
        }
        else {
            level = this.levelUnlocked();
        }
        return level;
    },

    levelUnlocked: function()
    {
        return Math.max(this.currentLevel, this.session_level, this.user_level);
    },

    milestoneUnlocked: function()
    {
        return this.milestone(this.levelUnlocked());
    },
    
    /**
     * @return  WG-1410 Usually display text of milestone to tiles.  Special case: If milestone is the highest, and it is locked, display "(?)".  2014-08-28 To clarify that highest milestone has not yet been unlocked, Aaron expects to obfuscate tile count at max.  Wouter expects a simple milestone display.  2014-08-14 To simplify, Abhi expects current tile count on current milestone.
     */
    milestoneToLevelText: function(milestone)
    {
        return this.milestoneToLevel(milestone).toString();
    },

    milestoneMax: function()
    {
        return CafeModel.visibleLevelMax;
    },

    /**
     * @param   tilesPerMinute  Confident of how many tiles can be cleared in 60 seconds.  
     * @return  Index into TroubleBrewing 1 levels [0..11], to unlock a milestone.  
     */
    milestone: function(level)
    {
        var index = Math.max(0, level);
        return index;
    }
});
