/*jslint node: true */
"use strict";

require("src/Game/Scene.js" );
require("src/Utility/LumosityGlobals.js" );
require("src/Utility/LumosityUtils.js" );

var TutorScene = Scene.extend(
{
    ctor: function(onCompleteTutorialDelegate, parent)
    {
        this.parent = parent;
        this.game = onCompleteTutorialDelegate;
        this.beginCcbFileName = "TutorBegin";
        this.ccbFileName = "Tutor";
        this.owner = this;
    },

    onBegin: function()
    {
        this._super();
        this.game.scene.hide();
        this.updateMessage(this.current.text);
    },

    /**
     * @param   tutor   Has string property "next" and "dim".
     * GAME-1874 In Layout0.ccb I added name "dim".  When "trialOrders" sequence ends, this plays.  When "trialOrdersEnd" sequence ends this dim plays "end".
     */
    update: function(tutor)
    {
        this._super();
        if (!tutor || !tutor.isUpdating)
        {
            return;
        }
        var nextState = tutor.next;
        var dimState = tutor.dim;
        if (nextState) 
        {
            var currentState = this.getCurrentState();
            if (nextState != currentState)
            {
                ll.log('TutorScene.update: from "' + currentState
                    + '" to "' + nextState 
                    + '" at ' + this.game.controller.model.now + ' milliseconds.');
                cc.BuilderReaderHelper.setState(this.root, nextState);
                this.updateMessage(this.current.text);
            }
        }
        if (dimState)
        {
            this.updateDim(dimState);
        }
    },

    updateDim: function(dimState)
    {
        var dim = this.root.board.rooms.layout_0.dim;
        dim.setVisible("none" != dimState);
        cc.BuilderReaderHelper.setState(dim, dimState);
    },

    /**
     * If next state ends with "End" or "End2" then do not change string.
     * Test case:  2014-12-16 Tutor 2 ingredients.  Do not expect to flash Tutor 1 ingredient message.
     */
    isNextMessage: function(nextState)
    {
        if (!nextState || "none" == nextState || "playBegin" == nextState) {
            return false;
        }
        else {
            var countless = nextState.split("2")[0];
            return !String.endsWith(countless, "End");
        }
    },

    /**
     * If no string, set to empty.
     * Test case:  2014-12-16 Tutor third trial.  Expect message.  Got overflow message.
     */
    updateMessage: function(text)
    {
        var nextState = this.getCurrentState();
        if (this.isNextMessage(nextState)) 
        {
            var string = "tutor_" + nextState;
            // ll.log("TutorScene.updateMessage: " + string);
            if (string in LumosityGlobals.localizedStrings) 
            {
                string = localize(string);
            }
            else 
            {
                string = "";
            }
            text.setString(string);
        }
    },

    getCurrentState: function()
    {
        var state;
        if (this.current) 
        {
            state = this.current.animationManager.getRunningSequenceName();
        }
        return state;
    },

    /**
     * GAME-1885 After loading tutorial level, hide dim layer.  Test case:  2015-01-11 How to play.  Dim.  Skip.  How to play.  Expect bright.  Got dim.
     * GAME-1885 After loading tutorial level, reset animation state.  Test case:  2015-01-12 How to play mode at first screen.  Trash.  Skip tutorial.  How to play mode.  Idle for 10 minutes.  Expect tutorial.  Coffee machine loops sliding into view.
     * When tutor loads its level, do not begin the main scene; instead begin playing the tutorial.  Test case:  2015-02-11 How to Play Mode.  Observe intro.  After intro Mark Torii expects to see a tutorial coffee machine.  Got level select screen that does not respond.
     */
    loadLevel: function()
    {
        this.load();
        this.game.loadLevel(0, this.root.board);
        this.game.controller.populate(this.root.board);
        if (this.game.scene.begin) {
            this.game.scene.begin.setVisible(false);
        }
        if (this.game.scene.root) {
            this.game.scene.root.setVisible(false);
        }
        this.root.setVisible(true);
        this.updateDim("none");
        this.playBegin();
    },

    /**
     * During practice, do not complete tutorial on end of gameplay.  After tutorial trials end animation finishes ("trialEnd"), show end tutorial pop-up.  Test case:  2015-02-13 After serve last of tutorial, play animation to scroll coffee machine offscreen.  Got coffee machine disappears abruptly.
     *
     * After any animation, localize message corresponding to next state.
     *
     * Tutor state would be cleaner if no chained timelines for state transitions in CocosBuilder,
     * and if CafeTutor could infer when to change state by end of a label.
     * Detecting end of label in Flash would be more work.
     */
    completedAnimationSequenceNamed: function(sequenceName)
    {
        var model = this.game.controller.model;
        var isUpdating = model.tutor && model.tutor.isUpdating;
        ll.log('TutorScene.completedAnimationSequenceNamed: "' + sequenceName 
            + '" at ' + model.now + ' milliseconds. isUpdating ' + isUpdating);
        var countless = sequenceName.split("2")[0];
        if (sequenceName == "gameBegin")
        {
            this.loadLevel();
        }
        else if (0 == sequenceName.indexOf("gameEnd")) 
        {
        }
        else if (0 == sequenceName.indexOf("trialEnd")) 
        {
            this.game.onCompleteTutorial(false);
        }
        else if (sequenceName == "playBegin") 
        {
            this.game.controller.onStartGameplay();
            this.game.view.cue_showSkipTutorialButton();
        }
        else 
        {
            var nextState = this.getCurrentState();
            var running = this.current.animationManager.getRunningSequenceName();
            ll.log('TutorScene.completedAnimationSequenceNamed: "' + sequenceName 
                + '". Next: "' + nextState 
                + '". Running: "' + running + '".');
            this.current.currentState = nextState || sequenceName;
            if (nextState && sequenceName != nextState) 
            {
                if (running != nextState) 
                {
                    this.runAnimations(nextState);
                }
            }
        }
        this.updateMessage(this.current.text);
    }
});

/**
 * @return  If name ends with ending.  If name is shorter than ending, name cannot end with ending.
 */
String.endsWith = function(name, ending)
{
    return ending.length <= name.length 
            && name.indexOf(ending) 
            == name.length - ending.length;
}

/**
 * @return  Object of {nodeName: stringKey}.  Value of "room_1_text" corresponds to the label to play.
 */
TutorScene.specifyRoomTexts = function(isRoom2or3)
{
    var labelPrefix = isRoom2or3 ? "room_2" : "room_3";
    var key = "tutor_" + labelPrefix;
    var nameTexts = {};
    nameTexts.room_0_text = key + "_0";
    nameTexts.room_1_text = key + "_1";
    nameTexts.room_2_text = key + "_2";
    nameTexts.wrongOrder = "";
    return nameTexts;
};

TutorScene.level = function(root, isRoom2or3)
{
    var nameTexts = TutorScene.specifyRoomTexts(isRoom2or3);
    for (var name in nameTexts) {
        root[name].setString(localize(nameTexts[name]));
    }
    cc.BuilderReaderHelper.runAnimations(root, "begin");
};
