/*global cc*/
/*global require*/
/*global lumosity*/
/*global Random*/
/*global lg*/

var GameModel;
( function()
{
    "use strict";
    
    GameModel = cc.Class.extend(
    {
        /** @lends GameModel.prototype */
        
        /** Whether or not this game should block input. */
        isBlockingInput: true,

        /** Whether or not this game should block input. */
        isGameRunning: false,

        /** A flag that tells us whether the game is over or not. */
        isGameOver: false,

        /** Whether or not this game is paused. */
        isGamePaused: false,

        /** True iff tutorial mode has ended.
         *  Used to distinguish between using onRestart to start the fitTest,
         *  and actually restarting the game. */
        tutorialEnded: false,

        /** This is confusing state-wrangling stuff. It is a janky bool that's inserted to let us
         know when onBegin was called. It's important for edge cases with the How To Play menu
         and where that navigates users depending on global state. Shitty. */
        isOnBeginCalled: false,

        /** True if this is a practice/tutorial game */
        isPracticeMode: false,

        /** True if the tutorial is skipped */
        tutorialSkipped: false,

        /** True if this is a practice/tutorial game */
        restartToPracticeMode: false,

        /** True if this is a fit test game */
        isFitTestMode: false,

        /** True if you should show the timer during a fit test game */
        shouldCreateTimer: null,
        
        /** Timer component */
        timeRemaining: 0,

        /** Time Bonus Component */
        timeBonus: 0,

        /** Score */
        score: 0,
        bonus: 0,
        prevScore: 0,
        
        /** Score multiplier component */
        multiplier: 0,
        prevMultiplier: 0,
        scoreMultiplierIndex: 0,

        meterValue: 0,
        prevMeterValue:0,

        pauseTimes: [],
        resumeTimes: [],
        
        //COMPTODO some other scoring component?
        totalResponseCount: 0,
        totalCorrectResponseCount: 0,
        consecutiveCorrectCount: 0,
        correctAnswerStreak: 0,
        
        //COMPTODO speed scoring component?
        lastAnswerTime: lumosity.game.frametime,
        averageReactionTime: 0.0,

        /** Dictionary object that will be sent back to the app with the results. */
        metadata: {},

        /** A dictionary object that is meant to hold keys for metadata. Use in derived classes. */
        currentTrial: {},

        /** Describes keys that are to be used for the stats object when exporting results. */
        StatsKeys:
        {
            CorrectStat: "correct", // Should represent the number of correct trials.
            TrialsStat:  "trials",  // The total number of trials a player reached in their playtime.
            TilesStat:   "tiles",   // The maximum number of tiles a player uncovered in a given trial (board).
            LevelStat:   "level",   // The maximum level a player reached.
            CardStat:    "cards",   // The total correct cards.
            BallStat:    "balls"    // The maximux number of balls player managed in the game.
        },

        persistentData:
        {
            isPracticeMode: false
        },
        
        /**
         *  A GameModel is a dumb object that contains all members a game may use.
         *  Each time it is run, a new model will be created. Once the game is complete,
         *  its model will be sent to the C++ layer with game completion information.
         *
         *  @constructs
         */
        ctor:function()
        {
            this.score = 0;
        },
        
        isPlaying:function()
        {
            return this.isGameRunning && !this.isGamePaused && !this.isGameOver;
        },

        addPause: function( pause )
        {
            this.pauseTimes.push( Math.round( pause * 1000 ) );
        },

        addResume: function ( resume )
        {
            this.resumeTimes.push( Math.round( resume * 1000 ) );
        },

        /**
         * Creates a results JSON string to send back to the application layer.
         * 
         * @param  {Object} stats       An object that contains keys describing particular stats.
         * @param  {String} bestStatKey A key with which to grab the "best stat" from.
         * @return {Object}             Results object.
         *
         * GAME-2504 Move session_level and user_level from metadata to game results.  Therefore, analytics can scrape session_level and user_level into game results database and gamesave events, which conventionally reserved a column for 'session_level' and 'user_level' for every game.  Aaron advises that games with levels save the metadata key 'session_level' (for the level being played) and 'user_level' for the level unlocked.  That has been the convention on web games since at least 2012; maybe since 2009.  Test case:  2015-01-23 Download gamesave of Disillusion or Trouble Brewing from s3.  Ethan's gamesave script expects a number in user_level and session_level column.  Got the string 'null'.  Analogous to the Flash bridge, [~aaron] expects Cocos2d base game to move keys to top level of game results.  Code review.  [~medwards] expects to user level appears once in metadata.
         */
        createGameResultsData: function ( stats, bestStatKey )
        {
            var results = {};

            results.score = this.score;
            results.bestStatKey = bestStatKey;
            results.gameStats = stats;
            if (undefined !== this.metadata.session_level)
            {
                results.session_level = this.metadata.session_level;
                delete this.metadata.session_level;
            }
            if (undefined !== this.metadata.user_level)
            {
                results.user_level = this.metadata.user_level;
                delete this.metadata.session_level;
            }
            results.game_result_data = this.metadata;
            this._populateDefaultGameResultData( results.game_result_data );

            if( stats && bestStatKey )
            {
                results.stat = stats[ bestStatKey ];  //NOTE: I stepped on Mark's toes with this one. We may adopt this change later, but it requires a good amount of app-side work.
            }
            else
            {
                results.stat = 0;
            }

            return results;
        },

        /**
         * Retrieves "user_level" from last game results (passed in from app).
         * Defaults to first level.
         * Adapted from Penguin Pursuit, except using metadataAggregateKeys.userLevel as key.
         * GAME-2504 Attempts to find user level in last result.  If not found, then attempts to find in game stats.  Test case:  2015-01-27 Code review.  [~medwards] expects to user level appears once in metadata.  A user plays in previous metadata format.  The user upgrades and has a result in previous format.  The program will expect to find the previously formatted user level.
         */
        loadFromLastResults: function(firstLevel)
        {
            var lastLevel = firstLevel;
            var isFound = false;
            if (lumosity && lumosity.game && lumosity.game.params && lumosity.game.params.last_results) {
                var last_results = lumosity.game.params.last_results;
                // ll.verbose("loadFromLastResults" + JSON.stringify(last_results));
                if (metadataAggregateKeys.userLevel in last_results) {
                    var foundLevel = last_results[metadataAggregateKeys.userLevel];
                    isFound = lu.isNumeric( foundLevel );
                    if (isFound) {
                        lastLevel = foundLevel;
                    }
                }
                if (!isFound) {
                    var gameStats = last_results.gameStats;
                    if (gameStats) {
                        if (metadataAggregateKeys.userLevel in gameStats) {
                            var foundLevel = gameStats[metadataAggregateKeys.userLevel];
                            isFound = lu.isNumeric( foundLevel );
                            if (isFound) {
                                lastLevel = foundLevel;
                            }
                        }
                    }
                }
            }
            return lastLevel;
        },

        /**
         * Creates game results using 'createGameResultsData' call  and stringifies them.
         *
         * @param  {Object} stats       An object that contains keys describing particular stats.
         * @param  {String} bestStatKey A key with which to grab the "best stat" from.
         * @return {String}             String version of results object.
         */
        exportResults:function( stats, bestStatKey )
        {
            var results = this.createGameResultsData ( stats, bestStatKey );

            cc.log( "results: " + JSON.stringify(results) );

            return JSON.stringify(results);
        },
        
        _populateDefaultGameResultData:function( data )
        {
            data.locale_id = lumosity.game.locale;
            data.default_locale_id = "en";
            data.version = String( lg.buildNumber ); //the verifier expects a string because that's how web is
            data.seed = Random.seed;
        }
    });

}());
